/*
 * ADDW SCCS: @(#)fileio.c	1.81 09/22/20 22:44:24
 */

/*
 * The routines in this file read and write ASCII files from the disk. All of
 * the knowledge about files are here. A better message writing scheme should
 * be used.
 */
#include	"estruct.h"
#include	"utf8.h"
#include        "edef.h"

#if	SHOWMODREL
char	filiovers[] = "@(#)fileio.c	1.81 09/22/20 22:44:24";
#endif

#if	UNIX
#include	<signal.h>
#include	<sys/wait.h>
#include	<fcntl.h>
#endif

static	mode_t	f_mode;
static	uid_t	f_uid;
static	gid_t	f_gid;

/* Under UNIX we must provide our own versions of the popen() and pclose()
 * routines so that the child which is spawned can be interrupted.
 *
 * The parent (MicroEMACS) elects to catch signals, the forked child
 * immediately exec's and so the signal catching routine of the child is
 * overlayed. The kernel has no option but to set signals for the exec'd
 * process to be ignored.
 *
 * This version of popen (called pi_open) switches signals in the child to
 * the default value (ie being interruptable) before doing the exec.
 *
 * Because popen uses its own internal array of process ids, we must provide
 * a pclose equivalent.
 *						afb  20th August 1987
 */
#define	READ_DESC	0	// stdin
#define	WRT_DESC	1	// stdout
#define	POPPIDSZ	17

#if	UNIX == 0
extern FILE*	fdopen();
#endif

/* PID corresponding to an open pipe fd.
 * Indexed by filed -- this assumes that we will never use fd > 17,
 * should be OK, but ...
 */
static int	pi_pid[POPPIDSZ];

#if	UNIX

/* Open a pipe onto ffp.
 * cmd	The command
 * mode	"r" or "w"
 * io_fd is a file descriptor that is to be connected to stdin or stdout
 *	of the pipe - the other end from which we read/write according to mode.
 *	Set this to -1 if this is not wanted.
 *	io_fd should not be 0 or 1, but that is checked against (but dup2()).
 *
 * The purpose of io_fd is to allow Emacs to open a file and then not
 * have to make the command in the pipe do so again by giving the file
 * name. This avoids (unlikely):
 * * race conditions where the file become unopenable
 * * permissions problems where Emacs can open it but the command cannot
 *
 * It is up to code elsewhere to deal with closing io_fd in the rest of Emacs.
 *
 * NB: ffropen() & ffwopen() also open to ffp.
 *
 * Set appropriately: eofflag, ffpWrite, ffpOnPipe
 * ffName will be set to pipe "pipe".
 *
 * Return:
 *	FIOSUC if OK
 *	FIOMEM - out of memory
 *	FIOERR - pipe() or fork() fail
 */
int	pi_open(const char* cmd, const char* mode, int io_fd)
{
	int	 p[2];
	register int i;
	register int parent_fds, child_fds, pid;

	if(pipe(p) < 0)
		return(FIOERR);

	mlwrite("%#Opening '%s' pipe to '%s'", mode, cmd);

	eofflag = FALSE;

	/* Set ffpWrite to: 0 to read, 1 to write */
	ffpWrite = (*mode == 'w');

	if(ffpWrite) {
		// write
		parent_fds = p[WRT_DESC];
		child_fds  = p[READ_DESC];
		PipeWriteFail = 0;	// Reset - only on write
	} else {
		parent_fds = p[READ_DESC];
		child_fds  = p[WRT_DESC];
	}

	switch(pid = doFork()) {
	  case -1:
		close(p[0]);
		close(p[1]);
		return(FIOERR);
	  case 0:
		// Were we given a FD to make the child's STDIN/STDOUT ?
		if(io_fd >= 0)
			dup2(io_fd, ffpWrite);

		/* Make child's STDOUT/STDIN the other end of the pipe that
		 * the parent holds:
		 */
		dup2(child_fds, 1 - ffpWrite);

		doExecl(cmd, TRUE);
		// break;	/* To keep the compiler happy */
	}

	pi_pid[parent_fds] = pid;
	(void) close(child_fds);

	tcpys(ffName, "pipe");	// Won't be used, but is a valid string

	// Wrap the fd up in a FILE*
	if((ffp = fdopen(parent_fds, mode)) != (FILE*)NULL) {
		ffpOnPipe = TRUE;
		return(FIOSUC);
	}

	// Unlikely, but don't lose an open fd:
	i = errno;
	close(parent_fds);
	errno = i;

	return(FIOMEM);
}

/* Open a file that is possibly gzipped.
 * If fname has a .gz suffix invoke: gzip -d
 *
 * fname will prob have come from findFile().
 *
 * Open the file ourselves & pass the filed to gzip as stdin:
 * * makes it much easier to notice a failure to open the file
 * * don't need to worry about the file name containing 'strange'
 *   characters like space or '&'.
 *
 * Return: FIOSUC on OK
 *	   FIO??? on error
 */
int	openPossGz(const BYTE* fname)
{
	int	rc;
	FILE*	fp;
	int	e;

	if((rc = ffropen(fname)) != FIOSUC)
		return(rc);

	// If it is not a gzipped file, all done:
	if(strcmp(strchr(fname, '\0') - 3, ".gz") != 0)
		return(rc);

	// Open the pipe reading from the opened filed:
	fp = ffp;
	ffp = (FILE*)NULL;

	rc = pi_open("gzip -d", "r", fileno(fp));

	// We know the name of the file, use that not "pipe" that pi_open() sets
	tcpys(ffName, fname);

	// Close the handle on the file without changing errno
	e = errno;
	fclose(fp);
	errno = e;

	return(rc);
}

/* Close a handle on a pipe in ffp.
 * This can be called by ffclose() if ffp is open on a pipe.
 *
 * ffp will be set NULL by this.
 *
 * If the exit code is non zero a message is appended to the Message History.
 * playError will be modified - with said message.
 *
 * The return values are consisten with what ffclose() returns,
 * not that this is properly tested for.
 *
 * This preserves the value of errno, ie it is not changed by closing the pipe.
 *
 * Returns:
 *	FIOSUC - closed OK & exit code of the child is 0
 *	FIOERR - child exit code non zero or some other error,
 *		 the pipe should be closed anyway
 *
 * The process exit code is stored in global rval.
 * Be consistent with what happens in respawn() & filterLines().
 */
int	pi_close(void)
{
	register int f;
	int	status;
	int	savErrno = errno;

	tsetempty(playError);

	f = fileno(ffp);
	(void) fclose(ffp);
	ffp = (FILE*)NULL;
	eofflag = TRUE;
	PipeWriteFail = 0;	// Reset

	/* Wait for child to die */
	if(reapChild(pi_pid[f]) == 0) {
		status = FIOSUC;
	} else {
		status = FIOERR;
		mlwrite("%#Pipe process %t", playError);
	}

	/* mark this pipe closed */
	pi_pid[f] = 0;

	errno = savErrno;

	return(status);
}

#endif	/* UNIX */


/*
 * Open a file for reading.
 *
 * Reset f_relink otherwise we muck about with files that were once
 * read only but now are not and when we do not need to, such as closing
 * a file that is open for read.
 *
 * isSlink is updated - regardless of the open eventually succeeding.
 *
 * fn is stored in ffName - for use in messages.
 *
 * Return:
 *	FIOSUC	Open worked
 *	FIOFNF	File not opened, not exist, perms, ...
 */
int	ffropen(const char* fn)
{
#if	UNIX
	struct	stat	readStat;	// Stat of stream being read
	numLinks = 0;			// Set 0 in case file does not open
#endif

	f_relink = FALSE;

	isSymlink(fn);

	mlwrite("%#Opening '%s' '%s'", "r", fn);
        if((ffp = fopen(fn, "r")) == NULL) {
		mlwrite("%ECannot open '%s' as: %e", fn, errno);
                return(FIOFNF);
	}

#if	UNIX
	fstat(fileno(ffp), &readStat);
	numLinks = (int)(readStat.st_nlink);
#endif

	eofflag = FALSE;
	ffpWrite = FALSE;
	ffpOnPipe = FALSE;
	tcpys(ffName, fn);

        return(FIOSUC);
}

/*
 * "Relink" a file. This means moving "fname" (which cannot be written)
 * sideways and creating a file of the same name which can be written.
 *
 * This is done by linking the file "fname" to another file, unlinking
 * "fname" and then creat-ing another with mode 666 (mode to be changed
 * later).
 *
 * If fname is /usr/tmp/foo.c the backup file becomes /usr/tmp/%foo.c
 *
 * Return a FIOxxx value
 *	FIOSUC on success
 *	FIOERR link or unlink failed
 *	FIOMEM out of memory
 *	If there is a failure - a message will be written
 */
#if	UNIX
int	relink(const char* fname)
{
	char*	base;
	int	fd;
	LineLn	pcOffset;	// Offset of where we put '%' in fname

	// Where abouts does that '%' go ?
	if((base = strrchr(fname, '/')) == (char*) NULL)
		pcOffset = 0;
	else
		pcOffset = (LineLn)(base - fname) + 1;

	// Create the name with the '%' in it:
	if(tcpys(f_backup, fname) == FALSE || tinsc(f_backup, pcOffset, '%') == FALSE)
		return(FIOMEM);

	if(link(fname, ttext(f_backup)) == -1) {
		mlwrite("%ELink(%s, %t) failed: %e", fname, f_backup, errno);
		return(FIOERR);
	}
	if(unlink(fname) == -1) {
		mlwrite("%EUnlink of %s failed: %e", fname, errno);
		(void) unlink(ttext(f_backup));
		return(FIOERR);
	}
	if((fd = creat(fname, 0666)) == -1) {
		/*
		 * The creat() failed, quickly link the files back again.
		 */
		if(link(ttext(f_backup), fname) == -1)
			mlwrite("%ECreat failed (%e): file saved in %t",
					errno, f_backup);
		else {
			mlwrite("%ECreate of %s failed: %e", fname, errno);
			unlink(ttext(f_backup));
		}
		return(FIOERR);
	}
	close(fd);
	return(FIOSUC);
}

#include	<errno.h>

#endif	/* UNIX */

/* Open a file for writing.
 * If append: open to write at the end of existing content.
 *
 * When running under unix, not having write permission on the file does not
 * necessarily mean that we cannot write it. Check write permission in the
 * directory and relink the file if possible.
 * Do not try this if appending.
 *
 * isSlink is updated - only if the open succeeds
 *
 * fn is stored in ffName - for use in messages.
 *
 * Reset the f_relink flag; set it again later if needs be.
 * 
 * Return a FIOxxx value:
 *	FIOSUC	if all is well
 *	FIOERR	link or unlink failed
 *		not writable file & user choses not to write
 *		open write/append fails - probably permission error
 *	FIOMEM	out of memory
 *	If there is a failure - a message will be written
 */
int	ffwopen(const char* fn, int append)
{
#if	UNIX
	char*		p;
	int		i;
	char*		openHow;	// write/append

	openHow = append ? "append" : "write";

	f_relink = FALSE;

	// Get into playText the name of the directory that the
	// file is in:
	tcpys(playText, fn);
	if((p = strrchr(ttext(playText), '/')) != (char*) NULL)
		*++p = '\0';
	else
		tcpys(playText, ".");

	if( !append && relinking && access(ttext(playText), W_OK) != -1 &&
				access(fn, W_OK) == -1 && errno == EACCES) {

		if(tcpys(f_fname, fn) != TRUE) {
			mlwrite(MsgOutOfMemory);	// This might also fail
			return(FIOMEM);
		}

		CacheStat(fn);
		if(mlyesno("%s: not writable (mode %o): really write [y/n]? ", fn,
						statbuf.st_mode & 0777) != TRUE) {
			// This is to replace the prompt with something nicer:
			mlwrite("%WNot opening (%s): %s", openHow, fn);

			return(FIOERR);
		}

		if((i = relink(fn)) != FIOSUC)
			return(i);

		f_mode = statbuf.st_mode & 0777;
		f_uid  = statbuf.st_uid;
		f_gid  = statbuf.st_gid;
		f_relink = TRUE;
	}
#endif

	/* Tell user that we are writing.
	 * This is because open & writing can take a long time.
	 * Think: slow disks (retries), network delays.
	 * The message may get trashed if the open has problems.
	 */
	mlwrite("%I""%FOpening (%s): %s", openHow, fn);

//	mlwrite("%#Opening '%s' '%s'", (append ? "a" : "w"), fn);

#if	VMS
	register int	fd;

	// Not sure if this is right for append:
	if((fd = creat(fn, 0666, "rfm=var", "rat=cr")) < 0 ||
	   (ffp = fdopen(fd, (append ? "a" : "w"))) == NULL) {
#else
	if ((ffp = fopen(fn, (append ? "a" : "w"))) == NULL) {
#endif
#if	UNIX
		mlwrite("%E""%s: cannot %s: %e", fn, openHow, errno);
#else
		mlwrite("%ECannot %s %s", openHow, fn);
#endif
		f_relink = FALSE;
		return(FIOERR);
	}

	isSymlink(fn);
	eofflag = FALSE;
	ffpWrite = TRUE;
	ffpOnPipe = FALSE;
	tcpys(ffName, fn);

	return(FIOSUC);
}

/* Test if file name is a symbolic link.
 * This is testing if the last component is a symlink, not something
 * earlier in the path.
 *
 * Set isSlink to true if it is a symlink.
 * If the file does not exist or cannot be reached this is set to 0.
 *
 * Return:
 *	1  file is a symlink
 *	0  file is not a symlink
 *	-1 the file does not exist or could not be accessed
 *
 * On non UNIX systems this always returns 0 & set isSlink to 0.
 */
int	isSymlink(const char* name)
{
#if	UNIX
	struct	stat	readStat;

	isSlink = 0;

	if(lstat(name, &readStat) == -1)
		return(-1);

	isSlink = S_ISLNK(readStat.st_mode) != 0;

	return(isSlink);
#else
	isSlink = 0;

	return(0);
#endif
}

/*
 * Close a file. Should look at the status in all systems.
 * Close the file open on ffp
 *
 * Global ffName is the name of the file - used error message
 *
 * If ffp is open on a pipe - close the pipe. UNIX only.
 *
 * This preserves the value of errno on entry - except if there is
 * a problem closing, when it returns FIOERR.
 *
 * Returns:
 *	FIOSUC - closed OK
 *	FIOERR - a problem with relinking, the file will be closed anyway
 */
int	ffclose(void)
{
	int	ret = FIOSUC;	// Return value
	int	i;
	int	savErrno = errno;

	// Should not happen, but just succeed if ffp is currently closed:
	if( !ffp)
		return(ret);

#if	UNIX
	if(ffpOnPipe)
		return(pi_close());
#endif

#if	MSDOS && CTRLZ
	if(ffpWrite)
		fputc(26, ffp);	/* add a ^Z at the end of the file */
#endif
	
#if	UNIX
	/* The trouble with a straight sync() is that it flushes every file on the machine,
	 * this can take a long time - especially if there is a slow write device in use.
	 * This does not distinguish beween closing files that were open for read or write.
	 *
	 * Only fflush() & fsync() if the file was opened for writing.
	 */
	if(syncmc == TRUE && ffpWrite) {
		// Complete the write so that it happens before the sync message:
		fflush(ffp);

		mlwrite("%I""%FSyncing: %t", ffName);
		fsync(fileno(ffp));
	}
#endif

	// This could take a long time - eg network issues
	mlwrite("%I""%FClosing: %t", ffName);

#if	UNIX || (MSDOS && (LATTICE || MSC))

	i = fclose(ffp);
	ffp = (FILE*)NULL;	// Else findFile() gets confused

	if(i != 0) {
#if	UNIX
		if(f_relink == TRUE)
			mlwrite("%EError closing file (%e), backup in %t",
					errno, f_backup);
		else
#endif
			mlwrite("%EError closing file %t", ffName);
#if	UNIX
		f_relink = FALSE;
#endif
		return(FIOERR);
	}
#if	UNIX
	if(f_relink) {
		unlink(ttext(f_backup));
		chmod(ttext(f_fname), f_mode);
		if(suidRoot) {
			// Use return value to prevent warning - but we don't report this as an error
			IGNORE(chown(ttext(f_fname), f_uid, f_gid))
		}

		statbufInvalid = 1;	// We just did chmod()
		if(CacheStat(ttext(f_fname)) != -1 && statbuf.st_uid != f_uid)
			w_relink = TRUE;
	}
#endif
#else	// UNIX || (MSDOS && (LATTICE || MSC))
	fclose(ffp);
#endif

	errno = savErrno;

	return(ret);
}

/*
 * Write a line to the already opened file. The "buf" points to the buffer,
 * and the "nbuf" is its length, less the free newline. Return the status.
 * Check only at the newline.
 * Suppress the newline if nonl, this is to stop \n @ EOF of a binary file.
 * If DosLine add a \r before the \n.
 *
 * Return:
 *	FIOSUC - OK
 *	FIOERR - something went wrong
 */
int	ffputline(const BYTE* buf, LineLn nbuf, int nonl, int DosLine)
{
	LineLn	i;

#if	CRYPT
	char c; 	/* character to translate */

	if (cryptflag) {
		for (i = 0; ++i <= nbuf;) {
			c = (char)*buf++;
			cryptBuffer(&c, 1);
			putc(c, ffp);
		}
	} else
		fwrite(buf, 1, (size_t)nbuf, ffp);

#else
	fwrite(buf, 1, (size_t)nbuf, ffp);
#endif

	ST_fileWroteBytes += nbuf;

	if(!nonl) {
		if(DosLine) {
			fputc('\r', ffp);
			ST_fileWroteBytes++;
		}

		fputc('\n', ffp);
		ST_fileWroteBytes++;
	}

	/* Hopefully the ferror() check makes the pipe write check
	 * redundant - be safe:
	 */
	if(ferror(ffp) || PipeWriteFail != 0) {
		mlwrite("%EWrite I/O error");
		return (FIOERR);
	}

	return (FIOSUC);
}

/* Write a region to ffp
 *
 * reg	The region. This function is interested in the Origin & Ending
 *	LINEs, it is not interested in the size or # lines.
 *	Ie you can fake it.
 *
 * It is OK of the region does not start/end at the start/end of line.
 * If it does not end at end of the last line then there will not be a \n
 * after the last bytes output.
 *
 * Return:
 *	FIOSUC - all worked
 *	FIOERR - I/O error - an error message will be generated
 */
int	ffWriteRegion(REGION* reg)
{
	LINE*	line = reg->r_linep;
	LineLn	stOff = reg->r_offset;

	while(line != reg->r_endLinep && !ferror(ffp) && PipeWriteFail == 0) {

		// Write the line, starting at whatever offset
		fwrite(lgetcaddr(line, stOff), 1, (size_t)(lused(line) - stOff), ffp);
		fputc('\n', ffp);

		// Go the next line
		line = lforw(line);
		stOff = 0;
	}

	// Deal with any partial part of the last line&
	if(reg->r_endOffset > stOff && !ferror(ffp) && PipeWriteFail == 0)
		fwrite(ltext(line), 1, (size_t)(reg->r_endOffset - stOff), ffp);

	if(ferror(ffp)) {
		mlwrite("%EWrite I/O error");
		return(FIOERR);
	}

	return(FIOSUC);
}

/*
 * Read a line from a file, put it into a newly allocated *newline.
 *
 * Complain about long lines and lines at the end of the file that don't have
 * a newline present. Check for I/O errors too. Return status.
 *
 * Read from FILE* ffp.
 *
 * This does not depend on curbp or curwp being anything in particular.
 *
 * Reporting via globals (yuck!)
 * If there is a UTF8 error - set global ffutf8err. This is cleared on entry.
 * On EOF set global eofflag
 *
 * newline:	pointer to new LINE struct.
 *		This is allocated on return: FIOSUC FIOFUN
 * utf8:	True if MB chars are allowed
 *
 * If reading from a pipe and pi_echo is true: echo input to the screen.
 * The test on ffpOnPipe below should be redundant, but does no harm.
 *
 * Return:
 *	FIOSUC - success
 *	FIOFNF FIOEOF FIOERR FIOMEM FIOFUN - all errors
 */
int	ffgetline(LINE** newline, int utf8)
{
	int	c;			// Char just read
	LINE*	line;
	LineLn	llen = 0;		// Length of line so far
	size_t	chunkLen = 0;		// Bytes in ibuf
	int	echo;			// Should we echo chars read to the terminal ?
	BYTE	ibuf[200];		// Read in chunks into here - should be enough for most lines

	ffutf8err = 0;

	/* if we are at the end...return it */
	if (eofflag)
		return(FIOEOF);

	// Allocate a new line
	if( !(line = lalloc(0)))
		return(FIOMEM);
	*newline = line;

	echo = ffpOnPipe && pi_echo;

	while((c = getc(ffp)) != EOF && c != '\n') {
		/* Echo of characters is done at this (the lowest) level
		 * so that things like cryptBuffer will work properly when it
		 * outputs the "Enter key:" prompt.
		 * 
		 * Mapping of control chars protects the terminal (someone
		 * might be cating a binary).
		 * If utf8 - chars with top bit is OK. If this results in a
		 * bad char - let the display handle it.
		 */
		if(echo) {
			register int	x = c;

			if(x >= 0x80 && !utf8) {
				TTputc('M');
				TTputc('-');
				x &= ~0x80;
			}
// TODO meta/ctrl display:
			if((x >= ' ' && (utf8 || x < 0x7f)) || x == '\t')
				TTputc(x);
			else {
				TTputc('^');
				TTputc((x == CharDelete) ? '?' : (x | '@'));
			}

			/* This flushes out character by character which is quite
			 * expensive. Unfortunately as what is writing down the pipe
			 * could pause at any moment we could have a part written
			 * buffer to the screen. We could deal with select()/poll()
			 * but that means abandoning stdio - not worth it for here.
			 */
			TTflush();
		}

		/* Check/ensure that there is enough room in the temp buffer.
		 * If not copy the chunk to the line.
		 */
		if(chunkLen == sizeof(ibuf)) {
			if( !lineMoreBytes(line, llen + (LineLn)sizeof(ibuf))) {
				FreeLine(line);
				return(FIOMEM);
			}
			memcpy(lgetcaddr(line, llen), ibuf, sizeof(ibuf));
			llen += (LineLn)sizeof(ibuf);
			chunkLen = 0;
		}

		ibuf[chunkLen++] = (BYTE)c;
        }
	if(echo) {
		TTputc('\n');
		TTputc('\r');
		TTflush();
	}

	/* Copy/flush the rest of the chunk in the temp buffer into the line.
	 * That is if there is anything to copy.
	 */
	if(chunkLen > 0) {
		if( !lineMoreBytes(line, llen + (LineLn)chunkLen)) {
			FreeLine(line);
			return(FIOMEM);
		}
		memcpy(lgetcaddr(line, llen), ibuf, chunkLen);
		llen += (LineLn)chunkLen;
	}
	lsetused(line, llen);

	ST_fileReadBytes += llen;

#if	ST520
	if(lgetc(line, llen - 1) == '\r')
		lsetused(line, lused(line) - 1)
#endif

#if	CRYPT
	if(cryptflag)
		cryptBuffer(ltext(line), lused(line));
#endif

	/* Check that the line is validly UTF8 encoded ...
	 * If the buffer is in UTF8 mode.
	 */
	if(utf8) {
		if(!utf8_chkbuf(ltext(line), (size_t)lused(line))) {
			ffutf8err = TRUE;
		} else {
			// No error - record if line has utf8 chars
			if(lused(line) != utf8_chars)
				lsetUtf8(line);
		}
	}

	if(c == EOF) {
		if(ferror(ffp)) {
			mlwrite("%EFile read error");
			FreeLine(line);
			return (FIOERR);
		}

		eofflag = TRUE;

		// Line at EOF does not end in \n ?
		if(llen != 0)
			return(FIOFUN);

		FreeLine(line);
		return (FIOEOF);
	}

	ST_fileReadBytes++;		// Count the \n not in stored the LINE

	return(FIOSUC);
}

/* Does <fname> exist on disk?
 *
 * Return TRUE if the files does exist.
 */
int	fexist(const char* fname)
{
#if	UNIX

	if(access(fname, F_OK) != -1)
		return(TRUE);

	return(FALSE);

#else	/* !UNIX */
	FILE *fp;

	/* try to open the file for reading */
	fp = fopen(fname, "r");

	/* if it fails, just return false! */
	if (fp == NULL)
		return(FALSE);

	/* otherwise, close it and report true */
	fclose(fp);
	return(TRUE);
#endif	/* ! UNIX */
}

#if	AZTEC && MSDOS
#undef	fgetc
/*	a1getc:		Get an ascii char from the file input stream
 *			but DO NOT strip the high bit
 */
int	a1getc(FILE* fp)
{
	int c;		/* translated character */

	c = getc(fp);	/* get the character */

	/* if its a <LF> char, throw it out  */
	while (c == 10)
		c = getc(fp);

	/* if its a <RETURN> char, change it to a LF */
	if (c == '\r')
		c = '\n';

	/* if its a ^Z, its an EOF */
	if (c == 26)
		c = EOF;

	return(c);
}
#endif

/* end */
