/* File handling, especially searching for
 *
 * ADDW SCCS: @(#)findFile.c	1.29 10/15/20 17:10:34
 */
#include	"estruct.h"
#include        "utf8.h"
#include        "edef.h"

#if	UNIX
#include <libgen.h>
#include <fcntl.h>
#endif

#if	SHOWMODREL
char	findfilevers[] = "@(#)findFile.c	1.29 10/15/20 17:10:34";
#endif

/* This describes everything that we need to know about finding
 * a file that we search for. This is used for:
 * * Startup files
 * * Documentation files
 * * Files containing macros
 * * Save files
 *
 * name is the used elsehow with the particular lookup
 *
 * path is a string of ':' separated directories that are searched in an attempt
 * to find the file name. These will have a '/' appended.
 * Specials:
 *    if the string starts with ~ or $HOME this is replaced with the value of
 *    the environment variable $HOME - if this is set.
 *    Whatever is after (if there is anything) is copied and should start with a '/'.
 *    Note that ~fred does NOT mean fred's home directory.
 * If the uservariable %_path-name exists this is used instead of the value here,
 * eg %_path-save.
 *
 * prefix is always inserted after the directory name & before the file name.
 *	If nothing is wanted this should be the empty string.
 *
 * suffix is added at the end of the path.
 *	If nothing is wanted this should be the empty string.
 *
 * Prefix should have a trailing '/' on it.
 *
 * If FI_LANG then the following are tried:
 *	.$language_$territory	eg: en_GB
 *	.$language		eg: en
 *	"" - empty string	ie: 
 *
 * If FI_GZ then an extra suffix of .gz is tried. If that is not found
 * look again without the .gz suffix.
 *
 * Once one language & gz suffix combination is found in a directory
 * then that is deemed enough - ie more combinations in that directory
 * will not be looked for (even if MULTIPLE).
 *
 * Once a file is found: stop looking, but if MULTIPLE continue looking
 * in other directories.
 *
 * Ie: <searched path - dirs>/<prefix><file name><[.$language[_$territory]]><suffix>[.gz]
 *
 * Yes: macro files not used yet.
 */
typedef	struct	{
	const	BYTE*	fi_name;	// Name of this search type
	const	char*	fi_path;	// ':' separated list of directories to try
	const	char*	fi_prefix;	// Before fname
	const	char*	fi_suffix;	// At the very end
	int		fi_flags;
#define	FI_LANG		0x01		// Try to insert $language_$territory
#define	FI_MULTIPLE	0x02		// Find as many as possibe
#define	FI_GZ		0x04		// Try a .gz suffix
} fTypeI;

// Search directories for the different types:
char	const	dotHome[]  = { "." PATHSTR "~"};
char	const	instPath[] = { "~/.microemacs.d" PATHSTR DOCDIR };
char	const	etcInst[]  = { "~/.microemacs.d" PATHSTR SYSCONFDIR PATHSTR DOCDIR };
char	const	homeOnly[] = { "~" };

/* Parameters for the different things that we look for.
 */
fTypeI	const	FindInfo[] = {
	[fT_Save]	= { "save",         dotHome,  "",        EMACS_SAVE_SUFFIX, 0 },
	[fT_Help]	= { "help",         instPath, "help/",   ".help",           FI_LANG | FI_GZ },
	[fT_HelpIndex]	= { "help",         instPath, "help/",   ".index",          FI_LANG | FI_GZ | FI_MULTIPLE },
	[fT_HelpTmpl]	= { "help",         instPath, "help/",   ".template",       FI_LANG | FI_GZ },
	[fT_StartGlob]	= { "startup",      etcInst,  "emacs",   ".rc",             0 },
	[fT_StartUser]	= { "user-startup", homeOnly, "",        ".emacsrc",        0 },
	[fT_Macro]	= { "macro",        instPath, "macros/", ".mac",            FI_GZ },
	[fT_MacroKeys]	= { "macro",        instPath, "macros/", ".index",          FI_LANG | FI_GZ | FI_MULTIPLE },
};

/* Find a file with a name based on fname.
 * type is a findTypes enum value; this implies: a search path; prefix (directory);
 * suffix; options:
 * + LANG	search using $language_$territory
 * + GZ		look for files with .gz suffix and decompress if found
 * + MULTIPLE	if a file is found, keep looking in other directories
 *
 * A set of directories are searched, the set depends on type. This will
 * be overridden if a user variable like %_path-help exists where 'help' is
 * the name of type in FindInfo[] above.
 *
 * If just where the prefix is added the prefix is already there, do
 * not append it (again). This enables an index file to reference a
 * file in the same directory. Maybe this ought to only happen if
 * fname starts "./".
 *
 * If fname contains a '/' then do not search directories. Suffix searching
 * still happens.
 *
 * If the file was found by adding a .gz suffix - a pipe is opened
 * from which the uncompressed file can be read.
 *
 * With the LANG option the values tried are ".$language_$territory" ".$language" ""
 *
 * Thus the per directory search could be (assuming GZ option & ".help" suffix):
 *	Buffer-modes.en_GB.help.gz
 *	Buffer-modes.en_GB.help
 *	Buffer-modes.en.help.gz
 *	Buffer-modes.en.help
 *	Buffer-modes.help.gz
 *	Buffer-modes.help
 *
 * fname is not normally expected to contain a '.', but if it does it
 * replaces the LANG search. Eg if "Buffer-modes.en" the per directory
 * search will be:
 *	Buffer-modes.en.help.gz
 *	Buffer-modes.en.help
 * A trailing '.' also suppresses LANG. Eg "Buffer-modes." searches:
 *	Buffer-modes.help.gz
 *	Buffer-modes.help
 *
 * It is assumed that ffp is NULL on entry.
 *
 * Use playText, playText2 & playDir - which have the NUL flag set.
 *
 * Return count of files found, ie 0 means nothing found.
 * Leave the first file open for reading on global var fpp.
 * The file name will be put in foundFiles (NUL terminated). If multiple
 * files are found there will be multiple strings in foundFiles.
 *
 * Beware: foundFiles is NUL terminated, this means that it can be passed to a
 * system call and work properly. But do not ask mlwrite() to format as %t as
 * the terminating NUL will be shown as ^@ -- pass as ttext() and format as %s.
 *
 * The name of the first file opened will be in ffName - even if it is
 * opened indirectly via a pipe to gzip -d.
 *
 * This is the new flook()
 */
int	findFile(const BYTE* fname, findTypes type)
{
	const	char*	path;			// List of directories to try
	const	char*	play;
	BYTE		userVar[30];		// "_path-"(ft->fi_name) - no leading %
	const	fTypeI*	ft = &FindInfo[type];

	// If there is a / then don't search directories:
	if((path = strrchr(fname, '/')) != (char*)NULL) {
		tcpym(playDir, fname, (LineLn)((const BYTE*)path - fname));
		fname = path + 1;
		path = ttext(playDir);
	} else {
		// What path do we use, override fixed with user variable - if set
		path = ft->fi_path;
		strcat(strcpy(userVar, "_path-"), ft->fi_name);
		if((play = gtusr(userVar)) != (char*)NULL)
			path = play;
	}

	return(findFilePath(path, fname, type));
}

/* This does the real work of findFile()
 * See there for more comments.
 *
 * path is a string that is a set of directories separated by PATHCHR.
 *	If there is not a PATHCHR then just look in one directory.
 *
 * Return count of files found, ie 0 means nothing found.
 */
int	findFilePath(const char* path, const BYTE* fname, findTypes type)
{
	int	found = 0;		// # files found
	char*	play;
	int	lenName;		// Length to end of fname
	FILE*	fptr;			// What we open onto
	const	fTypeI*	ft = &FindInfo[type];
	BYTE*	dot;			// Used as boolean to suppress language stuff
	char*	pathEnd = strchr(path, '\0');
	size_t	len;

	/* If fname contains '.' don't do LANG stuff.
	 * If the '.' is the last character then remove it from fname.
	 */
	if((dot = strrchr(fname, '.')) && dot[1] == '\0') {
		tcpym(playText2, fname, dot - fname);
		fname = ttext(playText2);
	}

	tsetempty(foundFiles);

	// Walk through the path
	while(path < pathEnd) {
		tcpys(playText, path);

		/* If there is a PATHCHR in what we have just copied
		 * truncate at that point - leave the rest for another iteration.
		 */
		if((play = strchr(path, PATHCHR)) != (char*)NULL) {
			tsetused(playText, (play - path));
			path = play + 1;
		} else	// No other directories to check:
			path = strchr(path, '\0');

		// Substitute a user or environment variable - start of path only
		play = ttext(playText);
		if((play[0] == '$' || play[0] == '%') &&
		   (len = strspn(play + 1, VarNameChars)) > 0 && len <= NVSIZE) {
			char	vname[NVSIZE + 3];
			memcpy(vname, play, len + 1);
			vname[len + 1] = '\0';

			if(getVarValue(vname, playText4) == TRUE) {
				tdeln(playText, 0, (LineLn)(len + 1));
				tinst(playText, 0, playText4);
			}
		}

		// Substitute ~ for the home directory ?
		if(home[0] != '\0' && tgetch(playText, 0) == '~') {
			tdeln(playText, 0, 1);
			tinss(playText, 0, home);
		}

		// Having two / confuses the prefix suppression below:
		if(tgetch(playText, tused(playText) - 1) != '/')
			tcatc(playText, '/');

		// If the prefix is already at the end of the growing path, don't add it again:
		if((size_t)tused(playText) < strlen(ft->fi_prefix) ||
		   strcmp(tgetcaddr(playText, ((size_t)tused(playText) - strlen(ft->fi_prefix))), ft->fi_prefix) != 0)
			tcats(playText, ft->fi_prefix);

		tcats(playText, fname);

		// We may want to come back to this length:
		lenName = (int)tused(playText);

		/* Set add to:
		 * 3 - try .en_GB then	.en then no language stuff
		 * 2 - try		.en then no language stuff
		 * 1 - try			 no language stuff
		 */
		int	add = 1;
		if(ft->fi_flags & FI_LANG) {
			if(*language && *territory)
				add = 3;
			else if(*language)
				add = 2;
			// else leave add = 1
		}

		if(dot)	// fname had trailing ".something", suppress language stuff
			add = 1;

		// This is true if we are to try appending a .gz suffix:
		int	gz = ft->fi_flags & FI_GZ;

		// This loop may go round up to 6 times:
		while(add > 0) {
			tsetused(playText, lenName);	// Reset

			// What language stuff gets added ?
			if(add >= 2) {
				tcatc(playText, '.');
				tcats(playText, language);
			}
			if(add == 3) {
				tcatc(playText, '_');
				tcats(playText, territory);
			}

			tcats(playText, ft->fi_suffix);	// Suffix

			if(gz)
				tcats(playText, ".gz");	// Maybe .gz

			// mlwrite("%#try to open %s", ttext(playText));
			if((fptr = fopen(ttext(playText), "r")) != (FILE*)NULL) {
				mlwrite("%#Found: '%t'", playText);
				// Assign the first that we succeed in opening to ffp, unless gz:
				if(ffp == (FILE*)NULL) {
					// Set flags, etc, as ffropen()
					f_relink = FALSE;
					eofflag = FALSE;
					ffpWrite = FALSE;
					ffpOnPipe = FALSE;

					if(gz) {
						/* Don't use zcat:
						 * * It can be a shell script: gzip -cd
						 * * On some systems it is a link to compress
						 */
						pi_open("gzip -d", "r", fileno(fptr));
						fclose(fptr);
					} else {
						ffp = fptr;
					}

					/* We know the name of the file, set in ffName.
					 * Beware: do after pi_open() as that sets ffName to "pipe"
					 */
					tcpyt(ffName, playText);
				} else
					fclose(fptr);

				found++;

				// Append what we found - add NUL in case multiple found
				tcatt(foundFiles, playText);
				tcatc(foundFiles, '\0');

				break;	// Don't look for other language or gz variants
			}

			// If did not add .gz try the next language combination:
			if( !gz) {
				add--;
				gz = ft->fi_flags & FI_GZ;
			} else
				gz = 0;
		}

		// Found something stop - unless want multiple names
		if(found && (ft->fi_flags & FI_MULTIPLE) == 0)
			break;
	}

	return(found);
}

/* For use within list-buffers -- show the various search paths
 *
 * Return 0 if all OK.
 */
int	WriteSearchPathsToBuffer(BUFFER* blp)
{
	int		ret = 0;
	const	fTypeI*	fp;

	ret |= !addline("", blp, 1);
	ret |= !addline("File search paths", blp, 1);
	ret |= !addline("-----------------", blp, 1);
	ret |= !addline("Override by eg: %_path-save", blp, 1);

	for(fp = FindInfo; fp < FindInfo + sizeof(FindInfo) / sizeof(FindInfo[0]); fp++) {
		BYTE	userVar[30];		// "_path-"(fp->fi_name) - no leading %
		char*	play;
		char*	path;

		printToText(playText, "   %s", fp->fi_name);
		ret |= !addLINE(playText, blp, 0);

		path = (char*)fp->fi_path;
		strcat(strcpy(userVar, "_path-"), fp->fi_name);
		if((play = gtusr(userVar)) != (char*)NULL)
			path = play;

		printToText(playText, "      path   %s", path);
		ret |= !addLINE(playText, blp, 0);

		printToText(playText, "      prefix %s", fp->fi_prefix);
		ret |= !addLINE(playText, blp, 0);

		printToText(playText, "      suffix %s", fp->fi_suffix);
		ret |= !addLINE(playText, blp, 0);
	}

	return(ret);
}
