/* Hashing code for MicroEmacs
 *
 * ADDW SCCS: @(#)hash.c	1.8 09/03/19 12:52:46
 *
 * I was not going to do this ... but got bored on a flight to Canada.
 *
 * These routines help to quickly find a member in a table (array of struct)
 * by obtaining its (array) index.
 *
 * The hash is an array of index values.
 *
 * The struct can be of arbitrary size but must contain:
 * * a key, a unique value that will be used to find the member,
 *   this value can be of any type.
 * * a pointer, this will be NULL only at the end of the table,
 *   this NULL value marks the end of table.
 *   The pointer can be to anything.
 *
 * For each type of value 2 functions are needed:
 * * generate a hash value
 * * compare two values for equality
 *
 * To use this on some table:
 * * create a hashSt that describes the table
 * * call hashInit() - this will hash index the table
 * * look up entries with hashLookup()
 *   Returns the index or HASH_MISSING if not found
 *
 * You may call call hashAlloc() to make the hash a certain minimum size, but
 * this is not necessary. Only do this if the table will grow and you have
 * some idea of its eventual size (the hash will grow anyway, but it is
 * more efficient if you pre-size it).
 *
 * If you delete something the index needs to be rebuilt:
 * * delete the item from the table
 * * hashEmpty()
 * * hashInit()
 *
 * The hash is implemented as an array of hashNx, short so 2 bytes/slot
 * and max value 32k. The empty array is filled with HASH_EMPTY.
 * A hash index is computed by (*hs_hash)(), this is specific to the type
 * in the table. It is modulo the size of the hash array.
 * The array is probed and if there is a collition the next element is
 * inspected. Linear good for CPU cache locality.
 *
 * On lookup type specific (*hs_cmp)() is used to decide if collision or match.
 *
 * Other hashSt members:
 * hs_keyOff	Offset of the unique key that identifies the member
 *		of a table
 * hs_ptrOff	Offset of the pointer in the table.
 *		A NULL value marks the end of the table.
 *		It can be a pointer to anything
 *		It is OK for hs_keyOff and hs_ptrOff to be the same.
 * hs_size	The size of one table element
 * hs_used	The number of non HASH_EMPTY members of the hash array
 * hs_slots	the the actual # slots, ie how many elements in the hash array.
 *		This is not the number asked of hashAlloc(), which will
 *		be smaller.
 *
 * hashFree()	Frees the hash array.
 *		You only need to do this if the table has shrunk a lot and
 *		you want to save space. Prob not necessary as hashNx is small.
 *
 * Give consideration to allowing the pointer to contain a special value
 * ((void*)1) that means 'this slot unused' - so it will not be indexed.
 * This would make deletion easier/quicker.
 *
 * Quite how much this is worth it is debateable. Although Emacs does a lot
 * of table lookups Emacs does not eat CPU because of this.
 * * Keyboard macros should be faster: kbdMacro stores KeyCodes that
 *   need looking up in keytab.
 * * Macros should be faster as Emacs commands need finding in CmdNames.
 *
 * String hashing algorithm. My original alogrithm has been replaced by one
 * written by Dan Bernstein: http://www.cse.yorku.ca/~oz/hash.html
 * I did try rehashing if the first hash produced a collision, the rehashed
 * value was computed by ignoring the first char in the string. It did not
 * make a big difference to the djb2 algorithm so I decided that it was
 * not worth the extra complication. The code looked like:
 *	char* v = *(char**)val + 1;
 *	indexp = td->hs_index +(td->hs_hash)(td, &v);
 *
 * My benchmarking was to run Emacs and immediately run list-buffers, by then
 * it will had done mainly string hashing, command lookup being just the
 * list-buffers lookup.
 * These all had:
 *        238  Key bindings, max 256
 *        290  Hash lookups
 *
 * Original:
 *        688  Hash probes
 * 
 * Original with rehash (str + 1):
 *        502  Hash probes
 * 
 * sdbm
 *        444  Hash probes
 * 
 * sdbm with rehash
 *        377  Hash probes
 * 
 * djb2
 *        382  Hash probes
 * 
 * djb2 + rehash
 *        364  Hash probes
 *
 * Some new tests [[ look at test.c ]] looking up everything in CmdNames once.
 * I get (probes how-many-this-number-of-probes):
 * 1   173
 * 2   27
 * 3   10
 * 4   9
 * 5   6
 * 6   4
 * 7   5
 * 8   0
 * 9   4
 * 10  0
 * 11  1
 *
 * Integer hashing [[ look at test.c ]]
 * Surprisingly my bash-a-few-shifts-and-xors-together works very well.
 * I tried several different ones that I found from the Internet, none
 * performed as well.
 * Resuts (test.c) using every key in the bind table & chars ' ' -> '~'
 * I get (probes how-many-this-number-of-probes):
 *  1   192
 *  2   64
 *  3   13
 *  4   9
 *  5   2
 *  6   1
 */
#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"

#if	SHOWMODREL
#include	"emdmap.h"
char	hashvers[] = "@(#)hash.c	1.8 09/03/19 12:52:46";
#endif

#define	HASH_EMPTY	SHRT_MAX	// Mark a hash slot as empty
#define	HASH_FREE_RATIO	3		// Keep 1/3 slots free

/* Add a new member to the hash
 *
 * td    	description of the table being hashed
 * hash     the hash, ie where to put index into the hash index
 * index    the position that str is found in the table
 *
 * This must not be used to add something that is already there.
 * So: if adding either:
 * * be sure that there is enough room to add without need to realloc
 * * call this before str is added to (the end of) the table
 *
 * The index table will be resized if needs be
 */
int	hashAdd(hashSt* td, void* val, hashNx index)
{
	if(td->hs_slots - td->hs_used < td->hs_slots / HASH_FREE_RATIO)
		hashInit(td);

	hashNx*	indexp;
	indexp = td->hs_index + (td->hs_hash)(td, val);

	while(*indexp != HASH_EMPTY)
		if(++indexp >= td->hs_index + td->hs_slots)
			indexp = td->hs_index;

	*indexp = index;
	td->hs_used++;

	return(1);
}

/* Count the number of entries that need to be hashed
 */
int	hashCount(hashSt* td)
{
	int	entries = 0;
	void*	actp;		// Pointer to the action

	actp = (void*)((char*)td->hs_table + td->hs_ptrOff);

	// Count # entries
	while(*(char**)actp != NULL) {
		entries++;
		actp = (void*)((char*)actp + td->hs_size);
	}

	return(entries);
}

/* Initialise a hash table:
 * * allocate it
 * * populate it
 *
 * Work out how many entries by looking at the action in the structure
 */
int	hashInit(hashSt* td)
{
	if(hashAlloc(td, hashCount(td)) == FALSE)
		return(FALSE);

	hashInsTable(td);

	return(TRUE);
}

/* Insert the whole table into the hash
 *
 * The value added (index) is the number of the entry in
 * the table that is being indexed.
 */
void	hashInsTable(hashSt* td)
{
	void**	actp;		// What this points at becomes NULL at table end
	void*	valp;		// Address of value
	hashNx	index = 0;

	actp = (void*)((char*)td->hs_table + td->hs_ptrOff);
	valp = (void*)((char*)td->hs_table + td->hs_keyOff);

	while(*actp != NULL) {
		// Compute a hash value and insert into the table
		hashAdd(td, valp, index);
		index++;

		/* hs_size is a size measured in multiples of 'char'
		 * Cast to char* to do pointer arithmetic in terms of char multiples.
		 * The cast to actp should really be (void**) but to do so
		 * gives warning:
		 * cast from 'char *' to 'void **' increases required alignment from 1 to 8 [-Wcast-align]
		 */
		valp = (void*)((char*)valp + td->hs_size);
		actp = (void*)((char*)actp + td->hs_size);
	}
}

/* Allocate the index and clear entries to HASH_EMPTY
 * Reallocate if it has already been allocated.
 *
 * slots    The number of slots wanted, must not be 0
 *	    This number is how many entries the table will contain,
 *	    this function increases it so that the hashing process
 *	    works well.
 *	    If the index is at least that big, do not reallocate
 *	    (but still clear to HASH_EMPTY).
 *
 * None of the tables are big, the size of hashNx is small, so doubling
 * slots does not use excessive memory.
 *
 * If this fails hs_index is left as it was.
 *
 * Return TRUE - OK
 *	  FALSE - out of memory
 */
int	hashAlloc(hashSt* td, int slots)
{
	hashNx*	index;

	slots *= 2;

	// Is the index already at least that big ?
	if(slots > td->hs_slots) {

		// New or reallocate ?
		if(td->hs_index != NULL)
			index = realloc(td->hs_index, (size_t)slots * sizeof(hashNx));
		else
			index = malloc((size_t)slots * sizeof(hashNx));

		if(index == NULL)
			return(FALSE);

		td->hs_index = index;
		td->hs_slots = slots;
	}

	hashEmpty(td);	// Clear the index

	return(TRUE);
}

/* Mark everything in the hash index as empty.
 * This is useful if you have deleted something and need to reindex.
 * It is OK to call this on a table that has not been initialised.
 */
void	hashEmpty(hashSt* td)
{
	hashNx*	index;
	int	slots;

	for(slots = td->hs_slots, index = td->hs_index; --slots >= 0; )
		*index++ = HASH_EMPTY;

	td->hs_used = 0;
}

/* Free the index associated with a hash.
 *
 * It is safe to call this if the hash has not been allocated.
 */
void	hashFree(hashSt* td)
{
	if(td->hs_index != NULL)
		free(td->hs_index);

	td->hs_index = NULL;
	td->hs_slots = 0;
}

/* Find whatever is at *val, return the index.
 * Return HASH_MISSING if it is not found.
 *
 * It is OK to call this on an uninitialised table - returns HASH_MISSING.
 */
hashNx	hashLookup(hashSt* td, const void* val)
{
	hashNx*	indexp;

	if(td->hs_index == NULL)
		return(HASH_MISSING);

	indexp = td->hs_index + (td->hs_hash)(td, val);

	ST_hashLookups++;		// Stats

	// Loop until we find it or come across an empty slot
	while(1) {
		ST_hashProbes++;	// Stats

		if(*indexp == HASH_EMPTY)
			return(HASH_MISSING);

		if((td->hs_cmp)(val, ((char*)td->hs_table + td->hs_keyOff + td->hs_size * *indexp)) == 0)
			return(*indexp);

		if(++indexp >= td->hs_index + td->hs_slots)
			indexp = td->hs_index;
	}
}


/* Strictly speaking these are not part of the hash code, but
 * they are necessary to make it work.
 *
 * These functions compute hash values and compare various types.
 * They will be used when filling in a hashSt
 */

/* Return a hash value for a string.
 *
 * valp	Pointer to the string pointer - but declared void*
 *	a this will be used to initialise hs_hash
 */
hashNx	hashStr(hashSt* td, const void* valp)
{
	int	value = 0;
	char*	str = *(char**)valp;

	while(*str)
		value = *str++ ^ value * 33;	// djb2 algorithm

	/* Other hashing algorithms tried (see top of this file):
	 * my original (i is character position in string)
	 *	value = *str++ ^ (value << 1 & 0xffff) ^ (value >> 7) ^ i++;
	 * sdbm
	 *	value = *str++ + (value << 6) + (value << 16) - value;
	 */

	value &= 0xffff;	// Else can go -ve & a -ve hash is bad
	value %= td->hs_slots;

	return((hashNx)value);
}

/* Return 0 if the 2 strings are the same
 *
 * Note that the arguments are void* but cast to char**
 * for use.
 */
int	hashStrcmp(const void* a, const void* b)
{
	return(strcmp(*(char**)a, *(char**)b));
}

/* Return a hash value for a key code
 */
hashNx	hashKeyCode(hashSt* td, const void* valp)
{
	KeyCode	key = *(KeyCode*)valp;

	return(hashNx)((((KeyCode_u)key << 4) ^ ((KeyCode_u)key >> 4) ^ ((KeyCode_u)key >> 24)) % (unsigned)td->hs_slots);
}

/* Return 0 if the 2 Keycodes are the same
 *
 * Note that the arguments are pointers to keycodes
 */
int	hashKeyCodecmp(const void*a, const void*b)
{
	return(*(KeyCode*)a != *(KeyCode*)b);
}
