/* Help routines
 * 
 * ADDW SCCS: @(#)help.c	1.69 10/08/20 20:05:25
 *
 * The buffer bmlHelpStack contains, in many ways, structured data.
 * It is viewable, but immutable - which saves many checks just in case
 * a user has changed it. A remaining problem is out of memory.
 * w_flag is not set - so changes might not be shown immediately.
 *
 * See help-on-help.help for a description of the format of help files.
 * Look at GenHelpIndices.pl - which generates the index files and does
 * some checking of the help files.
 *
 * Help can also be found in .mac files.
 *
 * Contents of [Help Stack] buffer:
 * Lines are added at the bottom when a link is followed.
 * The standard max # lines for a history buffer is done.
 * Return to the previous file/... results in the last line of the buffer being removed,
 * ie history is used to go back, going forwards is not possible.
 * 
 * Each line consists of a set of fields separated by NUL characters.
 * The first field identifies the line type:
 * 
 * 	T	h - Help file
 * 		Subsequent fields:
 * 		t	Line number of top line in the window
 * 		n	Line number where the cursor is
 * 		c	Column where the cursor is
 * 		f	path to the file
 * 			Path is a resolved path.
 * 
 * 	T	a - Apropos display
 * 		Subsequent fields:
 * 		t	Line number of top line in the window
 * 		n	Line number where the cursor is
 * 		c	Column where the cursor is
 * 		k	list of keys given to apropos command
 * 
 * 	T	k - Display of keyword search
 * 		Subsequent fields:
 * 		t	Line number of top line in the window
 * 		n	Line number where the cursor is
 * 		c	Column where the cursor is
 * 		k	list of keywords, space separated
 * 
 * Note how first 4 fields are always the same: T t n c
 * This number is #defined HelpStackNPref
 *
 *
 * Format of the _commands.index file - index to where Emacs commands are documented.
 * These files are read in to the [Help index] buffer, ie several may be
 * in the buffer, the below documents the buffer contents.
 *
 * Format of _commands.index file.
 * This is an index of Emacs commands, nothing else. It is used when \r is
 * pressed in the 'Binding list' buffer and when an ec: link is followed.
 * Line has NUL as last byte.
 * Invalid lines will be silently ignored.
 * Lines with NUL separated fields, 3 types:
 * 	#	-- Comment line, ignored.
 * 	d	-- Directory entry
 * 		Directory name
 * 	c	-- A command entry
 * 		command-name
 * 		file-name
 * 		The file-name is within the directory
 * 		and does not contain language_territory or the .help suffix
 * The 'd' type does not exist in the file, but does in the buffer.
 * The 'd' type is put there when the files are read.
 *
 * Format of _keywords.index file. This contains keywords, commands will also
 * be indexed here. Line has NUL as last byte. Invalid lines will be silently
 * ignored. Lines with NUL separated fields:
 * 	#	-- Comment line, ignored.
 * 	d	-- Directory entry
 * 		Directory name
 *	f	-- File name
 *		File name
 *	t	Title (description)
 *	k	Keywords, space separated
 *	c	Emacs command
 *		Might be followed by keywords
 * The 'd' type does not exist in the file, but does in the buffer.
 * The 'd' type is put there when the files are read.
 *
 * Entry points, ie functions called from elsewhere:
 * help() helpLink() helpNextPlace() helpPreviousPlace()
 * helpPushStack() helpPrevious()
 */

#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"

#define	HelpStackNPref	4	// See above: Contents of [Help Stack] buffer

#if	SHOWMODREL
char	helpvers[] = "@(#)help.c	1.69 10/08/20 20:05:25";
#endif

#define	HELP_KEY_INDEX	"_keywords"	// name of keywords file
#define	HELP_DEFAULT	"help-start"	// Default help file - all others linked from here
#define	HELP_CMD_INDEX	"_commands"	// Index to commands
#define	HELP_KEY_SHOW	"_key-show"	// Template file for show keywords page

// Keyword search flags:
#define	HELPKEYALL	0x01	// all keywords must match an item
#define	HELPKEYFULL	0x02	// keywords must be matched in full

/* Parameterisation for helpSubstValue()
 * Helps cut down of number of arguments for helpSubstValue()
 */
#define	HELP_SUBST_BIND	0
#define	HELP_SUBST_VAR	1
struct	sv_parm helpSvParm[] = {
	[HELP_SUBST_BIND] = { getCmdBindings, "None" , CMDMAXLEN, 1 },
	[HELP_SUBST_VAR]  = { getVarValue,    "Unset", NVSIZE+1,  0 },
};

// helpSubstValue() needs a buffer, how big should it be ?
#if	NVSIZE+1 > CMDMAXLEN
#define	MAX_SV_LEN	(NVSIZE+1)	// +1 as this includes the % or $
#else
#define	MAX_SV_LEN	CMDMAXLEN
#endif

// Recovered tnc values from [Help Stack]
BufLns	helpTopLine;	// Line number of top line in window
BufLns	helpCurLine;	// Current line number
LineLn	helpLinePos;	// Position in line

// Message produced several times:
static	char	HelpNotOnline[] = "%EHelp on '%s' is not online";
static	char	ExpectedWord[]  = "%E""%B""No link. Expected ||some-word||";

/* When reading a file we may want to ignore some (non help) lines, this
 * can be controlled by **!help-begin & **!help-end.
 * When reading a .help file this should be initialised to TRUE, but
 * when reading a .mac it should start FALSE.
 */
int	inHelp;		// TRUE if help sees lines

/* Chop a search string into substrings breaking on spaces.
 * The substrings are separate by NUL characters.
 *
 * Return the number of substrings.
 */
int	helpSSchop(TEXT* t)
{
	int	subStrings = 1;
	BYTE*	b;

	for(b = ttext(t); *b; b++) {
		if(*b == ' ') {
			*b = '\0';
			subStrings++;
		}
	}

	return(subStrings);
}


/* Determine if any of a set of strings is in another set of strings.
 *
 * srch    is what we are looking for, NUL terminated string of
 *         pattern strings (PS)
 * subj    is a candidate (subject). Set of subject strings (SS)
 * subjLen is length of subj, number of bytes
 *
 * srch	Starts with optional flags:
 *	'&'	each pattern string must match
 *	'='	pattern strings must match completely.
 *	'@'	a flag for the index generator. Emacs ignores it.
 *	'^'	need to match command name completely,
 *		this will be the first word.
 *	this will be followed by one or more pattern strings each separated by a space
 *
 * Eg will find "dog" where srch="cat dog" subj="cow pig dog fish"
 *
 * Neither srch nor subj will be modified.
 *
 * subj	may contain several subject strings each separated by a space
 *
 * No match is deemed if either srch or subj is empty (after being trimmed)
 *
 * Return a pointer to a string that matched.
 *		 This will not be a pointer in srch or subj as strsstrs()
 *		 works on copies of input.
 *	  NULL - there was not a match
 *	       - out of memory
 */
char*	strsstrs(const char* srch, const char* subj, int subjLen)
{
	BYTE*	b;
	int	all = 0;	// All patterns must match
	int	exact = 0;	// If true pattern must equal the subject, else just contain it
	int	cmd = 0;	// Command: first word of pattern must equal first word the subject
	TEXT	tmpt;		// Temporary text

	// Inspect for flags:
	while(1) {
		if(*srch == '@') {
			srch++;	// Ignore
			continue;
		}
		if(*srch == '&') {
			all = 1;
			srch++;
			continue;
		}
		if(*srch == '=') {
			exact = 1;
			srch++;
			continue;
		}
		if(*srch == '^') {
			cmd = 1;
			srch++;
			continue;
		}
		break;
	}

	/* Copy the argument strings to LINES so that we don't damage them.
	 * Break these up into NUL terminated substrings for easy matching.
	 * Count the number of components so that know how many NULs
	 * to look past.
	 * As we copy maybe convert to lower case for case independent matching.
	 */
	textToLower(strsSrch, tfakes(&tmpt, srch));
	textToLower(strsSubj, tfakem(&tmpt, subj, subjLen));

	ttrim(strsSrch);
	ttrim(strsSubj);

	if(textOOM)
		return(NULL);

	// No match if either string is empty:
	if(tused(strsSrch) == 0 || tused(strsSubj) == 0)
		return((char*)NULL);

	// Chop into NUL separated sub strings
	int	nSrch = helpSSchop(strsSrch);
	int	nSubj = helpSSchop(strsSubj);

	int	nsr;		// # srch strings looked for
	int	nsu;		// # subj strings looked for
	int	npm = 0;	// # pattern strings matched
	int	matchExact = cmd | exact;	// First word exact if command or all words exact

	// For all search strings (pattern)
	// nsr incremented in the loop body
	for(nsr = 0, b = ttext(strsSrch); nsr < nSrch; b = strchr(b, '\0') + 1) {
		BYTE*	sup;	// SUbject string Pointer

		// For all subject strings
		for(nsu = 0, sup = ttext(strsSubj); nsu < nSubj; nsu++, sup = strchr(sup, '\0') + 1) {

			// Compare using strcmp() or strstr() depending on exactness
			int	same = matchExact ? (strcmp(sup, b) == 0) : (strstr(sup, b) != (char*)NULL);

			matchExact = exact;	// First word exact might be different from subsequent

			if(same) {
				// If not all then one match is sufficient
				if( !all)
					return((char*)b);

				npm++;
				break;
			} else	// Not match
				if(cmd)			// Not find the command
					return(NULL);	// fail

			cmd = 0;		// Don't fail if not match subsequent words
		}

		nsr++;			// One more pattern tried
		cmd = 0;		// Don't fail if not match subsequent words

		/* If we get here it is either because the current pattern string
		 * did not match at all, or because it did but we need all to match.
		 *
		 * Fast return if we need to match all, return FALSE if the count
		 * of matches is less than the count of patterns tried.
		 */
		if(all && npm < nsr)
			return((char*)NULL);
	}

	/* If npm then we have found all of multiple srch strings
	 * Return the first one.
	 */
	return(npm ? (char*)ttext(strsSrch) : (char*)NULL);
}

/* Find the nth string in a set of NUL separated strings in text.
 *
 * nth	which string, counting from 0, ie the start of text is nth == 0
 *
 * Return the start of the nth string
 * 	  NULL if it could not be found
 */
char*	findNthNULstr(const TEXT* text, int nth)
{
	char*	str;	// Probe
	char*	end;	// End of line, must not reach here

	str = ttext(text);
	end = str + tused(text);

	for( ; nth > 0; nth--) {
		while(*str++ != '\0') {
			if(str >= end)
				return(NULL);
		}
	}

	return(str);
}

/* Find the [Help] buffer, return a pointer to it.
 *
 * create	Flags as per bfind()
 *
 * Set appropriate buffer flags.
 */
BUFFER*	helpGetBuffer(int create)
{
	BUFFER*	bp;

	if((bp = bfindCheck(HelpBname, create)) != (BUFFER*)NULL) {
		bp->b_flag |= BHELP | BFINVS;
		bp->b_mode |= MDVIEW;
	}

	return(bp);
}

/* Load a help file into the [Help] buffer.
 *
 * dir	The directory to find the file in.
 *	This path must not contain a PATHCHR (':')
 *	If this is NULL the standard help search path is used.
 * name	The name of the file.
 *	This name will have language, territory & .gz suffix tried
 *	to locate the file.
 *	If this starts with a '~' then the name is that of a procedure
 *	and so will be looked for using a different path and the
 *	.mac suffix. The '~' will be removed before searching.
 * 
 * Return:
 *	TRUE - OK
 *	FALSE  Help file not found
 *	       Could not find/create the help buffer
 *	       Could not split window
 *	       Could not read the file into the buffer
 */
int	helpLoadFile(const BYTE* dir, const BYTE* name)
{
	BUFFER*	bp;
	int	found;
	findTypes type = fT_Help;

	inHelp = 1;

	// Does this look like a link to a macro ?
	if(*name == '~') {
		name++;
		type = fT_Macro;
		inHelp = 0;
	}

	// Find some files, specify a directory if we have one
	if(dir == (BYTE*)NULL)
		found = findFile(name, type);
	else
		found = findFilePath(dir, name, type);

	if(found == 0) {
		mlwrite(HelpNotOnline, name);
		return(FALSE);
	}

	/* Don't use a dedicated buffer for help - the user then cannot
	 * delete it, so it chews up memory. Maybe auto clear it somehow.
	 */
	if((bp = helpGetBuffer(BFNCREATE|BFNACTIVE)) == (BUFFER*)NULL)
		goto fail;

	// If it is not displayed split the current window:
	if(bp->b_nwnd == 0 && splitwind(FALSE, 1) == FALSE)
		goto fail;

	if(readStream(HelpBname, NULL, FALSE) != TRUE)
		return(FALSE);

	setVisibleFname(bp, name);

	// Now: bp == curbp

	EnsureCurbpConnectedToCurwp();

	return(TRUE);

fail:
	// Something didn't work - ensure the stream is closed
	ffclose();
	return(FALSE);
}

/* Load all index files with name name.
 * This will be in buffer [Help index]
 *
 * The index file MUST be a set of command entry lines:
 *	c NUL command-name NUL file-name:key-word
 * Lines will be inserted like:
 *	d NUL directory name
 *
 * Search both help index and macro ones.
 * Macros do not have a command index, this silently fails.
 *
 * Return pointer to buffer if OK
 *	  NULL if error, after producing an error message.
 */
BUFFER*	helpLoadIndex(const char* name)
{
	BUFFER*	bnx;
	int	indexes;	// # Indexes: helps & macros
	int	totIndexes = 0;	// How many in all
	BYTE*	fname;		// Next file name
	int	i;
	static	findTypes	types[] = { fT_HelpIndex, fT_MacroKeys };

	// Find the index file(s):
	if((bnx = bfindCheck(HelpNxBname, BFNCREATE|BFNACTIVE)) == (BUFFER*)NULL)
		return(NULL);

	// Make invisible, read-only, immutable, should be transitory but ...
	bnx->b_flag |= BFINVS | BIMMUTABLE;
	bnx->b_mode |= MDVIEW;

	for(i = 0; i < 2; i++) {
		/* Macros do not have a command index ...
		 * so the find will fail.
		 * Could test for it, simpler not to - just waste a few file tests
		 */

		if((indexes = findFile(name, types[i])) == 0)
			continue;

		totIndexes += indexes;

		// This may have several NUL separated file names
		fname = ttext(foundFiles);

		// Read the index files:
		while(--indexes >= 0) {
			LINE*	dline;
			char*	dn;
			BYTE*	slash;
			int	dirLen;

			// Insert the directory line:
			if( !(dline = BufAppendLine(bnx)))
				goto fail;

			lcpyc(dline, 'd');
			lcatc(dline, '\0');
			if((slash = strrchr(fname, '/')) != (BYTE*)NULL) {
				dirLen = (int)(slash - fname + 1);
				dn = fname;
			} else {
				dirLen = 2;
				dn = "./";
			}
			lcatm(dline, dn, dirLen);
			lcatc(dline, '\0');	// Append a terminating NUL

			// Append to what we already have:
			if(ffp != (FILE*)NULL || openPossGz(fname) == FIOSUC)
				// Don't complain if this fails:
				readStream(HelpNxBname, NULL, TRUE);

			fname += strlen(fname) + 1;	// Next name, if any
		}
	}

	if(totIndexes > 0)
		return(bnx);

	mlwrite(HelpNotOnline, HELP_CMD_INDEX);

fail:	// Remove the buffer before failing
	zotbuf(bnx, TRUE);
	return((BUFFER*)NULL);
}

/* Give me some help!!!!
 * bring up a fake buffer and read the help file
 * into it with view mode
 *
 * Need to rework this - only want one help buffer in spite
 * of different files variously being shown.
 *
 * If given a count: prompt for keywords to search.
 * The default count is not set.
 *
 * Command:  help
 * Bound to: M-?
 *
 * Return:
 *	TRUE  success
 *	FALSE fail
 *	    Cannot open help file
 *	    Cannot create help buffer
 *	    Cannot split window
 *	    Fail to read file
 */
int	help(int f, Cnt n)
{
	register BUFFER* bp;	/* buffer pointer to help */

	if(f)
		return(helpKeyPrompt(TRUE));

	if( !findFile(HELP_DEFAULT, fT_Help)) {
		mlwrite(HelpNotOnline, HELP_DEFAULT);
		return(FALSE);
	}

	/* Don't use a dedicated buffer for help - the user then cannot
	 * delete it, so it chews up memory. Maybe auto clear it somehow.
	 */
	if((bp = helpGetBuffer(BFNCREATE|BFNACTIVE)) == (BUFFER*)NULL)
		goto fail;

	// If it is not displayed split the current window:
	if(bp->b_nwnd == 0 && splitwind(FALSE, 1) == FALSE)
		goto fail;

	if(readStream(HelpBname, HELP_DEFAULT, FALSE) != TRUE)
		return(FALSE);

	setVisibleFname(bp, HELP_DEFAULT);	// Remove associated file name

	inHelp = 1;

	helpBufferMassage();

	// Start the help stack - this is an entry point, so zap anything there
	bclear(bmlHelpStack);
	helpPushStack('h', ttext(ffName));

	// Now: bp == curbp

	Window*	wp = onscreen(curbp);

	if(wp != (Window*)NULL) {
		// It could be displayed more than once:
		ReconnectWindowsToBuffer(curbp);
		curwp->w_flag |= WFMODE;	// Ensure old window loses ornamentation
		curwp = wp;
	} else {
		DisconnectBuffer(curwp);
		ConnectBuffer(curbp, curwp, 0);
	}

	return(TRUE);

fail:
	// Something didn't work - ensure the stream is closed
	ffclose();
	return(FALSE);
}

/* Substitute a value as part of help file massaging.
 * lp	The LINE being worked on - setNUL
 * Off	Offset of '**something:' within lp
 * parm	pointer to parameter for what we are doing
 *
 * 'something' is a string (eg command) that does not contain a
 * space, tab, '*', ':', ';'. Ie any one of these ends it.
 *
 * Examples:
 *	**b:command  **
 *	**b:command:None thing;20  **
 *	**v:%name**
 *	**v:$TERM;30  **
 *
 * Return a pointer to use for searching for the next thing
 * to replace on the current line.
 *
 * If something goes wrong return a pointer to somewhere after
 * the ** - is won't go into an infinite loop if it is used
 * to start another search.
 */
BYTE*	helpSubstValue(LINE* lp, LineLn Off, struct sv_parm* parm)
{
	char	name[MAX_SV_LEN + 2];	// Name to look up
	char*	p;		// place
	char*	ep;		// end of place
	int	wantLen = 0;	// Length wanted

	p = strchr(lgetcaddr(lp, Off), ':') + 1;

	// Can this take a one-only option, is it there ?
	int	oneOnly = 0;
	if(parm->sv_onlyOne && *p == '-') {
		oneOnly = 1;
		p++;
	}

	// There must be an ending **
	if((ep = strstr(p, "**")) == (char*)NULL)
		return(p);	// Next search point

	/* Find where something ends
	 * This will be moved left as we find things that terminate it
	 */
	char*	est;		// End of something
	char*	tab;
	est = strchr(p, ' ');
	tab = strchr(p, '\t');
	if( !est && tab)
		est = tab;
	if(tab && est > tab)
		est = tab;
	if( !est)		// No space or tab, end something on **
		est = ep;

	// Good, seen ending **
	// Look for optional ":words"
	char*	def = strchr(p, ':');	// Start of default string
	if(def == (char*)NULL || def > ep)
		def = ep;
	if(est > def)		// End something on default ?
		est = def;

	char*	len;		// Length
	// Look for optional ;length
	if((len = strchr(p, ';')) != (char*)NULL && len < ep &&
	   ( Isdigit(len[1]) || (len[1] == '-' && Isdigit(len[2])))) {
		wantLen = atoi(len + 1);
		if(abs(wantLen) > 1000)
			wantLen = 0;
	}
	if( !len)
		len = est;
	if(est > len)		// End something on length ?
		est = len;

	if(est - p > parm->sv_maxlen || est - p >= (int)(sizeof(name) - 1))
		return(p);

	// Put the name into name:
	memcpy(name, p, (size_t)(est - p));
	name[est - p] = '\0';

	if(p == est) {
		// Nothing to substitute, ie here just for padding
		tsetempty(playText5);
	} else {
		// Find the value for this name:

		if((parm->sv_getval)(name, playText5) == FAILED)
			return(p);	// Unknown thing to substitute: leave text as is

		// No value found ?
		if(tused(playText5) == 0) {
			// Put custom or default message:
			if(def < len)
				tcpym(playText5, def + 1, len - def - 1);
			else
				tcpys(playText5, parm->sv_noval);
		} else if(oneOnly) {
			// Suppress more than one binding ?

			BYTE*	ms = (BYTE*)strstr(ttext(playText5), "   ");
			if(ms)
				tsetused(playText5, (LineLn)(ms - ttext(playText5)));
		}
	}

	// Specified length ?
	if(wantLen) {
		if(wantLen < 0) {
			wantLen = 0 - wantLen;
			if(tused(playText5) > wantLen)
				tsetused(playText5, wantLen);
		}
		// Pad to wanted length:
		while(tused(playText5) < wantLen)
			tcatc(playText5, ' ');
	}

	// Remove the **v:xxx** and replace with the value found:
	ldeln(lp, Off, (LineLn)((BYTE*)ep - lgetcaddr(lp, Off) + 2));
	if(linst(lp, Off, playText5) == FALSE)
		return(lgetcaddr(lp, Off));

	return(lgetcaddr(lp, Off + tused(playText5)));
}

/* Do substitutions, etc on raw .help lines, make fit for showing to the user.
 *
 * * Append an invisible NUL to every line
 * * Remove comment lines: **#
 * * Process directive lines: **!
 *	**!help-begin
 *	**!help-end
 *	All others ignore/discard
 * * Replace **b:command** with the command binding
 * * Replace **v:xxx** with variable value, xxx is like %env or $user
 * * Replace **\something with **something
 *
 * **b: and **v: may be followed by ;length (eg ';10'), which rights pads the
 * string to at least that length. Eg: **b:command;20**
 * ';-length' will pad but truncate to at most that length
 *     **b:command;-20**
 * The length may be followed by spaces which will be ignored, eg:
 *     **b:command;20  **
 *
 * The command may be omitted, this allows padding, eg:
 *     **b:;20         **
 * You cannot omit the variable with **v:
 *
 * Instances of ||\ will have the \ removed. This helps GenHelpIndices.pl
 *
 * This uses lunstitch() - so fixup needed
 *
 * Return the line to scan from - usually lp, but might
 *	be the line before if lp is removed.
 *	NULL on error.
 */
LINE*	helpLineMassage(LINE* lp)
{
	BYTE*	cp;
	char*	s1;	// Start of string of interest
	char*	s2;	// Start of string of interest
	char*	s3;	// Start of string of interest

	// Add a NUL at the end of every line, makes scanning easier:
	if( !lsetNUL(lp))
		return(NULL);

	s1 = ltext(lp);

	// Recognise **!help-begin and **!help-end
	// Remove comment line & all **!lines
	if(lused(lp) >= 3) {
		if(strncmp(s1, "**#", 3) == 0)
			return(lunstitch(lp));

		if(strncmp(s1, "**!", 3) == 0) {
			// Directives

			if(strncmp(s1 + 3, "help-begin", 10) == 0)
				inHelp = 1;
			if(strncmp(s1 + 3, "help-end", 8) == 0)
				inHelp = 0;

			// Above the only 2 directives that are recognised
			return(lunstitch(lp));
		}
	}

	// If not a help line - ignore
	if(inHelp == 0)
		return(lunstitch(lp));

	// Replace **b:xx** with the bindings of command xx
	while((s1 = strstr(s1, "**b:")) != (char*)NULL)
		s1 = helpSubstValue(lp, (LineLn)((BYTE*)s1 - ltext(lp)), helpSvParm + HELP_SUBST_BIND);

	// Replace ||\ by ||
	cp = ltext(lp);
	while((cp = strstr(cp, "||\\")) != NULL) {
		cp += 2;
		ldeln(lp, (LineLn)(cp - ltext(lp)), 1);
	}

	// Replace **v:$xx** or **v:%xx** with the value of the variable
	cp = ltext(lp);
	while(1) {
		s1 = strstr(cp, "**v:%");
		s2 = strstr(cp, "**v:$");
		s3 = strstr(cp, "**\\");

		if( !s1 && !s2 && !s3)	// Stop mangling this line if nothing to do
			break;

		/* Make s1 point to what we found, if unset use s2
		 *
		 * Might have both on one line, select the first:
		 */
		if( !s1 || (s2 && s2 < s1))
			s1 = s2;

		/* Replace **\XXX by **XXX
		 * This has to be mixed up with **v in case a variable
		 * contains a string like **\fred**
		 *
		 * **\foo**v:%fred** will (correctly) be seen as **foo followed by the value of %fred
		 */
		if(s3 && (!s1 || s3 < s1)) {
			// Remove the \ after **
			cp = s3 + 2;
			ldeln(lp, (LineLn)((BYTE*)cp - ltext(lp)), 1);
			continue;
		}

		// Substitute a variable:
		cp = helpSubstValue(lp, (LineLn)((BYTE*)s1 - ltext(lp)), helpSvParm + HELP_SUBST_VAR);
	}

	return(lp);
}

/* Process curbp so that it can be easily scanned.
 *
 * See helpLineMassage() - which does lunstitch() - thus need bLinesFix().
 *
 * inHelp must be set appropriately before this is called.
 *
 * Return TRUE if all is OK
 */
int	helpBufferMassage(void)
{
	LINE*	lp;

	for(lp = lforw(curbp->b_linep); lp != curbp->b_linep; lp = lforw(lp))
		if((lp = helpLineMassage(lp)) == NULL)
			return(FALSE);

	// Ensure that "." & topLine are valid - after lunstitch()ing
	bLinesFix(curbp);

	return(TRUE);
}

/* Find keyword in curbp.
 * These are at places: **: (place) or **^ (command)
 *
 * Start looking from:
 * + not displayed - the start of buffer
 * + if displayed in a window - from "."
 *   If not found, search again from the top
 *
 * If keyword starts with '-' look upwards,
 * if it starts '--' go to the one above the previous one.
 * This only works for helpWhatAny.
 *
 * The keyword may contain more than one word, each separated by ' '.
 * The place text may contain more than 1 word, each separated by ' '.
 *
 * If the keyword is the empty string find the next place.
 * A keyword of "-" find the previous place.
 *
 * Curbp might, or might not be currently displayed.
 *
 * Return TRUE if successful, otherwise FALSE.
 */
int	helpSearchKeys(const char* keyword)
{
	LINE*		lp;
	Window*		wp = onscreen(curbp);
	int		fromTop = 1;		// If search was from the top
	int		up = 0;			// Find previous
	int		findCmd = 0;		// TRUE if looking for a command
	const char*	kp;			// Keyword probe

	// Accept 1 or 2 '-':
	while(keyword[0] == '-' && up < 2) {
		keyword++;
		up++;
	}

	// Scan to see if looking for an Emacs command:
	kp = keyword;
	while(1) {
		switch(*kp) {
		  case '^':
			findCmd = 1;
			// DROP THROUGH
		  case '&':
		  case '=':
		  case '@':
			++kp;
			continue;
		}
		break;
	}

	// Start scanning at the top of the buffer,
	// but if displayed: the current line:
	lp = up ? lback(curbp->b_linep) : lforw(curbp->b_linep);
	if(wp) {
		lp = wp->w_dotp;
		fromTop = 0;
	}

// Come here if hit end of buffer & want another go
tryAgain:

	// For all buffer lines:
	for(; lp != curbp->b_linep; lp = up ? lback(lp) : lforw(lp)) {

		char*	sp;	// Start Search position
		char*	s1;	// Start of string of interest
		char*	s2 = NULL; // Start of string of interest
		char*	ep;	// end of place

		sp = ltext(lp);

		while((s1 = strstr(sp, "**^"))  != (char*)NULL ||
		      ( !findCmd &&
		       (s2 = strstr(sp, "**:"))   != (char*)NULL)) {

			// If no command, use place, also
			// might have both on one line, select the first:
			if(!s1 || (s2 && s2 < s1))
				s1 = s2;

			s1 += 3;	// Skip **^ or **:

			// Skip non-index directive, only on **:
			if(s2 && s1[0] == '-')
				s1++;

			s2 = NULL;		// For next loop iteration

			if((ep = strstr(s1, "**")) == (char*)NULL)
				break;		// Missing end ** on the line, next line

			sp = ep + 2;		// Where to start next search on this line

			// Any matches of what we are looking for and
			// what we just found ?
			// Accept anything with empty keyword.
			if(keyword[0] != '\0' &&
			   strsstrs(keyword, s1, (int)(ep - s1)) == (char*)NULL) {
				// Keyword(s) do not match
				continue;	// Next line
			}

			// If there is no match, leave "." at the top of file

			// match
			// Moving up 2 ?
			if(--up > 0)
				continue;	// Find next

			/* Stop at this place.
			 * Set the line & column to what we found.
			 *
			 * Try and show a bit of context above what we have moved to.
			 * 3 lines is about right.
			 *
			 * If the window is displayed let normal framing handle it,
			 * otherwise small jumps always reframe which is disturbing to view.
			 */
			if(wp){
				// Below needed as ConnectBuffer() does the reverse:
				wp->w_dotp = lp;
				wp->w_doto = (LineLn)((BYTE*)s1 - ltext(lp));
				SaveWindowPosToBuffer(curbp, wp);
			} else {
				int	i;
				curbp->b_dotp = lp;
				curbp->b_doto = (LineLn)((BYTE*)s1 - ltext(lp));
				// show some context above the place found.
				for(i = 3; --i >= 0 && lback(lp) != curbp->b_linep; )
					lp = lback(lp);
				curbp->b_topLinep = lp;
			}

			// Add this search to the stack unless doing this for next/previous
			if(keyword[0] != '\0')
				helpPushStack('h', ttext(ffName));

			EnsureCurbpConnectedToCurwp();

			return(TRUE);
		}
	}

	// Nothing found.
	// If did not start at the top, try from the top.
	if( !fromTop) {
		fromTop = 1;
		lp = up ? lback(curbp->b_linep) : lforw(curbp->b_linep);
		goto tryAgain;
	}

	return(FALSE);
}

/* Read a help file and find a keyword in it.
 * dir	is the directory that the file will be in
 * link	is a file name
 *	It must be possible to change link without ill effect.
 * keyword is what to search for. If this is NULL or empty leave
 *	"." on the first line of the file.
 *
 * Return TRUE if successful, otherwise FALSE.
 */
int	helpKeyFind(const BYTE* dir, const BYTE* link, const BYTE* keyword)
{
	if(helpLoadFile(dir, link) == 0)
		return(FALSE);

	inHelp = 1;

	helpBufferMassage();

	EnsureCurbpConnectedToCurwp();

	// If no search keyword - we are done
	if( !keyword || keyword[0] == '\0') {
		helpPushStack('h', ttext(ffName));
		return(TRUE);
	}

	return(helpSearchKeys(keyword));
}

/* Display help for the Emacs command cmd
 *
 * The will also try to find documentation on a procedure - which
 * is a word that is not a command.
 *
 * Return TRUE or FALSE.
 */
int	helpCmdShow(const BYTE* cmd)
{
	BUFFER*	indexBp;
	int	ret = FALSE;	// What to return

	if(fncmatch(cmd) == NULL) {
		// Not a command - probably a procedure name
		tcpyc(playText3, '~');
		tcats(playText3, cmd);
		
		if((ret = helpLoadFile(NULL, ttext(playText3))) == TRUE)
			helpBufferMassage();

		EnsureCurbpConnectedToCurwp();

		helpPushStack('h', ttext(ffName));

		return(ret);
	}

	// Load the index file(s):
	if( !(indexBp = helpLoadIndex(HELP_CMD_INDEX)))
		return(FALSE);

	// From now on we MUST kill the buffer before we return - including failure
	// It will/might be curbp

	// Look for the command entry
	LINE*	ln;
	LINE*	dir = NULL;	// Directory line, NULL to keep compiler happy

	for(ln = lforw(indexBp->b_linep); ; ln = lforw(ln)) {
		if(ln == indexBp->b_linep) {	// Hit end of index ?
			mlwrite(HelpNotOnline, cmd);
			goto fail;
		}

		if(lused(ln) <= 0)
			continue;

		/* Lines of interest look like:
		 * d NUL directory NUL. We inserted these in helpLoadIndex()
		 * c NUL command NUL file:key-word NUL
		 */
		if(lgetc(ln, 0) == 'd') {
			// Directory name
			dir = ln;
			continue;
		}
		if(lgetc(ln, 0) != 'c')
			continue;

		if(strcmp(cmd, lgetcaddr(ln, 2)) != 0)
			continue;

		/* Found it!
		 * Sanity checks that format is as above.
		 * We are going to be doing string things - much easier to be
		 * ended by '\0' rather than always checking the length.
		 */

		if(lused(ln) < 6 || lgetc(ln, lused(ln) - 1) != '\0')	// Sanity checks
			continue;

		/* Get the key word
		 * Sanity check: If this is beyond the line end, just ignore it
		 */
		BYTE*	link = strchr(lgetcaddr(ln, 2), '\0') + 1;
		if(link >= lgetcaddr(ln, lused(ln)) ||
		   *link == '\0')
			continue;

		tcpys(playText3, link);		// Copy to somewhere that is can be changed

		char	CmdSrch[CMDMAXLEN + 2];	// ^command-name -- search string
		CmdSrch[0] = '^';
		strcpy(CmdSrch + 1, cmd);

		ret = helpKeyFind(lgetcaddr(dir, 2), ttext(playText3), CmdSrch);

		break;
	}

fail:
	zotbuf(indexBp, TRUE);		// Done with it

	return(ret);
}


/* Prompt for search keys & show a list of matches.
 *
 * newHelp	TRUE if new entry to help system, so clear help stack
 *
 * Return TRUE on OK.
 *	  FALSE nothing available at prompt
 *	  ABORT C-g
 */
int	helpKeyPrompt(int newHelp)
{
	int status;		/* status return */

	// Want a history buffer at some point
	status = mlprompt("Search help keywords. Search string (* for all): ",
			MLSTRING | MLLN0END, bmlHelpSearch);

	if(status != TRUE)
		return(status);

	// If this is an entry point start the help stack, ie zap anything there
	if(newHelp)
		bclear(bmlHelpStack);

	return(helpSearch(ltext(bmlHelpSearch->b_dotp)));
}

/* Search template buffer for all lines of ltype & append to help buffer
 *
 * Found lines are massaged - ie variables, etc, substituted.
 *
 * hbp		The [Help] buffer
 * tmpl		The template buffer
 * ltype	The line type to find.
 */
void	helpAppendTmpl(BUFFER* hbp, const BUFFER* tmpl, char ltype)
{
	LINE*	line;

	for(line = lforw(tmpl->b_linep) ; line != tmpl->b_linep; line = lforw(line)) {
		if(lused(line) < 2 || lgetc(line, 1) != ':')
			continue;

		// Not what we are asked to generate ?
		if(lgetc(line, 0) != ltype)
			continue;

		LINE*   l = BufAppendLine(hbp);
		lcpyl(l, line);
		ldeln(l, 0, 2);		// Remove t: (ype)
		helpLineMassage(l);
	}
}


/* Search all _keywords files.
 *
 * keyw	start of keyword string
 *	A string '*' means find everything as does the empty string.
 *
 * Keyword string may start with '&' to make matching AND all keywords,
 * otherwise matching is OR ie any keyword. '=' means match words fully.
 *
 * Return TRUE or FALSE.
 */
int	helpSearch(const BYTE* keyw)
{
	BUFFER*	indexBp = NULL;		// For the index
	BUFFER*	tmplb = NULL;		// For the template
	BUFFER*	hbp = NULL;		// The Help buffer
	int	ret = FALSE;

	// Showing a help file, not a macro:
	inHelp = TRUE;

	if(setUserVar("%_keys", keyw) != TRUE)
		return(FALSE);

	// Get a value that will not be corrupted/overwritten - as %_keys is type STR
	keyw = gtusr("_keys");

	if(keyw[0] == '*' && keyw[1] == '\0')
		keyw = "";

	if( !(indexBp = helpLoadIndex(HELP_KEY_INDEX)))
		return(FALSE);

	// Look for the command entry
	LINE*	ln;
	LINE*	ldir = NULL;		// Last directory line seen
	LINE*	lfile = NULL;		// Last file line seen
	LINE*	ltitle = NULL;		// Last title line seen
	int	ndir = 0;		// Matches this directory
	int	nfile = 0;		// Matches this file
	int	SeenSomething;		// Seen at least one matching keyword
	int	Matches = 0;		// How many matches
	int	MultiDirs = 0;		// True if more than one directory matches

	SeenSomething = keyw[0] == '\0';// If no selection, assume that there is something

	// Only weed out what is not wanted if there are keywords
	for(ln = lforw(indexBp->b_linep); ln != indexBp->b_linep; ln = lforw(ln)) {

		if(lused(ln) < 2 ||		// At least type & NUL
		   !strchr("dftkc", lgetc(ln, 0))) { // Bad or unwanted type, eg comment
			ln = lunstitch(ln);	// Remove line - which is 
			continue;
		}

		if(keyw[0] == '\0')		// Want everything - done enough
			continue;


		/* Lines of interest look like:
		 * d NUL directory NUL. We inserted these in helpLoadIndex()
		 * f NUL file NUL
		 * k NUL keywords NUL
		 * c NUL command NUL
		 */
		if(lgetc(ln, 0) == 'd') {
			/* Free up d and f lines if nothing since last directory.
			 * Keep a count of # directories with something that matches
			 * (we want to know if it is one or many).
			 */
			if(ldir) {
				if(ndir == 0)
					lunstitch(ldir);
				else
					MultiDirs++;
			}

			if(lfile && ndir == 0) {
				lunstitch(lfile);
				lfile = NULL;
			}

			ldir = ln;
			ndir = 0;
			nfile = 0;

			continue;
		}
		if(lgetc(ln, 0) == 'f') {
			// Chuck last file if it did not contain a match
			if(nfile == 0) {
				if(lfile)
					lunstitch(lfile);
				if(ltitle)
					lunstitch(ltitle);
			}

			lfile = ln;
			ltitle = NULL;
			nfile = 0;

			continue;
		}

		// title/description lines if there are matches, so removed elsewhere
		if(lgetc(ln, 0) == 't') {
			if(ltitle)
				lunstitch(ltitle);
			ltitle = ln;
			continue;
		}

		// Last remaining line types are 'k' or 'c'
		// Does it match the keyword criteria ?
		if((lgetc(ln, 0) == 'k' || lgetc(ln, 0) == 'c') &&
		   strsstrs(keyw, lgetcaddr(ln, 2), (int)lused(ln) - 3)) {
			SeenSomething = 1;
			ndir++;
			nfile++;

			/* Record the first match - might be the only one
			 * We are reading a file, so it is (unlikely but)
			 * possible that 'f' has not been seen or that it is
			 * not NUL terminated.
			 * playText4 & playText3 are what are used to hold dir & file elsewhere.
			 */
			if(++Matches == 1) {
				if(lfile == NULL) {
					++Matches;	// Prevent if() below
				} else {
					tcpys(playText4, ltext(ldir) + 2);
					tcpym(playText3, ltext(lfile) + 2, lused(lfile) - 2);	// Copy NUL
				}
			}

			continue;	// Got a match
		}

		// It does not match what we are looking for, lose it:
		ln = lunstitch(ln);
	}

	// Tidy up: remove directories, files & titles that do not have any key matches
	if(nfile == 0) {
		if(lfile)
			lunstitch(lfile);
		if(ltitle)
			lunstitch(ltitle);
	}
	if(ldir && ndir == 0)
		lunstitch(ldir);

	/* All of the above lunstitch()ing might have left "." & topLine
	 * pointing to something that is no longer there.
	 * Fix as in helpBufferMassage()
	 */
	bLinesFix(indexBp);

	/* Exactly one match.
	 * Short circuit the user selecting it - just display the help page.
	 * Do not push the keys to the help stack as returning would bring the
	 * user straight back here.
	 */
	if(Matches == 1) {
		zotbuf(indexBp, TRUE);		// Done with it

		return(helpKeyFind(ttext(playText4), ttext(playText3), keyw));
	}

	/* Get the help buffer.
	 * Only do this now since if we get exactly one match we don't need it.
	 *
	 * Set the buffer name to the search keywords that we are being asked to find.
	 * The name is also used below to push to help history.
	 */
	if((hbp = helpGetBuffer(BFNCREATE|BFNACTIVE)) == (BUFFER*)NULL ||
	   bclear(hbp) == FALSE)
		goto fail;

	setVisibleFname(hbp, gtusr("_keys"));

	// If it is not displayed split the current window:
	if(hbp->b_nwnd == 0 && splitwind(FALSE, 1) == FALSE)
		goto fail;

	/* Read the template file that generates the keyword index
	 */
	if((tmplb = bfindCheck(HelpTmplBname, BFNCREATE|BFNACTIVE)) == (BUFFER*)NULL)
		goto fail;

	// Find the template file
	if(findFile(HELP_KEY_SHOW, fT_HelpTmpl) == 0) {
		// Something default to show the user:
		addline("H:Search results for keys **v:%_keys**", hbp, 1);
		addline("H:Choose the keyword link below:", hbp, 1);
		addline("k: ||**v:%_F**:&=**v:%_k**||", hbp, 1);
		addline("c: ||^ **v:%_c** ||", hbp, 1);
		addline("n:Nothing matched your keywords", hbp, 1);
	} else {
		if(readStream(HelpTmplBname, NULL, FALSE) != TRUE)
			goto fail;
	}

	// Copy header lines to the Help buffer:
	helpAppendTmpl(hbp, tmplb, 'H');

	if( !SeenSomething) {
		// No matches, put the nothing matched line:
		helpAppendTmpl(hbp, tmplb, 'n');
	} else {
		/* Work through the lines that search found.
		 * Generate, from the templates, lines of corresponding types.
		 * Set user variables with the information found in these lines.
		 */
		BYTE	vname[20];
		int	dircnt = 0;

		/* %_F is %_dn/%_f but with the %_dn/ suppressed if we only ever
		 * find a match in one directory. The point is that we were getting
		 * search results like: ||%_d0/buffers:&= internal buffers ||
		 * this is ugly and the %_d0/ is just confusing noise, if we
		 * don't need it suppress it.
		 *
		 * _Fprefix is %_D/ if we need it.
		 */
		BYTE*	_Fprefix;
		_Fprefix = MultiDirs ? "%_D/" : "";

		// Scan all lines resulting from the keyword search:
		for(ln = lforw(indexBp->b_linep); ln != indexBp->b_linep; ln = lforw(ln)) {
			BYTE	type = lgetc(ln, 0);

			// New directory or file ?
			if(type == 'd' || type == 'f') {
				// Ensure that values don't hang around when no longer valid
				if(setUserVar("%_t", "") != TRUE)
					goto fail;
				if(setUserVar("%_f", "") != TRUE)
					goto fail;
			}

			// Set var that has the name of the line type, value is next field. Eg: %_f
			vname[0] = '%';
			vname[1] = '_';
			vname[2] = type;
			vname[3] = '\0';
			if(setUserVar(vname, lgetcaddr(ln, 2)) != TRUE)
				goto fail;

			if(type == 'd') {
				// Directory, create a variable like %_d1 and %_D has value %_d1
				sprintf(vname, "%%_d%d", dircnt++);
				if(setUserVar(vname, lgetcaddr(ln, 2)) != TRUE)
					goto fail;
				if(setUserVar("%_D", vname) != TRUE)
					goto fail;
				if( !MultiDirs)	// Different template key if only one directory
					type = 'D';
			}

			if(type == 'f') {
				// Set %_F -- see above for comments
				tcpys(playText, _Fprefix);
				tcats(playText, lgetcaddr(ln, 2));
				if(setUserVar("%_F", ttext(playText)) != TRUE)
					goto fail;
			}

			// Add lines of the type just seen, massaging as we go
			helpAppendTmpl(hbp, tmplb, (char)type);
		}
	}

	// Copy the footer
	helpAppendTmpl(hbp, tmplb, 'F');

	Window*	wp = onscreen(hbp);

	if(wp != (Window*)NULL) {
		// It could be displayed more than once:
		ReconnectWindowsToBuffer(hbp);
		curwp->w_flag |= WFMODE;	// Ensure old window loses ornamentation
		curwp = wp;
	} else {
		DisconnectBuffer(curwp);
		ConnectBuffer(hbp, curwp, 0);
	}

	curbp = hbp;
	curwp->w_dotp = lforw(hbp->b_linep);
	curwp->w_doto = 0;

	// Add to help history the keys used to search:
	helpPushStack('k', hbp->b_fnameClean);

	ret = TRUE;

fail:

	if(indexBp)
		zotbuf(indexBp, TRUE);		// Done with it
	if(tmplb)
		zotbuf(tmplb, TRUE);

	return(ret);
}

/* Show help on an Emacs command within the apropos display.
 * \r has been entered, if "." is in the first word on the line
 * (matching [-a-z0-9]*) take it as an Emacs command name. Look
 * this up and display help about it.
 *
 * Return TRUE or FALSE.
 */
int	helpAprbuf(void)
{
	LineLn	lnp;		// Line position

	// Find the command name on the current line
	for(lnp = 0; ; lnp++) {
		if(lnp > lused(curwp->w_dotp))	// Should not happen
			return(FALSE);

		BYTE	ch = lgetc(curwp->w_dotp, lnp);

		if(Islower(ch) || Isdigit(ch) || ch == '-')
			continue;

		if(lnp == 0 || ch != ' ') {
			mlwrite("%E""%B""Line does not start with an Emacs command");
			return(FALSE);
		}
		if(curwp->w_doto > lnp) {
			mlwrite("%E""%B""Cursor not in first word in the line");
			return(FALSE);
		}

		break;
	}

	// Get it into a buffer:
	BYTE	cmd[CMDMAXLEN + 2];
	strncpy(cmd, ltext(curwp->w_dotp), CMDMAXLEN);
	*strchr(cmd, ' ') = '\0';

	return(helpCmdShow(cmd));
}

/* Do the work for a command like ||>close||
 *
 * p	points to the start of the command, ie after the '>'
 * ep	points to the ending ||
 *
 * The text at p will be considered read only.
 *
 * Return: TRUE or FALSE
 */
int	helpLinkCommand(const BYTE* p, const BYTE* ep)
{
	char*	ss;	// Search String

	if(strncmp(p, "close|", 6) == 0)
		return(delwind(FALSE, 0));

	if(strncmp(p, "back|", 5) == 0)
		return(helpPrevious(FALSE, 0));

	if(strncmp(p, "search:", 7) == 0 || strncmp(p, "s:", 2) == 0) {
		p = strchr(p, ':') + 1;	// span search: or s:

		tcpym(playText3, p, (LineLn)(ep - p));
		ss = ttext(playText3);

		// If the search is empty prompt for keywords
		if(ss[0] == '\0')
			return(helpKeyPrompt(FALSE));

		return(helpSearch(ss));
	}

	return(FALSE);
}

/* Follow a help Link.
 *
 * This does not clear the help stack
 *
 * The count is ignored. The default count is not set
 *
 * Command:  help-link-follow
 * Bound to: This is not in the standard bind table
 *
 * Return TRUE on success
 *	  FALSE if not - which could be that "." is not over a link
 */
int	helpLink(int f, Cnt n)
{
	BYTE*	p;		// The ||file:keyword|| that is at "."
	BYTE*	ep;		// EndP - the || at the end
	char*	keyword;

	if( !helpStorePosition())
		return(FALSE);

	inHelp = 1;		// Default type that we expect to end up in

	// The buffer generated by aprpos/describe-bindings
	// is treated specially:
	if(strcmp(curbp->b_fnameClean, AprFname) == 0)
		return(helpAprbuf());

	// Put a NUL at the line end, so that strstr(), etc, can be used below:
	if( !lsetNUL(curwp->w_dotp))
		return(FALSE);

	p = lgetcaddr(curwp->w_dotp, curwp->w_doto);

	// Backspace to the start of the link, ie find the "||"
	while(*p != '|') {
		if(p > lgetcaddr(curwp->w_dotp, 1)) {
			p--;
			continue;
		}

		mlwrite(ExpectedWord);
		return(FALSE);
	}

	// Backup if we found the 2nd '|'
	if(p > lgetcaddr(curwp->w_dotp, 0) && p[-1] == '|')
		p--;

	// Check that we are on '||':
	if(p[0] != '|' || p[1] != '|') {
		mlwrite(ExpectedWord);
		return(FALSE);
	}

	/* A sequence like ||file1||||file2|| may cause confusion
	 * if "." is not on or after the last of the 4 '|'.
	 */

	p += 2;		// Skip the starting "||"

	// Check that the "||" is not followed by space or tab:
	if(*p == ' ' || *p == '\t') {
		mlwrite(ExpectedWord);
		return(FALSE);
	}

	// Check that there is enough following, then look for the ending ||
	if(p + 1 >= lgetcaddr(curwp->w_dotp, lused(curwp->w_dotp)) ||
	   (ep = strstr(p, "||")) == (BYTE*)NULL) {
		mlwrite(ExpectedWord);
		return(FALSE);
	}

	// Trim trailing spaces from the link
	while(ep[-1] == ' ')
		ep--;

	// Deal with command directives starting >
	if(*p == '>')
		return(helpLinkCommand(p + 1, ep));

	// ec: Find/display documentation on Emacs command
	if(*p == '^') {
		BYTE	cmd[CMDMAXLEN + 1];
		BYTE*	endc;		// End of command

		// Skip '^' & trim leading spaces:
		while(*++p == ' ')
			;

		/* Find end of command name
		 * There might be keywords after the command, ignore them
		 */
		for(endc = p; endc < ep && *endc != '\0' && *endc != ' '; endc++)
			;

		// Sanity check
		if(p == endc || endc - p > CMDMAXLEN) {
			// Will only happen if help file is bad
			mlwrite("%E""%B""Bad Emacs command");
			return(FALSE);
		}

		memcpy(cmd, p, (size_t)(endc - p));
		cmd[endc - p] = '\0';

		return(helpCmdShow(cmd));
	}


	/* file:key
	 * file	= name - open that file, search for it
	 *	= ./name - look in current directory
	 *	= '' - look in current file
	 * key	= '' place at top of file
	 *	= word - search down from current location (top in new file)
	 *
	 * Put what we have in playText3, don't damage the current line
	 */

	tcpym(playText3, p, (LineLn)(ep - (BYTE*)p));
	p = ttext(playText3);

	/* Separate filename & keyword
	 * If the keyword is missing or empty, set it to NULL
	 */

	if((keyword = strchr(p, ':')) != (char*)NULL) {
		*keyword = '\0';
		if(p == (BYTE*)keyword) {
			p = NULL;		// Empty file name
		} else {
			if(*++keyword == '|')	// It is empty
				keyword = (char*)NULL;
		}
	}

	// p and/or keyword might be NULL
	if(p) {
		BYTE*	dir = NULL;	// NULL dir means use help search path

		if((dir = strrchr(p, '/')) != (BYTE*)NULL) {
			// Restrict search to directory that the current file is in

			char*	play;
			p = dir + 1;

			tcpyt(playText4, foundFiles);
			dir = ttext(playText4);

			if((play = strrchr(dir, '/')) != (char*)NULL)
				*play = '\0';
		}

		// We have a filename, go read it & maybe search:
		return(helpKeyFind(dir, p, keyword));
	}

	if(keyword) {
		// Find the keyword in the current buffer
		// Keyword currently points to ':'
		// keyword is in p which is writable

		return(helpSearchKeys(keyword + 1));
	}

	mlwrite(ExpectedWord);

	return(FALSE);
}

/* Find the next place, any place will do.
 *
 * Start looking on the next line (down the file).
 *
 * This assumed that the current window is on the [Help] buffer
 * and returns FALSE if this is not so.
 *
 * Because searching starts on the next/previous line, only one place will be
 * seen on the current line. Why: because trying to do so would make
 * helpSearchKeys() needlessly complicated. I doubt that most users will notice.
 *
 * Movement as a result of this is not recorded in the help stack.
 *
 * If the count is negative: next is up the file, start looking on the previous
 * line.
 * The default count is not set.
 *
 * Command:  help-next-place
 * Bound to: This is not in the standard bind table
 *
 * Returns result of searching.
 */
int	helpNextPlace(int f, Cnt n)
{
	BUFFER*	hbp;		// Help buffer pointer
	char*	keyString;	// Key string to use
	LINE*	savelp;		// Save current place in case of fail
	char*	what;		// What we are doing

	// Sanity check: unlikely
	if((hbp = helpGetBuffer(0x00)) == (BUFFER*)NULL ||
	    curbp != hbp ||
	    onscreen(hbp) != curwp)
		return(FALSE);

	savelp = curwp->w_dotp;

	if(!f || n >= 0) {
		keyString = "";
		curwp->w_dotp = lforw(curwp->w_dotp);
		what = "next";
	} else {
		keyString = "-";
		curwp->w_dotp = lback(curwp->w_dotp);
		what = "previous";
	}

	if(helpSearchKeys(keyString) == TRUE)
		return(TRUE);

	curwp->w_dotp = savelp;	// So that "." is not seen to move

	mlwrite("%E""%B""No %s place to go to", what);

	return(FALSE);
}

/* Find the previous place, any place will do.
 *
 * Start looking on the previous line.
 *
 * If there is a count: +ve search up, -ve search down
 * The default count is not set.
 *
 * Command:  help-previous-place
 * Bound to: This is not in the standard bind table
 *
 * Return what helpNextPlace() returns.
 */
int	helpPreviousPlace(int f, Cnt n)
{
	if(!f)
		n = 1;

	return(helpNextPlace(TRUE, -n));
}

/* Push a new line onto the end of the [Help Stack]
 *
 * The first argument is the type, thereafter the arguments depend on the type.
 *
 * Type is int, not char - freeBSD compiler complains that a promoted argument
 * is undefined.
 */
void	helpPushStack(int Type, ...)
{
	va_list	ap;
	TEXT*	text;

	if((text = &ltxt(BufAppendLine(bmlHelpStack))) == (TEXT*)NULL)
		return;

	va_start(ap, Type);

	// Start the text with T t n c:
	printToText(text, "%c%c1%c1%c1%c", Type, '\0', '\0', '\0', '\0');

	switch(Type) {
	  case 'a':	// Apropos
	  case 'h':	// Help file
	  case 'k':	// Keyword search - keyword(s)
		printToText(text, "%A""%s%c", va_arg(ap, char*), '\0');
		break;
	}

	va_end(ap);
}

/* Store the t n c values of the current window into the top [Help Stack] entry.
 *
 * These are the values that will allow the screen/cursor to be correctly
 * positioned on return fia helpPrevious()
 *
 * Return TRUE if OK
 */
int	helpStorePosition(void)
{
	char*	tp;	// Pointer to t value
	char*	cep;	// Pointer to end of c value, the NUL byte
	TEXT*	hsLast;	// Text of last line in [Help Stack]
	BUFFER*	hbp;	// Help buffer pointer

	if((hbp = helpGetBuffer(0x00)) == (BUFFER*)NULL)
		return(FALSE);	// Should not happen

	if(lback(bmlHelpStack->b_linep) == bmlHelpStack->b_linep)
		return(FALSE);	// Should not happen

	hsLast = &ltxt(lback(bmlHelpStack->b_linep));

	if( !(cep = findNthNULstr(hsLast, HelpStackNPref)))
		return(FALSE);	// If tested that cep exists, no need to do test on tp

	tp = findNthNULstr(hsLast, 1);

	printToText(playText, "%D""%c%D""%c%D", (long)getlinenum(hbp, curwp->w_topLinep), '\0',
		(long)getlinenum(hbp, curwp->w_dotp), '\0', (long)curwp->w_doto);

	tdeln(hsLast, (LineLn)((BYTE*)tp - ttext(hsLast)), (LineLn)(cep - 1 - tp));
	return(tinst(hsLast, (LineLn)((BYTE*)tp - ttext(hsLast)), playText));
}

/* Parse the t n c values in the top [Help Stack] entry.
 * Store in global variables: helpTopLine, helpCurLine, helpLinePos
 * so that they can be used later - after the stack entry has
 * been destroyed/rewritten.
 *
 * If there are errors: fail silently.
 */
void	helpParse_tnc(void)
{
	TEXT*	hsLast;	// Last line in [Help Stack]

	if(lback(bmlHelpStack->b_linep) == bmlHelpStack->b_linep)
		return;	// Should not happen

	hsLast = &ltxt(lback(bmlHelpStack->b_linep));

	if( !findNthNULstr(hsLast, HelpStackNPref))
		return;	// If tested that cep exists, no need to do test on tp

	helpTopLine = atol(findNthNULstr(hsLast, 1));
	helpCurLine = atol(findNthNULstr(hsLast, 2));
	helpLinePos = atol(findNthNULstr(hsLast, 3));
}

/* Restore a previous window framing to what it used to be.
 * Positions stored in global variables: helpTopLine, helpCurLine, helpLinePos.
 *
 * Works on the help buffer - so needs synching to the window later.
 */
void	helpRestore_tnc(void)
{
	LineLn	lnum;
	BUFFER*	hbp;	// Help buffer pointer
	LINE*	lp;

	if((hbp = helpGetBuffer(0x00)) == (BUFFER*)NULL)
		return;	// Should not happen

	// Find the lines by number:
	for(lnum = 1, lp = lforw(hbp->b_linep); lp != hbp->b_linep; lnum++, lp = lforw(lp)) {
		if(helpTopLine == lnum) {
			hbp->b_topLinep = lp;
			hbp->b_toplinenum = lnum;
		}
		if(helpCurLine == lnum) {
			hbp->b_dotp = lp;
			hbp->b_doto = helpLinePos;

			// Sanity check - highly unlikely
			if(hbp->b_doto > lused(hbp->b_dotp) ||
			   (lisUtf8(hbp->b_dotp) && BYTE_IS_MB_CONT(lgetc(hbp->b_dotp, hbp->b_doto))))
				hbp->b_doto = 0;
		}
	}
}

/* Go back to the previous help page viewed.
 *
 * If something goes wrong, just return FALSE.
 * Things are unlikely to go wrong but could do because:
 * * User has been editing [Help Stack] - they deserve what they get!
 * * A previously read help file has disappeared or become unreadable.
 * * Out of memory
 *
 * The count is ignored. The default count is not set.
 *
 * Command:  help-previous-page
 * Bound to: This is not in the standard bind table
 *
 * Return TRUE if this works.
 */
int	helpPrevious(int f, Cnt n)
{
	LINE*	line;
	LINE*	linep;
	char*	str;
	char*	name;
	BUFFER*	bp;

	line = lback(bmlHelpStack->b_linep);	// Last line

	if(line == bmlHelpStack->b_linep)	// Is the buffer empty ?
		return(FALSE);			// Should not happen

	linep = lback(line);			// Previous in help stack - what we return to

	if(linep == bmlHelpStack->b_linep) {
		mlwrite("%E""%B""No previous help page");
		return(FALSE);
	}

	/* Pop current context
	 * Keep a temp copy as may need it
	 */
	tcpyl(playText4, line);
	lfree(bmlHelpStack, line);

	if(lused(linep) < 1)
		return(FALSE);			// Should not happen

	helpParse_tnc();			// Before we overwrite the values

	switch(lgetc(linep, 0)) {
	  case 'a':	// Apropos
			// Pop the stack and rerun apropos

		tcpyl(playText4, linep);

		lfree(bmlHelpStack, linep);	// Pop apropos as about to recreate
		if((str = findNthNULstr(playText4, HelpStackNPref)) == (char*)NULL)
			return(FALSE);

		if( !buildlist(str))
			return(FALSE);

		helpRestore_tnc();

		ConnectBuffer(curbp, curwp, 1);

		return(TRUE);
	  case 'h':	// Help file
			// Reread the file

		// Get saved file name:
		if((name = findNthNULstr(&ltxt(linep), 4)) == (char*)NULL)
			return(FALSE);

		if((bp = bfind(HelpBname, BFNNONE, 0)) == (BUFFER*)NULL)
			return(FALSE);

		/* If the current file being displayed is the same as
		 * the one returning to, don't reread the file.
		 */
		if(tgetch(playText4, 0) != 'h' || strcmp(findNthNULstr(playText4, HelpStackNPref), name)) {

			if(openPossGz(name) != FIOSUC)
				return(FALSE);

			/* Get 'file name' - shows what part of help.
			 * Last component with suffix removed.
			 * Put in playText4 - don't harm what is in bmlHelpStack
			 */
			if((str = strrchr(name, '/')) != (char*)NULL)
				name = str + 1;
			tcpys(playText4, name);
			name = (char*)ttext(playText4);
			inHelp = 1;
			if((str = strchr(name, '.')) != (char*)NULL) {
				*str++ = '\0';

				/* Now that we have the suffix see if it is .mac (ie not .help)
				 * If so then we start off not being inHelp
				 * BN: .mac might be followed by .gz
				 */
				if(strncmp(str, "mac", 3) == 0)
					inHelp = 0;
			}

			if(readStream(HelpBname, name, FALSE) != TRUE)
				return(FALSE);

			setVisibleFname(bp, name);	// Remove associated file name

			helpBufferMassage();
		}

		helpRestore_tnc();

		ConnectBuffer(curbp, curwp, 1);

		return(TRUE);
	  case 'k':	// Key search
		tcpyl(playText4, linep);

		lfree(bmlHelpStack, linep);	// Pop key search as about to recreate
		if((str = findNthNULstr(playText4, 4)) == (char*)NULL)
			return(FALSE);

		if( !helpSearch(str))
			return(FALSE);

		helpRestore_tnc();

		ConnectBuffer(curbp, curwp, 1);

		return(TRUE);
	}

//	mlwrite("%E""%B""No previous - not implemented");
	return(FALSE);
}
