/*
 * ADDW SCCS: @(#)input.c	1.136 10/14/20 21:00:15
 */

/*	INPUT:	Various input routines for MicroEMACS 3.7
 *		written by Daniel Lawrence 5/9/86, heavily modified by
 *		Tony Bamford ever since.
 */
#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"

#if	SHOWMODREL
char	inputvers[] = "@(#)input.c	1.136 10/14/20 21:00:15";
#endif

/* BEWARE
 * TODO utf8 - mainly - needs checking.
 */


/* Ask a yes or no question in the message line.
 * Used any time a confirmation is required.
 *
 * The prompt (ie fmt) should contain something like "[y/n]".
 *
 * Give up after MaxQuestion times, unless this is 0.
 * BUG: if this eventually wraps round.
 *
 * Return shows the user:
 *	TRUE  - said yes
 *	FALSE - said no
 *	ABORT - bumped out of the question with a ^G
 *		or EOF on input or out of MaxQuestion.
 */
int	mlyesno(const char* fmt, ...)
{
	KeyCode	c;		/* input character */
	va_list	ap;
	int	asks = MaxQuestion;

	for(;;) {
		/* Build and prompt the user */
		va_start(ap, fmt);
		c = mlvGetKey(MGC_MAP, (BYTE*)NULL, fmt, ap);
		va_end(ap);

		if(termEOF || (c == (char)ectoc(abortc)) || --asks == 0)
			return(ABORT);	/* Bail out! */

		if(c=='y' || c=='Y')
			return(TRUE);

		if(c=='n' || c=='N')
			return(FALSE);
	}
}

/* Prompt and get a string answer either from the message line or macro command line.
 *
 * This is the new mlreply(); the real difference is that the reply is in a
 * history BUFFER's LINE rather than a fixed sized buffer.
 *
 * All works as it should for interactive prompts, but there is a 500 byte
 * buffer when reading macros & strings from there.
 *
 * This is the first version - see how it fares.
 *
 * prompt	prompt shown to the user
 * option	bit field which modifies our behaviour slightly, MLxxx
 *		It is good to give a type value to limit the length of what is obtained.
 *		MLINTERACTIVE if set prompt the user even if running a macro
 * history	the buffer that contains history of this prompt type
 *		The message will end up in here - no way of stopping this.
 *
 * The result will be in history->b_dotp - which will be NUL terminated (L_NUL)
 * If the MLNOECHO flag is set it will be in playCrypt
 *
 * Return TRUE   - got something: left in history->b_dotp
 *	  FALSE  - nothing available - ie zero length string
 *	  ABORT  - C-g or C-c entered
 *	  FAILED - out of memory
 */
int	mlprompt(const BYTE* prompt, int option, BUFFER* history)
{
	// If non interactive evaluate expression, variable, ...
	if(clexec == TRUE && (option & MLINTERACTIVE) == 0) {

		VAL*	valbuf = valGet(V_STR);

		int	ret = nextarg(prompt, &valbuf, option, history);

		if(ret != TRUE) {
			valRelease(valbuf);
			return(ret);
		}

		// Add as a new line to the history
		// Beware type in valbuf might not be string
		if( !addHistoryText(history,
				vTEXT(vtype(valbuf) == V_STR ? valbuf : valPrint(valbuf)), AHL_ADD_ZLEN)) {
			TTbeep();
			valRelease(valbuf);
			return(FAILED);
		}

		// In case the history buffer is in the current window:
		history->b_dotp = lback(history->b_linep);
		history->b_doto = lused(history->b_dotp);

		valRelease(valbuf);
		return(TRUE);
	}

	return(getstring(prompt, (VAL**)NULL, option, history));
}

/* Get a value for a command, either by prompt & user input or macro command line.
 *
 * Write a prompt into the message line, then read back a response. Keep
 * track of the physical position of the cursor. If we are in a keyboard
 * macro throw the prompt away, and return the remembered response. This
 * lets macros run at full speed. The reply is always terminated by a carriage
 * return. Handle erase, kill, and abort keys.
 *
 * prompt	prompt associated with what we're typing
 * buf		where it all goes at the end of the day
 * option	bit field which modifies our behaviour slightly, MLxxx
 * history	the buffer that contains history of this prompt type, NULL if none
 *
 * Return TRUE if got something
 *	  FALSE - nothing available
 *	  ABORT - C-g
 */
int	mlreply(const BYTE* prompt, VAL** buf, int option, BUFFER* history)
{
	return(nextarg(prompt, buf, option, history));
}

/* An error has occured that we want to tell the user about. Append the
 * error message "errmess" to the text already on the message line. If
 * it wont fit then make room for it.
 *
 * A space will clear the message, any other character gets returned.
 *
 * All the work gets done by mlappend().
 *
 * errmess	error message
 * syserr	sys_errlist[errno] or NULL
 *
 * These messages will timeout and disappear.
 *
 * Return the next input keystroke.
 */
KeyCode	mlerr(const BYTE* errmess, const BYTE* syserr)
{
	KeyCode	c;
	
	mlappend(errmess, TRUE, syserr);	/* show the error   */

	c = tgetc(MGC_MAP, mlerrMsgTimeout);

	mlappend((char*)NULL, TRUE, syserr);	/* restore the line */

	if(c == ' ' || c == KEY_TIMEOUT)
		c = tgetc(MGC_MAP, 0);

	return(c);
}

/* A simple string append to the message line.
 * A space is not placed there in front of str.
 * The length before the append is saved (static).
 *
 * If str is NULL the line length is set to the saved length.
 */
void	mlSappend(const BYTE* str)
{
	static	LineLn	savedLen;

	if(str == (BYTE*)NULL) {
		if(savedLen > lused(wmessp->w_dotp))
			savedLen = lused(wmessp->w_dotp);
		lsetused(wmessp->w_dotp, savedLen);
	} else {
		savedLen = lused(wmessp->w_dotp);
		lcats(wmessp->w_dotp, str);
	}

	wmessp->w_doto = lused(wmessp->w_dotp);
	wmessp->w_flag |= WFHARD;
}

/* Append 'mess' to whatever is on the message line. Store the old message
 * line before doing this, together with the cursor position.
 *
 * If 'mess' is NULL then restore the old message line and cursor position.
 *
 * If 'pretty' is true then put [ ] characters around the string and if
 * syserr (which is usually sys_errlist[errno]) if set.
 * 
 * NB the display can be very wide.
 */
void	mlappend(const BYTE* mess, int pretty, const BYTE* syserr)
{
	LINE*		mlp = wmessp->w_dotp;	// Should also be w_topLinep
	static	LineLn	oldLen;

	// Demand a refresh
	wmessp->w_flag |= WFHARD;

	// New spec: does not store old line, assumption that the line was appended to,
	// restore old length.

	// Need to look at this and not display/put into history if(discmd == FALSE) - TODO

	// Restore old line -- before appended to
	if(mess == (BYTE*)NULL) {
		if(oldLen > lused(mlp))	// Hopefully never
			oldLen = lused(mlp);

		// Restore old length
		lsetused(mlp, oldLen);
		wmessp->w_doto = oldLen;

		oldLen = 0;
		return;
	}

	oldLen = lused(mlp);	// Something to return to

	// Ignore what insStrLine() returns, it won't hurt to continue,
	// at worse some messed up message line.

	lcatm(mlp, "  [", (pretty ? 3 : 1));
	lcats(mlp, mess);

	if(syserr) {
		// Append ": " + syserr
		lcats(mlp, ": ");
		lcats(mlp, syserr);
	}

	if(pretty)
		lcats(mlp, "]");

	wmessp->w_doto = lused(wmessp->w_dotp);
}

/* Convert expanded character to character.
 * Collapse the CNTRL and SPEC flags back into an ascii code
 *
 * Return the changed character
 */
int	ectoc(KeyCode c)
{
	if(c & CNTRL)
		c = c & ~(CNTRL | 0x40);
	if(c & SPEC)
		c = c & 0xFF;
	return(c);
}

/* Convert character to extended character.
 * Pull out the CNTRL and SPEC prefixes (if possible)
 *
 * Return the changed character
 */
int	ctoec(int c)
{
	if(c>=0x00 && c<=0x1F)
		c = CNTRL | (c+'@');
	return(c);
}
 
/* Get a command name from the command line.
 *
 * Command completion means that pressing a <SPACE> or <TAB> will attempt to
 * complete an unfinished command name if it is unique.
 *
 * prompt	is the command prompt to use.
 * cmdname	is a buffer to use
 *
 * Return a pointer to the NBIND descriptor for the command.
 *	  NULL on: no input entered, C-g pressed
 */
NBIND*	getname(const char* prompt)
{
	if(mlprompt(prompt, MLEMACSCOMMAND, bmlEmCmdHistp) != TRUE)
		return(NULL);

	return(fncmatch(ltext(bmlEmCmdHistp->b_dotp)));
}

/* Prompt in the message line and read a single character from input.
 *
 * flags	Ored flags:
 *	MGC_NONE	Nothing
 *	MGC_MAP		Perform input key mapping - see map-key
 *	MGC_LEAVEP	Leave the prompt shown after this function returns
 *	MGC_NO_CURSOR	Do not touch CursorInMsg, cursor stays where it is
 *	MGC_RET_TIMEO	If there is a timeout, return KEY_TIMEOUT
 *	MGC_TIMOUT_MASK	If non zero here this is the delay in seconds
 *			Range: 1s to 1h 8m 15s
 *	MGC_APSTR_NOC	Timeout append str is coloured & cursor does not follow
 *			Important if cursor in message line & don't want it to move
 *	MGC_NO_REFRESH	Do not test input for screen refresh key
 *			This should only be given in rare cases like bind-to-key
 *			or when reading parts of multi key commands
 *	MGC_NO_KMACRO	When replaying from a keyboard macro - do not supply from
 *			the macro, this allows a prompted pause to really take place,
 *			eg after execute-program
 *			When recording: do not save in the recording
 *	MGC_PROMPTING	When prompting - to stop a message being put to the message line.
 *			Both lines are the same one, so stop messages mangling prompts.
 *			This involves the use of global variable CursorInPrompt :-(
 *			mlGetKey() sets this itself if MGC_LEAVEP
 *	MGC_DELAY_PROMPT Delay before writing the prompt/message (ie fmt)
 *			The idea is that there might be an (error) message and
 *			the user is given some time to read it.
 *			The delay is ddelay ($prefdlay)
 *	MGC_MACARG	If running in a macro return the first character from
 *			the next argument. Return KEY_NONE if no argument
 *
 * toStr	If this is not NULL then append this string to the prompt
 *		after $prefdlay (ddelay) seconds.
 *
 * fmt		Format string as for mlwrite() followed by arguments.
 *		The string should not contain %# as that will double msgPop().
 *		If this is NULL then some text should already have been put to
 *		wmessp, you might want to use MGC_LEAVEP.
 *
 * Return the character read
 *	  or KEY_TIMEOUT or KEY_NONE
 */
KeyCode	mlGetKey(int flags, const BYTE* toStr, const char* fmt, ...)
{
	KeyCode	c;

	if(clexec == TRUE && (flags & MGC_MACARG) != 0) {
		// Running a macro - grab next argument
		// ANY: could be a STRing or maybe an INT
		VAL*	vp = valGet(V_ANY);
		if(nextarg(fmt, &vp, MLNOOPT, bmlPromptHistp) == TRUE) {
			if(vtype(vp) == V_CHAR || vtype(vp) == V_INT || vtype(vp) == V_BOOL)
				c = (KeyCode)vint(vp);
			else	// Take first character:
				c = utf8_decode(vtext(valPrint(vp)));
		} else
			c = (KeyCode)KEY_NONE;

		valRelease(vp);

		return(c);
	}

	va_list	ap;
	va_start(ap, fmt);
	c = mlvGetKey(flags, toStr, fmt, ap);
	va_end(ap);

	return(c);
}

/* mlGetKey() but taking a va_list argument.
 *
 * Return the character read
 *	  or KEY_TIMEOUT
 */
KeyCode	mlvGetKey(int flags, const BYTE* toStr, const char* fmt, va_list ap)
{
	KeyCode	c;			// What we return
	int	cimSave = CursorInMsg;
	LineLn	doto = curwp->w_doto;
	LineLn	dotLen = lused(curwp->w_dotp);

	if((flags & MGC_NO_CURSOR) == 0)
		CursorInMsg = 1;

	/* Delay before writing the prompt/message ?
	 * See if we read something before then
	 */
	if((flags & MGC_DELAY_PROMPT) != 0 &&
	   (c = tgetc(flags | MGC_PROMPTING, ddelay * MILLION)) != KEY_TIMEOUT)
			goto off;	// Not timed out

	if(fmt != (char*)NULL)
		msgWrite(fmt, ap);

	// String to append after a timeout ?
	if(toStr) {
		int	timeout = ddelay;
		if(flags & MGC_TIMOUT_MASK)	// Specified timeout
			timeout = flags & MGC_TIMOUT_MASK;

		if((c = tgetc(flags | MGC_PROMPTING, timeout * MILLION)) != KEY_TIMEOUT)
			goto off;	// Not timed out

		// The caller wants to deal with the timeout itself ?
		if(flags & MGC_RET_TIMEO)
			goto off;	// Not timed out

		/* The appended string is to be coloured.
		 * Set where from and how long.
		 */
		if(flags & MGC_APSTR_NOC) {
			msgColourFrom = (int)dotLen;
			msgColourLen = (int)strlen(toStr);
		}

		mlSappend(toStr);

		/* Restore the cursor to where it was before the append
		 * But if it was at line end then leave it in its new position
		 * at the end of the extended line
		 */
		if((flags & MGC_APSTR_NOC) && doto != dotLen)
			curwp->w_doto = doto;
	}

	c = tgetc(flags | MGC_PROMPTING, 0);

	// If we added something, restore to as it was before:
	if(toStr) {
		curwp->w_doto = doto;
		lsetused(curwp->w_dotp, dotLen);
		wmessp->w_flag |= WFHARD;

		msgColourFrom = -1;	// No message line colouring
	}

off:
	// Is the prompt is to be left for the user to see as a message ?
	if((flags & MGC_LEAVEP) == 0) {
		msgTop++;		// Hmmm, a bit tacky
		msgPop(FALSE);
	}

	CursorInMsg = cimSave;

	return(c);
}

/* tgetc: Get a key from the terminal driver, resolve any keyboard macro action
 *
 * flags		Uses same values as mlGetKey() but not all recognised
 *	MGC_NONE	Nothing
 *	MGC_MAP		Perform keyboard/character mapping.
 *	MGC_APSTR_NOC	Timeout append str is coloured & cursor does not follow
 *			Important if cursor in message line & don't want it to move
 *	MGC_NO_REFRESH	Do not test input for screen refresh key
 *			This should only be given in rare cases like bind-to-key
 *			or when reading parts of multi key commands
 *	MGC_NO_KMACRO	When replaying from a keyboard macro - do not supply from
 *			the macro, this allows a prompted pause to really take place,
 *			eg after execute-program
 *			When recording: do not save in the recording
 *	MGC_PROMPTING	When prompting - to stop a message being put to the message line.
 *			Both lines are the same one, so stop messages mangling prompts.
 *			This involves the use of global variable CursorInPrompt :-(
 *
 * readTimeout is how long before the read times out and returns KEY_TIMEOUT
 *	readTimeout is measured in microseconds.
 *	readTimeout of 0 means no timeout.
 *
 * Note that we could be returning a Unicode Character or function/arrow key.
 *
 * Doing screen refresh here makes it easy to accept the command almost anywhere.
 * It is prob overkill to look at RefreshBinds[] (& maybe call refreshScreen()),
 * but it does mean that we can deal with rebindings - but ONLY if of a keystroke
 * so M-XX will not be recognised as we see that as 2 characters.
 *
 * Global CursorInMsg is TRUE if the cursor is in the message line, so this
 * will move it there before reading. This is a nasty kludge, should be an argument,
 * but too much to change -- for now.
 *
 * Keyboard macros:
 * PLAY		return the next key from the recording
 * RECORD	remember the key read into the recording
 *
 * If a visible buffer is in RENEW mode - checkage() to maybe read it in.
 *
 * Does an update() before reading.
 *
 * Return a KeyCode
 *	  or KEY_TIMEOUT - in RECORD this will be remembered in the macro
 */
KeyCode	tgetc(int flags, long readTimeout)
{
	KeyCode	c;	/* fetched character/token */
	long		ageTimeout = RenewTimeout * MILLION;	// How long to wait to check file aged

	/* If we are playing a keyboard macro back.
	 * Not if to force a real read from keyboard
	 */
	if(kbdmode == PLAY && (flags & MGC_NO_KMACRO) == 0) {

		/* if there is some left... */
		if(kbdptr < (KeyCode*)(void*)tgetcaddr(kbdMacro, tused(kbdMacro)))
			return(*kbdptr++);

		/* at the end of last repitition? */
		if(--kbdrep < 1) {
			kbdmode = STOP;
#if	VISMAC == 0
			/* force a screen update after all is done */
			update(FALSE);
#endif
		} else {

			/* reset the macro to the begining for the next rep */
			kbdptr = (KeyCode*)(void*)ttext(kbdMacro);
//			kbdptr = &kbdm[0];
			return(*kbdptr++);
		}
	}

	// Tell msgWrite() that the prompt is not to be disturbed
	if(flags & MGC_PROMPTING)
		CursorInPrompt = 1;

	/* Loop as we might be checking if buffers have been externally modified.
	 */
	while(1) {
		long	TimeoutBy = readTimeout;

		if(RenewBufferIsVisible) {
			if(time(NULL) >= lastCheckedAge + RenewTimeout) {

				checkage(TRUE);

				// Get the time again - in case checkage() takes ages
				lastCheckedAge = time(NULL);
				ageTimeout = RenewTimeout * MILLION;
			}

			// Timeout the smaller of the 2 possible times - we can always timeout a bit more later
			TimeoutBy = readTimeout != 0 && readTimeout < ageTimeout ? readTimeout : ageTimeout;
		}

getNextChar:	;
		Window*	savwpTgetc = curwp;	/* Available for restore */
		BUFFER*	savbpTgetc = curbp;	/* Save independently as curbp is not always curwp->w_bufp !! */
		LineLn	savdoto    = curwp->w_doto;

		if(CursorInMsg) {
			// Set so that the cursor is displayed in the message window
			// by update()
			curwp = wmessp;

			// Cursor at the end of the line
			if( !(flags & MGC_APSTR_NOC))
				curwp->w_doto = lused(curwp->w_dotp);

			curbp = curwp->w_bufp = bmlPromptHistp;
		}

		// Ensure that the message line is fully brought up to date:
		wmessp->w_flag |= WFHARD;
		update((flags & MGC_NO_KMACRO));

		/* Restore to saved window & buffer.
		 * Do it now before the read - which could time out and not return
		 * via here.
		 * Beware: possible buffer refresh - renewbuf(), check w_doto
		 */
		curwp->w_doto = savdoto;
		if(curwp->w_doto > lused(curwp->w_dotp))
			curwp->w_doto = 0;
		curwp = savwpTgetc;
		curbp = savbpTgetc;

		/* Fetch a character from the terminal driver.
		 * Map it if we are supposed to & if it is a real character, ie not
		 * a function key.
		 */
		c = TTgetc(TimeoutBy);

		// Did not timeout - do mapping, etc, & return to caller
		if(c != KEY_TIMEOUT) {
			/* Shall we try mapping the input character to some other one ?
			 *
			 * This is done here as the map might (unlikely) yeild the refresh key.
			 */
			if((flags & MGC_MAP) && c >= 0 && c <= UTF8_MAX_CODE_POINT) {
				UniChar	 oc = c;
				CharMap* cmp;

				if(c < (KeyCode)sizeof(keymap))	// Most common, probably
					c = keymap[oc];

				/* We can't leave it there ... it is possible that one of these US-ASCII
				 * (controls + Basic Latin) was mapped onto a Unicode character that will not
				 * fit in that range, so do a scan.
				 * This won't be that slow since the list is sorted so most of the time
				 * it will stop at the first one.
				 */
				for(cmp = CharMapChain ; cmp; cmp = cmp->chm_next) {
					if(cmp->chm_from < oc)
						continue;
					if(cmp->chm_from == oc)
						c = cmp->chm_to;
					break;
				}
			}

			// Check if the key entered is bound to refreshScreen()
			if((flags & MGC_NO_REFRESH) == 0) {
				int	nrbs;

				for(nrbs = 0; nrbs < NumRefreshBinds; nrbs++)
					if(c == RefreshBinds[nrbs]) {
						// Refresh the screen
						refreshScreen(FALSE, 0);

						// Might do wrong TimeoutBy - prob not an issue
						goto getNextChar;
					}
			}

			break;
		}

		/* Timeout.
		 * Could either be because this function was asked to timeout or
		 * we timed out because we need to check file age.
		 *
		 * Check if we reached the timeout that the function was asked to,
		 * if so return the timeout:
		 */
		if(readTimeout > 0 && (readTimeout -= TimeoutBy) <= 0)
			break;

		/* Just continue the loop to deal with timeout as we need to
		 * check if a file has been externally modified.
		 */
		if(ageTimeout > TimeoutBy)
			ageTimeout -= TimeoutBy;

	}

	CursorInPrompt = 0;

	/* record it for $lastkey */
	lastkey = c;

	/* Save it if we need to.
	 * Not if, at this point in the recording, we want to really
	 * wait for the user to type something.
	 */
	if(kbdmode == RECORD && (flags & MGC_NO_KMACRO) == 0)
		tcatm(kbdMacro, (BYTE*)&c, sizeof(c));

	/* and finally give the char back */
	return(c);
}

/*	GET1KEY:	Get one keystroke. The only prefixs legal here
 *			are the SPEC and CNTRL prefixes.
 *
 * Keyboard mapping will be done.
 * Termcap interpretation is done.
 * 
 * Return the character read
 *	  or KEY_TIMEOUT
 */
KeyCode	get1key(void)
{
	KeyCode	c;

	/* get a keystroke */
	c = tgetc(MGC_MAP, 0);

        return(c);
}

/* Get a canonical action value for a key that may be bound to any command.
 * Ie a binding transparent action key - an ISI_xxx value for a few bound keys.
 *
 * c has the SPEC, or other, bit set, ie is a function key.
 * This may be bound to something, return a value that represents that binding
 * (ie the function) that is invariant. This will be one of the INTL values.
 * The point is that these INTL values can then be put into a switch statement
 * or something.
 * The purpose is to allow the user to type actions and get the result that he
 * expects at a point when we are not using the normal interpretation, eg when
 * searching we want F1 (or whatever) to do a next search, we still want this to
 * work if the user has search bound to F2 (or something).
 *
 * If the current buffer is in help mode - those bindings will not be looked at.
 *
 * We do this by searching for the normal binding of c in keytab, then using
 * ismap to find an INTL value which we do by comparing the function pointer.
 * It is quite possible that we don't find a match in ismap, in this case
 * return c unchanged.
 *
 * Use getbind() as the way that bindings are organised might change in the future.
 *
 * Return the canonical ISI_xxx value
 *	the value c if it is not in the mapping table.
 */
KeyCode	ReverseMap(KeyCode c)
{
	RMTAB*	rm;
	NBIND*	bind;

	bind = getbind(c, FALSE);
	for(rm = ismap; rm->rm_code != 0; rm++)
		if(rm->rm_fp == bind->n_func)
			return(rm->rm_code);

	return(c);
}

/* A generalized prompt/reply function.
 *
 * prompt	prompt associated with what we're typing - NUL terminated
 * value	where the resulting string is put
 *		This can be NULL - value left in history
 * option	bit field which modifies our behaviour slightly - see estruct.h
 *		MLNOECHO will prevent the text being put to history, the
 *		text will be put into LINE playCrypt.
 * history	the buffer that contains history of this prompt type
 *		If this is NULL bmlPromptHistp will be used
 *
 * Return:
 *     TRUE  - OK. Value is in:
 *		*value
 *		history->b_dotp.
 *     FALSE - Zero length string entered
 *     ABORT - ^G or ^C or out of memory
 */
int	_getstring(const BYTE* prompt, int option, BUFFER* history, Window* win);
int	getstring(const BYTE* prompt, VAL** value, int option, BUFFER* history)
{
	static	int	depth;
	int	ret;
	Window*	savwp = curwp;
	BUFFER*	savbp = curbp;		/* Save independently as curbp is not always curwp->w_bufp !! */

	// fprintf(stderr, "\n");
	// fprintf(stderr, "getstring(''%s'' option=%#x messp_attrib=%#x\n", prompt, option, wmessp->w_attrib);

	// Make message line current
	curwp = wmessp;
	curbp = bmlPromptHistp;

	/* Push a new line in [Prompt Line] unless that is the history buffer
	 * that is being used.
	 */
//	if(history != bmlPromptHistp) {
		msgPush(1);
		depth++;
//	}

	// Kludge - TODO
	if( !history)
		history = bmlPromptHistp;

	NotToKillBuffer = 1;	// Don't want command line deletes in yank buffers

	ret = _getstring(prompt, option, history, savwp);

	if(ret == TRUE && value != (VAL**)NULL) {
		valCopy(value, valStrToVAL(""));
		// TEMP until everything takes the result from the history buffer
		LINE* res = history->b_dotp;
		tcpyl(vTEXT(*value), res);
	}

	NotToKillBuffer = 0;	// Reset to normal
	curwp->w_flag |= WFEDIT;

	// Out of memory ?
	if(ret == FAILED) {
		// generate an error message
		ret = ABORT;
	}

//	if(history != bmlPromptHistp) {
		msgPop(FALSE);
		// Evidentially not quite right: can end up with crud in the message line
		if(--depth == 0)
			mlerase();
//	}

	/* The above uses generic buffer editing functions.
	 * These will have marked the [Prompt Line] buffer as changed,
	 * we don't want an Internal buffer marked changed - its confusing to the user.
	 */
	bmlPromptHistp->b_flag &= ~BFCHG;

	curwp = savwp;
	curbp = savbp;

	return(ret);
}

#if	0
/* Debug version of mlwrite() for use here since if used in getstring()
 * the value of curwp->w_doto gets set to lused(curwp->w_dotp)
 */
void	MLWRITE(const CHAR* fmt, ...)
{
	LineLn	o = curwp->w_doto;
	va_list	ap;
	va_start(ap, fmt);
	msgWrite(fmt, ap);
	va_end(ap);
	curwp->w_doto = o;
}
#endif

/* This does the work of getstring() - interactive prompt the user for a string.
 *
 * It is only called from getstring() but this can return from several places,
 * so separating it keeps restoring the environment clean.
 *
 * curwp will be wmessp (ie prompt line at bottom of the screen)
 * curbp will be bmlPromptHistp - might be different from history
 *
 * Arguments as for getstring() above and:
 *
 * win	What was the current window before the prompt was issued.
 *	This is needed for finding things like the word under ".", also
 *	used by bnexpand()
 *
 * The result will be in history->b_dotp - which will be NUL terminated (L_NUL)
 * If the MLNOECHO flag is set it will be in playCrypt
 *
 *  Notes of issues:
 * * The line displayed is prompt + entered string
 *   must not allow line editing to change the prompt.
 *   This means that cannot naively use standard movement/... commands
 * * dotp - the line that we are editing. This is kept as some things (eg backword())
 *          could move curwp->w_dotp to another line, so useful to have.
 *	    Beware: whenever an update() is done the buffer can be rewritten
 *	    by renewbuf() and leave dotp pointing to garbage.
 *	    Thus it must be renewed whenever a key is read, remember things
 *	    like fnexpand().
 *
 * Return: TRUE it worked there is something to use
 *	   FALSE - Zero length string entered
 *	   ABORT - ^C or ^G entered
 *	   FAILED - Out of memory
 */
int	_getstring(const BYTE* prompt, int option, BUFFER* history, Window* win)
{
	int	eolchar;	// ESC or \n
	KeyCode	c;
	LineLn	cnt = 0;
	KeyCode	nextchar = -1;	/* next character kludge for fnexpand() */
	int	err;		/* error flag, causes RET to be ignored */
	int	type = option & MLTYPEMASK;	/* Purpose of the prompt */
	void*	Pos = NULL;	// Nothing previous
	LINE*	dotp = curwp->w_dotp;	// Line that we edit
	int	NextWasSpace = 0;	// Loop buster

	int	maxlen = MLmaxLengths[option & MLTYPEMASK];

	// Set default colours:
	mlNormalPallet();

	if(lsetNUL(dotp) == FALSE)
		return(FAILED);

	/* The prompt length is important as, for editing purposes, text
	 * starts after it. Ie move to start of line is really move to promptLen.
	 * 2 values: measured in bytes, measured in characters.
	 */
	int	promptBlen = (int)strlen(prompt);
	int	promptClen = (int)utf8_strlen(prompt);
	// MLWRITE("%#Blen=%d Clen=%d", promptBlen, promptClen);

	// Put the prompt at the start of the line - which is empty
	if(lcpys(curwp->w_dotp, prompt) == FALSE)
		return(FAILED);

	// The prompt is not output by tgetc()
	if(TTopened != TTfullyOpened) {
		fputs(prompt, stdout);
		fflush(stdout);
	}

	// If the history buffer mark is not set: place at history buffer end
	if(history->b_markp == (LINE*)NULL)
		history->b_markp = history->b_linep;
	history->b_marko = 0;

	// If the mark is not at history buffer end then the last use was
	// ended with C-o - so append the marked line as default value
	if(history->b_markp != history->b_linep) {
		lcatl(curwp->w_dotp, history->b_markp);
	} else if(option & MLDEFISLAST) {
		// Default is last result - ie dotp in history
		lcatl(curwp->w_dotp, history->b_dotp);

		/* Avoid a situation where we show a default, the user presses
		 * C-p and gets shown the same line.
		 * This is confusing so take it back one more.
		 */
		if(lback(history->b_markp) == history->b_dotp)
			history->b_markp = history->b_dotp;
	}

	curwp->w_doto = lused(curwp->w_dotp);	// Cursor at line end

	/* Tags need the first one displaying */
	if(type == MLTAG)
		TagDispInit(prompt);

	// With what is input ended ?
	eolchar = option & MLEOLESC ? CharEscape : '\n';

	for(;;) {
		int	curBlen = (int)lused(curwp->w_dotp) - promptBlen;
		int	curLen = option & MLU8LEN ?
				(int)utf8_chrnlen(ltext(curwp->w_dotp), (size_t)lused(curwp->w_dotp)) - promptClen :
				curBlen;

		// MLWRITE("%#curLen=%d nextchar=#%x doto=%d used=%d", curLen, nextchar, curwp->w_doto, lused(curwp->w_dotp));

		// Should not happen, but seems to. Fix up:
		if(curwp->w_doto > lused(curwp->w_dotp))
			curwp->w_doto = lused(curwp->w_dotp);

		/* If nextchar is set (ie it is not -1) then it is something
		 * from fnexpand(), bnexpand() or cmexpand().
		 *
		 * This gives us a sort of "ungetc()" feature. If neither are
		 * set, get a character from the keyboard.
		 *
		 * tgetc() also bring the screen up to date with update()
		 */
		if(nextchar != -1) {
			c = nextchar;
			nextchar = -1;
			if(c == ' ') {
				if(NextWasSpace) {
					NextWasSpace = 0;
					continue;
				}
				NextWasSpace = 1;
			}
		} else {
			wmessp->w_flag |= WFHARD;
			if(option & MLNOECHO) {
				/* Display what is being entered as ****
				 *
				 * Have to keep the cursor in the right place
				 * (maybe not at the end), interesting with
				 * multi byte chars.
				 *
				 * First: preserve what we have.
				 */
				LineLn	sdoto = curwp->w_doto;	// save "."
				size_t	len = utf8_chrnlen(ltext(curwp->w_dotp), (size_t)lused(curwp->w_dotp));
				size_t	dot = utf8_chrnlen(ltext(curwp->w_dotp), (size_t)sdoto);	// "." in the line (chars)
				curwp->w_doto = (LineLn)dot;
				tcpyl(playPrompt, curwp->w_dotp);
// chars not bytes TODO
				for(dot = len; (int)dot > promptBlen; )
					lputc(curwp->w_dotp, --dot, '*');
				lsetused(curwp->w_dotp, (LineLn)len);
				
				c = tgetc( ((option & MLNOMAP) ? MGC_NONE : MGC_MAP) | MGC_PROMPTING, 0);

				// Restore the real text
				lcpyt(curwp->w_dotp, playPrompt);
				curwp->w_doto = sdoto;
			} else {
				c = tgetc( ((option & MLNOMAP) ? MGC_NONE : MGC_MAP) | MGC_PROMPTING, 0);
			}
		}

		dotp = curwp->w_dotp;

		/* if it is a <ret>, change it to a <NL> */
		if(c == '\r')
			c = '\n';

		/* A space as first character can mean end of input is wanted */
		if((option & MLLN0END) && c == ' ' && curwp->w_doto == promptBlen)
			c = eolchar;

		if(c == eolchar || c == 0x0f) {
			/* 0x0f == ^O meaning: Enter and next - like bash */

			// Keep the promise of NUL termination
			lsetNUL(history->b_dotp);

			// Zero length entered ? (ignore prompt)
			if(lused(curwp->w_dotp) == promptBlen)
				return(FALSE);

#if	UNIX
			if(type == MLXFILE) {
				/* if a filename, try to expand */
				nextchar = fnexpand(prompt, 0, &err);

				if(err)
					continue;
			}
#endif

			
			// Append the line to the appropriate history buffer
			if(option & MLNOECHO) {
				/* If no-echo save in playCrypt
				 * This is to avoid it being readable to the user
				 */
				tcpys(playCrypt, ltext(curwp->w_dotp) + promptBlen);

				// Then blat it:
				cnt = lused(curwp->w_dotp) - promptBlen;
				if(cnt > 0) {
					lputc(curwp->w_dotp, promptBlen, 'x');
					memocpy(lgetcaddr(curwp->w_dotp, promptBlen + 1),
						lgetcaddr(curwp->w_dotp, promptBlen),
						(size_t)cnt);
				}
			} else {
				/* Append to history unless what we have is the same as the mark line.
				 */
				if(history->b_markp != history->b_linep &&
				   strcmp(ltext(history->b_markp), lgetcaddr(curwp->w_dotp, promptBlen)) == 0) {
					// The string used is from buffer's "." line
					history->b_dotp = history->b_markp;
				} else {
					// Add as a new line to the history
					if( !addHistoryLine(history, lgetcaddr(curwp->w_dotp, promptBlen), AHL_NO_DUP)) {
						TTbeep();
						return(FAILED);
					}
					history->b_dotp = lback(history->b_linep);
				}
			}

			lsetNUL(history->b_dotp);

			/* If in a command line macro echo what is entered at the prompt.
			 * This is so that it appears on output ... no need if stdin
			 * is connected to a tty - which will echo (presumably).
			 */
			if(TTopened != TTfullyOpened && !isatty(0)) {
				fputs(lgetcaddr(curwp->w_dotp, promptBlen), stdout);
				putchar('\n');
				fflush(stdout);
			}

			// "." at end text being made available
			history->b_doto = lused(history->b_dotp);

			/* This is all about where to put the mark in the history buffer,
			 * which affects what is shown the next time that the history buffer
			 * is used in a prompt.
			 *
			 * <ret> set the mark after the last line
			 * C-o   leave it on the line under where the last C-p or C-n left it.
			 */
			if(c == eolchar)
				history->b_markp = history->b_linep;
			else
				history->b_markp = lforw(history->b_markp);
			history->b_marko = 0;

			/* Just in case someone is looking at the history buffer
			 * ensure that the buffer's marks, etc, are shown in the window.
			 * Beware: buffer [Prompt Line] can have a window count without a window.
			 */
			if(history->b_nwnd != 0) {
				Window*	wp = onscreen(history);
				if(wp)
					ConnectBuffer(history, wp, 1);
			}

			// Move the cursor to start of the message line
			// So that the user knows that they have hit enter - if the action takes a long time to complete
			msglCpos(0);
			msglFlush();

			// Copy the text into the buffer is done by our caller
			return(TRUE);
		}

		if((option & MLSELECTONLY) != 0) {
			/* Choose from options. Entry of any text is not allowed.
			 * Translate the character entered to ' ' or CharRevSearch
			 * which will be recognised below and NextTag()/fnexpand()/...
			 * MGC_DELAY_PROMPT
			 */
			switch(c) {
			  case ' ':
			  case UTC_DOWN:
			  case ChrToCtrl('n'):	// C-n 0x0e
				c = ' ';
				break;

			  case CharRevSearch:
			  case UTC_UP:
			  case ChrToCtrl('p'):	// C-p 0x10
				c = CharRevSearch;
				break;

			  case ChrToCtrl('g'):	// C-g 0x0e - abort
				break;

			  default:	// Accept nothing else
				TTbeep();
				continue;
			}
		}

		/* Maybe search, back & forth on initial length - if the sequence
		 * is interrupted reset the position pointer
		 */
		if(c != ' ' && c != '\t' && c != CharRevSearch)
			Pos = NULL;

		switch(c) {
		  case UTC_HOME:
		  case 0x01:
			/* ^A : Move to start of buffer */
			curwp->w_doto = promptBlen;
			break;
#if	TONY_BINDINGS
		  case 0x02:
			/* ^B : Move back a word - this is only here because
			 *	I'm a pervert and have ^B mapped to backword
			 *	rather than its "correct" mapping of backchar
			 */
			// This could move to a previous line, bring it back
			backword(TRUE, 1);
			curwp->w_dotp = dotp;
			if(curwp->w_doto < promptBlen)
				curwp->w_doto = promptBlen;
			break;
#else
		// Can do back word with M-B
#endif
		  case 0x03:
			/* ^C : If a command, then save without executing */
			if((option & MLNOECHO) == 0)	// Not if a password - or something
				addHistoryLine(history, ltext(curwp->w_dotp) + promptBlen, AHL_NO_DUP);
			if(type == MLSHELLCOMM)
				return(ABORT);
			else
				TTbeep();
			break;
		  case UTC_DEL:
		  case 0x04:
			/* ^D : Delete the character under the cursor */
			if(curwp->w_doto < lused(curwp->w_dotp))
				ldelete(1, KL_DISCARD);
			else
				TTbeep();
			break;
		  case UTC_END:
		  case 0x05:
			/* ^E : Move to end of buffer */
			curwp->w_doto = lused(curwp->w_dotp);
		  	break;
		  case UTC_RIGHT:
		  case 0x06:
			/* ^F : Move forward one character */
			if(curwp->w_doto < lused(curwp->w_dotp))
				moveForw(1);
			else
				/* can't move past end of buffer */
				TTbeep();
			break;
		  case 0x07:
			/* ^G : Abort input and return */
			ctrlg(FALSE, 0);
			sgarbf = TRUE;    /* Force complete screen refresh -- only really needed after sub command, but .... */
			return(ABORT);
		  case UTC_LEFT:
#if	TONY_BINDINGS
		  case 0x08:	/* ^H for everything else */
#else
		  case 0x02:	/* ^B for IBMPC */
#endif
			if(curwp->w_doto > promptBlen)
				moveBack(1);
			else
				/* can't move past start of buffer */
				TTbeep();
			break;
		  case 0x09:	/* ^I - Tab
				 * If at end of line in file name expansion - expand what we have
				 * so far. Do this by putting an '*' and expanding.
				 * Otherwise just put tab in the name.
				 */

			if(type == MLXBUF) {
				nextchar = bnexpand(prompt, 1, win);
				break;
			}

			if(type == MLXMODE) {
				nextchar = mnexpand(&ModeBuf, prompt, 1, &err);
				break;
			}

			if(type == MLWOPTION) {
				nextchar = mnexpand(&WinOpt, prompt, 1, &err);
				break;
			}

			if(type == MLEMACSCOMMAND) {	// Internal Emacs command
				nextchar = NextCommand(&Pos, 1, prompt);
				break;
			}

			if(type == MLEMACSVAR) {	// Variable name
				nextchar = NextVarName(&Pos, 1, prompt);
				break;
			}

			if(type == MLXFILE && curwp->w_doto == lused(curwp->w_dotp))
				c = '*';

			// Check that it is not too long
			// c is either TAB or '*'
			if(maxlen != 0 && curLen >= maxlen) {
				TTbeep();
				break;
			}

			if(linsert(1, c) == FALSE) {
				TTbeep();
				break;
			}

			if(c == '*') {
				nextchar = fnexpand(prompt, FORW, &err);
			}
			break;
		  case 0xb:
			/* ^K : Kill to end of line */
			if(curwp->w_doto < lused(curwp->w_dotp))
				killtext(FALSE, 0);
			else
				TTbeep();
			break;

		  // 0xc ^L done in tgetc()

		  case UTC_UP:		/* ^P : previous line from history */
		  case 0x10:		/* Ditto */
			/* Move down, stop when reach the bottom of the history buffer.
			 * Replace message line text: prompt + line from history.
			 * The mark is place in the history buffer.
			 */
			if((option & MLNOECHO) == 0 &&
			   lback(history->b_markp) != history->b_linep) {
				history->b_markp = lback(history->b_markp);
				lcpys(curwp->w_dotp, prompt);
				lcatl(curwp->w_dotp, history->b_markp);
				curwp->w_doto = lused(curwp->w_dotp);
			} else
				TTbeep();
			break;
		  case UTC_DOWN:	/* ^N : next line from history */
		  case 0xe:		/* Ditto */
			if((option & MLNOECHO) == 0 &&
			   lforw(history->b_markp) != history->b_linep) {
				history->b_markp = lforw(history->b_markp);
				lcpys(curwp->w_dotp, prompt);
				lcatl(curwp->w_dotp, history->b_markp);
				curwp->w_doto = lused(curwp->w_dotp);
			} else
				TTbeep();
			break;
		  case 0x14:
			/* ^T :	Transpose the character before the cursor
			 * 	with the one under the cursor.
			 * Now: more like twiddle in any other buffer: if at line end
			 * it works on the last 2 chars in the line.
			 * Take care of multibyte chars.
			 */
			if(curwp->w_doto > promptBlen &&
			   lused(curwp->w_dotp) > promptBlen + CHAR_LEN(lgetc(curwp->w_dotp, promptBlen)))
				twiddle(FALSE, 0);
			else
				TTbeep();
			break;
		  case 0x15:
			/* ^U : Kill the whole line, this is stupid (I
			 * personally keep expecting it to set a count),
			 * but it does perserve backwards compatability
			 * with ksh - but not bash.
			 */
			// This is good enough to delete
			lsetused(curwp->w_dotp, promptBlen);
			curwp->w_doto = lused(curwp->w_dotp);
			break;
		  case 0x11:			/* ^Q */
		  case 0x16:			/* ^V */
			/* quote the next character */
			c = mlGetKey(MGC_MAP | MGC_LEAVEP | MGC_APSTR_NOC | MGC_NO_REFRESH, " Quote: ", NULL);
			dotp = curwp->w_dotp;

			/* Guard against it being a FNkey or something */
			if(c < '\0' || c > CharMax ||
			   linsert(1, c) == FALSE)
				TTbeep();
			break;

		  case 0x18:	/* ^X - prefix key -- changed from previous usage */
				// These all insert some context sensitive text
			;
			BYTE*	expBytes = 0;		// Expansion (ie inserted) bytes here
			int	expLen = 0;		// Length of expansion - bytes

			c = mlGetKey(MGC_MAP | MGC_LEAVEP | MGC_APSTR_NOC | MGC_NO_REFRESH, "C-x", NULL);
			dotp = curwp->w_dotp;
			switch(c) {

			  BYTE	numbuf[INTWIDTH];	// $ & . written here

			  case 'l':	// insert the line containing "."
				expBytes = ltext(win->w_dotp);
				expLen = (int)lused(win->w_dotp);
				break;
			  case 'i':	// insert C identifier under "."
				expBytes = lgetcaddr(win->w_dotp, win->w_doto);
				// Back off to the start of the identifier:
				while(expBytes > ltext(win->w_dotp) && (Isalnum(expBytes[-1]) || expBytes[-1] == '_'))
					expBytes--;

				// Count to end of the identifier:
				while(expBytes + expLen < ltext(win->w_dotp) + lused(win->w_dotp) &&
						(Isalnum(expBytes[expLen]) || expBytes[expLen] == '_'))
					expLen++;

				break;
			  case 'w':	// insert the word under "."
				expBytes = lgetcaddr(win->w_dotp, win->w_doto);
				// Back off to the start of the word -- is non space
				while(expBytes > ltext(win->w_dotp) && (expBytes[-1] != ' ' && expBytes[-1] != '\t'))
					expBytes--;

				// Count to end of the word:
				while(expBytes + expLen < ltext(win->w_dotp) + lused(win->w_dotp) &&
						(expBytes[expLen] != ' ' && expBytes[expLen] != '\t'))
					expLen++;

				break;
			  case 'd':	// insert the path of the current directory
				expBytes = ttext(cwDir);
				expLen = (int)tused(cwDir);
				break;
			  case 'n':	// insert the name (last component) of the current file
				if((expBytes = (BYTE*)strrchr(win->w_bufp->b_fname, '/')) != NULL)
					expBytes++;				// Skip the /
				else
					expBytes = win->w_bufp->b_fname;	// No / in the path
				expLen = (int)strlen(expBytes);
				break;
			  case 'f':	// insert the name (path) of the current file
				expBytes = win->w_bufp->b_fname;
				expLen = (int)strlen(expBytes);
				break;
			  case 'b':	// b: insert the name of the current buffer
				expBytes = win->w_bufp->b_bname;
				expLen = (int)strlen(expBytes);
				break;
			  case '.':	// insert the current line number
				// Temp swap back to do the count
				sprintf(numbuf, "%ld", (long)getlinenum(win->w_bufp, win->w_dotp));
				expBytes = numbuf;
				expLen = (int)strlen(expBytes);
				break;
			  case '$':	// insert the number of lines in the current buffer
				// Temp swap back to do the count
				sprintf(numbuf, "%ld", (long)getlinenum(win->w_bufp, (LINE*)NULL));
				expBytes = numbuf;
				expLen = (int)strlen(expBytes);
				break;
			  case 0x07:
				/* ^G : Abort input and return */
				ctrlg(FALSE, 0);
				sgarbf = TRUE;    /* Force complete screen refresh -- only really needed after sub command, but .... */
				return(ABORT);
			  default:
				mlerr("Sequence not bound", (char*)NULL);
				break;
			}

			// How much additional length is this - bytes or chars ?
			size_t	addLen = option & MLU8LEN ?
				utf8_chrnlen(expBytes, (size_t)expLen) :
				(size_t)expLen;

			if(maxlen != 0 && curLen + (int)addLen > maxlen) {
				// Trying to add more than allowed
				// How much can we add
				if(option & MLU8LEN) {
					/* The units of avail is characters
					 * Need to turn that into BYTEs
					 * 10000 is a stupidly large amount of supposed available
					 * space - which will never be reached.
					 *
					 * NB: cannot just subtract maxlen & curLen as avail is unsigned
					 */
					size_t	avail = 0;	// Available
					if(maxlen > curLen)
						avail = (size_t)(maxlen - curLen);

					expLen = (int)utf8_maxlen(expBytes, 10000, &avail);
				} else
					expLen = maxlen - curBlen;
			}

			// Not enough room ?
			if(expLen <= 0) {
				TTbeep();
				break;
			}

			// Insert the expanded string
			if(linsm(curwp->w_dotp, curwp->w_doto, expBytes, expLen) == FALSE)
				return(FAILED);

			curwp->w_doto += expLen;

			break;

		  case CharRevSearch:
			/* ^R : Give the last match in either a buffer
			 *	match a filename expansion match or a
			 *	mode match
			 */
			switch(type) {
			  case MLXFILE:
				nextchar = fnexpand(prompt, BACKW, &err);
				break;

			  case MLXBUF:
				nextchar = bnexpand(prompt, -1, win);
				break;

			  case MLXMODE:
				nextchar = mnexpand(&ModeBuf, prompt, -1, &err);
				break;

			  case MLWOPTION:
				nextchar = mnexpand(&WinOpt, prompt, -1, &err);
				break;

#if	CTAGS
			  case MLTAG:	/* C Tags */
				NextTag(prompt, -1);
				break;

#endif

			  case MLEMACSCOMMAND:	// Internal Emacs command
				nextchar = NextCommand(&Pos, -1, prompt);
				break;

			  case MLEMACSVAR:	// Variable name
				nextchar = NextVarName(&Pos, -1, prompt);
				break;

			  default:
				TTbeep();
			}
			break;
		  case ' ':			/* space */
			if(option & MLNOSPACE) {
				TTbeep();
				break;
			}

			switch(type) {
			  case MLXFILE:
				nextchar = fnexpand(prompt, FORW, &err);
				break;

			  case MLXBUF:
				nextchar = bnexpand(prompt, 1, win);
				break;

			  case MLXMODE:
				nextchar = mnexpand(&ModeBuf, prompt, 1, &err);
				break;

			  case MLWOPTION:
				nextchar = mnexpand(&WinOpt, prompt, 1, &err);
				break;

#if	CTAGS
			  case MLTAG:	/* C Tags */
				NextTag(prompt, 1);
				break;

#endif

			  case MLEMACSCOMMAND:	// Internal Emacs command
				nextchar = NextCommand(&Pos, 1, prompt);
				break;

			  case MLEMACSVAR:	// Variable name
				nextchar = NextVarName(&Pos, 1, prompt);
				break;

			  default:
				// Check that it is not too long
				if(maxlen != 0 && curLen >= maxlen) {
					TTbeep();
					break;
				}

				if(linsert(1, ' ') == FALSE)
					TTbeep();
			}

			break;
		  case CharEscape:
			/*
			 * Handle a limited set of Meta sequences. Note that
			 * there is no conflict when the end-of-line char is
			 * escape, since we will never reach here if it is.
			 *
			 * Note that the alphas are ONLY recognised in lower case.
			 */
			c = mlGetKey(MGC_MAP | MGC_LEAVEP | MGC_APSTR_NOC | MGC_NO_REFRESH, "M-", NULL);
			dotp = curwp->w_dotp;
			switch(c) {
			  case 'b':
				/* M-b : Move to start of previous word */
				// This could move to a previous line, bring it back
				backword(TRUE, 1);
				curwp->w_dotp = dotp;
				if(curwp->w_doto < promptBlen)
					curwp->w_doto = promptBlen;
				break;
			case 'c':
				/* M-c : Capitalise next word and move past it. */
				capword(TRUE, 1);
				if(curwp->w_dotp != dotp) {
					curwp->w_dotp = dotp;
					curwp->w_doto = lused(dotp);
				}
				break;
			  case 'l':
				/* M-l : Downcase the whole of the next
				 * 	 word and move past it.
				 */
				lowerword(TRUE, 1);
				if(curwp->w_dotp != dotp) {
					curwp->w_dotp = dotp;
					curwp->w_doto = lused(dotp);
				}
				break;
			  case 'u':
				/* M-u : Capitalise the whole of the next
				 * 	 word and move past it.
				 */
				upperword(TRUE, 1);
				if(curwp->w_dotp != dotp) {
					curwp->w_dotp = dotp;
					curwp->w_doto = lused(dotp);
				}
				break;
			  case UTC_DEL:
			  case 'd':
				/*
				 * Delete word forwards.
				 * Dont' use delfword() as it will go beyond line end
				 */
				cnt = 0;

				while(curwp->w_doto < lused(dotp) && inword() == FALSE) {
					cnt++;
					moveForw(1);
				}

				while(curwp->w_doto < lused(dotp) && inword() == TRUE) {
					cnt++;
					moveForw(1);
				}

				while(curwp->w_doto < lused(dotp) && inword() == FALSE) {
					cnt++;
					moveForw(1);
				}

				moveBack(cnt);
				if(ldelete(cnt, KL_DISCARD) != TRUE)
					return(FAILED);

				break;
			  case 'f':
				/* M-f : Move forward to start of next word.
				 */
				forwword(TRUE, 1);
				if(curwp->w_dotp != dotp) {
					curwp->w_dotp = dotp;
					curwp->w_doto = lused(dotp);
				}
				break;
			  case ' ':
				/*
				 * Do nothing - when I get around to
				 * write the code, it will set the mark at
				 * the current cursor position.
				 */
				break;
			  case CharDelete:
				/*
				 * Delete word backwards. If spaces before
				 * the input position then delete those too.
				 * delbword() will try to delete the prompt
				 */
				cnt = 0;
				while(curwp->w_doto > promptBlen && moveBack(1))
					if(inword() == FALSE) {
						cnt++;
					} else {
						moveForw(1);
						break;
					}

				while(curwp->w_doto > promptBlen && moveBack(1))
					if(inword() == TRUE) {
						cnt++;
					} else {
						moveForw(1);
						break;
					}

				if(ldelete(cnt, KL_DISCARD) != TRUE)
					return(FAILED);

				break;
			  case 0x07:
				/* ^G : Abort input and return */
				ctrlg(FALSE, 0);
				sgarbf = TRUE;    /* Force complete screen refresh -- only really needed after sub command, but .... */
				return(ABORT);
			  default:
			  	mlerr("Sequence not bound", (char*)NULL);
				// This does not work properly - it destroys the prompt
				// Want to put something that times out & disappears,
				// if something it typed early then push back on input,
				// implement TTungetc(KeyCode) - push back just one
				// immediately got with TTgetc
			  	break;
			} /* end of meta sequences */
			break;
#if	TONY_BINDINGS == 0
		  case 0x8:	/* ^H on an IBM PC */
#endif
		  case 0x7f:	/* ^? on everything else */
			/* Delete the character before the cursor.
			 * If there is no input, delete the prompt
			 * and return FALSE.
			 */
			if(lused(curwp->w_dotp) <= promptBlen)
				return(FALSE);
			if(curwp->w_doto <= promptBlen)
				TTbeep();
			else
				backdel(TRUE, 1);
			break;
		  default:
			if((c < ' ' && c > CharMax)) {
				TTbeep();
				break;
			}

			// Check that it is not too long
			if(maxlen != 0 && curLen >= maxlen) {
				TTbeep();
				break;
			}

			/*
			 * If its not a control character, map it if MLUPPER
			 * or MLLOWER are set ....
			 */
			if(option&MLUPPER && Islower(c))
				c = toupper(c);
			else if(option&MLLOWER && Isupper(c))
				c = tolower(c);
			/*
			 * And insert it ....
			 */
			// Failure is out of memory, not size of buffer any more
			if(linsert(1, c) == FALSE)
				return(FAILED);

			break;
		}
	}
}

/* end */
