/*
 * ADDW SCCS: @(#)isearch.c	1.70 09/22/20 22:44:36
 */

/*
 * The functions in this file implement commands that perform incremental
 * searches in the forward and backward directions.  This "ISearch" command
 * is intended to emulate the same command from the original EMACS 
 * implementation (ITS).  Contains references to routines internal to
 * SEARCH.C.
 *
 * REVISION HISTORY:
 *
 *	D. R. Banks 9-May-86
 *	- added ITS EMACSlike ISearch
 *
 *	John M. Gamble 5-Oct-86
 *	- Made iterative search use search.c's scanner() routine.
 *	  This allowed the elimination of bakscan().
 *	- Put isearch constants into estruct.h
 *	- Eliminated the passing of 'status' to scanmore() and
 *	  checknext(), since there were no circumstances where
 *	  it ever equalled FALSE.
 */

#include	"estruct.h"
#include	"utf8.h"
#include        "edef.h"

#if	SHOWMODREL
char	isrchvers[] = "@(#)isearch.c	1.70 09/22/20 22:44:36";
#endif

#if	ISRCH

/* A couple more "own" variables for the command string */

static	int	cmd_reexecute = -1;	/* > 0 if re-executing command */
static	int	repromptAfterTime;	// Reissue the prompt after a bit

// This is TRUE if the search string is the string searched the last i-search
// Ie nothing different has been entered
static	int	SearchLast;


/* Incremental reverse search.
 *
 * It actually uses the same code as the normal incremental search, as both can
 * go both ways.
 *
 * A count >= 0 - search reverse, negative count search forwards
 * Default count 1.
 *
 * Command:  reverse-incremental-search
 * Bound to: C-r   FN2
 *
 * Return TRUE if not aborted
 *	  FALSE on abort (C-g), will return "." to mark 0 and print an error.
 */
int	risearch(int f, Cnt n)
{
    int		ret;
    int		Dirn = n >= 0 ? REVERSE : FORWARD;

    /* remember the initial . on entry: */
    pnmark0();

    // Make sure the search doesn't match where we already are:

    moveBack(1);			// Back up a character

    /* Do the search.
     * On fail isearch() returns to mark 0, says fail.
     */
    if((ret = isearch(Dirn)) == TRUE) {	// Call ISearch forwards
					// If error in search:
	if(kbdmode != PLAY)
		mlerase();		// If happy, just erase the cmd line
	pushNmark();
    }

    return(ret);
}

/* Again, but for the forward direction.
 *
 * A count >= 0 - search forwards, negative count search reverse
 * Default count 1.
 *
 * Command:  incremental-search
 * Bound to: C-x /   FN1
 *
 * Return TRUE if not aborted
 *	  FALSE on abort (C-g), will return "." to mark 0 and print an error.
 */
int	fisearch(int f, Cnt n)
{
    int		ret;
    int		Dirn = n >= 0 ? FORWARD : REVERSE;

    /* remember the initial . on entry: */
    pnmark0();

    /* Do the search.
     * On fail isearch() returns to mark 0, says fail.
     */
    if((ret = isearch(Dirn)) == TRUE) {	// Call ISearch forwards
					// If error in search
	if(kbdmode != PLAY)
		mlerase();		// If happy, just erase the cmd line
	pushNmark();
    }

    return(ret);
}

/* Do an incremental search forwards & backwards.
 *
 * In general, this works similarly to the older micro-emacs search function,
 * except that the search happens as each character is typed, with the screen
 * and cursor updated with each new search character.
 *
 * While searching forward, each successive character will leave the cursor
 * at the end of the entire matched string.  Typing a Control-S or Control-X
 * will cause the next occurrence of the string to be searched for (where the
 * next occurrence does NOT overlap the current occurrence).  A Control-R will
 * change to a backwards search, META will terminate the search and Control-G
 * will abort the search.  Rubout will back up to the previous match of the
 * string, or if the starting point is reached first, it will delete the
 * last character from the search string.
 *
 * While searching backward, each successive character will leave the cursor
 * at the beginning of the matched string.  Typing a Control-R will search
 * backward for the next occurrence of the string.  Control-S or Control-X
 * will revert the search to the forward direction.  In general, the reverse
 * incremental search is just like the forward incremental search inverted.
 *
 * In all cases, if the search fails, the user will be beeped, and the search
 * will stall until the pattern string is edited back into something that
 * exists (or until the search is aborted).
 *
 * History: "." is the last search string. Move "." up/down in response to up/down arrows.
 * Update history as we go. Also save search options to the line's notes.
 *
 * Mark 0 should be set to "." before the search started.
 *
 * This needs long term LINE* values - but these are not allowed since they
 * could be trashed by buffer re-read. So use numeric marks 11 & 12 as they
 * will be preserved.
 *
 * Lots of complication introduced: move-window-up/down can move the Last
 * Found Place (LFP) out of the visible window, so "." sits on the top/bottom
 * line. The complication is when the move is reversed by the user - reattaching
 * "." to the line when it becomes visible again and putting "." in the
 * right line column.
 *
 * Direction	of search: FORWARD or REVERSE
 *
 * Return TRUE if not aborted
 *	  FALSE on abort (C-g), will return "." to mark 0 and print an error.
 */
int	isearch(int Direction)
{
    int		status;		/* Search status */
    KeyCode	c;		/* current input character */
    int		init_direction;	// The initial search direction
    int		keepCFCPCN;	// Preserve CFCPCN in lastflag to preserve w_leftoffset
    int		retVal = TRUE;	// What we return
    TEXT	fakeTEXT;

    int		lfpAbove = FALSE;	// "." on top    window line or above
    int		lfpBelow = FALSE;	// "." on bottom window line or below

    short	SearchOpts = 0;	// Options set for search

    /* Has the search string been saved in history in bmlSrchHistp ?
     * 0 - no
     * 1 - history line added
     * Allow search using history without adding to history,
     * getting search text from b_dotp, only create a new line if new
     * search text is entered.
     */
    int		histSaved = 0;

    NMARK*	lastNn = &curwp->w_nmarks[10];	// last non new search character place

    /* Somewhere to keep the place that the scanner found.
     * Needed as the user can use move-window-up/down/left/right
     * which can change w_dotp/w_doto to other lines/columns.
     * So restore to these before we do another scan else we risk
     * not finding something or finding the same thing twice.
     *
     * Long term, so use a numeric mark.
     */
    NMARK*	foundPlace = &curwp->w_nmarks[11];

    /* Initialize starting conditions */
    canSetEq_i = TRUE;
    canSetSpW  = TRUE;
    cmd_reexecute = -1;			/* We're not re-executing (yet?)      */
    tsetused(isearchKeys, 0);		// Clear the search keys buffer

    if(bmlSrchHistp->b_dotp)		// Known history ?
	tcpyl(searchPat_e, bmlSrchHistp->b_dotp);

    // Stash a copy of the initial string in case we need to start over
    tcpyt(searchRep, searchPat_e);

    l_changeNmark(lastNn, curwp->w_dotp, curwp->w_doto);	// Save the current place
    init_direction = Direction;		/* Save the initial search direction  */

    SearchLast = TRUE;
    RestoreSearchOpts(0);			// Search options start afresh
    CheckCaselessSearch();

    /* This is a good place to start a re-execution: */

start_over:

    status = TRUE;				/* Assume everything's cool   */

    /* Get the first character in the pattern.  If we get an initial Control-S
     * or Control-R, re-use the old search string and find the first occurrence.
     * If there is not a previous search strings ... we can't do this.
     */

    c = get_char(0);				/* Get the first character    */

    if (((c == ISI_FORWARD) ||
	 (c == ISI_REVERSE) ||
	 (c == IS_VMSFORW)) &&			/* Reuse old search string?   */
	 tused(searchPat_e) != 0)		/* If there is one	      */
    {
	/* Ignore warnings, here & below, about uninitialised col */
	if(kbdmode != PLAY)
	    mlSappend(ttext(searchPat_e));

	RestoreSearchOpts(lnotes(bmlSrchHistp->b_dotp));
	CheckCaselessSearch();

	if(c == ISI_REVERSE) {			// forward search ?
	    Direction = REVERSE;
	    moveBack(1);			// Be defensive about EOB
	} else
	    Direction = FORWARD;

	status = scanmore(searchPat, Direction);// Do the search
	SearchLast = FALSE;

	c = get_char(0);			// Get another character
    }

    // Place of first match or "." before search moves it
    l_changeNmark(foundPlace, curwp->w_dotp, curwp->w_doto);	// Save the current place
    keepCFCPCN = 0;		// w_leftoffset may need to be recalculated
    lastflag = 0;
    curgoalSet();		// Column that has "." - not same as byte or character offset
    lastflag = CFMARGIN0;	// Stop it being changed

    lfpBelow = lfpAbove = FALSE;	// After scan & at start "." must be visible

    /* Top of the per character loop */

    for (;;)					/* ISearch per character loop */
    {
	/* Check for special characters first:
	 * Most cases here change the search.
	 */

	if (c == CharEscape)			/* Want to quit searching?    */
	    goto ret;				/* Quit searching now	      */

	switch (c)				/* dispatch on the input char */
	{
	  case ISI_ABORT:			/* abort search request - abortc */

	    retmark0(FALSE);			// Return to entry point
	    curwp->w_flag |= WFMOVE;		// Say we've moved
	    update(FALSE);			// And force an update
	    mlwrite("%WSearch failed");		// Say we died
	    retVal = FALSE;
	    goto ret;				// Quit searching again

	  case ISI_FORWARD:
	  case ISI_REVERSE:
	  case IS_VMSFORW:			/* of either flavor - C-x     */
	    if(c == ISI_REVERSE)		/* If reverse search  */
		Direction = REVERSE;
	    else				/* Otherwise, 		      */
		Direction = FORWARD;

	    // Don't search if we don't have a search string:
	    if(tused(searchPat) != 0)
	    {
		    // Restore "." to as last search
		    curwp->w_dotp = foundPlace->n_pline;
		    curwp->w_doto = foundPlace->n_oline;

		    // If we matched, note where this was
		    l_changeNmark(lastNn, curwp->w_dotp, curwp->w_doto);	// Save the current place

		    status = scanmore(searchPat, Direction);	/* Start the search again     */
		    SearchLast = FALSE;

		    // "." is somewhere - remember where
		    keepCFCPCN = 0;		// w_leftoffset may need to be recalculated
		    l_changeNmark(foundPlace, curwp->w_dotp, curwp->w_doto);	// Save the current place
		    lastflag = 0;
		    curgoalSet();
		    lastflag = CFMARGIN0;
		    lfpBelow = lfpAbove = FALSE;	// After scan "." must be visible
	    } else {
		// Remove from replay a 'search next' command if there is nothing to search for
		tsetused(isearchKeys, tused(isearchKeys) - (LineLn)sizeof(KeyCode));
	    }

	    c = get_char(0);			/* Get the next char	      */

	    continue;				/* Go continue with the search*/

	  case '\r':
	    c = '\n';
	    break;

	  case ISI_QUOTE:			/* Quote character	      */
	  case IS_VMSQUOTE:			/*  of either variety	      */
	    c = get_char(1);			/* Get the next char	      */
	    if(gmode & MDUTF8 ? !utf8_isvalid(c): !IsChar(c))	/* Reprocess specials */
		continue;
	    break;

	  case '\t':				/* Generically allowed	      */
	  case '\n':				/*  controlled characters     */
	    break;				/* Make sure we use it	      */

	  case ISI_RUBOUT:			/* Back-space/rubout	      */
	    // Remove the last key & also the rubout
	    tsetused(isearchKeys, tused(isearchKeys) - 2 * (LineLn)sizeof(UniChar));

	    retmark0(FALSE);			/* Return to entry point      */

	    // Nothing left - ie cleared out search commands ?
	    if(tused(isearchKeys) <= 0)
		goto ret;			// Empty, just exit

	    Direction = init_direction;		/* Reset the search direction */
	    tsetempty(searchPat_e);		// Empty search string
	    cmd_reexecute = 0;			/* Start the whole mess over  */
	    goto start_over;			/* Let it take care of itself */

	  /* This is here because hitting ESC to quit searching used to 
	   * meta the next command too.
	   */
	  case IS_QUIT:				/* <ESC> Want to quit searching?    */
	    goto ret;				/* Quit searching now	      */

	  case ISI_UPLINE:	/* Previous line */
	  case ISI_DOWNLINE:	/* Next line */
		/* If there is a history buffer that we should use and there is
		 * something in it (next after head is not head).
		 * Move forwards/back from the mark.
		 * If the mark is not set - set it to the header line
		 * Copy in the text.
		 */
		if(lforw(bmlSrchHistp->b_linep) != bmlSrchHistp->b_linep) {
			// Get the line in the direction requested
			LINE*	nxt = (c == ISI_DOWNLINE) ?
				lforw(bmlSrchHistp->b_dotp) :
				lback(bmlSrchHistp->b_dotp);

			if(nxt == bmlSrchHistp->b_linep) {
				mlwrite("%W""%BEnd of search history");
				repromptAfterTime = 1;	// Clear msg after a bit
			} else {
				SearchLast = FALSE;	// Search string changed
				bmlSrchHistp->b_dotp = nxt;
			}

			int	len = (int)lused(bmlSrchHistp->b_dotp);
			len = NPAT > len ? len : NPAT;	// Don't copy beyond buffer size
			tcpym(searchPat_e, ltext(bmlSrchHistp->b_dotp), len);
			RestoreSearchOpts(lnotes(bmlSrchHistp->b_dotp));
			CheckCaselessSearch();
		} else {
			mlwrite("%W""%BNo search history");
			repromptAfterTime = 1;	// Clear msg after a bit
		}

		c = get_char(0);		/* Get the next char	*/

		continue;

	  // Allow the user to adjust the view of the window onto the buffer

	  case ISI_MOVE_DOWN:	// move-window-down - top line disappears, replaced by second
		
		lastflag = CFMARGIN0;	// Stop w_goalCol being messed with
		mvdnwind(TRUE, 1);
		repromptAfterTime = 1;	// Might be End message

		if(foundPlace->n_pline == WinBottomLine(curwp)) {
			curwp->w_dotp = foundPlace->n_pline;
			curwp->w_doto = foundPlace->n_oline;	// OK does not need checking
			lfpBelow = 0;	// dont need work below
		}
		if(foundPlace->n_pline == lback(curwp->w_topLinep))
			lfpAbove = 1;
		tsetused(isearchKeys, tused(isearchKeys) - (LineLn)sizeof(KeyCode));	// Not in replay
		goto next_char;

	  case ISI_MOVE_UP:	// move-window-up - top line becomes second line
		lastflag = CFMARGIN0;	// Stop w_goalCol being messed with
		mvupwind(TRUE, 1);
		repromptAfterTime = 1;	// Might be Top message

		if(lback(foundPlace->n_pline) == WinBottomLine(curwp))
			lfpBelow = 1;

		if(foundPlace->n_pline == lforw(curwp->w_topLinep))
			lfpAbove = 0;

		if(foundPlace->n_pline == curwp->w_topLinep) {
			curwp->w_dotp = foundPlace->n_pline;
			curwp->w_doto = foundPlace->n_oline;	// OK does not need checking
			lfpAbove = 0;	// dont need work below
		}
		tsetused(isearchKeys, tused(isearchKeys) - (LineLn)sizeof(KeyCode));	// Not in replay
		goto next_char;

	  case ISI_MOVE_LEFT:	// move-window-left
		moveLeftWind(TRUE, 1);
		keepCFCPCN = 1;	// Preserve w_leftoffset that this does
		tsetused(isearchKeys, tused(isearchKeys) - (LineLn)sizeof(KeyCode));	// Not in replay
		goto next_char;

	  case ISI_MOVE_RIGHT:	// move-window-right
		moveRightWind(TRUE, 1);
		keepCFCPCN = 1;	// Preserve w_leftoffset that this does
		tsetused(isearchKeys, tused(isearchKeys) - (LineLn)sizeof(KeyCode));	// Not in replay
		goto next_char;

	  case ChrToCtrl('o'):	// Set search option
		lsetNotes(bmlSrchHistp->b_dotp, EditSearchOpts(0));
		tsetused(isearchKeys, tused(isearchKeys) - (LineLn)sizeof(KeyCode));	// Not in replay
		goto next_char;

	  /* Presumably a quasi-normal character comes here */

	  default:				/* All other chars	      */
		/* Is it char & printable?    */
	    if(c < ' ' || (gmode & MDUTF8 ? !utf8_isvalid(c): !IsChar(c)))
	    {					/* Nope.		      */
		TTungetc(c);			/* So that the command/... is then actioned */
		goto ret;			/* And return the last status */
	    }
	}  /* Switch */

	/* I guess we got something to search for, so search for it	      */

	/* A search character has been typed.
	 * Clear out the last search string if that is what we were going to use
	 */
	if(SearchLast) {
		SearchLast = FALSE;
		tsetempty(searchPat_e);
	}

	tcatc(searchPat_e, c);	// Append to search pattern

	/* Maintain history.
	 * Add a new history line if we have not created history
	 * this time that we have searched. Force search history "."
	 * to this new last line - which makes it the default search
	 * for the next search.
	 * Any changes to search options would have been recorded in the
	 * current line - but we want them on the new history line.
	 */
	SearchOpts = lnotes(bmlSrchHistp->b_dotp);
	if(histSaved == 0 &&
	   addHistoryText(bmlSrchHistp, tfakem(&fakeTEXT, ttext(searchPat_e), tused(searchPat_e)), AHL_NO_DUP) == 1) {
		histSaved = 1;	// Line added, not duplicate

		bmlSrchHistp->b_dotp = lback(bmlSrchHistp->b_linep);
		bmlSrchHistp->b_doto = 0;
		lsetNotes(bmlSrchHistp->b_dotp, SearchOpts);
	} else {
		// Write the new (now longer) search string to the last line
		lcpyt(lback(bmlSrchHistp->b_linep), searchPat_e);
	}

	CheckCaselessSearch();
	mlSappend(utf8_encode(c));		// Show just entered search char on mode line

	if (!status)				/* If we lost last time	      */
	    TTbeep();				/* Beep again		      */
	else {					/* Otherwise, we must have won*/
	    /* We need to find the character just entered and possibly downcased
	     * by CheckCaselessSearch()
	     * It might be multibyte.
	     */
	    BYTE* nc = strchr(ttext(searchPat), '\0') - 1;
	    if(curbp->b_mode & MDUTF8)
		UTF8_BACK_FIND_START(nc);

	    keepCFCPCN = 0;	// w_leftoffset may need to be recalculated
	    curwp->w_dotp = foundPlace->n_pline;
	    curwp->w_doto = foundPlace->n_oline;

	    if((status = checknext(nc, searchPat, Direction)) == 0) { /* See if match */
		// It did not match - ie not match at ".", try scanning further afield
		LINE*	saveP = (LINE*)NULL;	// Initialise to stop compiler warning
		LineLn	saveO = 0;		// which it ought to realise is not needed

		if(Direction == REVERSE) {
			/* Reverse search:
			 * Reset "." to the last position where we did not add
			 * something new to the search string.
			 * Otherwise we dont find 'ss' by searching for 's' and
			 * then another 's'.
			 */
			saveP = curwp->w_dotp;
			saveO = curwp->w_doto;
			curwp->w_dotp = lastNn->n_pline;
			curwp->w_doto = lastNn->n_oline;
		}

		status = scanmore(searchPat, Direction);	// or find the next match

		/* Reset "." to where it was before we fiddled with it just above.
		 * The point is that if we leave a fiddled ".", then the cursor jumps back
		 * down to where we restarted the search from -- this is not where the user
		 * expects it to be, the only reason that we moved it was to find overlapping
		 * strings.
		 */
		if(status == FALSE && Direction == REVERSE) {
			curwp->w_dotp = saveP;
			curwp->w_doto = saveO;
		}
	    }

	    l_changeNmark(foundPlace, curwp->w_dotp, curwp->w_doto);	// Where scan took us to
	    lastflag = 0;
	    curgoalSet();
	    lastflag = CFMARGIN0;
	    lfpBelow = lfpAbove = FALSE;	// After scan "." must be visible
	}

next_char:

	if(lfpAbove) {
		// LFP on top line or above it. Nail it to the top line
		curwp->w_dotp = curwp->w_topLinep;
		curwp->w_doto = getgoal(curwp->w_topLinep);
	}
	if(lfpBelow) {
		// LFP on bottom line or below it. Nail it to the bottom line
		LINE*	bottomLine = WinBottomLine(curwp);
		if(bottomLine == NULL)
			bottomLine = curbp->b_linep;
		curwp->w_dotp = bottomLine;
		curwp->w_doto = getgoal(bottomLine);
	}
	if(keepCFCPCN)
		lastflag |= CFMARGIN0;

	c = get_char(0);

    } /* for {;;} */

// Come here to return retVal
// First: clear down foundPlace & lastNn
ret:
    l_changeNmark(foundPlace, NULL, 0);
    l_changeNmark(lastNn, NULL, 0);

    return(retVal);
}

/* Trivial routine to insure that the next character in the search string is
 * still true to whatever we're pointing to in the buffer.  This routine will
 * not attempt to move the "point" if the match fails, although it will 
 * implicitly move the "point" if we're forward searching, and find a match,
 * since that's the way forward isearch works.
 *
 * Complicated a bit: if eq() says that the current search character
 * might match another buffer character - try and do so.
 *
 * If the compare fails, we return FALSE and assume the caller will call
 * scanmore or something.
 *
 * Check next character in search string
 * chr		Address of next char to look for
 * patrn	The entire search string (incl chr)
 * dir		Search direction - FORWARD or REVERSE
 *
 * Return: TODO
 */
int	checknext(BYTE* chr, TEXT* patrn, int dir)
{
    LINE*	curline;		// current line during scan
    LineLn	curoff;			// position within current line
    BYTE*	buffp;			// character at current position
    int		status;			// how well things go
    int		statusLast;

    /* setup the local scan pointer to current "." */

    curline = curwp->w_dotp;		// Get the current line structure
    curoff  = curwp->w_doto;		// Get the offset within that line

    // Used to immediately set statusLast
    // We have not matched anything (yet)
    status = EQ_DIFFER;

    if(dir == FORWARD)				// If searching forward
    {
	do {
	    statusLast = status;		// last status

	    if (curoff == lused(curline))	// If at end of line
	    {
		curline = lforw(curline);	// Skip to the next line
		if(curline == curbp->b_abotline)
		    return(FALSE);		// Abort if at end of buffer
		curoff = 0;			// Start at the beginning of the line
		buffp = NewlineString;		// And say the next char is NL
	    } else {
		buffp = lgetcaddr(curline, curoff); // Get the next char
		curoff += (curbp->b_mode & MDUTF8) ? CHAR_LEN(*buffp) : 1;
	    }

	    // Match one or, maybe, more ?
	    if ((status = eq(buffp, chr)) != EQ_DIFFER)
	    {
		curwp->w_dotp = curline;	// Yes, set the buffer's point
		curwp->w_doto = curoff;		//  to the matched character
		curwp->w_flag |= WFHARD;	// Say that we've moved
	    }

	} while(status == EQ_MORE);

	// If we matched and went back to try another
	// then we matched something - return thus
	if(statusLast == EQ_MORE)
		status = EQ_MATCH;

	return(status);			// And return the status
    } else				// Else, if reverse search:
	return(match_pat(patrn));	// See if we're in the right place

}

/* This hack will search for the next occurrence of patrn in the buffer, either
 * forward or backward.
 *
 * If we can't find any more matches, "point" is left where it was before.
 *
 * If we do find a match, "point" will be at the end of the matched string for
 * forward searches and at the beginning of the matched string for reverse
 * searches.
 *
 * patrn	string to scan for
 * dir		direction to search - FORWARD or REVERSE
 *
 * Return:
 *	TRUE  Found the next patrn
 *	FALSE Did not find
 */
int	scanmore(TEXT* patrn, int dir)
{
	int	sts;			/* search status	      */

	if(kbdmode != PLAY)
		mlSappend("...");

	int	leaveDot = dir == REVERSE ? PTBEG : PTEND;

	sts = scanner(patrn, dir, leaveDot);	/* Nope. Go forward   */

	if(kbdmode != PLAY)
		mlSappend((BYTE*)NULL);

	if (!sts)
	{
		TTbeep();		/* Feep if search fails       */
		msglFlush();		/* see that the feep feeps    */
	}

	return(sts);			/* else, don't even try	      */
}

/*
 * The following is a worker subroutine used by the reverse search.  It
 * compares the pattern string with the characters at "." for equality. If
 * any characters mismatch, it will return FALSE.
 *
 * This isn't used for forward searches, because forward searches leave "."
 * at the end of the search string (instead of in front), so all that needs to
 * be done is match the last char input.
 *
 * patrn	String to match to buffer
 *
 * Return:
 *	TRUE  matched
 *	FALSE no match
 */
int	match_pat(const TEXT* patrn)
{
    LineLn	i;			/* Generic loop index/offset	      */
    BYTE*	buffp;			/* Address character at current position */
    LINE*	curline;		/* current line during scan	      */
    LineLn	curoff;			/* position within current line	      */
    int		utf8 = curbp->b_mode & MDUTF8;	// Current buffer UTF-8 mode

    /* setup the local scan pointer to current "." */

    curline = curwp->w_dotp;		/* Get the current line structure     */
    curoff  = curwp->w_doto;		/* Get the offset within that line    */

    /* top of per character compare loop: */

    for (i = 0; i < (LineLn)strlen(ttext(patrn)); )	/* Loop for all characters in patrn   */
    {
	if (curoff == lused(curline))	/* If at end of line		      */
	{
	    curline = lforw(curline);	/* Skip to the next line	      */
	    curoff = 0;			/* Start at the beginning of the line */
	    if (curline == curbp->b_abotline)
		return(FALSE);		/* Abort if at end of buffer	      */
	    buffp = NewlineString;	/* And say the next char is NL	      */
	} else {
	    buffp = lgetcaddr(curline, curoff); /* Get the next char	      */
	    curoff += utf8 ? CHAR_LEN(*buffp) : 1;
	}

	if (!eq(buffp, tgetcaddr(patrn, i)))	/* Is it what we're looking for?      */
	    return(FALSE);		/* Nope, just punt it then	      */

	i += utf8 ? CHAR_LEN(tgetch(patrn, i)) : 1;
    }

    return(TRUE);			/* Everything matched? Let's celebrate*/
}

/* Routine to prompt for I-Search string.
 * This isn't called if we are executing a keyboard macro.
 *
 * First prompt show the previous search string in [],
 * after they have typed anything show the search string at the end.
 * Use global: SearchLast
 */
void	promptpattern(void)
{
	if(SearchLast && tused(searchRep) != 0)
		mlwrite("%h(Opts C-o) ISearch: [%t]<META>: ", searchRep);
	else
		mlwrite("%h(Opts C-o) ISearch: <META>: %t", searchPat_e);

	if(gmode & MDUTF8) {
		tsetUtf8(searchPat);
		lsetUtf8(wmessp->w_dotp);
	}
}

/* Routine to get the next character from the input stream.
 *
 * Characters either come from the real terminal, or from LINE isearchKeys.
 * The point of isearchKeys is that if a RUBOUT is pressed then the way to get
 * to the last match is to return to the start and replay all search commands.
 *
 * raw	get the character unmunged
 *
 * If repromptAfterTime is TRUE refresh the search prompt after ddelay seconds.
 * Then clear repromptAfterTime.
 * The point is that mvupwind() & mvupwind() can obliterate the search prompt
 * with a top/end message.
 * Use promptpattern() to rewrite the prompt.
 *
 * Return: The next character
 */
KeyCode	get_char(int raw)
{
	KeyCode	c;			// A place to get a character
	long	timeout = 0;		// To redraw prompt

	// See if we're re-executing:
	if(cmd_reexecute >= 0) {	// Is there an offset ?

		if(cmd_reexecute < tused(isearchKeys)) {
			c = *(KeyCode*)tgetcaddr(isearchKeys, cmd_reexecute);
			cmd_reexecute += (int)sizeof(c);

			return(c);
		}
		cmd_reexecute = -1;	// No longer executing
	}

	// We're not re-executing (or aren't any more).  Try for a real char

	/* Set a timeout if needs be.
	 * If no timeout - ensure that the pattern shown is up to date.
	 */
	if(repromptAfterTime)
		timeout = ddelay * MILLION;
	else
		promptpattern();

	repromptAfterTime = 0;

	while(1) {
		c = raw ?
			tgetc(MGC_NONE, timeout) :
			ReverseMap(tgetc(MGC_MAP, timeout)); // Get the next character

		if(c != KEY_TIMEOUT)
			break;

		promptpattern();

		timeout = 0;			// Wait forever this time
	}


	// Append the character just read
	if( ! tcatm(isearchKeys, (BYTE*)&c, sizeof(c))) {
		mlwrite("%E? command too long"); // Complain loudly and bitterly
		return(CharEscape);		// And force a quit
	}

	return (c);				// Return the character
}

#else
isearch()
{
}
#endif

/* end */
