/*
 * ADDW SCCS: @(#)line.c	1.125 10/14/20 17:02:42
 */

#define	TRACE	0

#if	TRACE
#define	FPRINTF(x)	fprintf x
#else
#define	FPRINTF(x)
#endif

/* The functions in this file are a general set of line management utilities.
 * They are the only routines that touch the text. They also touch the buffer
 * and window structures, to make sure that the necessary updating gets done.
 * There are routines in this file that handle the kill buffer too. It isn't
 * here for any good reason.
 *
 * Note that this code only updates the dot and mark values in the window list.
 * Since all the code acts on the current window, the buffer that we are
 * editing must be being displayed, which means that "b_nwnd" is non zero,
 * which means that the dot and mark values in the buffer headers are nonsense.
 */

#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"

#if	SHOWMODREL
char	linevers[] = "@(#)line.c	1.125 10/14/20 17:02:42";
#endif


/* Naming of the following few functions:
 * DoprS()
 * D - Destination, they all copy to TEXTs, so start 't'
 *	Used to be LINEs so 'l'
 * opr - operation:
 *	cpy - copy, replace previous content
 *	cat - catenate, ie append
 *	ins - insert at some offset
 *	del - delete at some offset
 * S - Source, where the text comes from:
 *	l - LINE
 *	t - TEXT
 *	s - string, NUL terminated
 *	m - memory, a length must be supplied
 *	c - character, Unicode understood
 *	n - number of bytes
 * eg:
 * lcatm() - append bytes to a line
 *
 * The destination TEXT will be extended as needed - this may result in
 * the t_text being moved.
 *
 * These return TRUE on success, FALSE means out of memory on line extension.
 * You may, however, ignore the return value when doing a set of operations
 * one LINEs, it will be safe - as long as you don't do a line length
 * calculations based on what you think should have happened.
 *
 * If there was a memory allocation failure textOOM will be set TRUE,
 * so do check that at the end of the set of operations.
 *
 * Beware: these do not check if things like w_doto are affected by this,
 * so do NOT use if this is possible - ie anything in a displayed buffer.
 *
 * Neither do these simple routines check & adjust cursor or mark positions.
 */

/* Append len bytes from src into TEXT tdest.
 * A trailing NUL on tdest will be replaced.
 *
 * tdest's UTF8 status will be unset if the new bytes are not good UTF8.
 * These simple routines just unset UTF8 status, whereas insStrLine()
 * will replace a bad sequence with utf8_replace.
 *
 * Return:
 *	TRUE  on success.
 *	FALSE means out of memory.
 */
int	tcatm(TEXT* tdest, const BYTE* src, LineLn len)
{
	if(textMoreBytes(tdest, tused(tdest) + len) == FALSE)
		return(FALSE);

	memcpy(tgetcaddr(tdest, tused(tdest)), src, (size_t)len);
	tsetused(tdest, tused(tdest) + len);

	// Does the new stuff invalidate UTF8 status ?
	if(tisUtf8(tdest) && !utf8_chkbuf(src, (size_t)len))
		tunsetUtf8(tdest);

	return(TRUE);
}

/* Insert len bytes from src at offset in TEXT ldest.
 *
 * offset must be somewhere within the existing text of tdest; that
 * includes the start/end of the text.
 *
 * tdest's UTF8 status will be unset if the new bytes are not good UTF8.
 *
 * Return TRUE on success.
 *	  FALSE means out of memory or offset is not within tdest.
 */
int	tinsm(TEXT* tdest, LineLn offset, const BYTE* src, LineLn len)
{
	// Sanity check:
	if(offset < 0 || offset > tused(tdest))
		return(FALSE);

	if(textMoreBytes(tdest, tused(tdest) + len) == FALSE)
		return(FALSE);

	// Copy down, make space (if not inserting at the end):
	if(tused(tdest) > offset)
		memmove(tgetcaddr(tdest, offset + len), tgetcaddr(tdest, offset), (size_t)(tused(tdest) - offset));
	memcpy(tgetcaddr(tdest, offset), src, (size_t)len);
	tsetused(tdest, tused(tdest) + len);

	// Does the new stuff invalidate UTF8 status ?
	// Need to check more than what was inserted, in case we inserted into
	// the middle of a UTF-8 char
	if(tisUtf8(tdest) && !utf8_chkbuf(ttext(tdest), (size_t)tused(tdest)))
		tunsetUtf8(tdest);

	return(TRUE);
}

/* Insert TEXT src at offset in TEXT tdest.
 *
 * Return:
 *	TRUE  on success.
 *	FALSE means out of memory or offset not in tdest.
 */
int	tinst(TEXT* tdest, LineLn offset, const TEXT* src)
{
	return(tinsm(tdest, offset, ttext(src), tused(src)));
}

/* Insert a NUL terminated string from src at offset in TEXT tdest.
 *
 * Return:
 *	TRUE  on success.
 *	FALSE means out of memory or offset not in tdest.
 */
int	tinss(TEXT* tdest, LineLn offset, const BYTE* src)
{
	return(tinsm(tdest, offset, src, (LineLn)strlen(src)));
}

/* Copy len bytes from src to TEXT tdest.
 * Line will have length len, any NUL will be preserved.
 *
 * Return:
 *	TRUE  on success.
 *	FALSE means out of memory.
 */
int	tcpym(TEXT* tdest, const BYTE* src, LineLn len)
{
	tsetused(tdest, 0);
	return(tcatm(tdest, src, len));
}

/* Copy LINE lsrc to line ldest, replacing its contents.
 * ldest's NUL flag is preserved, if lsrc has the NUL flag it is copied to ldest.
 * ldest's UTF8 state is copied from that of lsrc.
 * 
 * Return:
 *	TRUE  on success.
 *	FALSE means out of memory.
 */
int	tcpyt(TEXT* tdest, const TEXT* tsrc)
{
	tsetused(tdest, 0);

	if(tisNUL(tsrc) && !tsetNUL(tdest))
		return(FALSE);

	if(tisUtf8(tsrc))
		tsetUtf8(tdest);	// This will cause the UTF8 status to be checked

	return(tcatm(tdest, ttext(tsrc), tused(tsrc)));
}

/* Append TEXT tsrc to TEXT tdest.
 * tdest's NUL flag is preserved, if tsrc has the NUL flag it is copied to tdest.
 *
 * tdest's UTF8 status might be unset if the new text is not good UTF8.
 * 
 * Return:
 *	TRUE  on success.
 *	FALSE means out of memory.
 */
int	tcatt(TEXT* tdest, const TEXT* tsrc)
{
	if(tisNUL(tsrc) && !tsetNUL(tdest))
		return(FALSE);

	return(tcatm(tdest, ttext(tsrc), tused(tsrc)));
}

/* Copy NUL terminated string src into TEXT tdest.
 * A trailing NUL on tdest will be replaced.
 *
 * Return:
 *	TRUE  on success.
 *	FALSE means out of memory.
 */
int	tcpys(TEXT* tdest, const BYTE* src)
{
	return(tcpym(tdest, src, (LineLn)strlen(src)));
}

/* Copy character c to tdest, replacing its contents.
 *
 * Return:
 *	TRUE  on success.
 *	FALSE means out of memory.
 */
int	tcpyc(TEXT* tdest, UniChar c)
{
	tsetused(tdest, 0);
	return(tcatc(tdest, c));
}

/* Append NUL terminated string src into TEXT tdest.
 * A trailing NUL on tdest will be replaced.
 *
 * Return:
 *	TRUE  on success.
 *	FALSE means out of memory.
 */
int	tcats(TEXT* tdest, const BYTE* src)
{
	return(tcatm(tdest, src, (LineLn)strlen(src)));
}

/* Append character c to tdest.
 *
 * Return:
 *	TRUE  on success.
 *	FALSE means out of memory.
 */
int	tcatc(TEXT* tdest, UniChar c)
{
	BYTE*	str = utf8_encode(c);

	return(tcatm(tdest, str, CHAR_LEN(*str)));
}

/* Insert character c into tdest at offset.
 *
 * Return:
 *	TRUE  on success.
 *	FALSE means out of memory or offset not in tdest.
 */
int	tinsc(TEXT* tdest, LineLn offset, UniChar c)
{
	BYTE*	str = utf8_encode(c);

	return(tinsm(tdest, offset, str, CHAR_LEN(*str)));
}

/* Delete len bytes at offset (bytes) in text.
 *
 * line's UTF8 status will be unset if the new bytes are not good UTF8.
 *
 * Return TRUE on success.
 *	  FALSE if trying to delete bytes that are not within the TEXT.
 */
int	tdeln(TEXT* text, LineLn offset, LineLn len)
{
	if(offset + len > tused(text))
		return(FALSE);

	memmove(tgetcaddr(text, offset), tgetcaddr(text, offset) + len, (size_t)(tused(text) - offset - len));
	tsetused(text, tused(text) - len);

	if(tisUtf8(text) && !utf8_chkbuf(ttext(text), (size_t)tused(text)))
		tunsetUtf8(text);

	return(TRUE);
}

/* Trim spaces, start/end.
 */
void	ttrim(TEXT* t)
{
	while(tused(t) > 0 && tgetch(t, tused(t) - 1) == ' ')
		tsetused(t, tused(t) - 1);

	while(tused(t) > 0 && tgetch(t, 0) == ' ')
		tdeln(t, 0, 1);
}

/* Compare the contents of TEXTs t1 & t2
 *
 * Return 0 if they are the same
 *	  > 0 if t1 sorts higher than t2
 *	  < 0 if t1 sorts lower  than t2
 * This is the same as strcmp()/memcmp()
 * Sorting is a binary sorting - no UTF-8 nonsense here!
 *
 * If they are different lengths & compare the same in the length
 * of the shorter one: then the longer one is deemed greater.
 */
int	textcmp(const TEXT* t1, const TEXT* t2)
{
	// Length of shortest
	LineLn	len = tused(t1) > tused(t2) ? tused(t2) : tused(t1);
	int	sort;

	if((sort = memcmp(ttext(t1), ttext(t2), (size_t)len)) != 0)
		return(sort);

	if(tused(t1) == tused(t2))
		return(0);

	return(tused(t1) > tused(t2) ? 1 : -1);
}

/* Make a fake LINE.
 * This is where a LINE is wanted but we don't want to go to the
 * effort of creating one and then removing it.
 * The LINE should prob be considered read only.
 *
 * ln	The LINE structure to fill, prob a stack variable
 * m	LINE's text
 * len	how long the text
 *
 * Return ln
 */
LINE*	lfakem(LINE* ln, const BYTE* m, LineLn len)
{
	memset(ln, 0, sizeof(LINE));

	ltxt(ln).t_text = (BYTE*)m;	// Prob read only

	lsetused(ln, len);
	lsetsize(ln, len);
	lsetback(ln, ln);
	lsetforw(ln, ln);

	return(ln);
}

/* Make a fake LINE.
 *
 * As for lfakem() but the memory is a NUL terminated string.
 *
 * The LINE will have the NUL flag.
 *
 * Return ln.
 */
LINE*	lfakes(LINE* ln, const BYTE* str)
{
	lfakem(ln, str, (LineLn)strlen(str));
	ltxt(ln).t_flags = L_NUL;

	return(ln);
}

/* Make a fake TEXT.
 * This is where a TEXT is wanted but we don't want to go to the
 * effort of creating one and then removing it.
 * The TEXT should prob be considered read only.
 *
 * t	The TEXT structure to fill, prob a stack variable
 * m	TEXT's text
 * len	how long the text
 *
 * Return t
 */
TEXT*	tfakem(TEXT* t, const BYTE* m, LineLn len)
{
	memset(t, 0, sizeof(TEXT));

	t->t_text = (BYTE*)m;	// Prob read only

	tsetused(t, len);
	tsetsize(t, len);

	return(t);
}

/* Make a fake TEXT.
 *
 * As for tfakem() but the memory is a NUL terminated string.
 *
 * The TEXT will have the NUL flag.
 *
 * Return t.
 */
TEXT*	tfakes(TEXT* t, const BYTE* str)
{
	tfakem(t, str, (LineLn)strlen(str));
	t->t_flags = L_NUL;

	return(t);
}

/* This routine allocates a block of memory large enough to hold a LINE
 * containing "used" characters (= bytes). The block is always rounded up a bit. Return
 * a pointer to the new block, or NULL if there isn't any memory left. Print a
 * message in the message line if no space.
 *
 * The space used is set to zero, so you need to set the length used when
 * putting something in it.
 *
 * Return a pointer to the new line
 *	  NULL on out of memory and textOOM will be set TRUE.
 */
LINE*	lalloc(LineLn used)
{
	register LINE*	lp;
	register LineLn	size;
	register BYTE*	text;

	/* Round up requested size */
	size = (used+NBLOCK-1) & ~(NBLOCK-1);
	if(size == 0)				/* Assume that an empty */
		size = NBLOCK;			/* line is for type-in. */

	if((lp = (LINE*)calloc(1, sizeof(LINE))) == NULL ||
	   (text = (BYTE*)malloc((size_t)size)) == (BYTE*)NULL) {
		textOOM = TRUE;

		if(lp != (LINE*)NULL)
			free((void*)lp);

		mlwrite(MsgOutOfMemoryAlloc, "line");
		return(NULL);
	}

	ltext(lp) = text;
	lsetsize(lp, size);

	return(lp);
}
TEXT*	talloc(LineLn used)
{
	TEXT*	tp;
	LineLn	size;
	BYTE*	text;

	/* Round up requested size */
	size = (used+NBLOCK-1) & ~(NBLOCK-1);
	if(size == 0)				/* Assume that an empty */
		size = NBLOCK;			/* line is for type-in. */

	if((tp = (TEXT*)calloc(1, sizeof(TEXT))) == NULL ||
	   (text = (BYTE*)malloc((size_t)size)) == (BYTE*)NULL) {
		textOOM = TRUE;

		if(tp != (TEXT*)NULL)
			free((void*)tp);

		mlwrite(MsgOutOfMemoryAlloc, "text");
		return(NULL);
	}

	ttext(tp) = text;
	tsetsize(tp, size);
	// Don't need to set used as tp is allocated with calloc() so it is zero.

	return(tp);
}

/* Free a LINE.
 * Works regardless of the text being 'char*' or an array, well easier to
 * change.
 */
void	FreeLine(LINE* lp)
{
	free(ltext(lp));
	free(lp);
}

/* Free a TEXT.
 */
void	FreeText(TEXT* tp)
{
	free(tp->t_text);
	free(tp);
}

/* Delete line "lp". Fix all of the links that might point at it (they are
 * moved to offset 0 of the next line. Unlink the line from whatever buffer it
 * might be in. Release the memory. The buffers are updated too; the magic
 * conditions described in the above comments don't hold here.
 *
 * The line is in bp - which may be NULL for non buffer lines.
 */
void	lfree(BUFFER* bp, LINE* lp)
{
	register Window *wp;

	// Search all windows, is it: top-line, ".", mark, last-mark-place
	for(wp = wmessp; wp; wp = wp->w_wndp) {
		if(wp->w_topLinep == lp) {
			wp->w_topLinep = lforw(lp);
			wp->w_flag |= WFMODE|WFHARD|WFNWID;
		}
		if(wp->w_dotp  == lp) {
			wp->w_dotp  = lforw(lp);
			wp->w_partoff = 0;
			wp->w_doto = 0;
			wp->w_flag |= WFMODE|WFHARD|WFNWID;
		}
		if(wp->w_markp == lp) {
			wp->w_markp = lforw(lp);
			wp->w_markpartoff = 0;
			wp->w_marko = 0;
			wp->w_flag |= WFMODE|WFHARD|WFNWID;
		}
		if(wp->w_lastdismarkp == lp) {
			wp->w_lastdismarkp = lforw(lp);
			wp->w_lastdismarko = 0;
		}
	}

	l_unNmark(bp, lp);

	// Buffer: ".", mark, top-line
	if(bp != NULL) {
		if(bp->b_nwnd == 0) {
			if(bp->b_dotp  == lp) {
				bp->b_dotp = lforw(lp);
				bp->b_doto = 0;
			}
			if(bp->b_markp == lp) {
				bp->b_markp = lforw(lp);
				bp->b_marko = 0;
			}
			if(bp->b_topLinep == lp)
				bp->b_topLinep = lforw(lp);
		}

		if(lmarked(lp))
			l_unmark(bp, lp);

		// Check that it is not top/bottom of a narrow
		if(SonLineB(bp, lp))
			bp->b_btopline = lback(lp);
		if(EonLineB(bp, lp))
			bp->b_abotline = lforw(lp);
	}

	lunstitch(lp);	// Unchain & zap
}

/* Remove a line from the list of lines in a buffer and free it.
 *
 * This does no checking that the LINE is referenced anywhere, ie
 * is a w_dotp or something.
 *
 * Return pointer to the line that was before the zapped line.
 */
LINE*	lunstitch(LINE* ln)
{
	LINE*	lbak = lback(ln);

	lsetforw(lbak, lforw(ln));
	lsetback(lforw(ln), lbak);

	FreeLine(ln);

	return(lbak);
}

/* The line "lp" is marked by an alphabetic mark. Look through
 * the linked list of marks for the one that references this
 * line and zap the pointer. This is marginally worse than
 * unsewing the mark from the linked list, but it has two
 * advantages in that it (i) saves time and (ii) we can put
 * out a "meaningful" error message when we try to go to
 * this mark.
 *
 * Note that a line may have more that one mark.
 */
void	l_unmark(BUFFER* bp, LINE* lp)
{
	register AMARK* p;

	for(p = bp->b_amark; p != (AMARK*) NULL; p = p->m_next) {
		if(p->m_pline == lp)
			p->m_pline = (LINE*) NULL;
	}

	lunsetMark(lp);
}

/* line will shortly be invalid, remove any reference to it
 * in any window that is showing the BUFFER bp.
 * Also bp's numeric marks.
 */
void	l_unNmark(BUFFER* bp, LINE* line)
{
	Window*	wp;
	NMARK*	nm;

	for(wp = wmessp; wp; wp = wp->w_wndp) {
		// Window not showing the buffer bp ?
		if(wp->w_bufp != bp)
			continue;

		for(nm = wp->w_nmarks; nm < wp->w_nmarks + NUM_NUM_MARKS; nm++)
			if(nm->n_pline == line)
				nm->n_pline = (LINE*)NULL;
	}

	/* Just in case it is referenced in the buffer's copy of a
	 * window's numeric marks.
	 * bp can be NULL in cexit() tidy up.
	 */
	if(bp != (BUFFER*)NULL)
		for(nm = bp->b_nmarks; nm < bp->b_nmarks + NUM_NUM_MARKS; nm++)
			if(nm->n_pline == line)
				nm->n_pline = (LINE*)NULL;
}

/* Check if the numeric flag L_NMARK on line is still needed.
 *
 * This is because the numeric reference within the window is going away.
 * The check is needed since a line can be referenced in more than one numeric mark.
 *
 * line will be in the current buffer.
 *
 * It is important that before calling this that the reference in w_nmarks[]
 * is removed.
 */
void	l_checkNmark(LINE* line)
{
	Window*	wp;

	lunsetNMark(line);

	for(wp = wmessp; wp; wp = wp->w_wndp) {
		NMARK*	nm = wp->w_nmarks;

		for( ; nm < wp->w_nmarks + NUM_NUM_MARKS; nm++)
			if(nm->n_pline == line) {
				// Found a reference, reset the mark
				lsetNMark(line);

				// No point in looking for another reference
				return;
			}
	}
}

/* Change what a numeric mark references.
 *
 * The point of this is to do the various housekeeping things needed on
 * un/setting numeric marks.
 *
 * mark	the mark that it being set
 * nlp	the new LINE that *lpp will refer to
 * noff	the offset within nlp
 *
 * Either/both *lpp and/or nlp may be NULL.
 */
void	l_changeNmark(NMARK* mark, LINE* nlp, LineLn noff)
{
	LINE*	lp = mark->n_pline;

	// If it is current set: clear it
	if(lp != (LINE*)NULL) {
		mark->n_pline = (LINE*)NULL;
		l_checkNmark(lp);
	}

	// Set a new value - if there is one
	if(nlp != (LINE*)NULL) {
		mark->n_pline = nlp;
		mark->n_oline = noff;
		lsetNMark(nlp);
	}
}

/* Update the required flags in the current buffer and window when something is changed.
 *
 * flag	  WF??? flag to be set
 *
 * If the buffer is being displayed in more than 1 window we change EDIT to
 * HARD.
 *
 * Set WFMODE if the mode line needs to be updated (the "*" has to be set).
 */
void	lchange(int flag)
{
	if(curbp->b_nwnd != 1)			/* Ensure hard.		*/
		flag = WFHARD;
	if((curbp->b_flag&BFCHG) == 0) {	/* First change, so	*/
		flag |= WFMODE; 		/* update mode lines.	*/
		curbp->b_flag |= BFCHG;
		if(curbp->b_flag&BFWEIRD)
			mlwrite(
			  "%W""%BReminder: unwriteable file in writable directory");
	}
	upmode(curbp, flag);
}

/* Insert n spaces forward into text.
 *
 * Default count 1, must not be -ve. Handled by FuncExec()
 *
 * Command:  insert-space
 * Bound to: unbound
 *
 * Return:
 *	TRUE  if it works
 *	FALSE insert or backspace fail - TODO
 */
int	insspace(int f, Cnt n)
{
	return(linsert(n, ' ') && moveBack(n));
}

/* Insert text at the current point.
 *
 * This is more than linsl() at if notices NL and does lnewline().
 *
 * Return:
 *	TRUE  if it works
 *	FALSE insert fail, out of memory
 */
int	linTEXT(const TEXT* instr)
{
	register int	status = TRUE;
	LineLn		pos = 0;				// Where we are in ostr
	int		utf8 = (curbp->b_mode & MDUTF8) != 0;	// UTF8 enabled buffer ?

	while(status && pos < tused(instr)) {
		BYTE*	nl;	// Address of any \n
		LineLn	len;	// Length to insert this time round
		LineLn	i;

		/* If nothing in the buffer or "." on b_linep: create a new line
		 * then move back so that we are on the line just created
		 */
		if(curwp->w_dotp == curbp->b_linep)
			if((status = lnewline(0)) == FALSE || (status = moveBack(1)) == FALSE)
				break;

		// How much to copy ?
		// Up to a newline (if present) else the rest of the string
		len = (nl = memchr(tgetcaddr(instr, pos), '\n', (size_t)(tused(instr) - pos))) ?
			(nl - tgetcaddr(instr, pos)) : (tused(instr) - pos);

		// Insert that bit
		if((i = insStrLine(curwp->w_dotp, curwp->w_doto, tgetcaddr(instr, pos), len, !utf8)) == -1)
			status = FALSE;
		else
			FixOffsets(curwp->w_dotp, curwp->w_doto, i, curbp, 0);

		if(utf8)
			lsetUtf8(curwp->w_dotp);

		pos += len;

		// If there was a newline, insert that & eat the \n in instr:
		if(status && nl) {
			status = lnewline(0);
			pos++;
		}
	}

	lchange(WFHARD);

	/* Insertion error? */
	if(status != TRUE)
		mlwrite(MsgOutOfMemoryWhile, "extending line");

	return(status);
}

/* linstr -- Insert a string at the current point
 *
 * The string could be contain UTF8 encoded characters. We don't have a clue,
 * deal with that set the UTF8 flag on the line if the buffer has UTF8 mode,
 * it don't do any harm *if* the buffer has UTF8 mode,
 * Part of that is validating that they are correctly encoded.
 *
 * Return:
 *	TRUE  if it works
 *	FALSE insert fail, out of memory
 */
int	linstr(const BYTE* instr)
{
	TEXT	t;

	return(linTEXT(tfakes(&t, instr)));
}

/* Insert bytes into a TEXT at a specific offset.
 * The point about this is that the string might be UTF8 encoded
 * Contrast to linsert() which inserts a Unicode character, if that were to be
 * used then the string would need to be decoded from UTF8 and then reencoded again.
 *
 * tp		TEXT to insert into
 * offset	insert point (bytes) from the start of the TEXT - ie insert after existing text
 * text		the bytes to insert
 * textLen	how many bytes
 * u8clean	the text is known to be UTF8 clean, or it does not matter (not a UTF8 buffer)
 *
 * Existing bytes after the insert point are preserved.
 * t_used is maintained, the t_text might be reallocated.
 *
 * However: if this is not told input is u8clean, the text inserted
 * is checked for UTF8 goodness and the Unicode bad codepoint replacement
 * character (UTF8_REPLACEMENT U+FFFD) inserted if an error is found.
 *
 * To save a bit of time, if u8clean it is known that UTF8 checking does not need to
 * be done; either because text is known to be good or the buffer that contains the
 * TEXT does not have the UTF8 flag.
 *
 * Unless the caller is sure that input does not contain UTF8, it might be good
 * to set l_utf8.
 *
 * This knows nothing about windows or buffers, thus w_doto/b_doto are not updated.
 * The buffer might not be curbp and might not even be displayed.
 *
 * If the inserted text contains \n then that will be inserted as a byte in the
 * TEXT, it will not cause a line break.
 *
 * You might need to call FixOffsets() after using this.
 *
 *
 * Return the number of bytes (not characters) inserted - could be zero.
 * 	  -1 on error (out of memory)
 */
LineLn	insStrText(TEXT* tp, LineLn offset, const BYTE* text, LineLn textLen, int u8clean)
{
	LineLn	endLen = tused(tp) - offset;	// Bytes already there after the insert offset
	LineLn	inserted;			// Bytes inserted

	// Just ensure that there is enough room first.
	// If not clean, allow for possibly having to insert a bad code point
	// character - as replacement for at least one byte:
	if( !textMoreBytes(tp, tused(tp) + textLen + (u8clean ? 0 : UTF8_REPLACEMENT_LENGTH - 1)))
		return(-1);

	// If there is any text after the insert offset, move it down:
	if(endLen > 0)
		memmove(tgetcaddr(tp, offset + textLen), tgetcaddr(tp, offset), (size_t)endLen);

	if(u8clean) {
		// OK to just copy in
		memcpy(tgetcaddr(tp, offset), text, (size_t)textLen);
		tsetused(tp, offset + textLen + endLen);
		inserted = textLen;
	} else {
		// Need to check the bytes for correct UTF8 encoding

		int	whoops = 0;	// No error so far
		int	len;		// Length of MB char copying

		inserted = 0;

		// Copy bytes into t_text[offset] - incrememt offset

		// Examine what is going in
		while(textLen > 0) {
			BYTE	c = *text++;	// Next byte
			int	putRep = 0;

			// Ensure that we have enough room
			// Only need to check this if there has been a previous error
			if(whoops && !textMoreBytes(tp, tused(tp) + textLen + endLen + UTF8_MAX_BYTES))
				return(-1);

			// Just copy in non multi byte characters:
			if( !BYTE_IS_MB_PART(c)) {
				tputc(tp, offset++, c);
				textLen--;
				inserted++;
				continue;
			}

			/* Look at what we have, assume the best to start with.
			 * Check for a valid sequence.
			 */
			if((len = CHAR_LEN(c)) > 1) {
				if(textLen >= len) {
					// Enough input bytes for the sequence
					// Copy over and look at what is next
					if(utf8_chkbuf(text - 1, (size_t)len)) {
						UTF8_MOVE(tgetcaddr(tp, offset), text - 1);
						offset += len;
						inserted += len;
						tsetUtf8(tp);
					} else {
						// Bad code point, replace and eat it.
						// Insert the bad codepoint replacement and eat the rest of the line
						putRep = 1;

						// One likely cause is a good first byte which is not followed
						// by enough continuation bytes.
						// Eat the valid continuation bytes but not what follows
						int cb = len;	// What should be the # cont bytes
						len = 1;

						while(len <= cb) {
							if( !BYTE_IS_MB_CONT(text[len - 1]))
								break;
							len++;
						}
					}
					textLen -= len;
					text += len - 1;
				} else {
					// We have the start of a MB sequence but there is not enough
					// for the rest of it in the buffer that we are copying from.
					// Insert the bad codepoint replacement and eat the rest of the line
					putRep = 1;
					textLen = 0;	// Will cause immediate end of loop
					len=1;
				}
			} else {
				// Length < 1 ... rogue continuation byte probably
				// Just eat it, put a bad code point character
				putRep = 1;
				textLen--;
				len = 1;
			}

			// Something gone wrong above, insert the Unicode replacement character.
			// We are replacing len bytes with UTF8_REPLACEMENT_LENGTH, we might
			// need to move any prior stuff down the buffer.
			// This should be rare.
			if(putRep) {
				if(endLen > 0 && len < UTF8_REPLACEMENT_LENGTH)
					memmove(tgetcaddr(tp, offset + textLen + len - UTF8_REPLACEMENT_LENGTH),
						tgetcaddr(tp, offset + textLen), (size_t)endLen);

				UTF8_MOVE(tgetcaddr(tp, offset), utf8_replace);

				inserted += UTF8_REPLACEMENT_LENGTH;
				offset += UTF8_REPLACEMENT_LENGTH;
				whoops = 1;	// Force recheck of buffer size
			}
		}

		tsetused(tp, offset + endLen);	// Bytes now in the line
	}

	return(inserted);
}

/* Expand a TEXT - more bytes needed.
 * BytesNeeded is how many bytes the TEXT needs to be able to store.
 * It might be that a reallocation is not needed.
 *
 * This does NOT change the used size - this simply makes the TEXT able
 * to store BytesNeeded.
 *
 * An extra, not accounted for, and not requested in BytesNeeded, byte is
 * allocated if the TEXT has the NUL flag.
 *
 * Return TRUE if it works.
 *	  FALSE - out of memory and textOOM will be set TRUE.
 */
int	textMoreBytes(TEXT* tp, LineLn BytesNeeded)
{
	BYTE*	cp;
	LineLn	newSize;

	// Trailing NUL ?
	if(tisNUL(tp))
		BytesNeeded++;

	if(tsize(tp) >= BytesNeeded)
		return(TRUE);		// Big enough as it is

	/* Make it big enough.
	 * Allocate a bit more (NBLOCK rounded) to stop excessive reallocations.
	 */
	newSize = (BytesNeeded + NBLOCK - 1) & ~(NBLOCK - 1);
	if((cp = (BYTE*)realloc(ttext(tp), (size_t)newSize)) == NULL) {
		textOOM = TRUE;
		return(FALSE);
	}

	ttext(tp) = cp;
	tsetsize(tp, newSize - tisNUL(tp));

	return(TRUE);
}

/* Set the L_NUL flag on the text and append a NUL byte to the text in the text.
 * This might involve a reallocation.
 *
 * This is because, sometimes, a text is used with a system call that expects
 * a NUL terminated string.
 * This NUL will be preserved in the various text manipulations but does
 * not appear as part of the text length.
 *
 * Calling this more than once is benign.
 *
 * Return TRUE if OK, FALSE - out of memory.
 */
int	tsetNUL(TEXT* tp)
{
	if(tisNUL(tp))
		return(TRUE);

	if(tused(tp) >= tsize(tp) &&
	   textMoreBytes(tp, tused(tp) + 1) != TRUE)
		return(FALSE);

	// The byte at the end becomes unaccounted for in tsize():
	tsetsize(tp, tsize(tp) - 1);

	tp->t_flags |= L_NUL;	// Set L_NUL flag
	tputc(tp, tused(tp), '\0');

	return(TRUE);
}

/* Insert "n" copies of the character "c" at the current location of dot.
 *
 * In the easy case all that happens is the text is stored in the line.
 * In the hard case, the line has to be reallocated.
 *
 * When the window list is updated, take special care; I screwed it up once.
 * You always update dot in the current window. You update mark, and a dot in
 * another window, if it is greater than the place where you did the insert. c
 * could be a Unicode character, insert it.
 *
 * Return TRUE if all is well, and FALSE on errors.
 */
int	linsert(BufChrs n, UniChar c)
{
	register BYTE*	cp1;
	register LINE*	lp1;
	register LINE*	lp2;
	register LINE*	lp3;
	register LineLn	doto;
	register Window*wp;
	int		utf8 = (curbp->b_mode & MDUTF8) != 0;	// UTF8 enabled buffer ?
	int		CharSize;	// Size (bytes) of character being inserted - utf8
	BYTE*		Utf8String = "";// UTF8 string of length CharSize - init to keep compiler happy
	LineLn		NewBytesNeeded;	// To hold the new characters

	lchange(WFEDIT);		/* Mode line, etc */

	lp1 = curwp->w_dotp;		/* Current line		*/

	// UTF8 mode buffer ?
	if(utf8) {
		if( !utf8_isvalid(c)) {
			mlwrite("%Elinsert: error inserting U-%X: %s", c, utf8_perror());
			return FALSE;
		}
		Utf8String = utf8_encode(c);
		CharSize = CHAR_LEN(*Utf8String);	// Might be 1
	} else {	// Silently truncate to a byte
		c &= 0xFF;
		CharSize = 1;
	}

	NewBytesNeeded = CharSize * n;

	if(lp1 == curbp->b_abotline) {	/* At the end of buffer or narrow: special	*/

		if(curwp->w_doto != 0) {
			/* 'cos the last line should never contain anything */
			mlwrite("%Ebug: linsert");
			return(FALSE);
		}

		if((lp2=lalloc(NewBytesNeeded)) == NULL)	/* Allocate new line	*/
			return(FALSE);

		/* Link the new line in before the header/footer line */
		lsetforw(lp2, lp1);
		lp3 = lback(lp1);
		lsetback(lp2, lp3);
		lsetforw(lp3, lp2);
		lsetback(lp1, lp2);

		curbp->b_numLines++;

		/* If the special line contained the mark, ensure that the mark
		 * stays where the user thinks it ought to - before the first
		 * character that we are about to insert.
		 * Only if this window isn't displayed.
		 */
		if(curbp->b_markp == lp1)
			curbp->b_markp = lp2;

		curwp->w_dotp = lp2;
		curwp->w_doto = NewBytesNeeded;

		/* Insert the character that many times */
		cp1 = ltext(lp2);
		if(CharSize == 1) {
			while( --n >= 0 )
				*cp1++ = (BYTE)c;
		} else {
			// Insert multi byte chars
			// Ignore Utf8String uninitialized warning
			for( ; --n >= 0; cp1 += CharSize)
				UTF8_MOVE(cp1, Utf8String);

			lsetUtf8(lp2);	// Note that we have some MB chars in the line
		}
		lsetused(lp2, NewBytesNeeded);

		/* We are inserting chars at the end of the buffer, there
		 * isn't a \n there (yet). Note this so we can zap it later &
		 * not have the user winge about a \n that he typed.
		 */
		curbp->b_flag |= BFIMPLICIT;

		/* Update windows - the line may have been the mark */
		for(wp = wmessp; wp != NULL; wp = wp->w_wndp)
			if(wp->w_markp == lp1) {
				FPRINTF((stderr, "rset mark %#x -> %#x\n",
					wp->w_markp, lp2));
				wp->w_markp = lp2;
			}

		// There might have been an alpha mark on this last line.
		// Unlikely, but ...
		if(lmarked(lp1)) {
			AMARK*	amp;
			for(amp = curbp->b_amark; amp; amp = amp->m_next)
				if(amp->m_pline == lp1) {
					amp->m_pline = lp2;
					lsetMark(lp2);
				}
			lunsetMark(lp1);
		}

		// Maybe a numeric mark
		// Check all windows - only where they show the current buffer
		// This code fixes the current window at the same time as the others
		if(lisNmarked(lp1)) {
			NMARK*	nmp;

			for(wp = wmessp; wp != NULL; wp = wp->w_wndp) {
				if(wp->w_bufp != curbp)
					continue;

				for(nmp = wp->w_nmarks; nmp < wp->w_nmarks + NUM_NUM_MARKS; nmp++)
					if(nmp->n_pline == lp1) {
						nmp->n_pline = lp2;
						lsetNMark(lp2);
					}
			}

			lunsetNMark(lp1);
		}

		return(TRUE);
	}

	FPRINTF((stderr, "linsert doto %#x/%ld mark %#x/%ld -> ", 
		curwp->w_dotp, (long)curwp->w_doto, curwp->w_markp, (long)curwp->w_marko));

	/* Is there enough room in the buffer ? */
        if(lused(lp1) + NewBytesNeeded > lsize(lp1)) {
		if( !lineMoreBytes(lp1, lused(lp1) + NewBytesNeeded))
			return(FALSE);
	}

	doto = curwp->w_doto;			/* Save for later	*/
	cp1 = ltext(lp1);

	/* Open up a gap */
	cp1 += doto;
	if(lused(lp1) - doto > 0)
		memmove(cp1 + NewBytesNeeded, cp1, (size_t)(lused(lp1) - doto));

	lsetused(lp1, lused(lp1) + NewBytesNeeded);

	/* Pop the characters in.
	 * CharSize is 1 if either non UTF8 mode or the char being inserted is US-ASCII
	 */
	if(CharSize == 1) {
		for(; --n >= 0; )
			*cp1++ = (BYTE)c;
	} else {
		// Insert multi byte chars
		// Ignore Utf8String uninitialized warning
		for( ; --n >= 0; cp1 += CharSize)
			UTF8_MOVE(cp1, Utf8String);

		lsetUtf8(lp1);		// Note that we have some MB chars in the line
	}

	FixOffsets(lp1, doto, NewBytesNeeded, curbp, 0);

	FPRINTF((stderr, "doto %#x/%ld mark %#x/%ld\n", 
		curwp->w_dotp, (long)curwp->w_doto, curwp->w_markp, (long)curwp->w_marko));

	return(TRUE);
}

/* Undo change to buffer
 * Not yet implemented.
 *
 * Command:  undo
 * Bound to: C-_
 *
 * Return FALSE - always
 */
int	undo(int f, Cnt n)
{
	mlwrite("%W""%BUndo - not yet implemented");

	return(FALSE);
}

/* Inserted bytes have been inserted into line lp at offset.
 * Fix up position of "." & marks, in all windows and alpha marks in the buffer.
 * The line is in buffer bp.
 * If inserted is negative it is because that many bytes have been removed.
 *
 * This function makes no assumption that bp is curbp or is displayed (curpw).
 *
 * If an adjustment results in the position meeting the end of line, leave
 * it there.
 *
 * What does fixup mean when characters are inserted at:
 * "."  - is moved right
 * mark - is left where it is
 * Thus if insert 123 into: abcXdef yielding abc123Xdef,
 * "." and mark is at X, end up with:
 * "." at X and mark at 1
 *
 * Thus a mark could be said to be after the character before its position.
 * This applies for all marks.
 *
 * If BeforeMark treat marks as "." - ie the inserted offset is before marks.
 * This is for paragraph reformatting which needs this.
 *
 * The commented out code is not needed and in the case marked 'v' valgrind
 * says that it can access a byte after what was allocated. Prob needed a test
 * xxxo < lused(lp)
 */
void	FixOffsets(const LINE* lp, LineLn offset, LineLn inserted, BUFFER* bp, int BeforeMark)
{
	Window*	wp;
	LineLn	markOffset = offset;

	if(BeforeMark)
		markOffset--;

	// bp might not be displayed, so b_dotp & b_markp are important
	if(bp->b_dotp == lp && bp->b_doto >= offset) {
		bp->b_doto += inserted;
		if(bp->b_doto < offset)
			bp->b_doto = offset;

		// Ensure that "." is not in the middle of UTF-8 character:
		// if((bp->b_mode & MDUTF8) != 0 && bp->b_doto > 0)
		//	UTF8_BACK_FIND_START_NX(lgetc(lp, bp->b_doto), bp->b_doto);
	}

	if(bp->b_markp == lp && bp->b_marko > markOffset) {
		bp->b_marko += inserted;
		if(bp->b_marko < offset)
			bp->b_marko = offset;

		// if((bp->b_mode & MDUTF8) != 0 && bp->b_marko > 0)
		//	UTF8_BACK_FIND_START_NX(lgetc(lp, bp->b_marko), bp->b_marko);
	}

	/* Check all windows, if we insert/remove on it's marked line, before
	 * the mark, move the mark right/left; similar with "." if we
	 * insert/remove before or at its position.
	 */
	for(wp = wmessp; wp != NULL; wp = wp->w_wndp) {
		if(wp->w_bufp != bp)
			continue;

		if(wp->w_dotp == lp && wp->w_doto >= offset) {
			wp->w_doto += inserted;
			if(wp->w_doto < offset)		// Keep at end of line
				wp->w_doto = offset;

			// if((bp->b_mode & MDUTF8) != 0 && wp->w_doto > 0)	// v
			//	UTF8_BACK_FIND_START_NX(lgetc(lp, wp->w_doto), wp->w_doto);
		}

		if(wp->w_markp == lp && wp->w_marko > markOffset) {
			wp->w_marko += inserted;
			if(wp->w_marko < offset)
				wp->w_marko = offset;

			// if((bp->b_mode & MDUTF8) != 0 && wp->w_marko > 0)
			//	UTF8_BACK_FIND_START_NX(lgetc(lp, wp->w_marko), wp->w_marko);
		}
	}

	// Alpha marked line ?
	// Adjust the line offset if needs be:
	if(lmarked(lp)) {
		AMARK*	amp;
		for(amp = bp->b_amark; amp; amp = amp->m_next)
			if(amp->m_pline == lp && amp->m_oline > markOffset) {
				amp->m_oline += inserted;
				if(amp->m_oline < offset)
					amp->m_oline = offset;

				// if((bp->b_mode & MDUTF8) != 0 && amp->m_oline > 0)
				//	UTF8_BACK_FIND_START_NX(lgetc(lp, amp->m_oline), amp->m_oline);
			}
	}

	// Ditto numeric marked
	if(lisNmarked(lp)) {
		NMARK*	nmp;

		for(wp = wmessp; wp != NULL; wp = wp->w_wndp) {
			if(wp->w_bufp != bp)
				continue;

			for(nmp = wp->w_nmarks; nmp < wp->w_nmarks + NUM_NUM_MARKS; nmp++)
				if(nmp->n_pline == lp && nmp->n_oline > markOffset) {
					nmp->n_oline += inserted;
					if(nmp->n_oline < offset)
						nmp->n_oline = offset;

					// if((bp->b_mode & MDUTF8) != 0 && nmp->n_oline > 0)
					//	UTF8_BACK_FIND_START_NX(lgetc(lp, nmp->n_oline), nmp->n_oline);
				}
		}
	}
}

/* Append a line to the end of buffer bp.
 * This does not do any clever screen work or checking where the mark is,
 * do not even make it the current line in the buffer
 * - just allocate, link in and be done
 * - does not mark the buffer as changed
 *
 * The line will be empty, but given the minimum allocation.
 *
 * Return the line or NULL on fail (out of memory).
 */
LINE*	BufAppendLine(BUFFER* bp)
{
	LINE*	newline = lalloc(0);

	if( !newline)
		return((LINE*)NULL);

	/* Link it into the chain at the end of the buffer */
	lsetforw(newline, bp->b_linep);
	lsetback(newline, lback(bp->b_linep));
	lsetforw(lback(newline), newline);
	lsetback(bp->b_linep, newline);

	bp->b_numLines++;

	return(newline);
}

/* Overwrite a character into the current line at the current position
 *
 * r:	replacement character to overwrite on current position
 * If r is a double width character - replace 2 characters - unless
 * what we replace is double width itself.
 * If we are replacing a tab - try to get it right.
 *
 * The width calculations regarding tabs are horrible!
 *
 * If r is tab ... we get it wrong.
 *
 * Return TRUE on success - which could mean making no change as the
 * replaced character is the same as the original; although we will
 * move the cursor one character to the right.
 */
int	lowrite(UniChar r)
{
	// Don't delete (so that we overwrite) if at end of line
	if(curwp->w_doto < lused(curwp->w_dotp)) {
		int	charsToDelete = 0;

		// If we are replacing a character with itself - do nothing
		if((curbp->b_mode & MDUTF8) == 0 ? (r == lgetc(curwp->w_dotp, curwp->w_doto)) :
				utf8_cheq(utf8_encode(r), lgetcaddr(curwp->w_dotp, curwp->w_doto))) {
			moveForw(1);
			return(TRUE);
		}

		// Get the character width, minimum 1 position
		int	rWidth = (curbp->b_mode & MDUTF8) == 0 ? 1 : mk_wcwidth(r);
		if(rWidth < 1)
			rWidth = 1;

		int	curWidth = (curbp->b_mode & MDUTF8) == 0 ? 1 :
				mk_wcwidth(utf8_decode(lgetcaddr(curwp->w_dotp, curwp->w_doto)));
		if(curWidth < 1)
			curWidth = 1;

		if(lgetc(curwp->w_dotp, curwp->w_doto) == '\t') {
			/* Find the column position that the tab takes us to
			 * Subtract the end column position of the character before
			 * That gives the width that the tab spans
			 * Thus work out if we need to delete the tab
			 */
			LineLn	tabEnd = getdiscol(curwp, curwp->w_dotp, curwp->w_doto+1, DisCol_Char);
			LineLn	beforeEnd  = getdiscol(curwp, curwp->w_dotp, curwp->w_doto, DisCol_Char);

			if(tabEnd - beforeEnd <= rWidth)
				if( !ldelete(1, KL_DISCARD))
					return(FALSE);

			// Double width char, might need to delete the next char as well
			if(rWidth > 1 && lgetc(curwp->w_dotp, curwp->w_doto) != '\t') {
				tabEnd = getdiscol(curwp, curwp->w_dotp, curwp->w_doto+1, DisCol_Char);
				beforeEnd  = getdiscol(curwp, curwp->w_dotp, curwp->w_doto, DisCol_Char);

				if(tabEnd > beforeEnd)
					charsToDelete = 1;
			}
		} else
			if(curWidth < rWidth) {
				// Replacing width 1 with width 2 - delete 2 chars
				if( !ldelete(1, KL_DISCARD))
					return(FALSE);
				if(curwp->w_doto < lused(curwp->w_dotp) && lgetc(curwp->w_dotp, curwp->w_doto) != '\t')
					charsToDelete = 1;
			}
		else
			charsToDelete = 1;

		if(charsToDelete > 0 && !ldelete(charsToDelete, KL_DISCARD))
			return(FALSE);
	}

	return(linsert(1, r));
}

/* lover -- Overwrite a string at the current point.
 * The big point about this is that it preserves horizontal position, this recognises
 * that not all characters are the same width: some characters are double width
 * the tab is variable width.
 *
 * This is used in ovstring() (overwrite-string).
 *
 * Return: TODO
 */
int	lover(const LINE* ostr)
{
	register int	status = TRUE;
	int		utf8 = (curbp->b_mode & MDUTF8) != 0;	// UTF8 enabled buffer ?
	LineLn		pos = 0;				// Where we are in ostr
	BYTE		tmpc;

	if(ostr == NULL)
		return(TRUE);

	while(pos < lused(ostr) && status == TRUE) {
		tmpc = lgetc(ostr, pos);
		if(tmpc == '\n') {
			status = lnewline(0);
			ostr++;
			continue;
		}

		// Simple case: byte replacement
		if( !utf8 || CHAR_LEN(tmpc) == 1) {
			status = lowrite(tmpc);
			pos++;
			continue;
		}

		// A multi-byte character:
		status = lowrite(utf8_decode(lgetcaddr(ostr, pos)));

		pos += CHAR_LEN(tmpc);
	}

	/* Insertion error? */
	if(status != TRUE)
		mlwrite(MsgOutOfMemoryWhile, "overwriting");

	return(status);
}

/* Lines have been inserted/deleted in curwp. If the buffer is
 * displayed more than once then line numbers in the other windows
 * might need adjusting.
 *
 * This is to deal with lines inserted/deleted above the top line
 * in these other windows. Lines ins/del below the window last line,
 * we don't care about; lines ins/del within what is display redisplay
 * handles this.
 *
 * What we adjust is w_toplinenum, this is used to:
 * *	calculate line numbers displayed WO_NUMBER
 * *	show number with WO_POSITION_SHOW
 *
 * inserted	The number of lines inserted at "."
 *		This will be -ve if lines were removed.
 *
 * Note that curwp is OK, the display routines handle that.
 */
void	adjustTopLineNum(BufLns inserted)
{
	Window*	wp;
	BufLns	dotNum;		// Line # of "."
	LINE*	lp;

	// Only displayed once - nothing to do
	if(curbp->b_nwnd < 2)
		return;

	// Find the line # where the ins/del happens
	for(dotNum = curwp->w_toplinenum, lp = curwp->w_topLinep;
	    lp != curwp->w_dotp && lp != curbp->b_linep;
	    dotNum++, lp = lforw(lp))
		;

	for(wp = wmessp; wp != NULL; wp = wp->w_wndp) {
		if(wp == curwp)		// curwp is catered for
			continue;

		if(wp->w_bufp != curbp)	// Only current buffer
			continue;

		/* If the change was above the first line shown in this window
		 * adjust it's line number and refresh lots of things.
		 */
		if(wp->w_toplinenum > dotNum) {
			wp->w_toplinenum += inserted;

			wp->w_flag |= WFHARD | WFMODE | WFNWID;
		}
	}
}

/* Insert a newline into the buffer at the current location of dot in the
 * current window. The update of dot and mark is a bit easier than
 * in the above case, because the split forces more updating.
 * 
 * Don't allocate if we are at the end of the last line, just move down a line,
 * it already exists. This is to stop spurious extra lines appearing at the end
 * of the file. If (even in this case) you want a line, set Force true.
 *
 * If "." is on b_linep then b_linep is set on to the newly created line
 * and "." follows it there. It is just simpler to do this rather than do
 * differently in this special case to avoid changing b_lines. KV: top/bottom
 * of narrow may be reassigned if this happens.
 *
 * The UTF8 flag is copied to the new line.
 * Any NUL flag set on the current line is not copied to the new line.
 *
 * Return
 *	TRUE if everything works out
 *	FALSE on error (memory allocation failure).
 */
int	lnewline(int Force)
{
	LINE*	lp1;		// Line that contains "." before newline inserted
	LINE*	lp2;		// New line created, "." will end at the start of this line
	LineLn	doto;
	LineLn	needed;
	Window* wp;

	lchange(WFHARD);

	lp1  = curwp->w_dotp;			/* Get the address and	*/
	doto = curwp->w_doto;			/* offset of "."	*/

	/* Do we need a new line buffer ?
	 * We are basically trying to avoid creating one at EOF when we don't
	 * need it & the implicit marker does it all.
	 * We do if we aren't at EOF (ie next line isn't EOF), or if the next
	 * is EOF we are going to put some characters onto it (the magic line
	 * must always be empty).
	 * We must create a new line if the buffer is empty - that is what the
	 * second line tests for.
	 *
	 * A special case to check for: zero lines in the buffer, insert a line.
	 * b_numLines tells us this.
	 */
	needed = lused(lp1) - doto;
	FPRINTF((stderr,
	"Needed %d, l1_used %ld, doto %ld lp1 %p curlinep %p lp1.nxt %p back %p Force %d impl %d\n",
		needed, (long)lused(lp1), (long)doto, lp1, curbp->b_linep,
		lforw(lp1), lback(lp1), Force, (curbp->b_flag & BFIMPLICIT)));
	if(lforw(lp1) != curbp->b_linep || needed > 0 || Force || curbp->b_numLines == 0 ||
	   (lback(lp1) == curbp->b_linep && (curbp->b_flag & BFIMPLICIT))) {

		/* New second half line */
		if((lp2 = lalloc(needed)) == NULL)
			return(FALSE);

		/* Shuffle text around - split between lines.
		 * lp1 ends up with the text at the start of the line, lp2
		 * the end.
		 * Copy line's UTF8 & L_NUL flags if there is any is any text there.
		 */
		if(needed > 0) {
			memcpy(ltext(lp2), lgetcaddr(lp1, doto), (size_t)needed);

			lsetused(lp2, needed);

			if(lisUtf8(lp1))
				lsetUtf8(lp2);
		}

		lsetused(lp1, doto);			/* Not so much now */

		/* Link into chain with lp2 after lp1 (which is ".") */
		lsetback(lp2, lp1);
		lsetforw(lp2, lforw(lp1));
		lsetforw(lp1, lp2);
		lsetback(lforw(lp2), lp2);

		// Increase buffer line count
		curbp->b_numLines++;

		// Check for other windows affected:
		adjustTopLineNum(1);

		FPRINTF((stderr,
			"Old line %p, new %p new.next %p head %p::",
			lp1, lp2, lforw(lp2), curbp->b_linep));

		/* If the line was the special start/end buffer, note that it
		 * has changed & that it is just a marker.
		 */
		if(curbp->b_linep == lp1) {
			curbp->b_linep = lp2;

			curbp->b_flag |= BFIMPLICIT;
		}

		// That might have been the bottom of a narrow
		// Maybe top also, I think not, but check anyway
		if(curbp->b_btopline == lp1)
			curbp->b_btopline = lp2;
		if(curbp->b_abotline == lp1)
			curbp->b_abotline = lp2;
	}
	else {
		/* We can get away with an implicit line */
		lp2 = curbp->b_linep;
		curbp->b_flag |= BFIMPLICIT;
	}

	FPRINTF((stderr, "Old line %p, new %p new.next %p head %p\n",
				lp1, lp2, lforw(lp2), curbp->b_linep));

	/* Do not need to look at curbp as this function assumes that the
	 * buffer is displayed in a window - thus b_dotp & b_markp will be overwritten later.
	 */

	/* Windows */
	for(wp = wmessp; wp != NULL; wp = wp->w_wndp) {
		if(wp->w_dotp == lp1 && wp->w_doto >= doto) {
			wp->w_dotp = lp2;
			wp->w_doto -= doto;
		}

		if(wp->w_markp == lp1 && wp->w_marko > doto) {
			wp->w_markp = lp2;
			wp->w_marko -= doto;
		}
	}

	// Marked line ?
	if(lmarked(lp1)) {
		AMARK*	amp;
		int	markStays = 0;

		// Alpha marks:
		for(amp = curbp->b_amark; amp; amp = amp->m_next)
			if(amp->m_pline == lp1) {
				/* Found a mark that refers to this line
				 * Check if it has moved to the next line, or stays.
				 */
				if(amp->m_oline > doto) {
					amp->m_pline = lp2;
					amp->m_oline -= doto;
					lsetMark(lp2);
				}
				else
					markStays = 1;
			}

		// All marks moved to next line, clear marked flag this line:
		if( !markStays)
			lunsetMark(lp1);
	}

	// Numeric marked line ?
	if(lisNmarked(lp1)) {
		NMARK*	nmp;
		int	markStays = 0;

		// See comments above
		for(wp = wmessp; wp != NULL; wp = wp->w_wndp) {
			if(wp->w_bufp != curbp)
				continue;

			for(nmp = wp->w_nmarks; nmp < wp->w_nmarks + NUM_NUM_MARKS; nmp++)
				if(nmp->n_pline == lp1) {
					if(nmp->n_oline > doto) {
						nmp->n_pline = lp2;
						nmp->n_oline -= doto;
						lsetNMark(lp2);
					}
					else
						markStays = 1;
				}
		}

		if( !markStays)
			lunsetNMark(lp1);
	}

	return(TRUE);
}

/* This function deletes "n" characters, starting at dot.
 * It understands how do deal with end of lines, etc.
 *
 * n		# of chars to delete - must not be -ve, can be 0
 * kflag	What to do with the characters that are deleted (killed)
 *		KL_DISCARD	Throw them away
 *		KL_KILL_APPEND	Put at the end   of the kill buffer text
 *		KL_KILL_START	Put at the start of the kill buffer text
 *
 * If utf8 - a character could be more than 1 byte
 *
 * Return
 *	TRUE  on success.
 *	FALSE Failure could be:
 *	      cannot remove implicit \n at buffer end;
 *	      cannot insert into kill buffer
 *	      cannot join 2 lines (ie delete \n)
 */
int	ldelete(BufChrs n, int kflag)
{
	BYTE*	cp1;
	BYTE*	cp2;
	BYTE*	start;			// Where we start deleting
	LINE*	dotp;
	LineLn	doto;
	int	deleted_bytes;
	int	utf8_buffer = (curbp->b_mode & MDUTF8) != 0;   // UTF8 enabled buffer ?

	int	kf = kflag == KL_KILL_START;	// Tells kinsert() to put at start of killed text

	curwp->w_partoff = 0;
	kill_lcount = 0;		/* reset number of lines killed */

	while(n > 0) {
		dotp = curwp->w_dotp;
		doto = curwp->w_doto;

		if(EonLine(dotp)) {
			// Hit end of buffer/narrow.

			if( ! EofLine(dotp))
				return(FALSE);

			if(EofLine(dotp)) {
				/* We are allowed to remove the \n at the end of buffer -- if
				 * there is one.
				 */
				if(!(curbp->b_flag & BFIMPLICIT))
					return(FALSE);

				/* We don't actually delete the newline 'cos we are at
				 * the special end-of-buffer & to do so would create
				 * havoc. But we mark the \n as not there (not implicit)
				 * & pop a \n into the kill buffer.
				 */
				curbp->b_flag &= ~BFIMPLICIT;
			}

			lchange(WFHARD);

			return(n == 1 &&
				(kflag == KL_DISCARD || kinsert(NewlineString, kf) == TRUE) ?
				TRUE : FALSE);
		}

		/* End of line, merge.
		 * If we are joining 2 lines, one might have the utf8 hint - copy it.
		 * Getting the value is harder than it might seem. Need to merge for
		 * the current line and what is currently the next - both of which
		 * could have their structures replaced - so OR both values now.
		 */
		if(lused(dotp) == doto) {
			char	u8 = lisUtf8(dotp) | lisUtf8(lforw(dotp));

			lchange(WFHARD);

			if(ldelnewline() == FALSE ||
			   (kflag != KL_DISCARD && kinsert(NewlineString, kf) == FALSE))
				return(FALSE);
			--n;

			if(u8)
				lsetUtf8(curwp->w_dotp);

			continue;
		}

		lchange(WFEDIT);

		/* Beware: n is in units of characters
		 * deleted_bytes is the number of bytes
		 */
		start = cp1 = lgetcaddr(dotp, doto);	/* Where we start deleting */
		cp2 = lgetcaddr(dotp, lused(dotp));	/* End of line */
		deleted_bytes = 0;

		/* Delete until either:
		 * + we have deleted everything that we have been asked to delete, or
		 * + we have reached the end of line
		 */
		while(n && cp1 < cp2) {
			// Put into kill buffer ?
			if(kflag != KL_DISCARD && kinsert(cp1, kf) == FALSE)
				return(FALSE);

			deleted_bytes += utf8_buffer ? CHAR_LEN(*cp1) : 1;
			cp1 += utf8_buffer ? CHAR_LEN(*cp1) : 1;

			// Another character deleted
			n--;
		}

		/* Now move the characters beyond those just deleted down.
		 * If there are any characters left - ie not deleted to the end of line
		 */
		while(cp1 < cp2)
			*start++ = *cp1++;
		lsetused(dotp, lused(dotp) - deleted_bytes);

		/* If we empty the line & it has an implicit \n at the end,
		 * zap the line, we get left sitting on the dummy line.
		 */
		if(lused(dotp) == 0 && EofLine(lforw(dotp)) &&
				(curbp->b_flag & BFIMPLICIT)) {
			if(ldelnewline() != TRUE)
				return(FALSE);
			continue;
		}

		/* Fix "." and marks */
		FixOffsets(dotp, doto, -deleted_bytes, curbp, 0);
	}

	return(TRUE);
}

/* getctext:	grab and return a string with the text of
 *		the current line
 *
 * The text is NUL terminated in a static buffer size NSTRING.
 * Return an out of memory message - if that happens.
 */
BYTE*	getctext(void)
{
	/* find the contents of the current line and its length */
	return(getlinestr(curwp->w_dotp));
}

/* Return the text in LINE lp, NUL terminated.
 * Set the NUL flag on the line - so the returned string is in the LINE.
 *
 * Return an out of memory string if that is what happens.
 */
BYTE*	getlinestr(LINE* lp)
{
	if( !lsetNUL(lp))
		return(OutOfMemory);

	return(ltext(lp));
}

/* putctext:	replace the current line with the passed in text
 *
 * iline	contents of new line
 *
 * Return
 *	TRUE  on success
 *	FALSE on error: killtext() linstr() lnewline() TODO
 */
int	putctext(const BYTE* iline)
{
	register int status;

	/* delete the current line */
	curwp->w_doto = 0;	/* starting at the beginning of the line */
	if((status = killtext(TRUE, 1)) != TRUE)
		return(status);

	/* insert the new line */
	if((status = linstr(iline)) != TRUE)
		return(status);

	status = lnewline(0);
	moveUp(1);
	return(status);
}

/* Delete a newline.
 * Join the current line with the next line.
 *
 * If the next line is the magic header line or end of narrow always return
 * TRUE; merging the last line with the header line can be thought of as always
 * being a successful operation, even if nothing is done, and this makes the
 * kill buffer work "right". Easy cases can be done by shuffling data around.
 * Hard cases require that lines be moved about in memory.
 *
 * Called by "ldelete" only.
 *
 * This function assumes that the buffer is curbp, but does not assume curwp.
 *
 * Return
 *	TRUE if all looks ok
 *	FALSE on error. Reasons. TODO
 */
int	ldelnewline(void)
{
	register LINE*	lp1;
	register LINE*	lp2;
	register Window* wp;
	LineLn	oldLen;			/* Length of line before next appended */

	kill_lcount++;			/* update count of lines killed */

	lp1 = curwp->w_dotp;
	lp2 = lforw(lp1);

	if(EofLine(lp2)) {			/* At the buffer end.	*/
		if(lused(lp1) == 0) {		/* Blank line.		*/
			lfree(curbp, lp1);
			curbp->b_numLines--;	// Decrease buffer line count
		}

		curbp->b_flag |= BFIMPLICIT;

		return(TRUE);
	}

	if(EonLine(lp2))	// End of Narrow ?
		return(TRUE);

	/* Append the text from the next line (lp2) onto the end of current (lp1)
	 * lp2 will then be freed.
	 * Ensure that we have enough room for the joined line
	 * If lp1 is not NUL but will need to be - ask for the extra byte, mainly to save
	 * complication if we can't allocate the room.
	 */
	if( !lineMoreBytes(lp1, lused(lp1) + lused(lp2) + (lisNUL(lp2) && !lsetNUL(lp1))))
		return(FALSE);

	if(lisNUL(lp2))
		lsetNUL(lp1);

	oldLen = lused(lp1);
	memcpy(lgetcaddr(lp1, oldLen), ltext(lp2), (size_t)lused(lp2));
	lsetused(lp1, oldLen + lused(lp2));

	// Find references to the old line (lp2) and fix them to refer
	// to lp1, adjusting the offset.

	// "." in the buffer that the line is in ?
	// This code is only needed if the buffer not displayed in a window, unlikely but possible.
	if(curbp->b_dotp == lp2) {
		curbp->b_dotp = lp1;
		curbp->b_doto += oldLen;
	}

	if(curbp->b_markp == lp2) {
		curbp->b_markp = lp1;
		curbp->b_marko += oldLen;
	}

	if(curbp->b_topLinep == lp2)
		curbp->b_topLinep = lp1;

	// Marked line ? Patch offsets:
	if(lmarked(lp2)) {
		// Alpha marks:
		AMARK*	amp;

		for(amp = curbp->b_amark; amp; amp = amp->m_next)
			if(amp->m_pline == lp2) {
				amp->m_pline = lp1;
				amp->m_oline += oldLen;
				lsetMark(lp1);
			}

		l_unmark(curbp, lp2);
	}

	// Numeric marked line ? Patch offsets:
	if(lisNmarked(lp2)) {
		NMARK*	nmp;

		for(wp = wmessp; wp != NULL; wp = wp->w_wndp) {
			if(wp->w_bufp != curbp)
				continue;

			for(nmp = wp->w_nmarks; nmp < wp->w_nmarks + NUM_NUM_MARKS; nmp++)
				if(nmp->n_pline == lp2) {
					nmp->n_pline = lp1;
					nmp->n_oline += oldLen;
					lsetNMark(lp1);
				}
		}

		l_unNmark(curbp, lp2);
	}

	// Is this buffer displayed ? Assume that it is, it prob will be; code below benign if it isn't
	// Display "." and marks in windows that display this buffer:
	for(wp = wmessp; wp; wp = wp->w_wndp) {
		if(wp->w_bufp == curbp) {
			if(wp->w_dotp == lp2) {
				wp->w_dotp = lp1;
				wp->w_doto += oldLen;
			}
			if(wp->w_markp == lp2) {
				wp->w_markp = lp1;
				wp->w_marko += oldLen;
			}
			if(wp->w_lastdismarkp == lp2)
				wp->w_lastdismarkp = NULL;
			if(wp->w_topLinep == lp2) {
				wp->w_topLinep = lp1;
				wp->w_flag |= WFFORCE;
			}
		}
	}

	// Unstitch from the list of lines:
	lunstitch(lp2);

	curbp->b_numLines--;		// Decrease buffer line count

	// Check for other windows affected:
	adjustTopLineNum(-1);

	return(TRUE);
}

/* Initialise the kill buffer so that it there is nothing to yank back.
 * Just clear the kill buffer.
 *
 * Count is not unused.
 *
 * Command:  initialise-kill-buffer
 * Bound to: unbound
 *
 * Return TRUE if everything looks good.
 *        FALSE on error
 *	  ABORT if prompt is aborted
 */
int	initkill(int f, Cnt n)
{
	return(bclear(bmlKill));
}


/* Insert a new, empty line into the kill buffer.
 *
 * Return:
 *	TRUE  success
 *	FALSE out of memory
 */
int	kNewline(void)
{
	LINE*	lp;

	/* A new kill or set of deletes, create a new line to store them.
	 * Put at the end of the buffer, make it the current line.
	 */
	if( !(lp = BufAppendLine(bmlKill)))
		return(FALSE);

	bmlKill->b_dotp = lp;
	bmlKill->b_doto = 0;

	/* Ensure that the kill buffer does not get too large.
	 * Avoid a potential nasty issue: consider what could happen if
	 * we are doing copy-region on the first line in the Kill Buffer
	 * and we try to remove the line that is being copied.
	 */
	if(curbp != bmlKill)
		trimbuffer(bmlKill, -SaveMlHist);

	// Just in case someone is viewing the kill buffer
	if(bmlKill->b_nwnd)
		upmode(bmlKill, WFHARD);

	return(TRUE);
}

/* Insert a character into the kill buffer.
 * Allocate a new LINE if the last command was not a DELETE. The effect of this
 * is that an uninterrupted sequence of delete/kill commands will all put their
 * deleted text into the same LINE in the kill buffer. This text is put at
 * the start or end of the line (see back argument) building text as it was
 * in the buffer that it was deleted from - thus when it is yanked it reappears
 * as it was before it was deleted.
 *
 * Works with either UTF8 buffers/characters or not so. The UTF8 flag
 * is taken from the current struct BUFFER mode. This may label pure
 * ASCII with the utf8 flag - but that is good since it means that there
 * are not any problems pasting it into another buffer.
 *
 * cp	Possibly UTF8 character bytes to insert in the kill buffer
 * back	means deleting backwards, so insert character at the start of the line
 *	in the kill buffer - otherwise insert at the end.
 *
 * If NotToKillBuffer then quick return - this is intended for prompting,
 * characters deleted there are not wanted in the yank buffer.
 *
 * Return
 *	TRUE if all is well, including flag NotToKillBuffer set
 *	FALSE on out of memory
 */
int	kinsert(const BYTE* cp, int back)
{
	LINE*	lp;
	int	utf8 = (curbp->b_mode & MDUTF8) != 0;   // UTF8 enabled buffer ?

	// Global suppress insert into yank buffer
	if(NotToKillBuffer)
		return(TRUE);

	// Only treat as UTF8 if the buffer is - regardless of what it looks like
	int	csize = utf8 ? CHAR_LEN(*cp) : 1;	// Character size - how many bytes

	// New kill/deletes: start a new line
	if( ((thisflag | lastflag) & CFDELETE) == 0)
		if(kNewline() == FALSE)
			return(FALSE);

	thisflag |= CFDELETE;

	lp = bmlKill->b_dotp;
	if(utf8)
		lsetUtf8(lp);

	/* Ensure that there is there enough room in the line */
	if(lused(lp) + csize > lsize(lp) &&
	   !lineMoreBytes(lp, lused(lp) + csize))
		return(FALSE);

	// Make room at the start of the LINE if that is where the new text is to go:
	if(back)
		memmove(lgetcaddr(lp, csize), lgetcaddr(lp, 0), (size_t)lused(lp));

	// Insert the character - possibly multiple bytes:
	if(utf8) {
		UTF8_MOVE(lgetcaddr(lp, (back ? 0 : bmlKill->b_doto)), cp);
	} else {
		lputc(lp, (back ? 0 : bmlKill->b_doto), *cp);
	}

	lsetused(lp, lused(lp) + csize);
	bmlKill->b_doto += csize;

	/* In the unlikely situation that someone is viewing the kill buffer while
	 * deleting elsewhere, or something, or even copying text in the kill buffer ...
	 * ensure that the screen is updated to reflect what has just been put into it.
	 */
	if(bmlKill->b_nwnd)
		upmode(bmlKill, WFHARD);

	return(TRUE);
}

/* Do the actual work of yanking.
 * Text is from the current line in the kill buffer.
 *
 * n	number of times to yank
 * Msg	A mlwrite() format string of the informational message to output.
 *
 * Return
 *	TRUE  on success
 *	FALSE linsert() failed
 */
int	yanksDo(Cnt n, const char* Msg)
{
	BYTE	mangbuf[50];		/* part of a message */
	BufLns	y_lcount = 0;		/* Count of number of lines yanked */
	LineLn	characters = 0;		/* Characters yanked */
	BufChrs	y_mangled_chars = 0;	/* How many chars had encoding probs in the last yank */
	char*	what = "line";		/* What we report as having yanked */
	int	utf8_buffer = (curbp->b_mode & MDUTF8) != 0;	// UTF8 enabled buffer ?
	LINE*	lp = bmlKill->b_dotp;


	/* Set the mark here, so that we can delete the region if the next
	 * command is a reyank.
	 */
	setmark(FALSE, FALSE);

	/* remember that this was a yank command */
	thisflag |= CFYANK;

	while(--n >= 0) {
		/* Bring this in from yankfrom() to make it easier to
		 * suppress the last line insert on the last iteration.
		 */

		LineLn	lnOff = 0;	// Offset in the line that we are copying from

		/* It is important that sp is a pointer to *unsigned* char so that
		 * *sp will yeild a value in the range 0..255.
		 * With signed, chars 0x80+ will yeild negative code points
		 */
		while(lnOff < lused(lp)) {
			UniChar	c;

			c = lgetc(lp, lnOff);
			FPRINTF((stderr, "yf '%c'\n", c));

			// Create a newline if we see a \n:
			if(c == '\n') {
				/* Avoid putting an extra/unwanted line at the end of the buffer/narrow.
				 * Put a \n unless it is the last char to be yanked
				 * and we are inserting at the end of buffer/narrow.
				 *
				 * However: move to the start of the line after the yanked text,
				 * this is what causes least surprise.
				 */
				if(n == 0 && lnOff == (lused(lp) - 1) && EonLine(lforw(curwp->w_dotp)) && curwp->w_doto == lused(curwp->w_dotp))
					moveForw(1);
				else if(lnewline(1) == FALSE)
					return(FALSE);

				y_lcount++;
				lnOff++;

				continue;	// next character
			}

			// What is important is are what we are copying utf8 ?
			// Ie the kill buffer. Then does it match the text buffer ?
			if(lisUtf8(lp)) {
				/* Copying from UTF8 line in the kill buffer.
				 * If the text buffer is not UTF8 - insert the raw bytes into the buffer.
				 * If UTF8 TB - convert to code point and insert, this is
				 * a long way round - might optimise at some point.
				 */
				int	u8_len = CHAR_LEN(c);
				if(u8_len == 1) {
					// Optimise common case
					if(linsert(1, c) == FALSE)
						return(FALSE);
					lnOff++;
					characters++;
				} else {
					// Dealing with multi byte character
					if(utf8_buffer) {
						// UTF8 buffer - perfect!
						if(linsert(1, utf8_decode(lgetcaddr(lp, lnOff))) == FALSE)
							return(FALSE);
						lnOff += u8_len;
						characters++;
					} else {
						/* MB char, ASCII text buffer
						 * Insert the raw bytes
						 */
						while(--u8_len >= 0)
							if(linsert(1, lgetc(lp, lnOff++)) == FALSE)
								return(FALSE);
						y_mangled_chars++;
						characters += u8_len;
					}
				}
			} else {
				/* Copying from plain ASCII kill buffer.
				 * If the character has the top bit set and the current text buffer
				 * is in UTF8 mode - confusion will result, so replace with UTF8 bad char.
				 * Count such masks offs & report to the user.
				 */
				if(c > UTF8_MAX_CP1 && utf8_buffer) {
					c = UTF8_REPLACEMENT;
					y_mangled_chars++;
				}
				if(linsert(1, c) == FALSE)
					return(FALSE);
				lnOff++;
				characters++;
			}
		}
	}

	/* Only how many chars are mangled is some were mangled */
	mangbuf[0] = '\0';
	if(y_mangled_chars)
		sprintf(mangbuf, ", %ld multi-byte chars mangled", (long)y_mangled_chars);

	/* All on one line, talk about the characters yanked */
	if(y_lcount == 0) {
		y_lcount = (BufLns)characters;
		what = utf8_buffer ? "character" : "byte";
	}

	mlwrite(Msg, (long)y_lcount, what, (y_lcount == 1 ? "" : "s"), mangbuf);

	displaySomething();

	return(TRUE);
}

/* Yank text back from the kill buffer.
 *
 * This is really easy. All of the work is done by the standard insert
 * routines. All you do is run the loop, and check for errors.
 *
 * Insert count copies of the text.
 * Default count 1. Must be >= 0
 *
 * Command:  yank
 * Bound to: C-y
 *
 * Return:
 *	TRUE  success, including nothing to yank
 *	FALSE linsert() failed
 */
int	yank(int f, Cnt n)
{
	// This can cause problems, anyway very little point:
	if(curbp == bmlKill) {
		mlwrite("%W""%BCannot yank into the Kill Buffer");
		return(FALSE);
	}

	/* Make sure there is something to yank */
	if(bmlKill->b_numLines <= 0) {
		mlwrite("%WNothing to yank");
		return(TRUE);		/* Not an error, just nothing */
	}

	/* Use the current line (".") as where to copy from.
	 * reyank yanks from the line before ".".
	 * If someone is viewing the buffer it should not make the cursor jump
	 * as the one in the window is used.
	 */
	bmlKill->b_dotp = lback(bmlKill->b_linep);
	bmlKill->b_doto = 0;

	return(yanksDo(n, "[%D %s%s yanked%s]"));
}

/* Remove the just yanked region and replace it by the line in the kill buffer
 * just before the one that had been yanked.
 *
 * If the count is negative: do the same but move to more recent killed text
 *
 * This MUST only be called after a yank or reyank.
 *
 * Insert count copies of the text.
 * Default count 1. Must be >= 0
 *
 * Command:  reyank
 * Bound to: M-y
 *
 * Return:
 *	TRUE  success
 *	FALSE last command was not yank/reyank
 *	      there is not a region
 *	      memory allocation error (linsert()) check
 *	      ldelete()
 */
int	reyank(int f, Cnt n)
{
	REGION	region;			/* dot to mark region		*/
	int	ret;
	char*	yankMsg = "[%D %s%s reyanked%s]";

	if((lastflag & CFYANK) == 0) {
		/* Last command was not a yank or reyank. Complain and return
		 * immediately, this is because the first thing we do is to
		 * delete the current region without saving it. If we've just
		 * done a yank, or reyank, the mark and cursor will surround
		 * the last yank so this will be ok, but otherwise the user
		 * would loose text with no hope of recovering it.
		 */
		mlwrite("%W""%BReyank must IMMEDIATELY follow a yank or reyank");
		return(FALSE);
	}

	/* Delete the current region.
	 */
	if((ret = getregion(&region)) != TRUE)
		return(ret);

	curwp->w_dotp = region.r_linep;
	curwp->w_doto = region.r_offset;
	if( !ldelete(region.r_size, KL_DISCARD))
		return(FALSE);

	/* If we've reached the top of the Kill Buffer there is nothing older to reyank.
	 * Wrap around to the most recent delete - the last line in the Kill Buffer.
	 * This makes it appear as if the linked list of kill buffers is actually
	 * implemented as the GNU style "kill ring".
	 *
	 * If the count is negative: the same but move to more recent killed text
	 *
	 * Beep out a warning as well.
	 */
	if(n >= 0) {
		if((bmlKill->b_dotp = lback(bmlKill->b_dotp)) == bmlKill->b_linep) {
			bmlKill->b_dotp = lback(bmlKill->b_dotp);
			yankMsg = "%B[%D %s%s reyanked from start of kill-ring%s]";
		}
	} else {
		if((bmlKill->b_dotp = lforw(bmlKill->b_dotp)) == bmlKill->b_linep) {
			bmlKill->b_dotp = lforw(bmlKill->b_dotp);
			yankMsg = "%B[%D %s%s reyanked from end of kill-ring%s]";
		}
		n = -n;
	}
	bmlKill->b_doto = 0;

	return(yanksDo(n, yankMsg));
}

/* end */
