/* Emacs test harness.
 *
 * This does an assortment of tests and should be run regularly.
 *
 * This is new and there are not many tests - so far.
 *
 * ADDW SCCS: @(#)meTest.c	1.8 07/14/20 01:14:29
 *
 * Copyright (C) Alain D D Williams <addw@phcomp.co.uk>, 2019
 * This is released as free software.
 * Use as you wish, no guarantee for anything at all is given.
 * License: GPL-2 or any later version.
 */

/* This program does not actually call emacs() but it calls many of the initialising
 * functions that are called when the editor starts and so we need a compiled main.c
 * that has emacs() rather than main().
 * CALLED must be defined below since it calls a version of cexit() that returns,
 * if it is not defined then calling it below screws the stack and CheckCommandNames()
 * is run again without being called ... it just end up in the code, somehow, and
 * then this falls over and core dumps.
 */
#define CALLED 1

#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"
#include	<getopt.h>

int	CheckCommandNames(void);
void	Error(char* msg, ...);
void	Note(char* msg, ...);
void	read_opts(int argc, char* const argv[]);
void	usage(int exitc);

int	quiet = 0;	// If true - less messages
int	verbose = 0;	// Be noisy

char*	usagem[] = {
	"This program does some MicroEmacs testing",
	"",
	"Usage: %s [options]",
	"Options:",
	"-h	Help",
	"-q	Quiet",
	"-v	Verbose, multiple times increases verbosity",
	"Author: Alain D D Williams <addw@phcomp.co.uk> April 2019, licenced GPL 2",
	(char*)NULL	/* Must end the list */
};

/* Error messages that should be looked at all start '!! ',
 * ie these will be unexpected errors.
 */
int	main(int argc, char* argv[])
{
	int	fails = 0;	// Count of unexpected results

	ST_wallStart = time((time_t*)NULL);

	/* Get the program name, last component of path */
	if((progname = strrchr(argv[0], '/')) != NULL)
		progname++;
	else
		progname = argv[0];

	/* Maybe some/all of this startup stuff should be in a function
	 * that is shared with Emacs startup
	 */
	home = getenv("HOME");	/* Might be NULL */
	if(!home)
		home = "";

	if((TermName = getenv("TERM")) == NULL)
		TermName = "";
	TermName = strmake_t(TermName);	// Allocate so that it can be changed

#ifdef	LINUX5
	uid_t	suid;		// Saved UID

	getresuid(&uid, &euid, &suid);
	suidRoot = uid == 0 || euid == 0 || suid == 0;
#else
	uid = getuid();
	euid = geteuid();
	suidRoot = uid == 0 || euid == 0;
#endif

	Utf8Init();
	mapinit();
	LinkBindings();		/* Initialise table linkage */
	edinit(mainBname);	/* Buffers, windows */
	varinit();		/* user variables */
	RestInit();		// Get shell name

	read_opts(argc, argv);

	fails += CheckCommandNames();


	if(fails)
		Error("Number of unexpected results: %d", fails);
	else
		Note("Number of unexpected results: %d", fails);

	// _exit() is called below
	fflush(stdout);
	fflush(stderr);

	/* Run the deallocation version of exit.
	 * This is especially good when this test is run under valgrind.
	 */
	cexit(fails ? 1 : 0);

	_exit(fails ? 1 : 0);
}

/* Run some checks in names[] - the NBIND array.
 *
 * Return non zero if there is an issue.
 * The value is the number of problems found.
 */
int	CheckCommandNames(void)
{
	int	ret = 0;
	NBIND*	np;

	// Scan everything in the table:
	for(np = CmdNames; np->n_name != NULL; np++) {
		NBIND*	pp;
		int	IsSynonym = 0;

		/* Look for synonyms and inverses
		 * Do this by comparing the function that the command runs.
		 */
		for(pp = CmdNames; pp->n_name != NULL; pp++) {

#if SHOWMODREL == 0
			/* There is a fudge that ctrlg is the command run
			 * when show-module-release is not compiled in.
			 */
			if(strcmp(np->n_name, "show-module-release") == 0 ||
			   strcmp(pp->n_name, "show-module-release") == 0)
				continue;

#endif

			if(np->n_func == pp->n_func) {

				if(np == pp)
					continue;

				IsSynonym = 1;	// Note the np is a synonym

				// Only mention the synonym once
				if(np < pp)
					Note("Found command synonyms: %s %s",
						np->n_name, pp->n_name);

				if((np->n_flag & N_SYNONYM) == 0) {
					Error("Missing SYNONYM flag: %s		%s", np->n_name, pp->n_name);
					ret++;
				}

				// Compare some things that should be the same
				if(np->n_rfunc != pp->n_rfunc) {
					Error("Synonyms do not have same reverse function: %s %s",
						np->n_name, pp->n_name);
					ret++;
				}

				// Compare some things that should be the same
				if(np->n_flag != pp->n_flag) {
					Error("Synonyms do not have same flags %s %s (%#.3x %#.3x)\n",
						np->n_name, pp->n_name,
						np->n_flag, pp->n_flag);
					ret++;
				}
			} else

			// Inverse
			if(np->n_func == pp->n_rfunc) {

				if(np->n_rfunc != pp->n_func) {
					Error("Inverse commands do not have the same inverses: %s %s",
						np->n_name, pp->n_name);
					ret++;
				}
			}
		}

		if(pp->n_name != NULL && !IsSynonym && (pp->n_flag & N_SYNONYM)) {
			Error("SYNONYM flag but none found: %s", pp->n_name);
			ret++;
		}
	}

	return(ret);
}


/* Interpret '-' arguments -options.
 * If there is an error in the arguments, print the options list and exit.
 *
 * Use standard getopt(), not GNU extended - for portability.
 *
 * It would be nice to declare the second argument: char const* const argv[]
 * This causes warnings, see:
 * https://stackoverflow.com/questions/35319842/why-c-doesnt-allow-implicit-conversion-from-char-to-const-char-const-and#
 */
void	read_opts(int argc, char* const argv[])
{
	register int	c;
	register int	error;
	char*		GetOptStr = ":hqv";

	error = 0;

	opterr = 1;	// Stop getopt_long printing error messages - we do it

	// Command line arguments:
	while((c = getopt(argc, (char* const*)argv, GetOptStr)) != -1) {
		switch(c) {
		  case 'h':
			usage(0);
		  case 'q':	// Quiet
			quiet = 1;
			break;
		  case 'v':
			verbose++;
			break;
		  case ':':
			Error("Missing argument to option: %c", optopt);
			error = 1;
			break;
		  case '?':
			Error("Unknown option: %c", optopt);
			error = 1;
			break;
		  default:
			Error("Unexpected value when parsing options: %#x", c);
			error = 1;
			break;
		}
	}

	if(error)
		usage(2);
}

/* Output a usage message and end with exit status exitc.
 */
void	usage(int exitc)
{
	register char**	mes;
	
	for(mes = usagem; *mes != (char*)NULL; mes++) {
		(void) fprintf(stderr, *mes, progname);
		fputc('\n', stderr);
	}
	fprintf(stderr, "Version: %s\n", "1.8 07/14/20");

	exit(exitc);
	/*NOTREACHED*/
}


/* Write an error message to stderr.
 * Prefix with program name, append a newline, otherwise much like printf()
 */
void	Error(char* msg, ...)
{
	va_list	ap;

	fprintf(stderr, "%s ", progname);
	va_start(ap, msg);
	vfprintf(stderr, msg, ap);
	va_end(ap);
	fprintf(stderr, "\n");
}

/* Write an error message to stdout.
 * Append a newline, otherwise much like printf()
 *
 * Do not output the message if quiet.
 */
void	Note(char* msg, ...)
{
	va_list	ap;

	if(quiet)
		return;

	va_start(ap, msg);
	vfprintf(stdout, msg, ap);
	va_end(ap);
	fprintf(stdout, "\n");
}
