/*
 * ADDW SCCS: @(#)search.c	1.89 09/22/20 22:44:40 - second
 */

/*
 * The functions in this file implement commands that search in the forward
 * and backward directions.  There are no special characters in the search
 * strings.  Probably should have a regular expression search, or something
 * like that.
 *
 * Aug. 1986 John M. Gamble:
 *	Made forward and reverse search use the same scan routine.
 *
 *	Added a limited number of regular expressions - 'any',
 *	'character class', 'closure', 'beginning of line', and
 *	'end of line'.
 *
 *	Replacement metacharacters will have to wait for a re-write of
 *	the replaces function, and a new variation of ldelete().
 *
 *	For those curious as to my references, i made use of
 *	Kernighan & Plauger's "Software Tools."
 *	I deliberately did not look at any published grep or editor
 *	source (aside from this one) for inspiration.  I did make use of
 *	Allen Hollub's bitmap routines as published in Doctor Dobb's Journal,
 *	June, 1985 and modified them for the limited needs of character class
 *	matching.  Any inefficiences, bugs, stupid coding examples, etc.,
 *	are therefore my own responsibility.
 *
 * April 1987: John M. Gamble
 *	Deleted the "if (n == 0) n = 1;" statements in front of the
 *	search/hunt routines.  Since we now use a do loop, these
 *	checks are unnecessary.  Consolidated common code into the
 *	function delins().  Renamed global mclen matchlen,
 *	and added the globals matchline, matchoff, patmatch, and
 *	mlenold.
 *	This gave us the ability to unreplace regular expression searches,
 *	and to put the matched string into an evironment variable.
 *	SOON TO COME: Meta-replacement characters!
 *
 *	25-apr-87	DML
 *	- cleaned up an unneccessary if/else in forwsearch() and
 *	  backsearch()
 *	- savematch() failed to malloc room for the terminating byte
 *	  of the match string (stomp...stomp...). It does now. Also
 *	  it now returns gracefully if malloc fails
 *
 *	July 1987: John M. Gamble
 *	Set the variables matchlen and matchoff in the 'unreplace'
 *	section of replaces().  The function savematch() would
 *	get confused if you replaced, unreplaced, then replaced
 *	again (serves you right for being so wishy-washy...)
 *
 *	August 1987: John M. Gamble
 *	Put in new function rmcstr() to create the replacement
 *	meta-character array.  Modified delins() so that it knows
 *	whether or not to make use of the array.  And, put in the
 *	appropriate new structures and variables.
 *
 * matchlen & things are variously int and unsigned int. Really ought to be
 * LineLn - but in practice will be quite enough.
 *
 * It is not worth the effort/cost of converting to/from UniChar/BYTE* when scanning,
 * so leave everything as a BYTE*. Most of the time it does not matter, but we do need
 * to know in MAGIC mode so that '.' == 1 character (maybe == several bytes).
 */

#include	"estruct.h"
#include        "utf8.h"
#include        "edef.h"

#if	SHOWMODREL
char	srchvers[] = "@(#)search.c	1.89 09/22/20 22:44:40";
#endif

#if	LATTICE
#define	void	int
#endif

static	int	readpattern(char* prompt, int srch);
static	int	replaces(int kind, int f, Cnt n);
static	BYTE*	nextch(LINE** pcurline, LineLn* pcuroff, int dir);
#if	MAGIC
static	int	cclmake(BYTE** ppatptr, MC* mcptr);
static	int	mcstr(void);
static	int	rmcstr(void);
static	int	mceq(BYTE* bc, const MC* mt);
static	void	setbit(int bc, BITMAP cclmap);
static	int	amatch(const MC* mcptr, int direct, LINE** pcwline, LineLn* pcwoff);
static	int	biteq(int bc, BITMAP cclmap);
static	BITMAP	clearbits(void);
#endif

/* Search forward.
 *
 * Get a search string from the user, and search for the string. If found,
 * reset the "." to be just after the match string, and (perhaps) repaint the
 * display.
 *
 * Find the count'th occurence of the string.
 *
 * Default count 1. Must be >= 0.
 *
 * Command:  search-forward
 * Bound to: C-s   M-/
 *
 * Return: TODO
 */
int	forwsearch(int f, Cnt n)
{
	int	status = TRUE;

	/* remember the initial . on entry: */
	pnmark0();

	/* Ask the user for the text of a pattern.  If the
	 * response is TRUE (responses other than FALSE are
	 * possible), search for the pattern for as long as
	 * n is positive (n == 0 will go through once, which
	 * is just fine).
	 */
	if((status = readpattern("Search: ", TRUE)) == TRUE) {
		do {
#if	MAGIC
			if((magical && curwp->w_bufp->b_mode & MDMAGIC) != 0)
				status = mcscanner(&mcpat[0], FORWARD, PTEND);
			else
#endif
				status = scanner(searchPat, FORWARD, PTEND);
		} while((--n > 0) && status);

		/* Save away the match, or complain
		 * if not there.
		 */
		if(status == TRUE)
			savematch(patmatch);
		else
			mlwrite(MsgNotFound);
	}

	if(status == TRUE)
		pushNmark();

	return(status);
}

/* Search forward for a previously acquired search string.
 *
 * If found, reset the "." to be just after the match string, and (perhaps)
 * repaint the display.
 *
 * Find the count'th occurence of the string.
 *
 * Default count 1. Will be >= 0.
 *
 * Command:  hunt-forward
 * Bound to: unbound
 *
 * Return: TODO
 */
int	forwhunt(int f, Cnt n)
{
	int	status = TRUE;

	/* remember the initial . on entry: */
	pnmark0();

	/* Make sure a pattern exists, or that we didn't switch
	 * into MAGIC mode until after we entered the pattern.
	 */
	if(tused(searchPat) == 0)
	{
		mlwrite("%ENo pattern set");
		return FALSE;
	}
#if	MAGIC
	if((curwp->w_bufp->b_mode & MDMAGIC) != 0 &&
		 mcpat[0].mc_type == MCNIL)
	{
		if(!mcstr())
			return FALSE;
	}
#endif

	/* Search for the pattern for as long as
	 * n is positive (n == 0 will go through once, which
	 * is just fine).
	 */
	do
	{
#if	MAGIC
		if((magical && curwp->w_bufp->b_mode & MDMAGIC) != 0)
			status = mcscanner(&mcpat[0], FORWARD, PTEND);
		else
#endif
			status = scanner(searchPat, FORWARD, PTEND);
	} while((--n > 0) && status);

	/* Save away the match, or complain
	 * if not there.
	 */
	if(status == TRUE) {
		pushNmark();
		savematch(patmatch);
	} else {
		mlwrite(MsgNotFound);
	}

	return(status);
}

/* Reverse search.
 *
 * Get a search string from the user, and search, starting at "." and
 * proceeding toward the front of the buffer. If found "." is left pointing at
 * the first character of the pattern (the last character that was matched).
 *
 * Find the count'th occurence of the string.
 *
 * Default count 1. Will be >= 0.
 *
 * Command:  search-reverse
 * Bound to: unbound
 *
 * Return: TODO
 */
int	backsearch(int f, Cnt n)
{
	int	status = TRUE;

	/* remember the initial . on entry: */
	pnmark0();

	/* Ask the user for the text of a pattern.  If the
	 * response is TRUE (responses other than FALSE are
	 * possible), search for the pattern for as long as
	 * n is positive (n == 0 will go through once, which
	 * is just fine).
	 */
	if((status = readpattern("Reverse search: ", TRUE)) == TRUE) {
		do {
#if	MAGIC
			if((magical && curwp->w_bufp->b_mode & MDMAGIC) != 0)
				status = mcscanner(&tapcm[0], REVERSE, PTBEG);
			else
#endif
				status = scanner(searchPat, REVERSE, PTBEG);
		} while((--n > 0) && status);

		/* Save away the match, or complain
		 * if not there.
		 */
		if(status == TRUE)
			savematch(patmatch);
		else
			mlwrite(MsgNotFound);
	}

	if(status == TRUE)
		pushNmark();

	return(status);
}

/* Reverse search for a previously acquired search string.
 *
 * Starting at "." and proceeding toward the front of the buffer. If found "."
 * is left pointing at the first character of the pattern (the last character
 * that was matched).
 *
 * Find the count'th occurence of the string.
 *
 * Default count 1. Will be >= 0.
 *
 * Command:  hunt-backward
 * Bound to: unbound
 *
 * Return: TODO
 */
int	backhunt(int f, Cnt n)
{
	int	status = TRUE;

	/* Make sure a pattern exists, or that we didn't switch
	 * into MAGIC mode until after we entered the pattern.
	 */
	if(tused(searchPat) == 0)
	{
		mlwrite("%ENo pattern set");
		return FALSE;
	}
#if	MAGIC
	if((curwp->w_bufp->b_mode & MDMAGIC) != 0 &&
		 tapcm[0].mc_type == MCNIL)
	{
		if(!mcstr())
			return FALSE;
	}
#endif

	/* remember the initial . on entry: */
	pnmark0();

	/* Go search for it for as long as
	 * n is positive (n == 0 will go through once, which
	 * is just fine).
	 */
	do
	{
#if	MAGIC
		if((magical && curwp->w_bufp->b_mode & MDMAGIC) != 0)
			status = mcscanner(&tapcm[0], REVERSE, PTBEG);
		else
#endif
			status = scanner(searchPat, REVERSE, PTBEG);
	} while((--n > 0) && status);

	/* Save away the match, or complain
	 * if not there.
	 */
	if(status == TRUE) {
		pushNmark();
		savematch(patmatch);
	} else {
		mlwrite(MsgNotFound);
	}

	return(status);
}

#if	MAGIC
/* Search for a meta-pattern in either direction.
 *
 * If found, reset the "." to be at the start or just after the match string,
 * and (perhaps) repaint the display.
 *
 * mcpatrn	pointer into pattern
 * direct	which way to go: REVERSE or FORWARD
 * beg_or_end	put point at beginning or end of pattern.
 *
 * Return: TRUE found a match
 *	   FALSE did not find a match. "." will be were it was on entry
 */
int	mcscanner(const MC* mcpatrn, int direct, int beg_or_end)
{
	LINE*	curline;		/* current line during scan */
	LineLn	curoff;			/* position within current line */
	LINE*	startline;		// "." at start, before moved
	LineLn	startoff;

	/* If we are going in reverse, then the 'end' is actually
	 * the beginning of the pattern.  Toggle it.
	 */
	beg_or_end ^= direct;

	/*
	 * Save the old matchlen length, in case it is
	 * very different (closure) from the old length.
	 * This is important for query-replace undo
	 * command.
	 */
	mlenold = matchlen;

	/* Setup local scan pointers to global ".".
	 */
	curline = curwp->w_dotp;
	curoff  = curwp->w_doto;
	startline = curline;
	startoff  = curoff;

	/* Ensure that following any move due to search that the display routines
	 * will scroll left/right as needed - clear CFCPCN.
	 */
	lastflag = 0;

	/* Scan each character until we hit the head link record.
	 */
	while(!boundry(curline, curoff, direct))
	{
		/* Save the current position in case we need to
		 * restore it on a match, and initialize matchlen to
		 * zero in case we are doing a search for replacement.
		 */
		matchline = curline;
		matchoff = curoff;
		matchlen = 0;

		if(amatch(mcpatrn, direct, &curline, &curoff))
		{
			/* A SUCCESSFULL MATCH!!!
			 * reset the global "." pointers.
			 */
			if(beg_or_end == PTEND)	/* at end of string */
			{
				curwp->w_dotp = curline;
				curwp->w_doto = curoff;
			}
			else		/* at beginning of string */
			{
				curwp->w_dotp = matchline;
				curwp->w_doto = matchoff;
			}

			curwp->w_flag |= WFMOVE; /* flag that we have moved */

			return(TRUE);
		}

		/* Advance the cursor.
		 */
		nextch(&curline, &curoff, direct);
	}

	curwp->w_dotp = startline;
	curwp->w_doto = startoff;

	return FALSE;	/* We could not find a match.*/
}

/* Search for a meta-pattern in either direction.
 *
 * Based on the recursive routine amatch() (for "anchored match") in
 * Kernighan & Plauger's "Software Tools".
 *
 * mcptr	string to scan for
 * direct	which way to go: REVERSE or FORWARD
 * pcwline	current line during scan
 * pcwoff	position within current line
 *
 * BEWARE: possible bug: might not detect hitting end of buffer
 *
 * Return TODO
 */
static int	amatch(const MC* mcptr, int direct, LINE** pcwline, LineLn* pcwoff)
{
	BYTE*	cp;			/* Start of current character */
	LINE*	curline;		/* current line during scan */
	LineLn	curoff;			/* position within current line */
	LineLn	nchars;			/* Number of characters (not BYTEs) matched */

	/* Set up local scan pointers to ".", and get
	 * the current character.  Then loop around
	 * the pattern pointer until success or failure.
	 */
	curline = *pcwline;
	curoff = *pcwoff;

	/* The beginning-of-line and end-of-line metacharacters
	 * do not compare against characters, they compare
	 * against positions.
	 * BOL is guaranteed to be at the start of the pattern
	 * for forward searches, and at the end of the pattern
	 * for reverse searches.  The reverse is true for EOL.
	 * So, for a start, we check for them on entry.
	 */
	if(mcptr->mc_type == BOL)
	{
		if(curoff != 0)
			return FALSE;
		mcptr++;
	}

	if(mcptr->mc_type == EOL)
	{
		if(curoff != lused(curline))
			return FALSE;
		mcptr++;
	}

	while(mcptr->mc_type != MCNIL)
	{
		cp = nextch(&curline, &curoff, direct);

		if(mcptr->mc_type & CLOSURE)
		{
			/* Try to match as many characters as possible
			 * against the current meta-character.  A
			 * newline never matches a closure.
			 */
			nchars = 0;
			while(*cp != '\n' && mceq(lgetcaddr(curline, curoff), mcptr))
			{
				cp = nextch(&curline, &curoff, direct);
				nchars++;
			}

			/* We are now at the character that made us
			 * fail.  Try to match the rest of the pattern.
			 * Shrink the closure by one for each failure.
			 * Since closure matches *zero* or more occurences
			 * of a pattern, a match may start even if the
			 * previous loop matched no characters.
			 */
			mcptr++;

			for(;;)
			{
				cp = nextch(&curline, &curoff, direct ^ REVERSE);

				if(amatch(mcptr, direct, &curline, &curoff))
				{
					matchlen += nchars;
					goto success;
				}

				if(nchars == 0)
					return FALSE;
				nchars--;
			}
		}
		else			/* Not closure.*/
		{
			/* The only way we'd get a BOL metacharacter
			 * at this point is at the end of the reversed pattern.
			 * The only way we'd get an EOL metacharacter
			 * here is at the end of a regular pattern.
			 * So if we match one or the other, and are at
			 * the appropriate position, we are guaranteed success
			 * (since the next pattern character has to be MCNIL).
			 * Before we report success, however, we back up by
			 * one character, so as to leave the cursor in the
			 * correct position.  For example, a search for ")$"
			 * will leave the cursor at the end of the line, while
			 * a search for ")<NL>" will leave the cursor at the
			 * beginning of the next line.  This follows the
			 * notion that the meta-character '$' (and likewise
			 * '^') match positions, not characters.
			 */
			if(mcptr->mc_type == BOL) {
				if(curoff == lused(curline)) {
					cp = nextch(&curline, &curoff,
						   direct ^ REVERSE);
					goto success;
				}
				else
					return FALSE;
			}

			if(mcptr->mc_type == EOL) {
				if(curoff == 0) {
					cp = nextch(&curline, &curoff,
						   direct ^ REVERSE);
					goto success;
				}
				else
					return FALSE;
			}

			/* Neither BOL nor EOL, so go through
			 * the meta-character equal function.
			 */
			if(!mceq(cp, mcptr))
				return FALSE;
		}

		/* Increment the length counter and
		 * advance the pattern pointer.
		 */
		matchlen++;
		mcptr++;
	}			/* End of mcptr loop.*/

	/* A SUCCESSFULL MATCH!!!
	 * Reset the "." pointers.
	 */
success:
	*pcwline = curline;
	*pcwoff  = curoff;

	return TRUE;
}
#endif

/* Search for a pattern in either direction.
 *
 * If found, reset the "." to be at the start or just after the match string,
 * and (perhaps) repaint the display.
 *
 * spatrn	string to scan for
 * direct	which way to go: FORWARD or REVERSE
 * beg_or_end	put point at beginning or end of pattern.
 *
 * This used to require a different spatrn depending on direct,
 * ie when REVERSE the pattern was reversed. This was rewritten when
 * multi-byte character searching was done. Now it just needs the
 * forward string and simply scans  it in reverse direction.
 *
 * This will understands eq() saying EQ_MORE - space matching white space.
 * But a pattern of a single space will not match a newline.
 *
 * Global matchlen is assigned the number of characters matched - can vary
 * as a space can match one or more characters.
 *
 * This is also used in C-tag (both via command and also command line -t option),
 * goto-matching-fence (via movefunc()) & ShowMatchingBrace().
 *
 * Return TRUE if found something, "." will be near it
 *	 FALSE if did not find something, "." will be where it was
 */
int	scanner(TEXT* spatrn, int direct, int beg_or_end)
{
	BYTE*	cp;			/* Address character at current position */
	BYTE*	patptr;			/* pointer into pattern */
	LINE*	curline;		/* current line during scan */
	LineLn	curoff;			/* position within current line */
	LINE*	scanline;		/* current line during scanning */
	LineLn	scanoff;		/* position in scanned line */
	BYTE*	patLast;		/* Last byte in spatrn */
	BYTE*	FirstMatch;		/* The first byte that we will match in spatrn */
	int	utf8 = curbp->b_mode & MDUTF8;	// Current buffer UTF-8 mode
	LINE*	startline;		// "." at start, before moved
	LineLn	startoff;

	int	patIsSp;		// The pattern is a single space
	patIsSp = tused(spatrn) == 1 && tgetch(spatrn, 0) == ' ';

	/* Note that FirstMatch is the last BYTE when direct == REVERSE
	 */
	patLast = tgetcaddr(spatrn, tused(spatrn) - 1);
	if(utf8)
		UTF8_BACK_FIND_START(patLast);

	FirstMatch = ((direct == FORWARD) ? tgetcaddr(spatrn, 0) : patLast);

	/* If we are going in reverse, then the 'end' is actually
	 * the beginning of the pattern.  Toggle it.
	 */
	beg_or_end ^= direct;

	/* Set up local pointers to global ".".
	 */
	curline = curwp->w_dotp;
	curoff  = curwp->w_doto;
	startline = curline;
	startoff  = curoff;

	/* Ensure that following any move due to search that the display routines
	 * will scroll left/right as needed - clear CFCPCN.
	 */
	lastflag = 0;

	/* Scan each character until we hit the head link record.
	 */
	while(!boundry(curline, curoff, direct))
	{
		int	eqRes;			// What eq() returns
		int	lastRes = EQ_DIFFER;	// eqRes last time

		/* Save the current position in case we match
		 * the search string at this point.
		 */
		matchline = curline;
		matchoff = curoff;

		/* Get the character resolving newlines, and
		 * test it against first char in pattern - in the direction
		 * that we are going.
		 */
		cp = nextch(&curline, &curoff, direct);

		// Does it match ?
		// Pattern " " must not match '\n'.
		eqRes = patIsSp && *cp == '\n' ? EQ_DIFFER : eq(cp, FirstMatch);
		// mlwrite("%#scanner patIsSp=%d spatrn='%t' eq('%u', '%u') -> %d", patIsSp, spatrn, cp, FirstMatch, eqRes);

		if(eqRes)	// It matches
		{
			/* Setup scanning pointers.
			 */
			scanline = curline;
			scanoff = curoff;
			patptr = direct == FORWARD ? tgetcaddr(spatrn, 0) : patLast;

			lastRes = eqRes;

			matchlen = 1;	// Matched first of, hopefully, many

			/* Scan through the pattern for a match.
			 * Match a buffer character with a pattern character.
			 * If eq() says EQ_MORE then try to match the next buffer
			 * character with the current pattern character.
			 *
			 * When EQ_MORE becomes EQ_DIFFER re-try the next pattern character
			 * with the current buffer character.
			 *
			 * getNextCh is set false when we want to do the re-try.
			 */
			int	getNextCh = 1;
			LINE*	lastScanline;		// scanline before last nextch()
			LineLn	lastScanoff;		// scanoff  before last nextch()

			lastScanline = scanline;	// Purely to keep the compiler happy
			lastScanoff = scanoff;

			while(1) {
				int	WantEndOfPattern = 0;
				// Detect/prevent wrap start/end of buffer:
				if(*cp == '\n' && scanline == (direct == FORWARD ? curbp->b_abotline : curbp->b_btopline)) {
					// mlwrite("%#Search hit end of buffer");

					// Forwards: Might have matched the \n at last line with \n in pattern
					// EQ_MORE - only OK if now at pattern end (tested below)
					if(eqRes == EQ_MATCH && *patptr != '\n' && direct == FORWARD)
						goto failSearch;

					// Reverse: no excuse for looking beyond start of buffer
					if(direct == REVERSE)
						goto failSearch;

					eqRes = EQ_MATCH;
					WantEndOfPattern = 1;
				}

				// Get pointer to the next character in the pattern
				// When we reach the patter end - break takes us to success
				if(eqRes == EQ_MATCH) {
					if(direct == FORWARD) {
						patptr += utf8 ? CHAR_LEN(*patptr) : 1;
						if(patptr > patLast)
							break;	// Reached end of pattern - so matched

						// Hit end of buffer, not end of pattern: fail
						if(WantEndOfPattern)
							goto failSearch;

					} else {
						// Reverse search, find the previous character in the pattern
						if(patptr <= ttext(spatrn))
							break;	// Reached start of pattern - so matched

						patptr--;

						if(utf8)
							UTF8_BACK_FIND_START(patptr);
					}
				}

				// Get next if not retrying the last one
				lastScanline = scanline;
				lastScanoff = scanoff;
				if(getNextCh)
					cp = nextch(&scanline, &scanoff, direct);

				// Does the next character match ?
				eqRes = eq(cp, patptr);
				// mlwrite("%#scanner eq('%u', '%u') -> %d lastRes=%d", cp, patptr, eqRes, lastRes);

				if(eqRes == EQ_DIFFER) {
					// Last character did not match
					// Was that after one pattern character matching multiple buffer characters ?
					if(lastRes == EQ_MORE) {
						// See if the current char matches the next pat char
						eqRes = lastRes = EQ_MATCH;

						/* We are retesting the last buffer char. If we hit forward end of
						 * pattern we undo the last buffer forward movement
						 */
						getNextCh = 0;

						continue;
					}
					goto fail;
				}

				lastRes = eqRes;
				getNextCh = 1;
				matchlen++;
			}

			/* The string/pattern matches.
			 *
			 * Do we need to check that it is not adjacent to alphanumeric or '_' ?
			 */
			if(matchNoAdjAlnum) {
				UniChar	ch;	// Character to test

				/* Going forwards: get the character before the start of string
				 * going backwards: get the same character but referenced differently
				 */
				ch = direct == FORWARD ?
				    (matchoff <= 0 ? '\n' : lgetc(matchline, matchoff - 1)) :
				    (scanoff  <= 0 ? '\n' : lgetc(scanline,  scanoff - 1));

				if(Isalnum(ch) || ch == '_')
					continue;	// Continue the search

				// Check end of pattern
				ch = direct == FORWARD ?
				    (scanoff  >= lused(scanline)  ? '\n' : lgetc(scanline,  scanoff)) :
				    (matchoff >= lused(matchline) ? '\n' : lgetc(matchline, matchoff));

				if(Isalnum(ch) || ch == '_')
					continue;	// Continue the search
			}

			// A SUCCESSFULL MATCH!!!

			/* If last test was a DIFFER after a MORE sequence
			 * undo the last buffer movement forward/backwards
			 */
			if(getNextCh == 0) {
				scanline = lastScanline;
				scanoff = lastScanoff;
			}

			// Reset the global "." pointers
			if(beg_or_end == PTEND)	/* at end of string */
			{
				curwp->w_dotp = scanline;
				curwp->w_doto = scanoff;
			}
			else		/* at beginning of string */
			{
				curwp->w_dotp = matchline;
				curwp->w_doto = matchoff;
			}

			/* Flag that we have moved.
			 * Reframe (WFMOVE) would be good, to get line number recalculate - if they are being shown;
			 * but this does not look good on interactive search.
			 * BUG if line numbers are being shown.
			 */
			curwp->w_flag |= WFMOVE;

			return(TRUE);

		}

fail:;		/* continue to search */
	}

failSearch:	// Fail - did not find anything
	curwp->w_dotp = startline;
	curwp->w_doto = startoff;

	return(FALSE);	/* We could not find a match */
}

/* Compare two characters.
 * 
 * subc	is from the subject string - from the buffer/where-ever
 * patc	is from the pattern.
 *
 * If global eqCompare_i compare in a case insensitive manner - if so it is
 * assumed that *patc has been downcased.
 * Unicode recommends downcase when case folding for comparison.
 *
 * If global spaceMatchWhite a pattern space will match a subject space, tab or newline.
 * This will return EQ_MORE if it matches - request to try the next subject
 * character as that also might match -- so a space can thus match several
 * spaces, tabs and newlines.
 *
 * Case folding is a bit naff, it will be until we can deal with downcasing
 * anything other than Unicode Basic Latin.
 *
 * Beware: UTF-8 mode is taken from the current buffer, this might not be correct
 * eg when called from a macro (eval.c).
 *
 * Return:
 * 	EQ_DIFFER	if they differ
 *	EQ_MATCH	if the two characters match
 *	EQ_MORE		they match, try again with the next subject character
 */
int	eq(const BYTE* subc, const BYTE* patc)
{
	/* If multibyte: compare them
	 */
	if((curbp->b_mode & MDUTF8) && (CHAR_LEN(*subc) != 1 || CHAR_LEN(*patc) != 1)) {
		if(CHAR_LEN(*subc) != CHAR_LEN(*patc))
			return(EQ_DIFFER);

		return( !memcmp(subc, patc, CHAR_LEN(*subc)) ? EQ_MATCH : EQ_DIFFER);
	}

	BYTE	sub = *subc;
	BYTE	pat = *patc;

	// If not EXACT then pattern space matches space, tab or newline
	if(pat == ' ' && matchSpaceWhite && (sub == ' ' || sub == '\t' || sub == '\n'))
		return(EQ_MORE);

	// Downcase the subject (buffer) character if needs be:
	if(eqCompare_i && Isupper(sub))
		sub = (BYTE)tolower(sub);

	return(sub == pat ? EQ_MATCH : EQ_DIFFER);
}

/* Used to examine the search, matched & replacement strings.
 *
 * The purpose is upper/lower case so that we can transform what
 * we substitute to have matching caseness.
 *
 * desc		description filled in
 * text		the line that is examined
 *
 * 2 different definitions of words: set of alpha, white space separated stuff
 *
 * The transform flag is set true if only lower case characters are seen.
 *
 * This only works with ASCII upper/lower
 */
void	getRepDesc(wDesc* desc, TEXT* text)
{
	BYTE*	txt = ttext(text);
	BYTE*	txtEnd = txt + tused(text);
	int	word = 0;	// 0: first word, 1: all subsequent words
	int	words = 0;	// # words in text
	Case	wd;		// Case of 1st or subsequent words

	desc->wd_line = text;

	desc->wd_newlines = 0;
	desc->wd_case[0] = desc->wd_case[1] = Np;	// Not known (yet)

	// Look at letter separated words:
	while(txt < txtEnd) {
		int	Fu, Fl, Su, Sl;

		// Find first letter
		while(txt < txtEnd && !Isletter(*txt)) {
			txt++;
		}

		Fu = Fl = Su = Sl = 0;	// None of these seen (yet)

		wd = Np;		// Not known/present

		// Look at the first letter
		if(txt < txtEnd) {
			if(Isupper(*txt))
				Fu = 1;
			if(Islower(*txt))
				Fl = 1;
			txt++;
		}

		// Look at any subsequent letters
		while(txt < txtEnd && Isletter(*txt)) {
			if(Isupper(*txt))
				Su = 1;
			if(Islower(*txt))
				Sl = 1;
			txt++;
		}

		if(Fu)
			wd = Su && Sl ? Oc : Su ? Uc : Tc;

		if(Fl)
			wd = Su ? Oc : Lc;
				
		// Record what we learned about the first word or subsequent words
		if(desc->wd_case[word] == Np)
			desc->wd_case[word] = wd;
		else if(desc->wd_case[word] != wd)
			desc->wd_case[word] = Oc;

		word = 1;
	}

	desc->wd_noUpper = desc->wd_case[0] == Lc &&
		(desc->wd_case[1] == Lc || desc->wd_case[1] == Np);

	// Second scan to learn about white space separated words:
	txt = ttext(text);
	while(txt < txtEnd) {
		while(txt < txtEnd && (*txt == ' ' ||*txt == '\t' ||*txt == '\n')) {
			if(*txt == '\n')
				if(desc->wd_newlines++ == 0) {
					desc->wd_nlAddr = txt;
					desc->wd_nlWord = words;
				}

			txt++;
		}

		// Skip next word (non-white space)
		while(txt < txtEnd && *txt != ' ' && *txt != '\t' && *txt != '\n')
			txt++;

		words++;
	}
}

/* Mangle the replacement string so that its capitalisation matches the
 * capitalisation in the text that was found.
 *
 * replaceText	The text as entered by the user.
 *		Alphas will only be lower case.
 * fD		Description of the text that was found
 *
 * This only works with ASCII upper/lower
 * Title case is wrong for Unicode.
 * 
 * Return playText - which is where the munged version is put.
 */
TEXT*	CapsReplaceMangle(TEXT* replaceText, wDesc fD)
{
	BYTE*	txt = ttext(replaceText);
	int	word = 0;
	Case	conv;		// What we convert to

	tsetempty(playText);

	while(txt < tbend(replaceText)) {
		// Copy text before the next word
		// Ignore _ so can get Title case
		while(txt < tbend(replaceText) && !Isletter(*txt))
			if(tcatm(playText, txt++, 1) == FALSE)
				return(NULL);

		conv = fD.wd_case[word];
		while(txt < tbend(replaceText) && Isletter(*txt)) {
			BYTE	c = *txt++;

			if((conv == Tc || conv == Uc))
				c = (BYTE)toupper(c);

			if(tcatm(playText, &c, 1) == FALSE)
				return(NULL);

			// Done first: the rest is lower case
			if(conv == Tc)
				conv = Lc;
		}

		word = 1;	// Subsequent words
	}

	return(playText);
}

/* Mangle the replacement line inserting newlines if needed.
 *
 * replaceText	The text that will replace the matched text
 *		This is in a temporary buffer and will be modified
 * fD		Description of the text that was found by search
 *
 * Return TODO
 */
int	NlReplaceMangle(TEXT* replaceText, wDesc fD)
{
	BYTE*	txt = ttext(replaceText);
	int	words = 0;	// # words in text

	while(txt < tbend(replaceText)) {
		// Found the word that the newline needs to be inserted before 
		if(words == fD.wd_nlWord) {
			// Delete spaces
			while(txt < tbend(replaceText) && *txt == ' ') {
				// Unlikely to have more than 1, so don't care speed
				tdeln(replaceText, (LineLn)(txt - ttext(replaceText)), 1);
			}

			// Replace them by white space from the found string
			while(fD.wd_nlAddr < tbend(fD.wd_line) &&
			      (*fD.wd_nlAddr == ' ' || *fD.wd_nlAddr == '\t' || *fD.wd_nlAddr == '\n'))
				if(tinsm(replaceText, (LineLn)(txt++ - ttext(replaceText)), fD.wd_nlAddr++, 1) == FALSE)
					return(FALSE);
			break;
		}

		// Move to next word (white space separated)
		while(txt < tbend(replaceText) && (*txt == ' ' || *txt == '\t' || *txt == '\n'))
			txt++;

		words++;

		// Gobble word
		while(txt < tbend(replaceText) && *txt != ' ' && *txt != '\t' && *txt != '\n')
			txt++;
	}

	return(TRUE);
}


/* Decide if searching should be caseless.
 * Set eqCompare_i & searchPat & patternHasUpper.
 *
 * This looks as searchPat_e - so call this after it has been set.
 * RestoreSearchOpts() should probably be called just before.
 *
 * Also set matchSpaceWhite - (does space match white space) ?
 * Default is spaceMatchWhite, but set FALSE if a TAB or newline is seen.
 *
 * Two adjacent spaces unsets matchSpaceWhite. These two could not
 * match anything anyway (the first would eat the second's lunch).
 *
 * If the buffer mode EXACT is NOT set then check that the search
 * pattern does not contain any upper case character
 *
 * Part of the point of this is to construct searchPat_d - a downcased
 * version of searchPat_e. This is so that eq() does not need to downcase
 * the pattern - downcasing will be more expensive with chars > U+80.
 */
void	CheckCaselessSearch(void)
{
	LineLn	i;
	int	utf8 = (curbp->b_mode & MDUTF8);
	int	eq_i;

	patternHasUpper = FALSE;
	patternIsIdent  = TRUE;		// The pattern is like a C identifier

	tsetempty(searchPat_d);

	/* Set the option that SPC matches white space.
	 * Not if already set by a C-o option
	 * Only if not MAGIC or EXACT:
	 */
	if(canSetSpW)
		matchSpaceWhite = ((curwp->w_bufp->b_mode & (MDMAGIC | MDEXACT)) == 0) && spaceMatchWhite;

	/* Scan the entered pattern (searchPat_e) for upper case characters,
	 * copying them to searchPat_d.
	 */
	for(i = 0; i < tused(searchPat_e); ) {
		BYTE	c = tgetch(searchPat_e, i);
		LineLn	chLen = utf8 ? CHAR_LEN(c) : 1;

		// If TAB or NL then space only matches space:
		if(c == '\t' || c == '\n')
			matchSpaceWhite = FALSE;

		// Two adjacent spaces ?
		if(c == ' ' && i > 0 && tgetch(searchPat_e, i - 1) == ' ')
			matchSpaceWhite = FALSE;

		if(Isupper(c)) {
			// Only for ASCII for now. Unicode TODO
			// Thus length will be 1
			patternHasUpper = TRUE;
			tcatc(searchPat_d, tolower(c));
		} else {	// Copy the character
			tcatm(searchPat_d, tgetcaddr(searchPat_e, i), chLen);
		}

		i += chLen;
	}

	// Should we do caseless searching ?
	// What we set eqCompare_i to if we can
	if((curwp->w_bufp->b_mode & MDEXACT) != 0 || patternHasUpper) {
		// Do case sensitive
		eq_i = FALSE;
	} else {
		// Do case insensitive
		eq_i = TRUE;
	}

	// Can we set eqCompare_i - ie not set by user option
	if(canSetEq_i)
		eqCompare_i = eq_i;

	// Set search pattern as entered or downcased one:
	searchPat = eqCompare_i ? searchPat_d : searchPat_e;
}

/* Prompt and change search options in effect.
 *
 * opts	says which ones can be changed
 *
 * Uses/sets global variables:
 *	canSetEq_i eqCompare_i
 *	canSetSpW matchSpaceWhite
 *	matchNoAdjAlnum
 *
 * Return the options encoded a bit pattern - SO_ values.
 */
BYTE	EditSearchOpts(int opts)
{
	KeyCode	c;
	int	i;

	while(1) {
		c = mlGetKey(MGC_MAP, NULL,
		    "i)nsensitive s)pace is space w)ord match o)pts-show : ");
		if(c == abortc || c == ' ' || c == '\n' || c == '\r')
			return((BYTE)(
			    (canSetEq_i ? 0 : (SO_EQI_flag | (eqCompare_i     ? SO_EQI_val : 0))) |
			    (canSetSpW  ? 0 : (SO_SMW_flag | (matchSpaceWhite ? SO_SMW_val : 0))) |
			    (matchNoAdjAlnum ? SO_WORD_val : 0)
			));

		switch(c) {
		  case 'i':
			i = mlyesno("Case insensitive match [y/n]? ");
			if(i != ABORT) {
				canSetEq_i = FALSE;
				eqCompare_i = i;
			}
			break;
		  case 's':
			i = mlyesno("Space matches one or more: SPC, TAB or NL [y/n]? ");
			if(i != ABORT) {
				canSetSpW = FALSE;
				matchSpaceWhite = i;
			}
			break;
		  case 'w':
			i = mlyesno("Match must not have alphanumeric or _ before/after [y/n]? ");
			if(i != ABORT) {
				// This function can always set it - not settable anywhere else
				matchNoAdjAlnum = i;
			}
			break;
		  case 'o':	// Show current options
				// Use ltoyn() as that is what the prompts ask
			printToText(playText3, "Ignore case difference: %s;", ltoyn(eqCompare_i));
			printToText(playText3, "%A Space match white: %s ($space-match-white=%s)",
				ltoyn(matchSpaceWhite), ltos(spaceMatchWhite));
			printToText(playText3, "%A No adjacent alphnum=%s", ltoyn(matchNoAdjAlnum));
			mlGetKey(MGC_MAP, NULL, "%t", playText3);
			break;
		}
	}
}

/* Restore search opts as encoded in the return value of EditSearchOpts()
 *
 * This sets the same global variables.
 */
void	RestoreSearchOpts(short opts)
{
	canSetEq_i = canSetSpW = TRUE;
	matchNoAdjAlnum = FALSE;

	if((opts & SO_EQI_flag) != 0) {
		canSetEq_i = FALSE;
		eqCompare_i = (opts & SO_EQI_val) != 0;
	}

	if((opts & SO_SMW_flag) != 0) {
		canSetSpW = FALSE;
		matchSpaceWhite = (opts & SO_SMW_val) != 0;
	}

	if((opts & SO_WORD_val) != 0)
		matchNoAdjAlnum = TRUE;
}

/* Read a pattern.  Stash it in history->b_dotp.
 *
 * If it is the search string, create the reverse pattern and the magic pattern,
 * assuming we are in MAGIC mode (and defined that way).
 * These are written to globals: searchPat_e or searchRep,.
 *
 * If srch is true then the pattern is a search pattern, else a replacement.
 *
 * Change: dfault param not given; can always edit the S & R strings
 *
 * Display the old pattern, in the style of Jeff Lomicka.
 *
 * Return as mlprompt():
 *     TRUE  - OK to process
 *     FALSE - no input
 *     ABORT - C-g or C-c entered
 *     FAILED- indicates out of memory.
 *		complain and return as ABORT.
 */
static	int	readpattern(char* prompt, int srch)
{
	int	status;
	BUFFER*	history = srch ? bmlSrchHistp : bmlRepHistp;

	// default: always give (if there) & always allow edit
	if((status = mlprompt(prompt, MLSTRING | MLDEFISLAST, history)) == TRUE) {
		if(srch)	/* If we are doing the search string.*/
		{
			tcpyl(searchPat_e, history->b_dotp);

			if(srch) {
				// The history line might have associated search options:
				RestoreSearchOpts(lnotes(bmlSrchHistp->b_dotp));

				CheckCaselessSearch();
			}

			// Remember the length (characters) for substitution purposes.
			mlenold = matchlen = (curbp->b_mode & MDUTF8) ? (LineLn)utf8_chrnlen(ttext(searchPat), (size_t)tused(searchPat)) : tused(searchPat);
		}
		else
			tcpyl(searchRep, history->b_dotp);

#if	MAGIC
		/* Only make the meta-pattern if in magic mode,
		 * since the pattern in question might have an
		 * invalid meta combination.
		 */
		if((curwp->w_bufp->b_mode & MDMAGIC) == 0)
		{
			mcclear();
			rmcclear();
		}
		else
			status = srch ? mcstr(): rmcstr();
#endif
	}

	return(status);
}

/* We found the pattern?  Let's save it away to saveto.
 *
 * saveto is NUL terminated.
 *
 * The original purpose was done so that we have the matched string in case an
 * UNDO is requested by the user; it is also used in magic replace where
 * mcr_type = DITTO -- which means insert the text replaced.
 *
 * It is now also used in query-replace-string (qreplace()) so that the text
 * that was actually matched can be shown as part of the prompt.
 *
 * The loop is needed since nextch() can return '\n', ie the
 * matched string can span lines.
 */
void	savematch(TEXT* saveto)
{
	int	j;
	LINE*	curline;	/* line of last match */
	LineLn	curoff;		/* offset "      "    */
	BYTE*	cp;

	tsetempty(saveto);
	curoff = matchoff;
	curline = matchline;

	for(j = 0; j < matchlen; j++) {
		cp = nextch(&curline, &curoff, FORWARD);
		tcatm(saveto, cp, (curbp->b_mode & MDUTF8) ? CHAR_LEN(*cp) : 1);
	}

	// saveto is NUL terminated, no need for:
	// tcatm(saveto, "", 1);

	// The above will fail gracefully on OOM
}

/* Search and replace.
 *
 * n:	# of repetitions wanted. If unset no limit.
 *
 * Count, if set, must be >= 0.
 * The default count is not set.
 *
 * Command:  replace-string
 * Bound to: M-r
 *
 * Return: TODO
 */
int	sreplace(int f, Cnt n)
{
	return(replaces(FALSE, f, n));
}

/* qreplace -- search and replace with query.
 *
 * n:	# of repetitions wanted. If unset no limit.
 *
 * Count, if set, must be >= 0.
 * The default count is not set.
 *
 * Command:  query-replace-string
 * Bount to: M-C-r
 *
 * Return: TODO
 */
int	qreplace(int f, Cnt n)
{
	return(replaces(TRUE, f, n));
}

/* Search for a string and replace it with another string.
 *
 * Count may not be -ve.
 *
 * queryRep	If TRUE each replace will be queried.
 *		FALSE - unconditional replace of everything that matches
 * f		default flag. If unset no limit on replacements.
 * n		# of repetitions wanted
 *
 * Once the prompts have happened, place mark 1 before we (try to) move.
 *
 * Do as with isearch() in repositioning "." when it becomes visible again
 * after move-window-up/down. This is not nice code.
 *
 * Return:
 *	TRUE	if it works
 *	FALSE	on failure:
 *		    which could be out of memory - no message will be produced
 *	ABORT	C-g entered
 *	FAILED	indicates out of memory.
 */
static	int	replaces(int queryRep, int f, Cnt n)
{
	int	foundMatch;		/* success flag on pattern inputs */
	LineLn	rlength = 0;	/* length of replacement string - chars, not bytes */
	int	numsub;		/* number of substitutions - not inc notNeeded */
	int	nummatch;	/* number of found matches */
	int	nlflag;		/* last char of search string a <NL>? */
	int	nlrepl;		/* was a replace done on the last line? */
	KeyCode	c;		/* input char for query */
	LINE*	origline;	/* original "." position */
	LineLn	origoff;	/* and offset (for . query option) */
	LINE*	lastline = NULL;/* position of last replace and */
	LineLn	lastoff = 0;	/* offset (for 'u' query option) */
	int	onemore = FALSE;/* only do one more replace */
	TEXT*	PromptToUse;	/* The prompt to use */
	int	notNeeded = 0;	// Replace not needed, text already as wanted
	int	keepCFCPCN = 0;	// Preserve CFCPCN in lastflag to preserve w_leftoffset
	int	repNn = 0;	// notNeeded at time of last substitution - for 'u'
	int	dirn = FORWARD;	// Search direction
	int	i;

	int	lfpAbove = FALSE; // "." on top    window line or above
	int	lfpBelow = FALSE; // "." on bottom window line or below
	NMARK	foundPlace;

	/* Initialize starting conditions */
	canSetEq_i = TRUE;
	canSetSpW  = TRUE;
	tsetused(isearchKeys, 0);		// Clear the search keys buffer

	/* Ask the user for the text of a pattern.
	 */
	if((i = readpattern(
	    (queryRep == TRUE ? "Query replace: " : "Replace: "), TRUE)) != TRUE)
		return(i);

	/* Ask for the replacement string.
	 * Clear out first else we will not be able to delete (ie replace by empty string)
	 */
	tsetempty(searchRep);
	if((i = readpattern("with: ", FALSE)) == ABORT || i == FAILED)
		return(i);

	/* Set up flags so we can make sure not to do a recursive
	 * replace on the last line.
	 *
	 * Avoid problem if matchlen is zero. It is unsigned, so subtract 1
	 * and get a very large number.
	 */
	nlflag = (matchlen > 0) ? (tgetch(searchPat, tused(searchPat) - 1) == '\n') : 0;
	nlrepl = FALSE;

	/* Analyse the search string, replacement done later as it might change.
	 * Things like: embedded newlines & capitalisation
	 * Info specifically for search/replace. CheckCaselessSearch() is about search.
	 */
	wDesc	SrchDesc;
	wDesc	RepDesc;
	getRepDesc(&SrchDesc, searchPat_e);

	/* Save original . position, init the number of matches and
	 * substitutions, and scan through the file.
	 */
	origline = curwp->w_dotp;
	origoff = curwp->w_doto;
	numsub = 0;
	nummatch = 0;

	/* We are going to start moving.
	 * Place mark 1 where we started.
	 * This destroys mark 0 - good - see 's' command below
	 */
	pnmark1();

	while( (f == FALSE || n > nummatch) &&
		(nlflag == FALSE || nlrepl == FALSE) && onemore == FALSE ) {

		char*	promptComment = "";	// A comment for the start of the prompt

		/* Somewhere to keep the place that the scanner found.
		 * Needed as the user can use move-window-up/down/left/right
		 * which can change w_dotp/w_doto to other lines/columns.
		 * So restore to these before we do things.
		 * Eg: if not restored before replace, the wrong text is replaced.
		 */
		LINE*	foundLine = curwp->w_dotp;
		LineLn	foundOff = curwp->w_doto;

		/* If a reverse search fails "." will stay where it is now.
		 * We need to be able to restore matchlen of the search that
		 * brought us here
		 */
		LineLn	lastMatchlen = matchlen;

		/* Search for the pattern.
		 * If we search with a regular expression,
		 * matchlen is reset to the true length of
		 * the matched string.
		 *
		 * Don't fiddle with foundMatch except just here, we'll
		 * test it again after goto SetReplace.
		 */
#if	MAGIC
		if((magical && curwp->w_bufp->b_mode & MDMAGIC) != 0)
			foundMatch = mcscanner(&mcpat[0], dirn, PTBEG);
		else
#endif
			foundMatch = scanner(searchPat, dirn, PTBEG);

		foundPlace.n_pline = curwp->w_dotp;
		foundPlace.n_oline = curwp->w_doto;

		lastflag = 0;
		curgoalSet();		// Column that has "." - not same as byte or character offset
		lastflag = CFMARGIN0;	// Stop it being changed

		/* Scanned & found something new so OK to do next search if the
		 * replacement is the same.
		 * The point is that we might have moved through replacement history &
		 * by chance hit something that matched the search string, don't auto
		 * replace as the user might have wanted somewhere else in history.
		 *
		 * OK: also here if the scan failed at end of buffer. Minor point.
		 */
		int	newMatch = 1;

		/* Analyse the the replacement - upper/lower, white space.
		 * Only do this now as searchRep might be changed interactively below
		 * and we want to do the automatic mangling.
		 */
SetReplace:	getRepDesc(&RepDesc, searchRep);

		/* Copy what we replace with (if replacing) to playText2
		 * This is because we might munge it: capitalisation & newlines
		 */
		tcpyt(playText2, searchRep);
		TEXT*	replaceText = playText2;


		// Did the scan fail (ie not found) ?
		if(foundMatch == FALSE) {
			// Nothing found
			if(dirn == FORWARD)
				break;		// All done

			/* Backwards: "." is where it was, set match there.
			 * There must be a match there else the user would not have
			 * been prompted - which is the only way that he could have
			 * asked to go backwards.
			 */
			matchline = curwp->w_dotp;
			matchoff  = curwp->w_doto;
			matchlen  = lastMatchlen;

			savematch(search_m);
			curwp->w_dotp = matchline;
			curwp->w_doto = matchoff;

			// Let them know that there is nothing upwards
			bel();
			promptComment = "No match above. ";
		} else {
			// Got a match - or have a new replacement string

			/* Increment # matches
			 * But only if a new match, not because the user wants a new
			 * replacement string.
			 */
			if(newMatch)
				++nummatch;

			keepCFCPCN = 0;	// w_leftoffset may need to be recalculated
			foundLine = curwp->w_dotp;
			foundOff = curwp->w_doto;

			/* Check if we are on the last line.
			 */
			nlrepl = (lforw(curwp->w_dotp) == curwp->w_bufp->b_abotline);

			// Going back ('b' entered) - need to back up match vars which
			// otherwise point to the end of what we matched (which is
			// start going backwards)
			if(dirn == REVERSE) {
				int	j;
				for(j = 0; j < matchlen; j++)
					nextch(&matchline, &matchoff, REVERSE);
			}

			// Put what we found into search_m and analyse
			savematch(search_m);

			wDesc	FoundDesc;	// Info about search_m
			getRepDesc(&FoundDesc, search_m);

			/* Does the replacement need to be mangled so that its
			 * upper/lower case is as that of the string found.
			 * * not EXACT mode
			 * * search/replace strings do not contain upper case
			 * * the found string contains upper case
			 */
			if((curwp->w_bufp->b_mode & MDEXACT) == 0 &&
			   (SrchDesc.wd_noUpper & RepDesc.wd_noUpper) &&
			   (FoundDesc.wd_case[0] == Uc || FoundDesc.wd_case[0] == Tc || FoundDesc.wd_case[1] == Uc || FoundDesc.wd_case[1] == Tc)) {
				if((replaceText = CapsReplaceMangle(replaceText, FoundDesc)) == NULL)
					return(FALSE);
			}

			/* Similarly mangle the replacement string inserting newlines:
			 * * not EXACT mode
			 * * search/replace strings do not contain newlines
			 * * the found string contains one newline
			 */
			if((curwp->w_bufp->b_mode & MDEXACT) == 0 &&
			   SrchDesc.wd_newlines == 0 && RepDesc.wd_newlines == 0 &&
			   FoundDesc.wd_newlines == 1)
				if(NlReplaceMangle(replaceText, FoundDesc) == FALSE)
					return(FALSE);

			/* If the replacement is the same as what is there already
			 * go look for the next possible change.
			 * * Don't bug the user with such a 'change'
			 * * Don't modify the buffer if there is not a real change
			 * newMatch - see above for why.
			 */
			if( !textcmp(search_m, replaceText) && !AskReplaceSame && newMatch) {
				if(dirn == FORWARD) {
					++notNeeded;
					moveForw((curbp->b_mode & MDUTF8) ? (LineLn)utf8_chrnlen(ttext(replaceText), (size_t)tused(replaceText)) : tused(replaceText));
				}

				continue;
			}
		}

		// Going backwards only when asked to, a short term thing
		dirn = FORWARD;

		// Don't auto skip - see comments above
		newMatch = 0;

		/* Check for query.
		 */
		if(queryRep)
		{
			// Several help messages available via 'm' or '?'
			int	more = 0;	// More prompt index
			static	const	char*	Helps[] = {
				"(' ')yes (y)es (n)o (l)ast (!)do rest (u)ndo last (C-g)quit (q)uit (m)ore",
				"(.)abort back (b)ack search. Use window move keys to adjust view. (m)ore",
				"(E)dit the replacement. Modify words case: (U)pper (L)ower (T)itle. (m)ore",
				"next/previous line keys, (s)how last change, (o)ptions, (m)ore",
			};

			/* Get the query.
			 */
pprompt:		PromptToUse = searchPrompt;

			/* Repeated manipulation can result in trying to read & write
			 * from/to playText at the same time - resulting in corruption.
			 * Avoid this
			 */
			if(replaceText == playText) {
				tcpyt(playText2, playText);
				replaceText = playText2;
			}

			/* Build query replace question string showing what was matched
			 * and what we will replace with - important due to case (in)sensitivity
			 */
			if(tused(searchRep) != 0) {
				printToText(searchPrompt, "%s(?)Help. Replace '%t' with '%t' ?", promptComment, search_m, replaceText);
			} else {
				printToText(searchPrompt, "%sDelete '%t' ?", promptComment, search_m);
			}

qprompt:
			// Leave the cursor where the change would happen.
			CursorInMsg = 0;

			if(keepCFCPCN)
				lastflag |= CFMARGIN0;

			// Do we need to move "." to the top/bottom line ?
			if(lfpAbove) {
				// LFP on top line or above it. Nail it to the top line
				curwp->w_dotp = curwp->w_topLinep;
				curwp->w_doto = getgoal(curwp->w_topLinep);
			}
			if(lfpBelow) {
				// LFP on bottom line or below it. Nail it to the bottom line
				LINE*	bottomLine = WinBottomLine(curwp);
				if(bottomLine == NULL)
					bottomLine = curbp->b_linep;
				curwp->w_dotp = bottomLine;
				curwp->w_doto = getgoal(bottomLine);
			}

			c = mlGetKey(MGC_MAP | MGC_NO_CURSOR, NULL, "%t", PromptToUse);
			mlerase();			/* and clear it */

			// Look up some commands - bound sequences
			// If found it will return ISI_something
			c = ReverseMap(c);

			/* And respond appropriately.
			 */
			switch(c) {
			  case 'y':	/* yes, substitute */
			  case ' ':
				curwp->w_dotp = foundLine;
				curwp->w_doto = foundOff;

				keepCFCPCN = 0;	// w_leftoffset may need to be recalculated
				break;

			  case 'n':	/* no, onwards */
				curwp->w_dotp = foundLine;
				curwp->w_doto = foundOff;
				moveForw(1);
				keepCFCPCN = 0;	// w_leftoffset may need to be recalculated
				continue;

			  case 'b':	// Search backwards - just once
					// Back to something perhaps missed
			  case '^':	// GNU emacs recognises this for back to last match
				dirn = REVERSE;
				continue;

			  case 'l':	/* last one */
				curwp->w_dotp = foundLine;
				curwp->w_doto = foundOff;
				onemore = TRUE;
				break;

			  case '!':	/* yes/stop asking */
				curwp->w_dotp = foundLine;
				curwp->w_doto = foundOff;
				queryRep = FALSE;
				keepCFCPCN = 0;	// w_leftoffset may need to be recalculated
				break;

			  case 'E':	/* Edit the replacement
					 * Always interactive
					 */
				if((i = mlprompt("Replace by: ", MLSTRING | MLINTERACTIVE | MLDEFISLAST, bmlRepHistp)) != TRUE &&
				   i != FALSE)
					return(FALSE);
				replaceText = &ltxt(bmlRepHistp->b_dotp);
				// Doesn't do case mangling
				goto pprompt;

			  case 'u':	/* undo last and re-prompt */
				/* Restore old position.
				 */
				if(lastline == NULL) {
					/* There is nothing to undo.
					 */
					TTbeep();
					goto pprompt;
				}
				curwp->w_dotp = lastline;
				curwp->w_doto = lastoff;

				keepCFCPCN = 0;	// w_leftoffset may need to be recalculated

				// Delete the new string - ie inserted as replacement
				i = delins(rlength, patmatch, FALSE);
				if(i != TRUE)
					return(i);

				/* Record one less substitution, notNeeded at
				 * the time of last substitution, backup, and
				 * reprompt.
				 */
				--numsub;
				notNeeded = repNn;
				curwp->w_dotp = lastline;
				curwp->w_doto = lastoff;
				lastline = NULL;

				continue;

			  case 'o':	// Set search option
				lsetNotes(bmlSrchHistp->b_dotp, EditSearchOpts(0));
				goto qprompt;

			  case 's':	// Show what the last replacement did
					// Put "." at last change
				if( !retmark0(TRUE)) {
					// First replacement not yet done
					TTbeep();
					goto qprompt;
				}

				lchange(WFHARD);
				mlGetKey(MGC_NONE, NULL, "Press a key to continue ");

				// Go back to where we were
				retmark0(TRUE);
				lchange(WFHARD);
				goto qprompt;

			  case 'U':	// Convert replacement to UPPER case
				textToUpper(playText, replaceText);
				replaceText = playText;
				goto pprompt;

			  case 'L':	// Convert replacement to lower case
				textToLower(playText, replaceText);
				replaceText = playText;
				goto pprompt;

			  case 'T':	// Convert replacement to Title case
				textToTitle(playText, replaceText);
				replaceText = playText;
				goto pprompt;

			  // Allow the user to adjust the view of the window onto the buffer

			  case ISI_MOVE_DOWN:		// move-window-down - top line disappears, replaced by second
				lastflag = CFMARGIN0;	// Stop w_goalCol being messed with
				mvdnwind(TRUE, 1);

				if(foundPlace.n_pline == WinBottomLine(curwp)) {
					curwp->w_dotp = foundPlace.n_pline;
					curwp->w_doto = foundPlace.n_oline;	// OK does not need checking
					lfpBelow = 0;	// dont need work below
				}
				if(foundPlace.n_pline == lback(curwp->w_topLinep))
					lfpAbove = 1;
				goto qprompt;

			  case ISI_MOVE_UP:		// move-window-up - top line becomes second line
				lastflag = CFMARGIN0;	// Stop w_goalCol being messed with
				mvupwind(TRUE, 1);

				if(lback(foundPlace.n_pline) == WinBottomLine(curwp))
					lfpBelow = 1;

				if(foundPlace.n_pline == curwp->w_topLinep) {
					curwp->w_dotp = foundPlace.n_pline;
					curwp->w_doto = foundPlace.n_oline;	// OK does not need checking
					lfpAbove = 0;	// dont need work below
				}
				goto qprompt;

			  case ISI_MOVE_LEFT:	// move-window-left
				moveLeftWind(TRUE, 1);
				keepCFCPCN = 1;	// Preserve w_leftoffset that this does
				goto qprompt;

			  case ISI_MOVE_RIGHT:	// move-window-right
				moveRightWind(TRUE, 1);
				keepCFCPCN = 1;	// Preserve w_leftoffset that this does
				goto qprompt;

			  case ISI_UPLINE:	// Previous in replacement history buffer
				if(lback(bmlRepHistp->b_dotp) == bmlRepHistp->b_linep) {
					TTbeep();	// Hit end of history buffer
					goto qprompt;
				}
				// Set previous as current, copy as replacement
				bmlRepHistp->b_dotp = lback(bmlRepHistp->b_dotp);
				tcpyl(searchRep, bmlRepHistp->b_dotp);

				goto SetReplace;	// Does do replacement case mangling

			  case ISI_DOWNLINE:	// Next in replacement history buffer
				if(lforw(bmlRepHistp->b_dotp) == bmlRepHistp->b_linep) {
					TTbeep();	// Hit end of history buffer
					goto qprompt;
				}
				// Set previous as current, copy as replacement
				bmlRepHistp->b_dotp = lforw(bmlRepHistp->b_dotp);
				tcpyl(searchRep, bmlRepHistp->b_dotp);

				goto SetReplace;	// Does do replacement case mangling

			  case '.':	/* abort! and return */
				/* restore old position */
				curwp->w_dotp = origline;
				curwp->w_doto = origoff;
				curwp->w_flag |= WFMOVE;
				keepCFCPCN = 0;	// w_leftoffset may need to be recalculated

				// DROPTHROUGH
			  case ISI_ABORT:	// Abort the search
			  case 'q':		// Quit - same thing
				// abort and stay where we were
				mlwrite("%W""Quit. %d substitutions", numsub);
				if(notNeeded != 0)
					mlwrite("%A %d not needed", notNeeded);

				return(FALSE);

			  default:
				// Unknown: bitch and beep
				TTbeep();

				more = 0;	// Show first help prompt

				// DROPTHROUGH
			  case '?':	/* help me */
			  case 'm':	/* more opts - several helps */
				PromptToUse = playText;
				if(more >= (int)(sizeof(Helps) / sizeof(Helps[0])))
					more = 0;
				tcpys(playText, Helps[more]);
				more++;
				goto qprompt;

			}	/* end of switch */
		}	/* end of "if queryRep" */

		if(replaceText == NULL)	// Possible out of memory
			return(FALSE);

		/* Delete the sucker, and insert its replacement.
		 *
		 * First check if it is the same. Can't use previous compare
		 * as the user might have changed the replacement.
		 * Will only be here if $search-replace-same is TRUE
		 */
		if( !textcmp(search_m, replaceText)) {
			// Move to after the replacement before next search
			moveForw((curbp->b_mode & MDUTF8) ? (LineLn)utf8_chrnlen(ttext(replaceText), (size_t)tused(replaceText)) : tused(replaceText));
			++notNeeded;
		} else {
			// search_m is what was matched
			savematch(patmatch);	// Remember what was last matched and replaced

			// Save position of replacement & notNeeded - for 'u'
			repNn = notNeeded;
			lastline = curwp->w_dotp;
			lastoff = curwp->w_doto;
			pnmark0();

			// Get length of the replacement string
			rlength = (curbp->b_mode & MDUTF8) ? (LineLn)utf8_chrnlen(ttext(replaceText), (size_t)tused(replaceText)) : tused(replaceText);

			i = delins(matchlen, replaceText, TRUE);
			if(i != TRUE)
				return(i);

			numsub++;	/* increment # of substitutions */
		}
	}

	/* And report the results.
	 */
	mlwrite("%I""%d substitutions", numsub);
	if(notNeeded != 0)
		mlwrite("%A %d not needed", notNeeded);

	return(TRUE);
}

/* delins -- Delete a specified length from the current point
 *	then either insert the string directly, or make use of
 *	replacement meta-array.
 *
 * dlength	characters to delete (not bytes)
 * rep		replacement to insert
 * use_meta	consider rep to contain (maybe) '&' to insert matched string
 */
int	delins(LineLn dlength, TEXT* rep, int use_meta)
{
	int	status;
/*	int	replen = 1;		Length of replacement */
#if	MAGIC
	RMC	*rmcptr;
#endif

	/* Zap what we gotta,
	 * and insert its replacement.
	 */
	if((status = ldelete(dlength, KL_DISCARD)) != TRUE) {
		mlwrite("%EERROR while deleting");
		moveForw(1);
	} else {
#if	MAGIC
		if((rmagical && use_meta) &&
		     (curwp->w_bufp->b_mode & MDMAGIC) != 0) {
			rmcptr = &rmcpat[0];
			while(rmcptr->mcr_type != MCNIL && status == TRUE) {
				if(rmcptr->mcr_type == LITCHAR)
					status = linstr(rmcptr->rstr);
				else
					status = linTEXT(patmatch);
				rmcptr++;
			}
			/* replen = rmcptr - rmcpat; */
		} else
#endif
		{
			status = linTEXT(rep);

			/* replen = strlen(instr); */
		}
	}

#if 0
	/* This is not needed as the forward movement is done my linstr() */
	/* Skip over what we just inserted, this is so that we don't end
	 * up in an infinite replacement when we replace something like ':'
	 * with 'xx:'.
	 */
	if(replen)
		moveForw(replen);
#endif
	return(status);
}

/* boundry -- Return information depending on whether we may search no
 *	further.  Beginning of file and end of file are the obvious
 *	cases, but we may want to add further optional boundry restrictions
 *	in future, a' la VMS EDT.  At the moment, just return TRUE or
 *	FALSE depending on if a boundry is hit (ouch).
 */
int	boundry(LINE* curline, LineLn curoff, int dir)
{
	int	border;

	if(dir == FORWARD)
	{
		/* Note that it is now possible to be on the last line of the file
		 * but only if it is empty (so a test on linelength == 0 is
		 * redundent).
		 */
		border = (curline == curbp->b_abotline);
	}
	else
	{
		border = (curoff == 0) &&
			 (lback(curline) == curbp->b_btopline);
	}

	return(border);
}

/* nextch -- retrieve the next/previous byte in the buffer,
 *	and advance/retreat the point.
 *	The order in which this is done is significant, and depends
 *	upon the direction of the search.  Forward searches look at
 *	the current character and move, reverse searches move and
 *	look at the character.
 *
 * pcurline	the LINE that is being looked at. Will be changed
 *		when next/prev line
 * pcuroff	offset within the line
 * dir		direction of movement: FORWARD or REVERSE
 *
 * Return a pointer to the character. This is because there will be several of
 * them if it is a multi-byte character. It in not worth the expense of converting
 * to/from UniChar when searching.
 *
 * This takes note of the buffer's MDUTF8 flag.
 *
 * BYTE not character; when going REVERSE this will return the byte
 * at the end of a multi-byte UTF-8 character before the first one.
 */
static	BYTE*	nextch(LINE** pcurline, LineLn* pcuroff, int dir)
{
	LINE*		curline;
	LineLn		curoff;
	BYTE*		cp;

	curline = *pcurline;
	curoff = *pcuroff;

	if(dir == FORWARD)
	{
		if(curoff == lused(curline))		/* if at EOL */
		{
			curline = lforw(curline);	/* skip to next line */
			curoff = 0;
			cp = NewlineString;		/* and return a <NL> */
		}
		else {
			cp = lgetcaddr(curline, curoff);	/* get the char */
			curoff += (curbp->b_mode & MDUTF8) ? CHAR_LEN(*cp) : 1;
		}
	}
	else			/* Reverse.*/
	{
		if(curoff == 0)
		{
			curline = lback(curline);
			curoff = lused(curline);
			cp = NewlineString;
		}
		else {
			curoff--;

			// If UTF-8 ensure that we end up at the start of any
			// multi byte sequence:
			if((curbp->b_mode & MDUTF8))
				UTF8_BACK_FIND_START_NX(lgetc(curline, curoff), curoff);

			cp = lgetcaddr(curline, curoff);
		}

	}
	*pcurline = curline;
	*pcuroff = curoff;

	return(cp);
}

#if	MAGIC
/* mcstr -- Set up the 'magic' array.  The closure symbol is taken as
 *	a literal character when (1) it is the first character in the
 *	pattern, and (2) when preceded by a symbol that does not allow
 *	closure, such as a newline, beginning of line symbol, or another
 *	closure symbol.
 *
 *	Coding comment (jmg):  yes, i know i have gotos that are, strictly
 *	speaking, unnecessary.  But right now we are so cramped for
 *	code space that i will grab what i can in order to remain
 *	within the 64K limit.  C compilers actually do very little
 *	in the way of optimizing - they expect you to do that.
 *
 * TODO: does not work with UTF-8 - character classes
 */
static	int	mcstr(void)
{
	MC*	mcptr, *rtpcm;
	BYTE*	patptr;
 	int	mj;
 	BYTE	pchr;
 	int	status = TRUE;
 	int	does_closure = FALSE;

	/* If we had metacharacters in the MC array previously,
	 * free up any bitmaps that may have been allocated.
	 */
	if(magical)
		mcclear();

	magical = FALSE;
	mj = 0;
	mcptr = &mcpat[0];
	patptr = ttext(searchPat_e);
	patternHasUpper = FALSE;

	/* Scan the pattern looking for UPPER case
	 * Ignore the letter after a backslash
	 *
	 * Could make this horribly complicated: since [@-_] spans UPPERS
	 * we could say that that implies seeing upper case. But: no,
	 * only if an actual upper case char is seen do we note it.
	 * So: look inside character classes by not treating them specially.
	 */
	while((pchr = *patptr) != '\0') {
		if(Isupper(pchr)) {
			// Only for ASCII for now. Unicode TODO
			patternHasUpper = TRUE;
		} else if(pchr == MC_ESC) {
			/* Skip the char after backslash - unless NUL
			 * This will break, for now, on the likes of \S
			 * but in future that will mean something else.
			 */
			if(patptr[1])
				patptr++;
		}
		patptr++;
	}

	// So: do we do case insensitive comparison ?
	eqCompare_i = (curwp->w_bufp->b_mode & MDEXACT) == 0 && !patternHasUpper;

	patptr = ttext(searchPat_e);
	while((pchr = *patptr) != '\0' && status)
	{
		switch(pchr)
		{
			case MC_CCL:	// '['
				status = cclmake(&patptr, mcptr);
				magical = TRUE;
				does_closure = TRUE;
				break;
			case MC_BOL:	// '^'
				if(mj != 0)
					goto litcase;

				mcptr->mc_type = BOL;
				magical = TRUE;
				does_closure = FALSE;
				break;
			case MC_EOL:	// '$'
				if(*(patptr + 1) != '\0')
					goto litcase;

				mcptr->mc_type = EOL;
				magical = TRUE;
				does_closure = FALSE;
				break;
			case MC_ANY:	// '.'
				mcptr->mc_type = ANY;
				magical = TRUE;
				does_closure = TRUE;
				break;
			case MC_CLOSURE: // '*'
				/* Does the closure symbol mean closure here?
				 * If so, back up to the previous element
				 * and indicate it is enclosed.
				 */
				if(!does_closure)
					goto litcase;
				mj--;
				mcptr--;
				mcptr->mc_type |= CLOSURE;
				magical = TRUE;
				does_closure = FALSE;
				break;

			/* Note: no break between MC_ESC case and the default.
			 */
			case MC_ESC:	// '\\'
				if(*(patptr + 1) != '\0')
				{
					pchr = *++patptr;
					magical = TRUE;
				}
				/* FALLTHROUGH */
			default:
litcase:			mcptr->mc_type = LITCHAR;
				mcptr->u.lchar[0] = pchr;
				if(curbp->b_mode & MDUTF8) {
					UTF8_MOVE(mcptr->u.lchar, patptr);
					// -1 as increment below
					patptr += CHAR_LEN(*patptr) - 1;
				}
				does_closure = (pchr != '\n');
				break;
		}		/* End of switch.*/
		mcptr++;
		patptr++;
		mj++;
	}		/* End of while.*/

	/* Close off the meta-string.
	 */
	mcptr->mc_type = MCNIL;

	/* Set up the reverse array, if the status is good.  Please note the
	 * structure assignment - your compiler may not like that.
	 * If the status is not good, nil out the meta-pattern.
	 * The only way the status would be bad is from the cclmake()
	 * routine, and the bitmap for that member is guarenteed to be
	 * freed.  So we stomp a MCNIL value there, and call mcclear()
	 * to free any other bitmaps.
	 */
	if(status)
	{
		rtpcm = &tapcm[0];
		while(--mj >= 0)
		{
#if	LATTICE
			movmem(--mcptr, rtpcm++, sizeof(MC));
#endif

#if	MWC86 || AZTEC || MSC || TURBO || VMS || USG || BSD_UNIX || V7
			*rtpcm++ = *--mcptr;
#endif
		}
		rtpcm->mc_type = MCNIL;
	}
	else
	{
		(--mcptr)->mc_type = MCNIL;
		mcclear();
	}

	return(status);
}

/* rmcstr -- Set up the replacement 'magic' array.  Note that if there
 *	are no meta-characters encountered in the replacement string,
 *	the array is never actually created - we will just use the
 *	character array rpat[] as the replacement string.
 */
static	int	rmcstr(void)
{
	RMC*	rmcptr;
	BYTE*	patptr;
	int	status = TRUE;
	int	mj;		// Byte count since start of searchRep or last: '&' or '\'
	int	utf8 = curbp->b_mode & MDUTF8;	// Current buffer UTF-8 mode
	int	len;

	patptr = ttext(searchRep);
	rmcptr = &rmcpat[0];
	mj = 0;
	rmagical = FALSE;

	while(*patptr && status == TRUE)
	{
		switch(*patptr)
		{
			case MC_DITTO:	// '&'

				/* If there were non-magical characters
				 * in the string before reaching this
				 * character, plunk it in the replacement
				 * array before processing the current
				 * meta-character.
				 */
				if(mj != 0) {
					rmcptr->mcr_type = LITCHAR;
					if((rmcptr->rstr = (BYTE*)malloc((size_t)mj + 1)) == NULL)
					{
						mlwrite(MsgOutOfMemory);
						status = FALSE;
						break;
					}
					strncpy(rmcptr->rstr, patptr - mj, (size_t)mj);
					rmcptr->rstr[mj] = '\0';
					rmcptr++;
					mj = 0;
				}

				// Make a DITTO replacement, it contains no data
				rmcptr->mcr_type = DITTO;
				rmcptr++;
				rmagical = TRUE;
				patptr++;
				break;

			case MC_ESC:	// '\'
				rmcptr->mcr_type = LITCHAR;

				/* We malloc mj plus len + one here, instead
				 * of one, because we have to count the
				 * current character.
				 */
				len = utf8 ? CHAR_LEN(patptr[1]) : 1;
				if((rmcptr->rstr = (BYTE*)malloc((size_t)(mj + len + 1))) == NULL)
				{
					mlwrite(MsgOutOfMemory);
					status = FALSE;
					break;
				}

				// Copy everything up to the MC_ESC
				strncpy(rmcptr->rstr, patptr - mj, (size_t)mj);

				/* If MC_ESC is not the last character
				 * in the string, find out what it is
				 * escaping, and overwrite the last
				 * character with it.
				 */
				if(patptr[1] != '\0') {
					patptr++;
					if(utf8) {
						UTF8_MOVE(rmcptr->rstr + mj, patptr);
					} else {
						*((rmcptr->rstr) + mj) = *patptr;
					}
					patptr += len;
				}
				*((rmcptr->rstr) + mj + len) = '\0';

				mj = 0;
				rmcptr++;
				rmagical = TRUE;
				break;

			default:
				// Don't care if we copy part of a UTF-8 character,
				// all of it will be copied shortly.
				mj++;
				patptr++;
		}
	}

	// Anything left over ?
	if(rmagical && mj > 0)
	{
		rmcptr->mcr_type = LITCHAR;
		if((rmcptr->rstr = (BYTE*)malloc((size_t)mj + 1)) == NULL)
		{
			mlwrite(MsgOutOfMemory);
			status = FALSE;
		}
		strncpy(rmcptr->rstr, patptr - mj, (size_t)mj);
		rmcptr->rstr[mj] = '\0';
		rmcptr++;
	}

	rmcptr->mcr_type = MCNIL;

	return(TRUE);
}

/* mcclear -- Free up any CCL bitmaps, and MCNIL the MC search arrays.
 */
void	mcclear(void)
{
	MC*	mcptr;

	mcptr = &mcpat[0];

	while(mcptr->mc_type != MCNIL)
	{
		if((mcptr->mc_type & MASKCL) == CCL ||
		    (mcptr->mc_type & MASKCL) == NCCL)
			free(mcptr->u.cclmap);
		mcptr++;
	}
	mcpat[0].mc_type = tapcm[0].mc_type = MCNIL;
}

/* rmcclear -- Free up any strings, and MCNIL the RMC array.
 */
void	rmcclear(void)
{
	RMC*	rmcptr;

	rmcptr = &rmcpat[0];

	while(rmcptr->mcr_type != MCNIL)
	{
		if(rmcptr->mcr_type == LITCHAR)
			free(rmcptr->rstr);
		rmcptr++;
	}

	rmcpat[0].mcr_type = MCNIL;
}

/* mceq -- meta-character equality with a character.  In Kernighan & Plauger's
 *	Software Tools, this is the function omatch(), but i felt there
 *	were too many functions with the 'match' name already.
 */
static int	mceq(BYTE* bc, const MC* mt)
{
	int	result;

	switch(mt->mc_type & MASKCL)
	{
		case LITCHAR:
			result = eq(bc, mt->u.lchar);
			break;

		case ANY:
			result = (*bc != '\n');
			break;

		case CCL:
			if((result = biteq(*bc, mt->u.cclmap)) == 0)
			{
				if(eqCompare_i && Isletter(*bc))
				{
					result = biteq(CHCASE(*bc), mt->u.cclmap);
				}
			}
			break;

		case NCCL:
			result = !biteq(*bc, mt->u.cclmap);

			if(eqCompare_i && Isletter(*bc))
			{
				result &= !biteq(CHCASE(*bc), mt->u.cclmap);
			}
			break;

		default:
			mlwrite("%Emceq: what is %d?", mt->mc_type);
			result = FALSE;
			break;

	}	/* End of switch.*/

	return(result);
}

/* cclmake -- create the bitmap for the character class.
 *	ppatptr is left pointing to the end-of-character-class character,
 *	so that a loop may automatically increment with safety.
 */
static	int	cclmake(BYTE** ppatptr, MC* mcptr)
{
	BITMAP	bmap;
	BYTE*	patptr;		// Important: unsigned - when passed to setbit()
	BYTE	pchr, ochr;

	if((bmap = clearbits()) == NULL)
	{
		mlwrite(MsgOutOfMemory);
		return FALSE;
	}

	mcptr->u.cclmap = bmap;
	patptr = *ppatptr;

	/* Test the initial character(s) in ccl for
	 * special cases - negate ccl, or an end ccl
	 * character as a first character.  Anything
	 * else gets set in the bitmap.
	 */
	if(*++patptr == MC_NCCL)	// '^'
	{
		patptr++;
		mcptr->mc_type = NCCL;
	}
	else
		mcptr->mc_type = CCL;

	if((ochr = *patptr) == MC_ECCL)	// ']'
	{
		mlwrite("%ENo characters in character class");
		return(FALSE);
	}
	else
	{
		if(ochr == MC_ESC)
			ochr = *++patptr;

		setbit(ochr, bmap);
		patptr++;
	}

	while(ochr != '\0' &&(pchr = *patptr) != MC_ECCL)
	{
		switch(pchr)
		{
			/* Range character loses its meaning
			 * if it is the last character in
			 * the class.
			 */
			case MC_RCCL:	// '-'
				if(*(patptr + 1) == MC_ECCL)
					setbit(pchr, bmap);
				else
				{
					pchr = *++patptr;
					while(++ochr <= pchr)
						setbit(ochr, bmap);
				}
				break;

			/* Note: no break between case MC_ESC and the default.
			 */
			case MC_ESC:	// '\\'
				pchr = *++patptr;
				/*FALL THROUGH */
			default:
				setbit(pchr, bmap);
				break;
		}
		patptr++;
		ochr = pchr;
	}

	*ppatptr = patptr;

	if(ochr == '\0')
	{
		mlwrite("%ECharacter class not ended");
		free(bmap);
		return FALSE;
	}
	return TRUE;
}

/* biteq -- is the character in the bitmap?
 */
static	int	biteq(int bc, BITMAP cclmap)
{
	if(bc >= HICHAR)
		return FALSE;

	return((*(cclmap + (bc >> 3)) & BIT(bc & 7))? TRUE: FALSE );
}

/* clearbits -- Allocate and zero out a CCL bitmap.
 */
static	BITMAP	clearbits(void)
{
	BITMAP	cclstart, cclmap;
	int	j;

	if((cclmap = cclstart = (BITMAP) malloc(HIBYTE)) != NULL)
		for(j = 0; j < HIBYTE; j++)
			*cclmap++ = 0;

	return(cclstart);
}

/* setbit -- Set a bit (ON only) in the bitmap.
 */
static	void	setbit(int bc, BITMAP cclmap)
{
	if(bc < HICHAR)
		*(cclmap + (bc >> 3)) |= (char)BIT(bc & 7);
}
#endif

/* end */
