/*
 * ADDW SCCS: @(#)tag.c	1.52 10/02/20 21:27:23
 */

/*
 * Find tags from a tag file. Only defined if running some sort of UNIX.
 * Conditionally compiled on CTAGS.
 *
 * This code written by Mike Rendell formerly of ROOT Computers Ltd.
 * Modified to run on System V and use TAGS environment variable by
 * Tony Bamford.
 *
 * Look here for tags file format:
 * https://github.com/universal-ctags/ctags/blob/master/old-docs/website/FORMAT
 */
#include "estruct.h"
#include "utf8.h"
#include "edef.h"

#if	CTAGS		// Compiles to nothing unless this is set

#if	SHOWMODREL
char	tagvers[] = "@(#)tag.c	1.52 10/02/20 21:27:23";
#endif

/* Something to store alternatives for when we find more than one match.
 */
typedef	struct	_TagPos	{
	struct	_TagPos*ts_next;	/* Next in list */
	char*		ts_file;	/* File that it is in */
	char*		ts_tag;		/* What to search for */
} TagPos;

/* The current tag, etc, needs to be visible so that NextTag() can update
 * them. Beware - the pointers are only valid when a selection is in
 * progress, they point to garbage at other times.
 */
static	TagPos*	TagRoot;		/* Start of chain */
static	TagPos*	NewTag;			/* Current one */
static	int	NumTags;		/* How many in the list */

/* Find a C tag that is prompted for.
 *
 * If the count is set split the window to show the tag, else use this window.
 * The default count is not set.
 *
 * Widens any buffer - in case the tag in not in a narrow.
 *
 * If the count is set split the window to show the tag, else use this window.
 * The default count is not set.
 *
 * Command:  find-tag
 * Bound to: M-t
 *
 * Return:
 *	TRUE  found the tag
 *	FALSE tag not found in index files
 *	      window cannot be split
 *	      cannot open the file that contains the tag
 *	ABORT C-g
 */
int	findtag(int f, Cnt n)
{
	int	s;

	if((s = mlreply("Tag: ", &playValStr, MLTAG | MLNOSPACE, bmlTagHistp)) != TRUE)
		return(s);

	return(do_tag(f, playValStr));
}


/* Take the current word as the tag to search for.
 *
 * This assumes that the cursor is at the start of the word, it probably
 * ought to move back to the start of the word.
 * I think that the definition of word is a bit wide as it also includes
 * things like '.'.
 *
 * Widens any buffer - in case the tag in not in a narrow.
 *
 * If the count is set split the window to show the tag, else use this window.
 * The default count is not set.
 *
 * Command:  find-current-tag
 * Bound to: M-T
 *
 * Return:
 *	TRUE  found the tag
 *	FALSE tag not found in index files
 *	      window cannot be split
 *	      cannot open the file that contains the tag
 *	FAILED - out of memory
 *	ABORT C-g
 */
int	findcurtag(int f, Cnt n)
{
//	VAL*	tag = playValStr
	LineLn	cnt;
	BYTE*	p = lgetcaddr(curwp->w_dotp, curwp->w_doto);

	// Back off to the start of the identifier:
	while(p > ltext(curwp->w_dotp) && (Isalnum(p[-1]) || p[-1] == '_'))
		p--;

	// Count to end of the identifier:
	cnt = 0;
	while(p + cnt < ltext(curwp->w_dotp) + lused(curwp->w_dotp) &&
			(Isalnum(p[cnt]) || p[cnt] == '_'))
		cnt++;

	if(cnt == 0) {
		TTbeep();
		return(FALSE);
	}

	if(tcpym(vTEXT(playValStr), p, cnt) == FALSE)
		return(FAILED);

	return(do_tag(f, playValStr));
}

/* Clean up/convert the pattern from vi format.
 */
void	cleantag(TEXT* tag)
{
	BYTE*	p = ttext(tag);

	for( ; *p != '\0' ; p++)
		switch(*p) {
		  case '?':
		  	*p = '\0';
		  	break;
		  case '^':
			*p = '\n';
			break;
		  case '$':
			*p = '\n';
			break;
		  case '\\':
			*p = '\0';
			break;
		  case '/':
			*p = '\0';
			break;
		}

	tsetused(tag, (LineLn)strlen(ttext(tag)));
}

/* Just to be consistant between main() and here, we provide the following
 * routine.
 */
void	tagerr(const BYTE* tagname)
{
	mlwrite("%WI found the tags file, but the tag \"%s\" is missing", tagname);
}

/* Find a tag
 *
 * If SplitWindow the current window is split to display the found function/...
 *
 * Uses LINE searchPat.
 * Use LINE playText2 to store the name of the file that we will open
 *
 * The tag used is added to tag history
 *
 * Any narrowing is removed - the tag could be anywhere in the buffer.
 *
 * Return:
 *	TRUE  if found
 *	FALSE tag not found in index files
 *	      window cannot be split
 *	      cannot open the file that contains the tag
 *	ABORT C-g
 */
int	do_tag(int SplitWindow, VAL* tag)
{
	char*	ofiln = curbp->b_fname;		// old file
	BYTE	obufn[BUFFER_NAME_SIZE];	// Current buffer name
	int	status;

	strcpy(obufn, curbp->b_bname);

	addHistoryText(bmlTagHistp, vTEXT(tag), AHL_NO_DUP);

	// Find entry in tags file
	// playText2 will be set to the file containing the function
	// searchPat a search string
	if((status = find_fun(vTEXT(tag), playText2, searchPat, playDir)) != TRUE)
		return(status);

	if(SplitWindow == TRUE)
		if(splitwind(FALSE, 1) == FALSE)
			return(FALSE);

	/* If the file doesn't exist, try relative to the directory that
	 * contains the tags file.
	 */
	if(CacheStat(ttext(playText2)) == -1 && errno == ENOENT && tgetch(playText2, 0) != '/') {
		/* Try relative to tag name */
		tcatc(playDir, '/');
		tcatt(playDir, playText2);
		tcpyt(playText2, playDir);
	}

	if(getfile(ttext(playText2), TRUE) == FALSE) {
		mlwrite("%WCan't find %t", playText2);
		return(FALSE);
	}

	widen(FALSE, 0);		// Lose any narrowing

	/* If going to the tag will put us into another buffer, remember 
	 * the name of the current buffer in lbufn, else if we stay in the 
	 * same buffer, set the mark at the current position.
	 */
	if(strcmp(ofiln, curbp->b_fname) == 0) {
		// gotobob() sets numeric mark 1
		setmark(FALSE, 0);
	} else {
		strcpy(lbufn, obufn);
	}

	gotobob(TRUE, 0);	/* can't return error */

	/* Reframe to get line number recalculate - if they are being shown.
	 * There is a problem in scanner() - this fixes it for here.
	 */
	curwp->w_flag |= WFFORCE;

	/* searchPat contains an EX command. This is either:
	 * /regular-expression/		(could be ?regular-expression?)
	 * line-number
	 */
	if(Isdigit(tgetch(searchPat, 0))) {
		int	n = atoi(ttext(searchPat));

		if(n > curbp->b_numLines) {
			mlwrite("%Wctag file sayd %d, buffer only %D lines long", n, (long)curbp->b_numLines);
			return(FALSE);
		}

		return(moveDown(n - 1));

	}

	// It is a search string surrounded by '/' or '?' - the first one is removed
	cleantag(searchPat);
	eqCompare_i = FALSE;		// Want case sensitive comparison of function names
	matchSpaceWhite = FALSE;	// A space only matches a single space, not TAB or NL

	if(scanner(searchPat, FORWARD, PTBEG) == FALSE) {
		tagerr(vtext(tag));
		return(FALSE);
	}

	/* Since we started "patt" off with a newline character, we are
	 * now sitting one character off the "target", and so need to
	 * advance our postion.
 	 */
	moveForw(1);

	// Request that the found tag is in window line 1/3 from the top
	curwp->w_force = (char)(curwp->w_ntrows / 3);

	return(TRUE);
}

/* Read a word from file descriptor "fp" into "buf", NUL terminating.
 * A word is defined to be non space characters bounded by spaces tabs or
 * newlines.
 */
void	readword(FILE* fp, TEXT* buf)
{
	int	c;
	int	inword = 0;

	tsetused(buf, 0);

	while((c = getc(fp)) != EOF && c != '\n') {
		if(c == '\t' || c == ' ') {
			if(inword)
				break;
			else
				continue;
		}
		inword = 1;
		tcatc(buf, c);
	}
}

/* Parse a line from the tags file. Split it into the function name, the
 * filename and the search string.
 *
 * Put the search pattern into patt
 *
 * This is here because there seem to be problems using the (much more
 * elegant) line of code on System V.
 *
 *	(void) fscanf(fp, "%[^\t]\t%[^\t]\t/^%[^\n]$/\n", function,
 *							file, patt);
 *
 * Return false if EOF is reached, else return the # characters eaten.
 * If we hit EOF close the file.
 */
int	parse_tag(FILE* fp, TEXT* func, TEXT* file, TEXT* patt)
{
	int	c;
	int	eaten = 0;

	readword(fp, func);
	readword(fp, file);

	/* Throw away the "?" or "/" character.
	 * But if we find that it is a digit - put it back,
	 * we will want the whole line number.
	 */
	c = getc(fp);
	if(Isdigit(c))
		ungetc(c, fp);

	/* Read the remainder of the line into patt.
	 */
	tsetused(patt, 0);
	for(;;) {
		if((c = getc(fp)) == EOF)
			return(0);
		eaten++;
		if(c == '\n')
			break;
		tcatc(patt, c);
	}

	return(eaten);
}

/* Find the function in the tag file.  From Mike Rendell.
 *
 * Use a TAGS environment variable if there is one, and use the tags files
 * in there if the tag cannot be found in the tags file in the current
 * directory.
 *
 * Return the line of the tags file, and let do_tag handle it from here.
 * 
 * dirn is set to contain the name of the directory of the tags
 * file that contained the name.
 * If dirn is NULL don't touch it.
 *
 * Return TRUE  if we find it.
 *	  FALSE if we do not
 *	  ABORT - C-g
 */
int	find_fun(TEXT* key, TEXT* file, TEXT* patt, TEXT* dirn)
{
	int	status;
	int	count = 0;		// Count of files looked in and it is not there
	FILE*	fd;
	char*	envtag;
	
	/* Step 0. See if we have TAGS in the environment.
	 * Avoid crunch if it is too long.
	 */
	if((envtag = getenv("TAGS")) != (char*)NULL &&
	    strlen(envtag) >= TAGSIZE - 2) {
		mlwrite("%ETAGS envrionment variable is too long. Max %d", TAGSIZE - 2);
		envtag = (char*)NULL;
	}

	/* Step 1. Go for the tags file in the current directory.
	 */
	if((fd = fopen("tags", "r")) == (FILE *) NULL
					&& envtag == (char*) NULL) {
		mlwrite("%WNo tags file, no TAGS environment variable");
		return(FALSE);
	}
	if(fd != (FILE*) NULL) {
		status = ffind_fun(key, file, patt, fd);
		count++;

		/* If we found the tag, or something went drastically wrong,
		 *  return.
		 */
		if(status == TRUE || status == ABORT) {
			if(dirn != (TEXT*)NULL)
				tcpys(dirn, ".");
			return(status);
		}
	}

	/* Step 2. Open all the files pointed to by the TAGS variable.
	 * This is a list of tags file names that are separated by
	 * spaces or colons.
	 */
	for(; envtag != (char*) NULL && *envtag != '\0'; /*void*/) {
		size_t	len = strcspn(envtag, " :");
		tcpym(playText, envtag, (LineLn)len);
		envtag += len;

		while(*envtag == ' ' || *envtag == ':')
			envtag++;

		mlwrite("%IUsing %t", playText);
		
		if((fd = fopen(ttext(playText), "r")) != (FILE*) NULL) {
			status = ffind_fun(key, file, patt, fd);
			if(status == ABORT)
				return(ABORT);

			if(status == FALSE) {
				count++;
				continue;
			}

			/* It got something - save the directory name */
			if((envtag = strrchr(ttext(playText), '/')) != (char*)NULL)
				tsetused(playText, (LineLn)((BYTE*)envtag - ttext(playText)));

			if(dirn != (TEXT*)NULL)
				tcpyt(dirn, playText);

			return(TRUE);
		}
	}

	if(count == 1)
		mlwrite("%WCan't find %t in tags file", key);
	else
		mlwrite("%WCouldn't find %t in %d tag files", key, count);

	return(FALSE);
}

/* Find the function in the tag file. Rewritten - ADDW.
 * Return the line of the tags file, and let do_tag handle it from here.
 * Close file before we leave.
 * This works by binary chopping the file, maintain a start & end of the region
 * that we search, bring these closer together by 1/2 the range at every comparison.
 * When these are less than 1K appart or 1K from an end of file use a linear search.
 *
 * Return TRUE  if we find it.
 *	  FALSE if we do not
 *	  ABORT - C-g
 */
int	ffind_fun(const TEXT* key, TEXT* file, TEXT* patt, FILE* fp)
{
	long	start;			/* Start of area we look within */
	long	end;			/* End of file part to look within */
	long	FileEnd;		/* Size of the file */
	long	pos;			/* Current search position */
	int	linear;			/* True if in linear search mode */
	int	ch;
	int	AtLineStart;		/* Where we are now */
	int	Succ;			/* What this function returns */
	TEXT*	FunBuf = playText3;

	TagRoot = (TagPos*)NULL;

	start = 0;
	fseek(fp, 0L, SEEK_END);
	end = FileEnd = ftell(fp);
	AtLineStart = 0;
	linear = 0;
	pos = (start + end) / 2;	/* Where we should start */

	for(;;) {
		if(! linear) {
			AtLineStart = 0;	/* About to seek */

			/* Should we be going linear ? */
			if(end - start <= 1024 || pos <= 1024 ||
						FileEnd - pos <= 1024) {
				linear = 1;

				if(start <= 1024) {
					start = 0;
					AtLineStart = 1;
				}
				pos = start;
			}
			else	/* Binary chop */
				pos = (start + end) / 2;

			fseek(fp, pos, SEEK_SET);

			/* Find the start of the next line - if not already there.
			 */
			while(!AtLineStart &&
					(ch = fgetc(fp)) != '\n' && ch != EOF)
				pos++;
		}

		/* Break out the tags line */
		if((ch = parse_tag(fp, FunBuf, file, patt)) == 0) {
			fclose(fp);
			return(FALSE);		/* EOF */
		}

		/* Note where we got to in the file */
		pos += ch;

		/* Got a match ? */
		if((ch = strcmp(ttext(key), ttext(FunBuf))) == 0)
			break;

		/* Didn't match, do we look next in the 1st or 2nd half ? */
		if(ch < 0) {	/* Look towards start of file */
			if(linear) {
				fclose(fp);
				return(FALSE);
			}
			end = pos;
		}
		else {
			/* Prevent a seek to start of line else we could
			 * miss the start point that we are already at.
			 */
			start = pos;
			AtLineStart = 1;
		}
	}

	/* We have a match.
	 * We need to work out if it is unique, if it isn't we ought to
	 * present a list of alternatives so that the user can choose.
	 *
	 * Work back 1K at a time 'till we find the first line which contains
	 * the match.
	 */
	for(;;) {
		if((pos -= 1024) < 0)
			pos = 0;
		fseek(fp, pos, SEEK_SET);

		while(pos != 0 && (ch = fgetc(fp)) != '\n' && ch != EOF)
			;

		/* Break out the tags line */
		if(parse_tag(fp, FunBuf, file, patt) == 0) {
			fclose(fp);
			return(FALSE);		/* EOF - shouldn't happen */
		}

		/* If we went further back, start looking forwards */
		if(strcmp(ttext(key), ttext(FunBuf)) > 0 || pos == 0)
			break;
	}

	/* Move forwards 'till we find the first one to match */
	while(strcmp(ttext(key), ttext(FunBuf)) != 0)
		if(parse_tag(fp, FunBuf, file, patt) == 0) {
			fclose(fp);
			return(FALSE);		/* EOF - shouldn't happen */
		}

	NumTags = 0;
	/* Put them into the tag select buffer 'till we don't have a match */
	while(strcmp(ttext(key), ttext(FunBuf)) == 0) {
		if((NewTag = (TagPos*)calloc(1, sizeof(TagPos))) ==
								(TagPos*)NULL ||
		   (NewTag->ts_file = strmake(ttext(file))) == (char*)NULL ||
		   (NewTag->ts_tag = strmake(ttext(patt))) == (char*)NULL)
			break;		/* Ought really to winge I suppose */

		NewTag->ts_next = TagRoot;
		TagRoot = NewTag;
		NumTags++;

/*		fprintf(stderr, ">%s<>%s<\n", ttext(file), ttext(patt));	*/

		if(parse_tag(fp, FunBuf, file, patt) == 0)
			break;
	}

	fclose(fp);

	if(TagRoot == (TagPos*)NULL)
		return(FALSE);		/* Should not happen */

	/* If there is more than one of them, display a buffer with them
	 * so that the user can choose. If there is just one -- go for it.
	 */
	if(NumTags == 1) {
		NewTag = TagRoot;
		Succ = 1;
	}
	else
		Succ = SelectTag();

	/* Copy the choice back into the arrays passed to us */
	if( !tcpys(file, NewTag->ts_file) ||
	    !tcpys(patt, NewTag->ts_tag))
		Succ = FALSE;

	/* Zap the linked list */
	while(TagRoot != (TagPos*)NULL) {
		NewTag = TagRoot->ts_next;
		free((void*)TagRoot->ts_file);
		free((void*)TagRoot->ts_tag);
		free((void*)TagRoot);
		TagRoot = NewTag;
	}

	return(Succ);
}

/* Prompt the user to select a tag from the list in TagRoot.
 * Set NewTag to the member. See static variables at the top of this module.
 *
 * Return TRUE if got something
 *	  FALSE - nothing available, something went wrong - *ChosenTag will still be valid.
 *	  ABORT - C-g
 */
int	SelectTag(void)
{
	char	prompt[MAXCOL];

	sprintf(prompt, "%d tags: ", NumTags);

	return(mlreply(prompt, &playValStr, MLTAG | MLSELECTONLY, NULL));
}

/* Display the next tag in the list of possibilities.
 * Update NewTag to point to this.
 *
 * prompt	current prompt
 * dirn		direction to go, + forward, - backward
 *
 * Called, indirectly, from getstring() where wmessp->w_dotp will
 * have had the prompt put in it - so lsetused() is safe.
 */
void	NextTag(const char* prompt, int dirn)
{
	TagPos*	probe;
	LINE*	dotp = wmessp->w_dotp;

	if(dirn < 0)	/* Forwards */
		NewTag = NewTag->ts_next == (TagPos*)NULL ? TagRoot :
							    NewTag->ts_next;
	else {	/* Backwards */
		/* If we are the first, we want the last, ie where the next
		 * is NULL.
		 */
		if(NewTag == TagRoot)
			NewTag = (TagPos*)NULL;

		for(probe = TagRoot; probe->ts_next != NewTag;
							probe = probe->ts_next)
			;

		NewTag = probe;
	}

	/* Build up 'filename tag_search_string' in buf.
	 */
	lsetused(dotp, (LineLn)strlen(prompt));
	lcats(dotp, NewTag->ts_file);
	lcats(dotp, " ");
	lcats(dotp, NewTag->ts_tag);
}

/* Show first in the list on initial promt.
 *
 * If a list has not been made - do nothing.
 */
void	TagDispInit(const char* prompt)
{
	if(TagRoot == (TagPos*)NULL)
		return;

	NewTag = TagRoot;

	NextTag(prompt, 1);
}

#endif	/* CTAGS */

/* end */
