/*
 * ADDW SCCS: @(#)tcap.c	1.112 08/06/20 23:02:21
 */

/*	tcap:	Unix V5, V7 and BS4.2 Termcap video driver
 *		for MicroEMACS
 *
 * For documentation on this see:
 * http://www.gnu.org/software/termutils/manual/termcap-1.3/html_mono/termcap.html
 * http://www.delorie.com/gnu/docs/termcap/termcap_toc.html
 *
 * https://lists.gnu.org/archive/html/bug-ncurses/2015-12/msg00009.html
 *
 * A lot of modern terminal emulations are based on this:
 * https://en.wikipedia.org/wiki/ANSI_escape_code
 *
 * It was really hard to get full intensity display.
 * Because when we used colours they call came out half bright, so a white
 * background looks grey.
 * Making bold fixed this, but, on xterm at least, it also makes the characters
 * heavier - which is harder to read.
 * The final solution was to use OP (original pair) to reset to standard and then
 * set the colour.
 * Care was needed to not end up changing/resetting the colour too many times and not doing
 * anything between colour changes.
 *
 * Also problems with Gnome's terminal emulation which is why TERM=gnome is
 * recommended. Quick summary: F13 is got by S-F1, but Gnome terminal generates
 * a different sequence than documented in terminfo:xterm.
 * Same for F14 ...
 * http://invisible-island.net/ncurses/ncurses.faq.html#xterm_generic
 *
 * Insert/delete characters:
 * https://www.gnu.org/software/termutils/manual/termcap-1.3/html_chapter/termcap_4.html#SEC32
 */

#define	termdef	1			/* don't define "term" external */

#include	"estruct.h"
#include        "utf8.h"
#include        "edef.h"

#if	SHOWMODREL
char	tcapvers[] = "@(#)tcap.c	1.8 4/5/97 14:27:41 MicroEMACS";	/* SCCS */
#endif

// Trap config error. Must have TERMCAP configured if using CURSES
#if	CURSES
#if	TERMCAP == 0
#error	Terminal configuration option CURSES is set but TERMCAP is false.
#endif
#endif

#if	TERMCAP

/* Some <term.h> declare tparm() as taking 10 arguments.
 * The define below fixes it to use tiparm().
 * This seems NetBSD, also used on Minix.
 */
#define	TPARM_VARARGS	1

#if	CURSES
#include <curses.h>
#include <term.h>
#else
#if	TERMCAP_USE_TERM_H
#include <term.h>
extern char*	tgetstr(char *id, char **area);
#endif
#endif	// CURSES

/* Used in term definition.
 */
char*	tkeyname(KeyCode c);
int	tDoesWhat(int what);
KeyCode	tspecial(const BYTE* seq);
void	tcfixMargin(void);

#define	MARGIN	8			/* size of minimim margin and	*/
#define	SCRSIZ	64			/* scroll size for extended lines */
#define	NPAUSE	10			/* # times thru update to pause */

/*
 * The names of the following two functions were changed because of clashes
 * because the System V.0 C compiler only recognises the first 7 characters
 * as significant.
 */
#if	TERMCAP_USE_TERM_H == 0
char*	tgoto(char*, int, int);
char*	tparm(char* string, ...);
#endif

#if	CURSES == 0
// We need to give termcap a buffer of 1024
#define TCAPSLEN	1024	/* Some can be long */
char	tcapbuf[TCAPSLEN];
#endif

/* What we extract from termcap/terminfo
 * Some names not quite what you might expect - avoid clashes with some <termcap.h>
 *
 * UP clashes on Minix.
 */
static	char*	UP1;		/* Up one line */
static	char*	UPsl;		/* Up several lines */
static	char*	CM;		/* Cursor motion */
static	char*	HO;		/* Move cursor home: position 0,0 */
static	char*	DO1;		/* Down one line */
static	char*	DO;		/* Down several lines */
static	char*	LE1;		/* Move left one space */
static	char*	LE;		/* Move left several spaces */
static	char*	ND1;		/* Move right one space */
static	char*	ND;		/* Move right several spaces */
static	char*	CR;		/* Carriage return - move to column 0 */
static	char*	CE;		/* Clear to end of line */
static	char*	CL;		/* Clear screen */
static	char*	AL;		/* Insert N lines */
static	char*	al;		/* Insert 1 line */
static	char*	DL;		/* Delete N lines */
static	char*	dl;		/* Delete 1 line */
static	char*	cs;		/* Sets a scrolling region, top line N1 bottom line N2, leaves cursor position undefined */
static	char*	sf;		/* Scroll up 1 line, cursor must be at bottom of region */
static	char*	SF;		/* Scroll up n lines */
static	char*	sr;		/* Inserts 1 line at 0,0 (top of region) - cursor must be at 0,0 */
static	char*	SR;		/* Inserts N lines at 0,0 */
static	char*	dc;		/* Delete one character */
static	char*	DC;		/* Delete N characters */
static	char*	ic;		/* Insert one character */
static	char*	IC;		/* Insert N characters */
static	char*	im;		/* Enter insert mode */
static	char*	ei;		/* Exit insert mode */
static	char*	SO;		/* Begin standout mode */
static	char*	SE;		/* End standout mode */
static	char*	US;		/* Start underline mode */
static	int	gnomeSeUeBug;	/* SE & UE are the same. gnome terminfo bug */
static	char*	UE;		/* End underline mode */
static	char*	TI;		/* Begin program that uses cursor motion */
static	char*	VS;		/* Make cursor very visible */
static	char*	TE;		/* End program that uses cursor motion */
static	char*	ME;		/* Turn off all attributes */
static	char*	KS;		/* Keypad start */
static	char*	KE;		/* Keypad end */
static	char*	VB;		/* Visible bell */
static	char*	BL;		/* Bell sequence */
static	char*	BT;		/* Move to previous tab position */
static	char*	HT;		/* Move to next tab position */
static	char	PCh;		/* Pad character */
static	char*	PC_t;		/* Pad character - as a string */
static	char*	K_up;		/* Sent by up arrow */
static	char*	K_down;		/* Sent by down arrow */
static	char*	K_left;		/* Sent by left arrow */
static	char*	K_right;	/* Sent by right arrow */
static	char*	K_pgup;		/* Sent by page up */
static	char*	K_pgdown;	/* Sent by page down */
static	char*	K_home;		/* Home key */
static	char*	K_end;		/* End key */
static	char*	K_del;		/* Delete key */
static	char*	K_ins;		/* Insert key */
static	char*	K_btab;		/* Back-tab key */
static	char*	K_f0;		/* F0 key */
static	char*	K_f1;		/* F1 key */
static	char*	K_f2;
static	char*	K_f3;
static	char*	K_f4;
static	char*	K_f5;
static	char*	K_f6;
static	char*	K_f7;
static	char*	K_f8;
static	char*	K_f9;
static	char*	K_f10;
static	char*	K_f11;
static	char*	K_f12;
static	char*	K_f13;		// F13 key - terminfo: kf13 key_f13
static	char*	K_f14;		// F14 key - terminfo: kf14 key_f14
static	char*	K_f15;		// F15 key - terminfo: kf15 key_f15
static	char*	K_f16;		// F16 key - terminfo: kf16 key_f16
static	char*	K_f17;		// F17 key - terminfo: kf17 key_f17
static	char*	K_f18;		// F18 key - terminfo: kf18 key_f18
static	char*	K_f19;		// F19 key - terminfo: kf19 key_f19
static	char*	K_f20;		// F20 key - terminfo: kf20 key_f20
static	char*	K_f21;		// F21 key - terminfo: kf21 key_f21
static	char*	K_f22;		// F22 key - terminfo: kf22 key_f22
static	char*	K_f23;		// F23 key - terminfo: kf23 key_f23
static	char*	K_f24;		// F24 key - terminfo: kf24 key_f24
static	char*	K_f25;		// F25 key - terminfo: kf25 key_f25
static	char*	K_f26;		// F26 key - terminfo: kf26 key_f26
static	char*	K_f27;		// F27 key - terminfo: kf27 key_f27
static	char*	K_f28;		// F28 key - terminfo: kf28 key_f28
static	char*	K_f29;		// F29 key - terminfo: kf29 key_f29
static	char*	K_f30;		// F30 key - terminfo: kf30 key_f30
static	char*	K_f31;		// F31 key - terminfo: kf31 key_f31
static	char*	K_f32;		// F32 key - terminfo: kf32 key_f32
static	char*	K_f33;		// F33 key - terminfo: kf33 key_f33
static	char*	K_f34;		// F34 key - terminfo: kf34 key_f34
static	char*	K_f35;		// F35 key - terminfo: kf35 key_f35
static	char*	K_f36;		// F36 key - terminfo: kf36 key_f36
static	char*	K_f37;		// F37 key - terminfo: kf37 key_f37
static	char*	K_f38;		// F38 key - terminfo: kf38 key_f38
static	char*	K_f39;		// F39 key - terminfo: kf39 key_f39
static	char*	K_f40;		// F40 key - terminfo: kf40 key_f40
static	char*	K_f41;		// F41 key - terminfo: kf41 key_f41
static	char*	K_f42;		// F42 key - terminfo: kf42 key_f42
static	char*	K_f43;		// F43 key - terminfo: kf43 key_f43
static	char*	K_f44;		// F44 key - terminfo: kf44 key_f44
static	char*	K_f45;		// F45 key - terminfo: kf45 key_f45
static	char*	K_f46;		// F46 key - terminfo: kf46 key_f46
static	char*	K_f47;		// F47 key - terminfo: kf47 key_f47
static	char*	K_f48;		// F48 key - terminfo: kf48 key_f48
static	char*	K_f49;		// F49 key - terminfo: kf49 key_f49
static	char*	K_f50;		// F50 key - terminfo: kf50 key_f50
static	char*	K_f51;		// F51 key - terminfo: kf51 key_f51
static	char*	K_f52;		// F52 key - terminfo: kf52 key_f52
static	char*	K_f53;		// F53 key - terminfo: kf53 key_f53
static	char*	K_f54;		// F54 key - terminfo: kf54 key_f54
static	char*	K_f55;		// F55 key - terminfo: kf55 key_f55
static	char*	K_f56;		// F56 key - terminfo: kf56 key_f56
static	char*	K_f57;		// F57 key - terminfo: kf57 key_f57
static	char*	K_f58;		// F58 key - terminfo: kf58 key_f58
static	char*	K_f59;		// F59 key - terminfo: kf59 key_f59
static	char*	K_f60;		// F60 key - terminfo: kf60 key_f60
static	char*	K_f61;		// F61 key - terminfo: kf61 key_f61
static	char*	K_f62;		// F62 key - terminfo: kf62 key_f62
static	char*	K_f63;		// F63 key - terminfo: kf63 key_f63

static	char*	Sf;		// Set foreground 
static	char*	Sb;		// Set background 
static	char*	AF;		// Set foreground - ANSI colours
static	char*	AB;		// Set background - ANSI colours
static	char*	OP;		// Original pair
static	int	Co;		/* Max # colours supported */
static	int	TabWidth;	// init_tabs "it" Tabs every this number of positions

#ifdef	NCURSES_VERSION
/* Extended terminfo capabilities.
 * These are only available via terminfo, ie not termcap.
 * Wanting to use these was the motivation to support termcap.
 *
 * Why want them in the first place ? On a laptop with a glidepad I found that I was
 * getting many of these sequences generated accidentally by brushing the glidepad while
 * control/... was held down. These were not recognised and this resulted in bits of
 * escape sequence being inserted into the file. So the primary aim was to recognise them
 * which would result in Emacs complaining that they were unbound.
 * Being able to bind them might be useful as well.
 *
 * Fixing the problem on xterm was the motivator - should work elsewhere as well.
 *
 * Some of the comments below come-from/are-based-on ncurses-6.0/misc/terminfo.src
 *
 * These provide most of the xterm extensions are for function-keys.  Since patch #94 (in
 * 1999), xterm has supported shift/control/alt/meta modifiers which produce
 * additional function-key strings.
 *
 * Some names are extensions allowed by ncurses, e.g.,
 *       kDN, kDN5, kDN6, kLFT5, kLFT6, kRIT5, kRIT6, kUP, kUP5, kUP6
 *
 * The uppercase names are made up, since there are no standards that apply.
 * If they were limited to two characters, they could in principle be translated
 * to termcap.  However, termcap sizes are limited to 1023 bytes, so there is
 * little point in ensuring that extended key names can be translated to
 * termcap.  A terminfo file can be up to 4096 bytes; using all extended keys
 * that xterm can generate would in fact exceed that limit.
 *
 * The numbers correspond to the modifier parameters documented in Xterm
 * Control Sequences:
 *
 *	 2	 Shift
 *	 3	 Alt
 *	 4	 Shift + Alt
 *	 5	 Control
 *	 6	 Shift + Control
 *	 7	 Alt + Control
 *	 8	 Shift + Alt + Control
 *
 * Thus names like kDC3 and kDC4. There is not a kDC2, it is called kDC.
 * These relate to kdch1 - the delete key.
 * kDC3 .. kDC7 are defined, not kDC8 - can't bind S-A-C-xxx
 * So: the naming is not entirely obvious; see tcap_tab[] and TermToks[]
 *
 * Base names:
 *	kUP	up-arrow
 *	kDN	down-arrow
 *	kRIT	right-arrow
 *	kLFT	left-arrow
 *	kIC	insert
 *	kDC	delete
 *	kHOM	home
 *	kEND	end
 *	kNXT	page-down	(next page)
 *	kPRV	page-up		(previous page)
 *
 * Sept 2019: added a lot of optimisation for local movements, eg use TAB to
 * move along a line. This exposed a bug in Gnome terminal 2.31.2 (CentOS 6)
 * where a block cursor (inverted colouring) put on space drawn with a TAB -
 * the whole width made by that TAB was inverted. Doesn't look very nice. This
 * bug is fixed with Mate terminal 1.16.3 that comes with CentOS 7.
 * The bug is not seen in: xterm, screen, putty, ...
 */

static	char*	K_s_up;		// Sent by shift up arrow - terminfo: kUP
static	char*	K_a_up;		// Sent by alt up arrow - terminfo: kUP3
static	char*	K_sa_up;	// Sent by shift-alt up arrow - terminfo: kUP4
static	char*	K_c_up;		// Sent by control up arrow - terminfo: kUP5
static	char*	K_sc_up;	// Sent by shift-control up arrow - terminfo: kUP6
static	char*	K_ac_up;	// Sent by alt-control up arrow - terminfo: kUP7

static	char*	K_s_down;	// Sent by shift down arrow - terminfo: kDN
static	char*	K_a_down;	// Sent by alt down arrow - terminfo: kDN3
static	char*	K_sa_down;	// Sent by shift-alt down arrow - terminfo: kDN4
static	char*	K_c_down;	// Sent by control down arrow - terminfo: kDN5
static	char*	K_sc_down;	// Sent by shift-control down arrow - terminfo: kDN6
static	char*	K_ac_down;	// Sent by alt-control down arrow - terminfo: kDN7

static	char*	K_s_right;	// Sent by shift right arrow - terminfo: kRIT
static	char*	K_a_right;	// Sent by alt right arrow - terminfo: kRIT3
static	char*	K_sa_right;	// Sent by shift-alt right arrow - terminfo: kRIT4
static	char*	K_c_right;	// Sent by control right arrow - terminfo: kRIT5
static	char*	K_sc_right;	// Sent by shift-control right arrow - terminfo: kRIT6
static	char*	K_ac_right;	// Sent by alt-control right arrow - terminfo: kRIT7

static	char*	K_s_left;	// Sent by shift left arrow - terminfo: kLFT
static	char*	K_a_left;	// Sent by alt left arrow - terminfo: kLFT3
static	char*	K_sa_left;	// Sent by shift-alt left arrow - terminfo: kLFT4
static	char*	K_c_left;	// Sent by control left arrow - terminfo: kLFT5
static	char*	K_sc_left;	// Sent by shift-control left arrow - terminfo: kLFT6
static	char*	K_ac_left;	// Sent by alt-control left arrow - terminfo: kLFT7

static	char*	K_s_insert;	// Sent by shift insert key - terminfo: kIC
static	char*	K_a_insert;	// Sent by alt insert key - terminfo: kIC3
static	char*	K_sa_insert;	// Sent by shift-alt insert key - terminfo: kIC4
static	char*	K_c_insert;	// Sent by control insert key - terminfo: kIC5
static	char*	K_sc_insert;	// Sent by shift-control insert key - terminfo: kIC6
static	char*	K_ac_insert;	// Sent by alt-control insert key - terminfo: kIC7

static	char*	K_s_delete;	// Sent by shift delete key - terminfo: kDC
static	char*	K_a_delete;	// Sent by alt delete key - terminfo: kDC3
static	char*	K_sa_delete;	// Sent by shift-alt delete key - terminfo: kDC4
static	char*	K_c_delete;	// Sent by control delete key - terminfo: kDC5
static	char*	K_sc_delete;	// Sent by shift-control delete key - terminfo: kDC6
static	char*	K_ac_delete;	// Sent by alt-control delete key - terminfo: kDC7

static	char*	K_s_home;	// Sent by shift home key - terminfo: kHOM
static	char*	K_a_home;	// Sent by alt home key - terminfo: kHOM3
static	char*	K_sa_home;	// Sent by shift-alt home key - terminfo: kHOM4
static	char*	K_c_home;	// Sent by control home key - terminfo: kHOM5
static	char*	K_sc_home;	// Sent by shift-control home key - terminfo: kHOM6
static	char*	K_ac_home;	// Sent by alt-control home key - terminfo: kHOM7

static	char*	K_s_end;	// Sent by shift end key - terminfo: kEND
static	char*	K_a_end;	// Sent by alt end key - terminfo: kEND3
static	char*	K_sa_end;	// Sent by shift-alt end key - terminfo: kEND4
static	char*	K_c_end;	// Sent by control end key - terminfo: kEND5
static	char*	K_sc_end;	// Sent by shift-control end key - terminfo: kEND6
static	char*	K_ac_end;	// Sent by alt-control end key - terminfo: kEND7

static	char*	K_s_pgdown;	// Sent by shift page down key - terminfo: kNXT
static	char*	K_a_pgdown;	// Sent by alt page down key - terminfo: kNXT3
static	char*	K_sa_pgdown;	// Sent by shift-alt page down key - terminfo: kNXT4
static	char*	K_c_pgdown;	// Sent by control page down key - terminfo: kNXT5
static	char*	K_sc_pgdown;	// Sent by shift-control page down key - terminfo: kNXT6
static	char*	K_ac_pgdown;	// Sent by alt-control page down key - terminfo: kNXT7

static	char*	K_s_pgup;	// Sent by shift page up key - terminfo: kPRV
static	char*	K_a_pgup;	// Sent by alt page up key - terminfo: kPRV3
static	char*	K_sa_pgup;	// Sent by shift-alt page up key - terminfo: kPRV4
static	char*	K_c_pgup;	// Sent by control page up key - terminfo: kPRV5
static	char*	K_sc_pgup;	// Sent by shift-control page up key - terminfo: kPRV6
static	char*	K_ac_pgup;	// Sent by alt-control page up key - terminfo: kPRV7

static	char*	K_kp_ent;	// Sent by keypad Enter - kent key_enter
static	char*	K_kp_a1;	// Sent by keypad upper left - ka1 key_a1
static	char*	K_kp_a3;	// Sent by keypad upper right - ka3 key_a3
static	char*	K_kp_b2;	// Sent by keypad center - kb1 key_b2
static	char*	K_kp_c1;	// Sent by keypad lower left - kc1 key_c1
static	char*	K_kp_c3;	// Sent by keypad lower right - kc3 key_c3

#endif	// NCURSES_VERSION

/* Not needed for now:
 * static	int	NC;	// video attributes that cannot be used with colors. I ignore this - for now.
 *				// For example, on many IBM PC consoles, the underline attribute collides with the foreground color blue and is not available in color mode.  These should have an ncv capability of 2
 */

static	int	ControlLeadIn;	/* All Fn keys have ctrl char as 1st char */
static	int	KeyByte1Same;	// True if all sequences generated by function/... keys start with the same byte
static	char	KeyByte1;	// This is what that byte is. NUL means not known

/* Cost means string length in: DO1, LE, CR -- if they are defined, else a very large number */
static	int	CostDO1 = 999;	/* # chars in DOl - down 1 line */
static	int	CostDO  = 999;	/* # chars in DO  - down several lines */
static	int	CostLE1 = 999;	/* # chars in LE1 - left 1 space */
static	int	CostLE  = 999;	/* # chars in LE  - left several spaces */
static	int	CostCR  = 999;	/* # chars in CR  - move to column 0 */
static	int	CostND1 = 999;	/* # chars in ND1 - move right 1 space */
static	int	CostND  = 999;	/* # chars in ND  - move right several spaces */
static	int	CostUP1 = 999;	/* # chars in UPl - move up one line */
static	int	CostUP  = 999;	/* # chars in UPsl  - move up several lines */
static	int	CostHT  = 999;	/* # chars in HT  - Horisontal tab */
static	int	CostBT  = 999;	/* # chars in BT  - Back horisontal tab */
static	int	CostHO  = 999;	/* # chars in HO  - move home (0,0) */

static	int	CostCM;		/* # chars in typical cursor movement */
static	int	CmPricey = 0;	/* CmPricey means that CM costs more than DO & LE */

// Can we do some things ?
static	int	CanIlDl;	// Can insert/delete line
static	int	CanScroll;	// Can set up scroll region
static	int	CanInsC1;	// Can insert 1 char
static	int	CanInsCN;	// Can insert N chars
static	int	CanInsC;	// Can insert



/* Array of strings that change colours.
 * These are ordered COLOUR_BLACK..COLOUR_WHITE.
 * The values below are overwritten when termcap is read.
 * By writing empty strings: run time is simpler.
 */
static	char*	ColChange_fg[] = { "", "", "", "", "", "", "", "" };
static	char*	ColChange_bg[] = { "", "", "", "", "", "", "", "" };

/* Map a setf/setb colour to get the setaf/setab colour */
static	char	mapCols[] = { 0, 4, 2, 6, 1, 5, 3, 7 };

/* Current foreground and background colours.
 * Special (non colour) values are negative.
 */
#define	ATTRIB_UNSET	-2		// Set on startup when we don't know what the state is
#define	COLOUR_UNSET	ATTRIB_UNSET	// Set on startup when we don't know what the state is
static	int	nowfcol;		// Foreground colour
static	int	wantfcol;
static	int	nowbcol;		// Background colour
static	int	wantbcol;
static	int	nowrev;			// Reverse
static	int	wantrev;
static	int	nowul;			// Underline
static	int	wantul;

/* Something to help extract strings easily
 *
 * The names are the termcap names, and the capability names (see terminfo(5)).
 * These are different thus the use of the macro to decide which one is compiled
 * in - don't have both.
 */
typedef	struct	{
	char**	tc_str;		/* String value extracted from terminal database */
#if	CURSES
	char*	tc_name;	/* Terminfo name */
#else
	char	tc_name[3];	/* Termcap name */
#endif
} tcap_strs;

#if	CURSES
#define	GetTent(store, tcap, tinfo)	{store,	tinfo }
#else
#define	GetTent(store, tcap, tinfo)	{store,	tcap }
#endif

/* Init termcap/terminfo.
 * Terminated by NULL tc_str.
 */
static	tcap_strs	tcap_tab[] = {
	GetTent(&PC_t,		"pc",	"pad"	),	// Padding char - instead of \0 - terminfo: pad pad_char
	GetTent(&UP1,		"up",	"cuu1"	),	// Up one line - terminfo: cuu1 cursor_up
	GetTent(&UPsl,		"UP",	"cuu"	),	// Up several lines - terminfo: cuu parm_up_cursor
	GetTent(&DO1,		"do",	"cud1"	),	// Down one line - terminfo: cud1 cursor_down
	GetTent(&DO,		"DO",	"cud"	),	// Down several lines - terminfo: cud parm_down_cursor
	GetTent(&LE1,		"le",	"cub1"	),	// Move left one character - terminfo: cub1 cursor_left
	GetTent(&LE,		"LE",	"cub"	),	// Move left one character - terminfo: cub parm_left_cursor
	GetTent(&ND1,		"nd",	"cuf1"	),	// Move right one character (non destructive space) - terminfo: cuf1 cursor_right
	GetTent(&ND,		"ND",	"cuf"	),	// Move right one character (non destructive space) - terminfo: cuf parm_right_cursor
	GetTent(&CR,		"cr",	"cr"	),	// Carriage return - move to column 0 - terminfo: cr carriage_return
	GetTent(&CM,		"cm",	"cup"	),	// Cursor move - terminfo: cup cursor_address
	GetTent(&HO,		"ho",	"home"	),	// Cursor home - terminfo: home cursor_home
	GetTent(&CE,		"ce",	"el"	),	// Clear to end of line - terminfo: el clr_eol
	GetTent(&CL,		"cl",	"clear"	),	// Move cursor home and clear screen - terminfo: clear clear_screen
	GetTent(&AL,		"AL",	"il"	),	// Insert N lines - terminfo: il parm_insert_line
	GetTent(&al,		"al",	"il1"	),	// Insert 1 line - terminfo: il1 insert_line
	GetTent(&DL,		"DL",	"dl"	),	// Delete N lines - terminfo: dl parm_delete_line
	GetTent(&dl,		"dl",	"dl1"	),	// Delete 1 line - terminfo: dl1 delete_line
	GetTent(&cs,		"cs",	"csr"	),	// Sets a scrolling region, terminfo: csr change_scroll_region
	GetTent(&sf,		"sf",	"ind"	),	// Scroll up 1 line, cursor must be at bottom of region - terminfo: ind scroll_forward
	GetTent(&SF,		"SF",	"indn"	),	// Scroll up n lines - terminfo: indn parm_index
	GetTent(&sr,		"sr",	"ri"	),	// Inserts 1 line at 0,0, terminfo: ri scroll_reverse
	GetTent(&SR,		"SR",	"rin"	),	// Inserts N lines at 0,0 - terminfo: rin parm_rindex
	GetTent(&DC,		"DC",	"dch"	),	// Delete N characters - terminfo: dch parm_dch
	GetTent(&dc,		"dc",	"dch1"	),	// Delete 1 character - terminfo: dch1 delete_character
	GetTent(&IC,		"IC",	"ich"	),	// Insert N characters - terminfo: ich parm_ich
	GetTent(&ic,		"ic",	"ich1"	),	// Insert 1 character - terminfo: ich1 insert_character
	GetTent(&im,		"im",	"smir"	),	// Enter insert mode - terminfo: smir enter_insert_mode
	GetTent(&ei,		"ei",	"rmir"	),	// Exit insert mode - terminfo: rmir exit_insert_mode
	GetTent(&SO,		"so",	"smso"	),	// Begin standout mode - terminfo: smso enter_standout_mode
	GetTent(&SE,		"se",	"rmso"	),	// Exit standout mode - terminfo: rmso exit_standout_mode
	GetTent(&US,		"us",	"smul"	),	// Begin underline mode  - terminfo: smul enter_underline_mode
	GetTent(&UE,		"ue",	"rmul"	),	// End underline mode - terminfo: rmul exit_underline_mode
	GetTent(&VB,		"vb",	"flash"	),	// Visible bell - terminfo: flash flash_screen
	GetTent(&BL,		"bl",	"bel"	),	// Ring the bell - terminfo: bel bell
	GetTent(&BT,		"bt",	"cbt"	),	// Move to previous tab position
	GetTent(&HT,		"ta",	"ht"	),	// Move to next tab position
	GetTent(&TI,		"ti",	"smcup"	),	// Terminal Initialise - enter visual mode - terminfo: smcup enter_ca_mode
	GetTent(&VS,		"vs",	"cvvis"	),	// Make cursor very visible - terminfo: cvvis cursor_visible
	GetTent(&TE,		"te",	"rmcup"	),	// Terminal End - exit visual mode - terminfo: rmcup exit_ca_mode
	GetTent(&ME,		"me",	"sgr0"	),	// Turn off all attributes - terminfo: sgr0 exit_attribute_mode
	GetTent(&KS,		"ks",	"smkx"	),	// Keyboard Start - terminfo: smkx keypad_xmit
	GetTent(&KE,		"ke",	"rmkx"	),	// Keyboard End - terminfo: rmkx keypad_local
	GetTent(&K_up,		"ku",	"kcuu1"	),	// Up arrow key - terminfo: kcuu1 key_up
	GetTent(&K_down,	"kd",	"kcud1"	),	// Down arrow key - terminfo: kcud1 key_down
	GetTent(&K_left,	"kl",	"kcub1"	),	// Left arrow key - terminfo: kcub1 key_left
	GetTent(&K_right,	"kr",	"kcuf1"	),	// Right arrow key - terminfo: kcuf1 key_right
	GetTent(&K_pgup,	"kP",	"kpp"	),	// Previous page key - terminfo: kpp key_ppage
	GetTent(&K_pgdown,	"kN",	"knp"	),	// Next page key - terminfo: knp key_npage
	GetTent(&K_home,	"kh",	"khome"	),	// Home key - terminfo: khome key_home
	GetTent(&K_end,		"kH",	"kll"	),	// Lower left key (home down) - terminfo: kll key_ll
	GetTent(&K_end,		"@7",	"kend"	),	// End key - sometimes - terminfo: kend key_end
	GetTent(&K_del,		"kD",	"kdch1"	),	// Delete character key - terminfo: kdch1 key_dc
	GetTent(&K_ins,		"kI",	"kich1"	),	// Insert character key - terminfo: kich1 key_ic
	GetTent(&K_btab,	"kB",	"kcbt"	),	// Back-tab key key - terminfo: kcbt key_btab

	GetTent(&K_f0,		"k0",	"kf0"	),	// F0 key - terminfo: kf0 key_f0
	GetTent(&K_f1,		"k1",	"kf1"	),	// F1 key - terminfo: kf1 key_f1
	GetTent(&K_f2,		"k2",	"kf2"	),	// F2 key - terminfo: kf2 key_f2
	GetTent(&K_f3,		"k3",	"kf3"	),	// F3 key - terminfo: kf3 key_f3
	GetTent(&K_f4,		"k4",	"kf4"	),	// F4 key - terminfo: kf4 key_f4
	GetTent(&K_f5,		"k5",	"kf5"	),	// F5 key - terminfo: kf5 key_f5
	GetTent(&K_f6,		"k6",	"kf6"	),	// F6 key - terminfo: kf6 key_f6
	GetTent(&K_f7,		"k7",	"kf7"	),	// F7 key - terminfo: kf7 key_f7
	GetTent(&K_f8,		"k8",	"kf8"	),	// F8 key - terminfo: kf8 key_f8
	GetTent(&K_f9,		"k9",	"kf9"	),	// F9 key - terminfo: kf9 key_f9

	GetTent(&K_f10,		"k;",	"kf10"	),	// F10 key - terminfo: kf10 key_f10
	GetTent(&K_f11,		"F1",	"kf11"	),	// F11 key - terminfo: kf11 key_f11
	GetTent(&K_f12,		"F2",	"kf12"	),	// F12 key - terminfo: kf12 key_f12
	GetTent(&K_f13,		"F3",	"kf13"	),	// F13 key - terminfo: kf13 key_f13
	GetTent(&K_f14,		"F4",	"kf14"	),	// F14 key - terminfo: kf14 key_f14
	GetTent(&K_f15,		"F5",	"kf15"	),	// F15 key - terminfo: kf15 key_f15
	GetTent(&K_f16,		"F6",	"kf16"	),	// F16 key - terminfo: kf16 key_f16
	GetTent(&K_f17,		"F7",	"kf17"	),	// F17 key - terminfo: kf17 key_f17
	GetTent(&K_f18,		"F8",	"kf18"	),	// F18 key - terminfo: kf18 key_f18
	GetTent(&K_f19,		"F9",	"kf19"	),	// F19 key - terminfo: kf19 key_f19

	GetTent(&K_f20,		"FA",	"kf20"	),	// F20 key - terminfo: kf20 key_f20
	GetTent(&K_f21,		"FB",	"kf21"	),	// F21 key - terminfo: kf21 key_f21
	GetTent(&K_f22,		"FC",	"kf22"	),	// F22 key - terminfo: kf22 key_f22
	GetTent(&K_f23,		"FD",	"kf23"	),	// F23 key - terminfo: kf23 key_f23
	GetTent(&K_f24,		"FE",	"kf24"	),	// F24 key - terminfo: kf24 key_f24
	GetTent(&K_f25,		"FF",	"kf25"	),	// F25 key - terminfo: kf25 key_f25
	GetTent(&K_f26,		"FG",	"kf26"	),	// F26 key - terminfo: kf26 key_f26
	GetTent(&K_f27,		"FH",	"kf27"	),	// F27 key - terminfo: kf27 key_f27
	GetTent(&K_f28,		"FI",	"kf28"	),	// F28 key - terminfo: kf28 key_f28
	GetTent(&K_f29,		"FJ",	"kf29"	),	// F29 key - terminfo: kf29 key_f29

	GetTent(&K_f30,		"FK",	"kf30"	),	// F30 key - terminfo: kf30 key_f30
	GetTent(&K_f31,		"FL",	"kf31"	),	// F31 key - terminfo: kf31 key_f31
	GetTent(&K_f32,		"FM",	"kf32"	),	// F32 key - terminfo: kf32 key_f32
	GetTent(&K_f33,		"FN",	"kf33"	),	// F33 key - terminfo: kf33 key_f33
	GetTent(&K_f34,		"FO",	"kf34"	),	// F34 key - terminfo: kf34 key_f34
	GetTent(&K_f35,		"FP",	"kf35"	),	// F35 key - terminfo: kf35 key_f35
	GetTent(&K_f36,		"FQ",	"kf36"	),	// F36 key - terminfo: kf36 key_f36
	GetTent(&K_f37,		"FR",	"kf37"	),	// F37 key - terminfo: kf37 key_f37
	GetTent(&K_f38,		"FS",	"kf38"	),	// F38 key - terminfo: kf38 key_f38
	GetTent(&K_f39,		"FT",	"kf39"	),	// F39 key - terminfo: kf39 key_f39

	GetTent(&K_f40,		"FU",	"kf40"	),	// F40 key - terminfo: kf40 key_f40
	GetTent(&K_f41,		"FV",	"kf41"	),	// F41 key - terminfo: kf41 key_f41
	GetTent(&K_f42,		"FW",	"kf42"	),	// F42 key - terminfo: kf42 key_f42
	GetTent(&K_f43,		"FX",	"kf43"	),	// F43 key - terminfo: kf43 key_f43
	GetTent(&K_f44,		"FY",	"kf44"	),	// F44 key - terminfo: kf44 key_f44
	GetTent(&K_f45,		"FZ",	"kf45"	),	// F45 key - terminfo: kf45 key_f45
	GetTent(&K_f46,		"Fa",	"kf46"	),	// F46 key - terminfo: kf46 key_f46
	GetTent(&K_f47,		"Fb",	"kf47"	),	// F47 key - terminfo: kf47 key_f47
	GetTent(&K_f48,		"Fc",	"kf48"	),	// F48 key - terminfo: kf48 key_f48
	GetTent(&K_f49,		"Fd",	"kf49"	),	// F49 key - terminfo: kf49 key_f49

	GetTent(&K_f50,		"Fe",	"kf50"	),	// F50 key - terminfo: kf50 key_f50
	GetTent(&K_f51,		"Ff",	"kf51"	),	// F51 key - terminfo: kf51 key_f51
	GetTent(&K_f52,		"Fg",	"kf52"	),	// F52 key - terminfo: kf52 key_f52
	GetTent(&K_f53,		"Fh",	"kf53"	),	// F53 key - terminfo: kf53 key_f53
	GetTent(&K_f54,		"Fi",	"kf54"	),	// F54 key - terminfo: kf54 key_f54
	GetTent(&K_f55,		"Fj",	"kf55"	),	// F55 key - terminfo: kf55 key_f55
	GetTent(&K_f56,		"Fk",	"kf56"	),	// F56 key - terminfo: kf56 key_f56
	GetTent(&K_f57,		"Fl",	"kf57"	),	// F57 key - terminfo: kf57 key_f57
	GetTent(&K_f58,		"Fm",	"kf58"	),	// F58 key - terminfo: kf58 key_f58
	GetTent(&K_f59,		"Fn",	"kf59"	),	// F59 key - terminfo: kf59 key_f59

	GetTent(&K_f60,		"Fo",	"kf60"	),	// F60 key - terminfo: kf60 key_f60
	GetTent(&K_f61,		"Fp",	"kf61"	),	// F61 key - terminfo: kf61 key_f61
	GetTent(&K_f62,		"Fq",	"kf62"	),	// F62 key - terminfo: kf62 key_f62
	GetTent(&K_f63,		"Fr",	"kf63"	),	// F63 key - terminfo: kf63 key_f63

	GetTent(&OP,		"op",	"op"	),	// Original pair - resets foreground and background colors to their default values for the terminal - terminfo: op orig_pair
	GetTent(&AF,		"AF",	"setaf"	),	// Set foreground - ANSI colours - terminfo: setaf set_a_foreground
	GetTent(&AB,		"AB",	"setab"	),	// Set background - ANSI colours - terminfo: setab set_a_background
	GetTent(&Sf,		"Sf",	"setf"	),	// Set foreground - terminfo: setf set_foreground
	GetTent(&Sb,		"Sb",	"setb"	),	// Set background - terminfo: setb set_background

#ifdef	NCURSES_VERSION
	// These are only in terminfo, not termcap

	GetTent(&K_s_up,	NULL,	"kUP"	),	// Sent by shift up arrow - terminfo: kUP
	GetTent(&K_a_up,	NULL,	"kUP3"	),	// Sent by alt up arrow - terminfo: kUP3
	GetTent(&K_sa_up,	NULL,	"kUP4"	),	// Sent by shift-alt up arrow - terminfo: kUP4
	GetTent(&K_c_up,	NULL,	"kUP5"	),	// Sent by control up arrow - terminfo: kUP5
	GetTent(&K_sc_up,	NULL,	"kUP6"	),	// Sent by shift-control up arrow - terminfo: kUP6
	GetTent(&K_ac_up,	NULL,	"kUP7"	),	// Sent by alt-control up arrow - terminfo: kUP7

	GetTent(&K_s_down,	NULL,	"kDN"	),	// Sent by shift down arrow - terminfo: kDN
	GetTent(&K_a_down,	NULL,	"kDN3"	),	// Sent by alt down arrow - terminfo: kDN3
	GetTent(&K_sa_down,	NULL,	"kDN4"	),	// Sent by shift-alt down arrow - terminfo: kDN4
	GetTent(&K_c_down,	NULL,	"kDN5"	),	// Sent by control down arrow - terminfo: kDN5
	GetTent(&K_sc_down,	NULL,	"kDN6"	),	// Sent by shift-control down arrow - terminfo: kDN6
	GetTent(&K_ac_down,	NULL,	"kDN7"	),	// Sent by alt-control down arrow - terminfo: kDN7

	GetTent(&K_s_right,	NULL,	"kRIT"	),	// Sent by shift right arrow - terminfo: kRIT
	GetTent(&K_a_right,	NULL,	"kRIT3"	),	// Sent by alt right arrow - terminfo: kRIT3
	GetTent(&K_sa_right,	NULL,	"kRIT4"	),	// Sent by shift-alt right arrow - terminfo: kRIT4
	GetTent(&K_c_right,	NULL,	"kRIT5"	),	// Sent by control right arrow - terminfo: kRIT5
	GetTent(&K_sc_right,	NULL,	"kRIT6"	),	// Sent by shift-control right arrow - terminfo: kRIT6
	GetTent(&K_ac_right,	NULL,	"kRIT7"	),	// Sent by alt-control right arrow - terminfo: kRIT7

	GetTent(&K_s_left,	NULL,	"kLFT"	),	// Sent by shift left arrow - terminfo: kLFT
	GetTent(&K_a_left,	NULL,	"kLFT3"	),	// Sent by alt left arrow - terminfo: kLFT3
	GetTent(&K_sa_left,	NULL,	"kLFT4"	),	// Sent by shift-alt left arrow - terminfo: kLFT4
	GetTent(&K_c_left,	NULL,	"kLFT5"	),	// Sent by control left arrow - terminfo: kLFT5
	GetTent(&K_sc_left,	NULL,	"kLFT6"	),	// Sent by shift-control left arrow - terminfo: kLFT6
	GetTent(&K_ac_left,	NULL,	"kLFT7"	),	// Sent by alt-control left arrow - terminfo: kLFT7

	GetTent(&K_s_insert,	NULL,	"kIC"	),	// Sent by shift insert key - terminfo: kIC
	GetTent(&K_a_insert,	NULL,	"kIC3"	),	// Sent by alt insert key - terminfo: kIC3
	GetTent(&K_sa_insert,	NULL,	"kIC4"	),	// Sent by shift-alt insert key - terminfo: kIC4
	GetTent(&K_c_insert,	NULL,	"kIC5"	),	// Sent by control insert key - terminfo: kIC5
	GetTent(&K_sc_insert,	NULL,	"kIC6"	),	// Sent by shift-control insert key - terminfo: kIC6
	GetTent(&K_ac_insert,	NULL,	"kIC7"	),	// Sent by alt-control insert key - terminfo: kIC7

	GetTent(&K_s_delete,	NULL,	"kDC"	),	// Sent by shift delete key - terminfo: kDC
	GetTent(&K_a_delete,	NULL,	"kDC3"	),	// Sent by alt delete key - terminfo: kDC3
	GetTent(&K_sa_delete,	NULL,	"kDC4"	),	// Sent by shift-alt delete key - terminfo: kDC4
	GetTent(&K_c_delete,	NULL,	"kDC5"	),	// Sent by control delete key - terminfo: kDC5
	GetTent(&K_sc_delete,	NULL,	"kDC6"	),	// Sent by shift-control delete key - terminfo: kDC6
	GetTent(&K_ac_delete,	NULL,	"kDC7"	),	// Sent by alt-control delete key - terminfo: kDC7

	GetTent(&K_s_home,	NULL,	"kHOM"	),	// Sent by shift home key - terminfo: kHOM
	GetTent(&K_a_home,	NULL,	"kHOM3"	),	// Sent by alt home key - terminfo: kHOM3
	GetTent(&K_sa_home,	NULL,	"kHOM4"	),	// Sent by shift-alt home key - terminfo: kHOM4
	GetTent(&K_c_home,	NULL,	"kHOM5"	),	// Sent by control home key - terminfo: kHOM5
	GetTent(&K_sc_home,	NULL,	"kHOM6"	),	// Sent by shift-control home key - terminfo: kHOM6
	GetTent(&K_ac_home,	NULL,	"kHOM7"	),	// Sent by alt-control home key - terminfo: kHOM7

	GetTent(&K_s_end,	NULL,	"kEND"	),	// Sent by shift end key - terminfo: kEND
	GetTent(&K_a_end,	NULL,	"kEND3"	),	// Sent by alt end key - terminfo: kEND3
	GetTent(&K_sa_end,	NULL,	"kEND4"	),	// Sent by shift-alt end key - terminfo: kEND4
	GetTent(&K_c_end,	NULL,	"kEND5"	),	// Sent by control end key - terminfo: kEND5
	GetTent(&K_sc_end,	NULL,	"kEND6"	),	// Sent by shift-control end key - terminfo: kEND6
	GetTent(&K_ac_end,	NULL,	"kEND7"	),	// Sent by alt-control end key - terminfo: kEND7

	GetTent(&K_s_pgdown,	NULL,	"kNXT"	),	// Sent by shift page down key - terminfo: kNXT
	GetTent(&K_a_pgdown,	NULL,	"kNXT3"	),	// Sent by alt page down key - terminfo: kNXT3
	GetTent(&K_sa_pgdown,	NULL,	"kNXT4"	),	// Sent by shift-alt page down key - terminfo: kNXT4
	GetTent(&K_c_pgdown,	NULL,	"kNXT5"	),	// Sent by control page down key - terminfo: kNXT5
	GetTent(&K_sc_pgdown,	NULL,	"kNXT6"	),	// Sent by shift-control page down key - terminfo: kNXT6
	GetTent(&K_ac_pgdown,	NULL,	"kNXT7"	),	// Sent by alt-control page down key - terminfo: kNXT7

	GetTent(&K_s_pgup,	NULL,	"kPRV"	),	// Sent by shift page up key - terminfo: kPRV
	GetTent(&K_a_pgup,	NULL,	"kPRV3"	),	// Sent by alt page up key - terminfo: kPRV3
	GetTent(&K_sa_pgup,	NULL,	"kPRV4"	),	// Sent by shift-alt page up key - terminfo: kPRV4
	GetTent(&K_c_pgup,	NULL,	"kPRV5"	),	// Sent by control page up key - terminfo: kPRV5
	GetTent(&K_sc_pgup,	NULL,	"kPRV6"	),	// Sent by shift-control page up key - terminfo: kPRV6
	GetTent(&K_ac_pgup,	NULL,	"kPRV7"	),	// Sent by alt-control page up key - terminfo: kPRV7

	// Keypad when Num Lock is off
	// What is sent by xterm keypad '/' & '*' does not seem to be described
	GetTent(&K_kp_ent,	"@8",	"kent"	),	// Sent by keypad Enter - kent key_enter
	GetTent(&K_kp_a1,	"K1",	"ka1"	),	// Sent by keypad upper left - ka1 key_a1
	GetTent(&K_kp_a3,	"K3",	"ka3"	),	// Sent by keypad upper right - ka3 key_a3
	GetTent(&K_kp_b2,	"K2",	"kb2"	),	// Sent by keypad center - kb1 key_b2
	GetTent(&K_kp_c1,	"K4",	"kc1"	),	// Sent by keypad lower left - kc1 key_c1
	GetTent(&K_kp_c3,	"K5",	"kc3"	),	// Sent by keypad lower right - kc3 key_c3

#endif	// NCURSES_VERSION

	GetTent((char**)NULL,	NULL,	""	)
};

/* Numeric capability - maybe needed:
	GetTent(&NC,		"NC",	"ncv"	),	// Video attributes that cannot be used with colours - terminfo: ncv no_color_video
 */

/* This table holds all of the things that we may expect to get from the
 * user via the keyboard.
 *
 * The first element is the keyboard sequence itself.
 * the second is a user friendly name that can be displayed in key bindings.
 * the third is the integer token value that is returned when the sequence
 *	is recognised.
 * the fourth is the length of the keyboard sequence.
 * the fifth is (if not 0) a legacy name used in .emacsrc files, eg FNU for up arrow.
 *
 * The length is computed in tcapopen() - once it is known.
 *
 * These are arranged in roughly the order that one may expect to find them.
 *
 * FN11 & FN12 are out of order else they will not be found in tspecial().
 *
 * When scanning this table used tk_seq == NULL to determine when the end of
 * table is reached. This is due to how elmination of missing (in termcap)
 * entries is done in tcapopen().
 *
 * const on tk_name is to help the compiler optimise duplicates.
 */
static	struct	TkeyType {
	char**		tk_seq;		// Key sequence
	char*		tk_name;	// Name to describe the key with
	KeyCode		tk_token;	// Token to return
	char		tk_len;		// Length in *tk_seq
	char		tk_legacyName;	// Legacy name from when termcap first used
} TermToks[] = {
	{ &K_up,	"up-arrow",	UTC_UP,			0, 'U' },
	{ &K_down,	"down-arrow",	UTC_DOWN,		0, 'D' },
	{ &K_left,	"left-arrow",	UTC_LEFT,		0, 'L' },
	{ &K_right,	"right-arrow",	UTC_RIGHT,		0, 'R' },
	{ &K_pgup,	"page-up",	UTC_PGUP,		0, 'B' },
	{ &K_pgdown,	"page-down",	UTC_PGDOWN,		0, 'F' },
	{ &K_home,	"home",		UTC_HOME,		0, 'S' },	/* Home key */
	{ &K_end,	"end",		UTC_END,		0, 'E' },	/* End key */
	{ &K_del,	"delete",	UTC_DEL,		0, 'd' },
	{ &K_ins,	"insert",	UTC_INS,		0, 'I' },
	{ &K_f11,	"FN11",		UTC_FN_KEY(11),		0, 'a' },
	{ &K_f12,	"FN12",		UTC_FN_KEY(12),		0, 'b' },
	{ &K_btab,	"back-tab",	UTC_BAKTAB,		0, 0 },
//	{ &K_f0,	"FN0",		UTC_FN_KEY(0),		0, 0 },		// Name clash with FN10
	{ &K_f1,	"FN1",		UTC_FN_KEY(1),		0, 0 },
	{ &K_f2,	"FN2",		UTC_FN_KEY(2),		0, 0 },
	{ &K_f3,	"FN3",		UTC_FN_KEY(3),		0, 0 },
	{ &K_f4,	"FN4",		UTC_FN_KEY(4),		0, 0 },
	{ &K_f5,	"FN5",		UTC_FN_KEY(5),		0, 0 },
	{ &K_f6,	"FN6",		UTC_FN_KEY(6),		0, 0 },
	{ &K_f7,	"FN7",		UTC_FN_KEY(7),		0, 0 },
	{ &K_f8,	"FN8",		UTC_FN_KEY(8),		0, 0 },
	{ &K_f9,	"FN9",		UTC_FN_KEY(9),		0, 0 },
	{ &K_f10,	"FN10",		UTC_FN_KEY(10),		0, 0 },
	{ &K_f13,	"FN13",		UTC_FN_KEY(13),		0, 0 },
	{ &K_f14,	"FN14",		UTC_FN_KEY(14),		0, 0 },
	{ &K_f15,	"FN15",		UTC_FN_KEY(15),		0, 0 },
	{ &K_f16,	"FN16",		UTC_FN_KEY(16),		0, 0 },
	{ &K_f17,	"FN17",		UTC_FN_KEY(17),		0, 0 },
	{ &K_f18,	"FN18",		UTC_FN_KEY(18),		0, 0 },
	{ &K_f19,	"FN19",		UTC_FN_KEY(19),		0, 0 },
	{ &K_f20,	"FN20",		UTC_FN_KEY(20),		0, 0 },
	{ &K_f21,	"FN21",		UTC_FN_KEY(21),		0, 0 },
	{ &K_f22,	"FN22",		UTC_FN_KEY(22),		0, 0 },
	{ &K_f23,	"FN23",		UTC_FN_KEY(23),		0, 0 },
	{ &K_f24,	"FN24",		UTC_FN_KEY(24),		0, 0 },
	{ &K_f25,	"FN25",		UTC_FN_KEY(25),		0, 0 },
	{ &K_f26,	"FN26",		UTC_FN_KEY(26),		0, 0 },
	{ &K_f27,	"FN27",		UTC_FN_KEY(27),		0, 0 },
	{ &K_f28,	"FN28",		UTC_FN_KEY(28),		0, 0 },
	{ &K_f29,	"FN29",		UTC_FN_KEY(29),		0, 0 },
	{ &K_f30,	"FN30",		UTC_FN_KEY(30),		0, 0 },
	{ &K_f31,	"FN31",		UTC_FN_KEY(31),		0, 0 },
	{ &K_f32,	"FN32",		UTC_FN_KEY(32),		0, 0 },
	{ &K_f33,	"FN33",		UTC_FN_KEY(33),		0, 0 },
	{ &K_f34,	"FN34",		UTC_FN_KEY(34),		0, 0 },
	{ &K_f35,	"FN35",		UTC_FN_KEY(35),		0, 0 },
	{ &K_f36,	"FN36",		UTC_FN_KEY(36),		0, 0 },
	{ &K_f37,	"FN37",		UTC_FN_KEY(37),		0, 0 },
	{ &K_f38,	"FN38",		UTC_FN_KEY(38),		0, 0 },
	{ &K_f39,	"FN39",		UTC_FN_KEY(39),		0, 0 },
	{ &K_f40,	"FN40",		UTC_FN_KEY(40),		0, 0 },
	{ &K_f41,	"FN41",		UTC_FN_KEY(41),		0, 0 },
	{ &K_f42,	"FN42",		UTC_FN_KEY(42),		0, 0 },
	{ &K_f43,	"FN43",		UTC_FN_KEY(43),		0, 0 },
	{ &K_f44,	"FN44",		UTC_FN_KEY(44),		0, 0 },
	{ &K_f45,	"FN45",		UTC_FN_KEY(45),		0, 0 },
	{ &K_f46,	"FN46",		UTC_FN_KEY(46),		0, 0 },
	{ &K_f47,	"FN47",		UTC_FN_KEY(47),		0, 0 },
	{ &K_f48,	"FN48",		UTC_FN_KEY(48),		0, 0 },
	{ &K_f49,	"FN49",		UTC_FN_KEY(49),		0, 0 },
	{ &K_f50,	"FN50",		UTC_FN_KEY(50),		0, 0 },
	{ &K_f51,	"FN51",		UTC_FN_KEY(51),		0, 0 },
	{ &K_f52,	"FN52",		UTC_FN_KEY(52),		0, 0 },
	{ &K_f53,	"FN53",		UTC_FN_KEY(53),		0, 0 },
	{ &K_f54,	"FN54",		UTC_FN_KEY(54),		0, 0 },
	{ &K_f55,	"FN55",		UTC_FN_KEY(55),		0, 0 },
	{ &K_f56,	"FN56",		UTC_FN_KEY(56),		0, 0 },
	{ &K_f57,	"FN57",		UTC_FN_KEY(57),		0, 0 },
	{ &K_f58,	"FN58",		UTC_FN_KEY(58),		0, 0 },
	{ &K_f59,	"FN59",		UTC_FN_KEY(59),		0, 0 },
	{ &K_f60,	"FN60",		UTC_FN_KEY(60),		0, 0 },
	{ &K_f61,	"FN61",		UTC_FN_KEY(61),		0, 0 },
	{ &K_f62,	"FN62",		UTC_FN_KEY(62),		0, 0 },
	{ &K_f63,	"FN63",		UTC_FN_KEY(63),		0, 0 },

#ifdef	NCURSES_VERSION
	// Lots of repeated strings, but a decent compiler will optimise them to just one

	{ &K_s_up,	"up-arrow",	SHIFT |         UTC_UP,		0, 0 },
	{ &K_a_up,	"up-arrow",	META  |         UTC_UP,		0, 0 },
	{ &K_sa_up,	"up-arrow",	SHIFT | META  | UTC_UP,		0, 0 },
	{ &K_c_up,	"up-arrow",	CNTRL |         UTC_UP,		0, 0 },
	{ &K_sc_up,	"up-arrow",	SHIFT | CNTRL | UTC_UP,		0, 0 },
	{ &K_ac_up,	"up-arrow",	META  | CNTRL | UTC_UP,		0, 0 },

	{ &K_s_down,	"down-arrow",	SHIFT |         UTC_DOWN,	0, 0 },
	{ &K_a_down,	"down-arrow",	META  |         UTC_DOWN,	0, 0 },
	{ &K_sa_down,	"down-arrow",	SHIFT | META  | UTC_DOWN,	0, 0 },
	{ &K_c_down,	"down-arrow",	CNTRL |         UTC_DOWN,	0, 0 },
	{ &K_sc_down,	"down-arrow",	SHIFT | CNTRL | UTC_DOWN,	0, 0 },
	{ &K_ac_down,	"down-arrow",	META  | CNTRL | UTC_DOWN,	0, 0 },

	{ &K_s_left,	"left-arrow",	SHIFT |         UTC_LEFT,	0, 0 },
	{ &K_a_left,	"left-arrow",	META  |         UTC_LEFT,	0, 0 },
	{ &K_sa_left,	"left-arrow",	SHIFT | META  | UTC_LEFT,	0, 0 },
	{ &K_c_left,	"left-arrow",	CNTRL |         UTC_LEFT,	0, 0 },
	{ &K_sc_left,	"left-arrow",	SHIFT | CNTRL | UTC_LEFT,	0, 0 },
	{ &K_ac_left,	"left-arrow",	META  | CNTRL | UTC_LEFT,	0, 0 },

	{ &K_s_right,	"right-arrow",	SHIFT |         UTC_RIGHT,	0, 0 },
	{ &K_a_right,	"right-arrow",	META  |         UTC_RIGHT,	0, 0 },
	{ &K_sa_right,	"right-arrow",	SHIFT | META  | UTC_RIGHT,	0, 0 },
	{ &K_c_right,	"right-arrow",	CNTRL |         UTC_RIGHT,	0, 0 },
	{ &K_sc_right,	"right-arrow",	SHIFT | CNTRL | UTC_RIGHT,	0, 0 },
	{ &K_ac_right,	"right-arrow",	META  | CNTRL | UTC_RIGHT,	0, 0 },

	{ &K_s_pgup,	"page-up",	SHIFT |         UTC_PGUP,	0, 0 },
	{ &K_a_pgup,	"page-up",	META  |         UTC_PGUP,	0, 0 },
	{ &K_sa_pgup,	"page-up",	SHIFT | META  | UTC_PGUP,	0, 0 },
	{ &K_c_pgup,	"page-up",	CNTRL |         UTC_PGUP,	0, 0 },
	{ &K_sc_pgup,	"page-up",	SHIFT | CNTRL | UTC_PGUP,	0, 0 },
	{ &K_ac_pgup,	"page-up",	META  | CNTRL | UTC_PGUP,	0, 0 },

	{ &K_s_pgdown,	"page-down",	SHIFT |         UTC_PGDOWN,	0, 0 },
	{ &K_a_pgdown,	"page-down",	META  |         UTC_PGDOWN,	0, 0 },
	{ &K_sa_pgdown,	"page-down",	SHIFT | META  | UTC_PGDOWN,	0, 0 },
	{ &K_c_pgdown,	"page-down",	CNTRL |         UTC_PGDOWN,	0, 0 },
	{ &K_sc_pgdown,	"page-down",	SHIFT | CNTRL | UTC_PGDOWN,	0, 0 },
	{ &K_ac_pgdown,	"page-down",	META  | CNTRL | UTC_PGDOWN,	0, 0 },

	{ &K_s_home,	"home",		SHIFT |         UTC_HOME,	0, 0 },
	{ &K_a_home,	"home",		META  |         UTC_HOME,	0, 0 },
	{ &K_sa_home,	"home",		SHIFT | META  | UTC_HOME,	0, 0 },
	{ &K_c_home,	"home",		CNTRL |         UTC_HOME,	0, 0 },
	{ &K_sc_home,	"home",		SHIFT | CNTRL | UTC_HOME,	0, 0 },
	{ &K_ac_home,	"home",		META  | CNTRL | UTC_HOME,	0, 0 },

	{ &K_s_end,	"end",		SHIFT |         UTC_END,	0, 0 },
	{ &K_a_end,	"end",		META  |         UTC_END,	0, 0 },
	{ &K_sa_end,	"end",		SHIFT | META  | UTC_END,	0, 0 },
	{ &K_c_end,	"end",		CNTRL |         UTC_END,	0, 0 },
	{ &K_sc_end,	"end",		SHIFT | CNTRL | UTC_END,	0, 0 },
	{ &K_ac_end,	"end",		META  | CNTRL | UTC_END,	0, 0 },

	{ &K_s_delete,	"delete",	SHIFT |         UTC_DEL,	0, 0 },
	{ &K_a_delete,	"delete",	META  |         UTC_DEL,	0, 0 },
	{ &K_sa_delete,	"delete",	SHIFT | META  | UTC_DEL,	0, 0 },
	{ &K_c_delete,	"delete",	CNTRL |         UTC_DEL,	0, 0 },
	{ &K_sc_delete,	"delete",	SHIFT | CNTRL | UTC_DEL,	0, 0 },
	{ &K_ac_delete,	"delete",	META  | CNTRL | UTC_DEL,	0, 0 },

	{ &K_s_insert,	"insert",	SHIFT |         UTC_INS,	0, 0 },
	{ &K_a_insert,	"insert",	META  |         UTC_INS,	0, 0 },
	{ &K_sa_insert,	"insert",	SHIFT | META  | UTC_INS,	0, 0 },
	{ &K_c_insert,	"insert",	CNTRL |         UTC_INS,	0, 0 },
	{ &K_sc_insert,	"insert",	SHIFT | CNTRL | UTC_INS,	0, 0 },
	{ &K_ac_insert,	"insert",	META  | CNTRL | UTC_INS,	0, 0 },

	/* The main purpose of these is so that they can either
	 * be bound to nop & not cause problems, or otherwise ignored.
	 */
	{ &K_kp_ent,	"keypad-enter",			UTC_KP_ENT,	0, 0 },	// Keypad Enter
	{ &K_kp_a1,	"keypad-up-left",		UTC_KP_UL,	0, 0 },	// Keypad upper left
	{ &K_kp_a3,	"keypad-up-right",		UTC_KP_UR,	0, 0 },	// Keypad upper right
	{ &K_kp_b2,	"keypad-center",		UTC_KP_CEN,	0, 0 },	// Keypad center
	{ &K_kp_c1,	"keypad-low-left",		UTC_KP_LL,	0, 0 },	// Keypad lower left
	{ &K_kp_c3,	"keypad-low-right",		UTC_KP_LR,	0, 0 },	// Keypad lower right

#endif	// NCURSES_VERSION

	{ NULL,		NULL,				0,		0, 0 }
};


/* Something to help extract booleans easily */
typedef	struct	{
	BYTE*	tc_bool;		// Boolean value
#if	CURSES
	char*	tc_name;		// Terminfo name
#else
	char	tc_name[3];		// Termcap name
#endif
} tcap_bool;

static	BYTE	MI;			// Safe to move while in insert mode
static	BYTE	MS;			// Safe to move while in standout mode
static	BYTE	EO;			// Can erase over-strikes with a blank
static	BYTE	UT;			// Screen erased with background color
static	BYTE	XT;			// Tabs are destructive
static	BYTE	AM;			// Output char at RH edge moves cursor to start of next line
static	BYTE	IN;			// Insert NULL - we can't cope/do insert if this is set

/* mi we do not use - yet
 *
 * I don't know how accurate some of this is, eg
 */
static	tcap_bool	tcap_bools[] = {
	GetTent(&MI,		"mi",	"mir"	),	// move_insert_mode - Safe to move while in insert mode
	GetTent(&MS,		"ms",	"msgr"	),	// move_standout_mode - Safe to move while in standout mode
	GetTent(&EO,		"eo",	"eo"	),	// erase_overstrike - Can erase over-strikes with a blank
	GetTent(&UT,		"ut",	"bce"	),	// back_color_erase - Screen erased with background color
	GetTent(&XT,		"xt",	"xt"	),	// dest_tabs_magic_smso - Tabs are destructive
	GetTent(&AM,		"am",	"am"	),	// auto_right_margin - Output char at RH edge moves cursor to start of next line
	GetTent(&IN,		"in",	"in"	),	// insert_null_glitch - difference between empty space created by ' ' or moving over with cursor movement
	GetTent(NULL,		NULL,	NULL	)
};



TERM term = {
	0,	/* these four values are set dynamically at open time 	 */
	0,	/* they are each 0 and not NULL, because on some systems */
	0,	/* NULL is defined in terms of pointers, which gets the	 */
	0,	/* compiler somewhat confused				 */
	MARGIN,
	SCRSIZ,
	NPAUSE,
	tcapopen,
	ttclose,
	tcapkopen,
	tcapkclose,
	ttgetc,
	ttputc,
	ttflush,
	tcapmove,
	tcapeol,
	tcapeop,
	tcapbeep,
	tcaprev,
	tcapul,
	tcapcres,
	tcapfcol,
	tcapbcol,
	tkeyname,
	tDoesWhat,
	tcapscroll,
	tspecial,
	ttungetc,
	tcfixMargin,
	tcapInsert,
	tcapDelete,
	tcapEnterInsert,
	tcapExitInsert,
};

static	ST_val	eopc = -1;	// # chars out when last moved to screen origin


/* Read the termcap/terminfo database.
 *
 * If CheckSufficient check that the terminal provides a minimum that is needed
 * to support Emacs.
 */
void	termcapRead(int CheckSufficient)
{
	char*		t;
	char*		LacksWhat;	/* Termcap entry lacks this */
	tcap_strs*	tcpp;
	tcap_bool*	tbpp;
	struct TkeyType*tkp;
	struct TkeyType*tkp2;
	char		ch;
	char*		DbType;		// termcap or terminfo

	if(TermName[0] == '\0')
		Die("Environment variable TERM empty or not defined");

	TERMOUTTRACE((TermDebFD, "Debugging for terminal: %s\n", TermName))

#if	CURSES
	DbType = "terminfo";

#ifdef	NCURSES_VERSION
	/* Get user-defined capabilities - these include things like ^left-arrow
	 * I think that this *has* to be called before getting entries.
	 * http://invisible-island.net/ncurses/man/terminfo.5.html
	 */
	TERMOUTTRACE((TermDebFD, "NCURSES_VERSION='%s' curses_version()='%s'\n", NCURSES_VERSION, curses_version()))

	use_extended_names(1);
#endif

	{	/* The big problem with terminfo is how to populate the strings database
		 * without mungeing terminal settings. There is not a nice clean separation.
		 * Giving it a bogus file descriptor to play with is a fudge; works on Linux,
		 * hopefully elsewhere.
		 */
		int	err;

	        if(setupterm(TermName, -1, &err) == ERR)
			// The value in err is not useful, just complain generically
			Die("Terminal \"%s\" is not known or is not capable enough to run Emacs",
				TermName);
	}

	// Get string capabilities - avoid duplicates:
	for(tcpp = tcap_tab; tcpp->tc_str != (char**)NULL; tcpp++)
		if(*tcpp->tc_str == (char*)NULL) {
			// Likewise: there does not seem to be much distinction on -1 or NULL
			// returns of tigetstr(), treat both the same: not present & continue
			// fprintf(stderr, "Terminfo capability \"%s\" is not a string capability\n", tcpp->tc_name);
			char*	sv;
			if((sv = tigetstr(tcpp->tc_name)) != (char*)-1 && sv != NULL)
				*tcpp->tc_str = sv;
		}

	// Get booleans:
	for(tbpp = tcap_bools; tbpp->tc_name != NULL; tbpp++)
		*tbpp->tc_bool = (BYTE)tigetflag(tbpp->tc_name);

#else	// CURSES
    {
	DbType = "termcap";
	char*		p;
	char		tcbuf[1024];

	if((tgetent(tcbuf, TermName)) != 1)
		Die("Unknown terminal type \"%s\"", TermName);

	/* Read all the strings into the buffer tcapbuf.
	 * Deal with synonyms, some extra keys which weren't in the
	 * original termcap definition have been called different things.
	 *
	 * Test tcpp->tc_name as there are some terminfo only ones that we can't get here.
	 */
	p = tcapbuf;
	for(tcpp = tcap_tab; tcpp->tc_str != (char**)NULL; tcpp++)
		if(*tcpp->tc_str == (char*)NULL && tcpp->tc_name)
			*tcpp->tc_str = tgetstr(tcpp->tc_name, &p);

	// Get booleans:
	for(tbpp = tcap_bools; tbpp->tc_name != NULL; tbpp++)
		*tbpp->tc_bool = (BYTE)tgetflag(tbpp->tc_name);

	/* Did it fit ? */
	if(p >= &tcapbuf[TCAPSLEN])
		Die("%s termcap longer than %d characters",
						TermName, TCAPSLEN);
    }
#endif	// CURSES

	if(PC_t)		/* Pad character defined ? */
		PCh = *PC_t;	/* String -> character */

	/* Check that we have a minumum capability.
	 * li & co are no longer mandatory, they will get (re)set by GetTermDims()
	 * Set to a sensible default.
	 *
	 * All that we really need is: clear_screen and cursor_address
	 */
	LacksWhat = (char*)NULL;
	if((term.t_nrow = (short)tgetnum("li")) == -1)
		term.t_nrow = 24;
	if((term.t_ncol = (short)tgetnum("co")) == -1)
		term.t_ncol = 80;
	if(CL == (char*) NULL)
		LacksWhat = "cl";
	if(CM == (char*) NULL)
		LacksWhat = "cm";
	Co = tgetnum("Co");		// Max # colours on the screen
	if(CheckSufficient) {
		if(LacksWhat != (char*)NULL)
			Die("%s database for TERM=%s lacks \"%s\" entry",
					DbType, TermName, LacksWhat);

		/* Cost of a typical movement to somewhere in the middleish of the screen.
		 * Wrap up in CheckSufficient in case the term is dumb!
		 */
		CostCM = (int)strlen(tgoto(CM, 15, 20));
	}

	/* init_tabs "it" Tabs every this number of positions
	 * Assume 8 if unset -- do this until we find one where this is not true.
	 * Used to nail this as 8 - will prob never see one where it is different.
	 */
	TabWidth = tgetnum("it");
	if(TabWidth == -1)
		TabWidth = 8;

	/* Cost of down-one, left-one, also cost of carriage-return (move to column 0).
	 * CmPricey means that CM costs more than DO & LE.
	 * If the capabilities are missing - pretend that they are really costly,
	 * they are initialised to 999.
	 */
	if(LE1)
		CostLE1 = (int)strlen(LE1);

	if(LE)	// Assume nominal 5 places
		CostLE = (int)strlen(tparm(LE, 5));

	if(DO1)
		CostDO1 = (int)strlen(DO1);

	if(DO)
		CostDO = (int)strlen(tparm(DO, 5));

	if(ND1)
		CostND1 = (int)strlen(ND1);

	if(ND)
		CostND = (int)strlen(tparm(ND, 5));

	if(UP1)
		CostUP1 = (int)strlen(UP1);

	if(UPsl)
		CostUP = (int)strlen(tparm(UPsl, 5));

	if(HT)
		CostHT = (int)strlen(HT);

	if(BT)
		CostBT = (int)strlen(BT);

	CmPricey = (CostDO1 + CostLE) < CostCM;

	if(CR)
		CostCR = (int)strlen(CR);

	if(HO)
		CostHO = (int)strlen(HO);

	if( !US || !UE)	// Need both, use of one will just screw
		US = UE = (char*) NULL;

	/* SG added for tvi955 ajc */
	SG = tgetnum("sg");		// # blank chars left by smso or rmso

	// We can't cope if SO takes space on the screen - unlikely today
	if(SO && SG < 1 && SE)
		revexist = TRUE;
	else
		SE = SO = (char*) NULL;
	
	/* gnome terminfo entry has the same string to end standout & underline.
	 * In fact it resets a bunch of things, but the interaction of SO & SE
	 * is what concerns us.
	 */
	gnomeSeUeBug = UE && SE && !strcmp(UE, SE);

	if (CE == (char*) NULL)		/* will we be able to use clear to EOL? */
		eolexist = FALSE;

	// 2 ways of scrolling. Test if we can do each of them
	CanIlDl = (AL || al) && (DL || dl);
	CanScroll = cs && (sf || SF) && (sr || SR);

	// If insert_null_glitch don't do character insert:
	if(IN)
		dc = DC = ic = IC = im = (char*)NULL;

	// Can we do character insert, one way of the other:
	CanInsC1 = dc && ic;
	CanInsCN = (DC && IC) || (im && ei);
	CanInsC = CanInsCN || CanInsC1;

	/* Cost of line left/right scroll (character insert/delete) - hard to estimate as there are different terminal capabilities:
	 * These are global variables as used directly in display.c
	 * Movement along the line is pessimistic: assume > 9 places, also ignore
	 * optimisations that will be done like sending CR.
	 */
	if(DC) {
		TTCostDel = (int)strlen(tparm(DC, 5));		// Cost of delete several characters

		TTCostDel2Move = TTCostDel + 2 * CostLE;	// Cost of term delete & 2 moves along the line
	}

	// Cost of insertion is more difficult as 2 ways of inserting
	if(CanInsC) {
		// Cost of insert several characters
		if(im && ei)
			TTCostIns = (int)strlen(im) + (int)strlen(ei);
		if(IC)
			TTCostIns = (int)strlen(tparm(IC, 5));

		// Cost of term insert (or mode in/out) & 2 moves along the line
		TTCostIns2Move = TTCostIns + 2 * CostLE;
	}


#if	TERMDEBUG
	if(TermDebFD) {
		fprintf(TermDebFD, "CostLE=%d\nCostDO1=%d\nCostND=%d\nCostUP1=%d\nCostUP=%d\nCostCM=%d\nCostCR=%d\nCostHT=%d\nCostBT=%d\nCostHO=%d\nCmPricey=%d\n",
			CostLE, CostDO1, CostND, CostUP1, CostUP, CostCM, CostCR, CostHT, CostBT, CostHO, CmPricey);
		fprintf(TermDebFD, "revexist=%d\neolexist=%d\ngnomeSeUeBug=%d\nSG=%d\nCo=%d\nCanIlDl=%d\nCanScroll=%d\n",
			revexist, eolexist, gnomeSeUeBug, SG, Co, CanIlDl, CanScroll);
	}
#endif

	/* Colours.
	 * What we get back is a string that needs to be scanned/interpretted to
	 * effect a change. Do all of that now
	 * I assume a paint brush model for colours, not a start/end region model.
	 */
	if(Co > 0) {
		int	c;
		int	use_c = 0;	// Keep compiler from wingeing
		char*	s;

		// Test
		// AF = AB = NULL;

#if	TERMDEBUG
		if(TermDebFD)
			fprintf(TermDebFD, "Colour escape trace.\nAF='%s'\nSf='%s'\nAB='%s'\nSb='%s'\n",
				(AF ? AF : "null"),
				(Sf ? Sf : "null"),
				(AB ? AB : "null"),
				(Sb ? Sb : "null")
				);
#endif

		for(c = 0; c <= 7; c++) {
			// Try to find a string that we can interpret:
			s = NULL;

			if(AF) {
				s = AF;
				use_c = c;
			} else {
				if(Sf) {
					s = Sf;
					use_c = mapCols[c];
				}
			}

			/* Something that we can use ? */
			if(s)
				ColChange_fg[c] = strmake_t(tparm(s, use_c));

			// Same again for background
			s = NULL;
			if(AB) {
				s = AB;
				use_c = c;
			} else {
				if(Sb) {
					s = Sb;
					use_c = mapCols[c];
				}
			}

			/* Something that we can use ? */
			if(s)
				ColChange_bg[c] = strmake_t(tparm(s, use_c));

			TERMOUTTRACE((TermDebFD, "Colour %d fg='%s' bg='%s'\n", c, ColChange_fg[c], ColChange_bg[c]))
		}
	}

	/* Cost of attribute change.
	 * A kind of average of: standout, end standout, original pair, colour 0.
	 * Guess that these are the most commonly used.
	 * This is not right as the average is not the cost in a particular instance or
	 * frequency of use. Too hard to do anything else.
	 */
	int	i = 0;
	if(SO) {
		TTCostAttChange += (int)strlen(SO);
		i++;
	}
	if(SE) {
		TTCostAttChange += (int)strlen(SE);
		i++;
	}
	if(OP) {
		TTCostAttChange += (int)strlen(OP);
		i++;
	}
	if(*ColChange_fg[0]) {
		TTCostAttChange += (int)strlen(ColChange_fg[0]);
		i++;
	}
	if(i != 0)
		TTCostAttChange /= i;

	TERMOUTTRACE((TermDebFD, "TTCostAttChange=%d\nTTCostIns=%d\nTTCostDel=%d\nTTCostDel2Move=%d\nTTCostIns2Move=%d\n", TTCostAttChange, TTCostIns, TTCostDel, TTCostDel2Move, TTCostIns2Move))
	

	/* Now the TermToks array of structures. The "ok" flag is set to
	 * 0 (false) if any of the sequences does not start with a control
	 * character, ie "ok" is left as true if all of the sequences have
	 * a control lead-in.
	 *
	 * Compute the input string length, and eliminate elements for which
	 * there is no termcap string - makes for faster scanning later on.
	 */
	for(ControlLeadIn = 1, tkp2 = tkp = TermToks;
			(tkp2->tk_seq = tkp->tk_seq) != (char**)NULL; tkp++) {
		if((t = *(tkp2->tk_seq = tkp->tk_seq)) == (char*)NULL)
			continue;

		// Copy over entry. If no NULL entries so far: copy current to itself
		tkp2->tk_name = tkp->tk_name;
		tkp2->tk_token = tkp->tk_token;
		tkp2->tk_legacyName = tkp->tk_legacyName;
		tkp2->tk_len = (char)strlen(t);

		/* Look at the first character:
		 * * are they all the same (probably escape) ?
		 * * are they all a control character ?
		 */
		ch = *t;
		if(KeyByte1 == 0) {
			KeyByte1 = ch;		// We know what it is
			KeyByte1Same = 1;	// They might all be the same
		}

		if(KeyByte1 != ch)
			KeyByte1Same = 0;

		if( !isascii(ch) || !iscntrl(ch))
			ControlLeadIn = 0;

		tkp2++;
	}

	/* This may be needed for debugging */
#if	TERMDEBUG
	if(TermDebFD) {
		for(tcpp = tcap_tab; tcpp->tc_str; tcpp++)
			fprintf(TermDebFD, "%s -> '%s'\n", tcpp->tc_name,
				*tcpp->tc_str ? *tcpp->tc_str : "(null)");

		for(tbpp = tcap_bools; tbpp->tc_name != NULL; tbpp++)
			fprintf(TermDebFD, "%s -> %d\n", tbpp->tc_name, *tbpp->tc_bool);
	}
#endif

}

/* Read the termcap entry and open the terminal.
 *
 * This should really be void as there is not a meaningful value that is returned,
 * too much effort to fix for no real gain.
 */
int	tcapopen(void)
{
	/* This stops us re-reading /etc/termcap every time we shell out
	 * or suspend.
	 */
	if(TTopened == TTwasOpened)
		return(ttopen());

	termcapRead(1);
	
	return(ttopen());
}

/* Open the keyboard.
 * Send the string to set the keyboard into the right mode.
 * With some strange terminals (eg Xterm) this changes the escape sequence
 * that the function keys return !!
 * ksOnly is used to just send this key sequence - so that we recognise
 * what they arrow keys, etc, send.
 *
 * This is needed by shell out (respawn()) so that the likes of up-arrow
 * can by typed after a command has run when a full restart would also
 * clear the screen (or something) and make the output of the previous
 * command disappear.
 *
 * This function is misnamed as it can do more than just set the keyboard mode.
 */
void	tcapkopen(int ksOnly)
{
	TERMOUTTRACE((TermDebFD, "tcapkopen(%d)\n", ksOnly))

	if( !ksOnly) {
		// We don't know the state of the screen, so record that
		// and set a suitable default
		nowfcol  = COLOUR_UNSET;
		wantfcol = COLOUR_SYS_DEF;
		nowbcol  = COLOUR_UNSET;
		wantbcol = COLOUR_SYS_DEF;
		nowrev   = FALSE;
		wantrev  = FALSE;
		nowul    = FALSE;
		wantul   = FALSE;

		putpad(TI);	// Terminal Initialise
		putpad(CL);	// Needed as TI sometimes does not do so

		// Cursor visible - sometimes it disappears.
		// Maybe I should use cursor_normal (ve) instead of cursor_visible (vs)
		// which is supposed to make it very visible.
		putpad(VS);

		// So that we know what we are at & reflect the above:
		putpad(SE);
		putpad(UE);

		strcpy(sres, "NORMAL");
	}

	// Make arrow & function keys generate what we recognise:
	putpad(KS);
}

/* Close the keyboard.
 */
void	tcapkclose(void)
{
	TERMOUTTRACE((TermDebFD, "tcapclose()\n"))

	putpad(ME);	// Now that we are fiddling with colours - reset
	putpad(KE);
	putpad(TE);
	TTflush();
}

/* Move the cursor to row, col, set ttrow & ttcol.
 * ttputs_noNA() to avoid setting paintbrush just for a move, not really needed.
 * By NOT flushing any pending attributes before we do this; this can save
 * quite a few bytes of output, the higher level functions can set/unset attributes
 * and then do the reverse a little later not having put any display characters to
 * the screen..
 *
 * This tries several different ways of moving to (row, col): relative movement
 * using up/down-line, forw/back-tab, forw/back-char, home; cursor movement
 * (CM) if all else fails.
 *
 * If moving to origin [start of the screen (0, 0)], note the # bytes output.
 * This might be used in clear-screen to force colours to be reset.
 * # bytes is a cheap way of detecting if anything has happened since to move to the origin.
 *
 * Might possibly need to consider the testing of msgr capability - but I doubt that
 * there are any terminals that stupid still around.
 *
 * See comments display.c: movecursor() of problems moving when the cursor is on
 * a double width character. The fix taken is to prevent most of the optimisations
 * below and use a full CM. By setting ttcol to a large number use of CR will still be done.
 */
void	tcapmove(int row, int col)
{
	// Note debug line started here and finished below:
	TERMOUTTRACE((TermDebFD, "\ntcapmove(%d, %d) at (%d, %d", row, col, ttrow, ttcol))

	/* Writing to the last character position in a line can result in
	 * the column being undefined - or at least unsure.
	 * If it goes there force a full cursor movement.
	 * This fixes occasional screen corruption.
	 * Maybe AM should be checked.
	 */
	if(ttcol >= AllocColumns)
		ttcol = TERM_HUGE;

	// Relative Vertical & Horisontal distance:
	int	relV = row - ttrow;	// Ensure +ve numbers
	int	relH = col - ttcol;
	int	costH;			// Cost of horizontal movement
	int	costV;			// Cost of vertical movement
	int	useCR = 0;		// Use CR for horizontal movement to get to col 0
	int	useHO;			// Use HO to get to 0,0

	TERMOUTTRACE((TermDebFD, " [%d]) relV=%d relH=%d\n", ttcol, relV, relH))

#if	TERMDEBUG
	/* Understanding generated escape sequences can be hard.
	 * Make it a bit easier by keeping cursor movement simple.
	 */
	if(TermDebug > 1) {
		ttputs_noNA(tgoto(CM, col, row));
	} else {
#endif
	// Quick return if we are already where we need to be:
	if(relV == 0 && relH == 0)
		return;

	// Not safe to move in insert mode ?
	if(ttInInsert && !MI)
		tcapExitInsert();

	/* Safe to move in standout mode ?
	 * If not, unset - but leave wantrev as desired.
	 */
	if(nowrev && !MS) {
		ttputs_noNA(SE);
		nowrev = 0;
	}

	/* A common case is that movement will be by a small amount - ie to somewhere close.
	 * Another is that it will be to the start of a line: same one, or up/down close.
	 *
	 * Try to work out the cost of getting there using relative movement rather
	 * than a CM. The horizontal movement relative to where the cursor is.
	 * If that is too much try moving to the start of line and then relative right.
	 *
	 * Horizontal cost: could be cheaper to start with a CR.
	 *
	 * There is not a problem if DO/... is not defined as the corresponding CostXX will
	 * be high - so it won't be used. But: we must test CR & ND, consider
	 * col == 0 and/or relH > col.
	 *
	 * Remember: if CR is not available then CostCR will be very large.
	 *
	 * The relative movement sequences come in 2 forms: move1 and move-many.
	 * Beware: the cost of something like 'up' can be very different from the cost of 'down'.
	 */
	costV = (relV >= 0) ?
			( relV * CostDO1 > CostDO ? CostDO :  relV * CostDO1) :
			(-relV * CostUP1 > CostUP ? CostUP : -relV * CostUP1);
	costH = (relH >= 0) ?
			( relH * CostND1 > CostND ? CostND :  relH * CostND1) :
			(-relH * CostLE1 > CostLE ? CostLE : -relH * CostLE1);

	// Don't know where we are ? Force CM.
	if(ttrow >= TERM_HUGE - 10 || ttcol >= TERM_HUGE - 10)	// Test almost TERM_HUGE
		costV = costH = 999;

	/* Going to 0,0 ?
	 * Don't bother trying to be clever & use HO and DO to get to 1,0
	 * this won't happen very often.
	 */
	useHO = HO && row == 0 && col == 0;

	/* Need to move left/right.
	 * See if we can do horizontal movement by tab/back-tab
	 * These are defined as being width TabWidth (prob 8)
	 *
	 * Beware: TERM_HUGE - means that we do not know where the cursor is.
	 */

	int	useTab = 0;	// True if to use tabs
	int	Tabs = 0;	// Tabs forward
	int	BackTabs = 0;	// Tabs backward
	int	NDs = 0;	// # positions forward (after tab)
	int	LEs = 0;	// # positions backward (after tab)

	if( !useHO && !XT && col != ttcol) {
		TERMOUTTRACE3((TermDebFD, "RELH (%d -> %d) costH=%d\n", ttcol, col, costH))
		int	usingCR = 0;		// Cost of using CD - if it is used
		int	colfrom = ttcol;	// Where we move from
		int	bestCost = CostCM;	// Best cost so far

		if(col > ttcol) {
			/* Moving right
			 * See if can get there by use of HT and, maybe ND or LE
			 * Beware: Tb is changed and we might come back with the goto.
			 * Similarly vars like LEs might need resetting.
			 *
			 * When we knew that TabWidth was 8, used to do the nice:
			 * Tb = col & 0xFFF8;
			 */
tryCrForw:		;
			int	Tb = col / TabWidth * TabWidth;	// Position of tab before or at col
			int	tTabs;				// Don't assign to Tabs until we are sure

			tTabs = (Tb + TabWidth - 1 - colfrom) / TabWidth;
			int	tNDs = col - Tb;
			int	Tcost = tTabs * CostHT + (tNDs * CostND1 > CostND ? CostND : tNDs * CostND1) + usingCR;
			LEs = 0;

			if(costH > Tcost && Tcost + costV < bestCost) {
				// Tab and right a bit more
				Tabs = tTabs;
				NDs = tNDs;
				TERMOUTTRACE3((TermDebFD, "TAB usingCR=%d costH=%d Tb=%d Tabs=%d NDs=%d Tcost=%d\n", usingCR, costH, Tb, Tabs, NDs, Tcost))
				costH = Tcost;
				useTab = 1;
				BackTabs = 0;
				LEs = 0;
				useCR = usingCR;
			}
			// Try going beyond and then back with LE
			if((Tb += TabWidth) < AllocColumns) {
				int	tLEs = Tb - col;
				Tcost = (tTabs + 1) * CostHT + tLEs * CostLE1 + usingCR;
				if(costH > Tcost && Tcost + costV < bestCost) {
					Tabs = tTabs + 1;
					LEs = tLEs;
					TERMOUTTRACE3((TermDebFD, "TAB+BACK usingCR=%d costH=%d Tb=%d Tabs=%d LEs=%d Tcost=%d\n", usingCR, costH, Tb, tTabs, LEs, Tcost))
					costH = Tcost;
					useTab = 1;
					BackTabs = 0;
					NDs = 0;
					useCR = usingCR;
				}
			}
		} else {
			/* Moving left
			 * Try back tab and then forw/back positions 
			 */
			int	Tb = col / TabWidth * TabWidth;	// Position of tab before or at col
			int	tBackTabs = (ttcol + TabWidth - 1 - Tb) / TabWidth;
			int	tNDs = col - Tb;
			int	Tcost = tBackTabs * CostBT + tNDs * CostND1;
			if(costH > Tcost && Tcost + costV < bestCost) {
				TERMOUTTRACE3((TermDebFD, "BACKTAB costH=%d Tb=%d tBackTabs=%d NDs=%d Tcost=%d\n", costH, Tb, tBackTabs, NDs, Tcost))
				costH = Tcost;
				BackTabs = tBackTabs;
				NDs = tNDs;
				useTab = 1;
			}
			// Try going less far back with tabs & get there with LE
			if((Tb += TabWidth) < ttcol && tBackTabs > 0) {
				int	tLEs = Tb - col;
				Tcost = (tBackTabs + 1) * CostBT + tLEs * CostLE1;
				if(costH > Tcost && Tcost + costV < bestCost) {
					BackTabs = tBackTabs - 1;
					LEs = tLEs;
					TERMOUTTRACE3((TermDebFD, "BACKTAB+FORW costH=%d Tb=%d BackTabs=%d LEs=%d Tcost=%d\n", costH, Tb, BackTabs, LEs, Tcost))
					costH = Tcost;
					NDs = 0;
					useTab = 1;
				}
			}

			/* What if we move to the start of line & tab forwards ?
			 * Try again with the moving right code.
			 */
			usingCR = CostCR;
			colfrom = 0;
			TERMOUTTRACE3((TermDebFD, "TRY-CR\n"))
			goto tryCrForw;

			// Could also look at move Home & tab forwards, but see comments below.
		}
	}

#if	0
	/* Cost of HO & move from there.
	 * This is removed, not worth it.
	 * * it will not be a frequent occurence - except to (0,0) which is done below
	 * * CostHO is typically not much less than CostCM - so little scope for
	 *   moving from home.
	 */
	int	costHm = CostHO + row * CostDOl + col * CostND;
	if(costHm < costV + costH) {
		relV = row;
		NDs = col;
		// Need to make this less than CostCM - somehow
	}
#endif

	TERMOUTTRACE3((TermDebFD, "costV=%d costH=%d CostCM=%d useCR=%d useTab=%d Tabs=%d BackTabs=%d NDs=%d LEs=%d\n", costV, costH, CostCM, useCR, useTab, Tabs, BackTabs, NDs, LEs))


	if(useHO) {
		// Go to 0,0
		if(ttrow != 0 || ttcol != 0) {
			TERMOUTTRACE3((TermDebFD, "GO HO\n"))
			ttputs_noNA(HO);
		} else {
			TERMOUTTRACE3((TermDebFD, "NO GO HO - already there\n"))
		}
	} else if(costH + costV < CostCM) {
		// Local movement is cheaper than CM - Cursor Movement

		// Do vertical movement first
		char*	V1 = DO1;		// String for vertical change, one line
		char*	Vm = DO;		// String for multiple line vertical change
		int	jmp = relV * CostDO1 > CostDO;

		TERMOUTTRACE3((TermDebFD, "MOVELOCAL relV=%d jmp=%d\n", relV, jmp))

		if(relV < 0) {
			// Moving up
			relV = -relV;
			V1 = UP1;
			Vm = UPsl;
			jmp = relV * CostUP1 > CostUP;
			TERMOUTTRACE3((TermDebFD, "MOVELOCAL-UP relV=%d jmp=%d\n", relV, jmp))
		}

		if(jmp)	// jmp tells us if one multiple-lines sequence is cheaper than multiple one line changes
			ttputs_noNA(tparm(Vm, relV));
		else
			while(--relV >= 0)
				ttputs_noNA(V1);

		// Relative horizontal

		// Maybe move to start of line first and then move right
		if(useCR)
			ttputs_noNA(CR);

		// Not using tabs to get close, then all the movement is left/right:
		if( !useTab) {
			if(relH < 0)
				LEs = -relH;
			else
				NDs = relH;
		}

		TERMOUTTRACE3((TermDebFD, "MOVEHORIZ useCR=%d Tabs=%d BackTabs=%d NDs=%d LEs=%d\n", useCR, Tabs, BackTabs, NDs, LEs))

		// Use forward/back tab to get close, then forward/back to get there

		// Put forward/backward tab movements
		while(--Tabs >= 0)
			ttputs_noNA(HT);

		while(--BackTabs >= 0)
			ttputs_noNA(BT);

		// Send what is shorter: multiple move-1 or move-n

		// Move rights
		if(NDs * CostND1 > CostND)
			ttputs_noNA(tparm(ND, NDs));
		else
			while(--NDs >= 0)
				ttputs_noNA(ND1);

		// Move lefts
		if(LEs * CostLE1 > CostLE)
			ttputs_noNA(tparm(LE, LEs));
		else
			while(--LEs >= 0)
				ttputs_noNA(LE1);
	} else {
		// Non local cursor movement
		TERMOUTTRACE3((TermDebFD, "NONLOCAL row=%d col=%d\n", row, col))

		ttputs_noNA(tgoto(CM, col, row));
  	}
#if	TERMDEBUG
	}
#endif	

	// Help with clearing to end of page. See tcapeop()
	if(col == 0 && row == 0)
		eopc = ST_termOut;

	ttrow = row;
	ttcol = col;
}

/* Clear to end of line
 */
void	tcapeol(void)
{
#if	0	// We ought to do this:

	/* Zap global variable: eolexist
	 * This ought to be done entirely within the terminal emulation
	 */
	if (CE == (char*) NULL) {	/* will we be able to use clear to EOL? */

		int	i;
		int	col = ttcol;

		TTforg(gfcolor);
		pBrush = pens[PAL_MODE * INKS + INK_PLAIN];

		/* Beware wrapping, especially on the last line - "am" might help
		 * term.t_ncol might have made more sense than AllocColumns, but
		 * stay consistent.
		 */
		for(i = 0; i < AllocColumns - 1; i++) {
			TTputc(' ');
			ttcol++;
		}

		movecursor(ttrow, col);
	}
	else
#endif

	TERMOUTTRACE((TermDebFD, "CLEAR at (%d, %d)\n", ttrow, ttcol))

	putpad(CE);
}

/* Clear to end of page
 *
 * Terminfo defintion of CL also moves the cursor to 0,0.
 */
void	tcapeop(void)
{
	TERMOUTTRACE((TermDebFD, "tcapeop() at (%d, %d)\n", ttrow, ttcol))

	// If clearing at screen origin request colour reset since they
	// might have been lost.
	if(eopc == ST_termOut) {
		putpad(CL);

		nowfcol  = COLOUR_SYS_DEF;
		nowbcol  = COLOUR_SYS_DEF;
		AttribSet();
	} else {
		// Not at origin - just clear
		putpad(CL);
	}

	// Move cursor to 0,0:
	ttrow = 0;
	ttcol = 0;
}

/* The attributes are wrong, set them.
 * Attributes: for/back colours, reverse, underline
 */
void	AttribSet(void)
{
	/* This will soon be true. Do it now to avoid infinite recusion
	 * via fflush(), ttputs(), ttputc().
	 */
	needNewAttrib = 0;

	// If either of the FG/BG colours want setting to none (== original)
	// do so, but remember that it sets both.
	if((nowfcol != wantfcol && wantfcol == COLOUR_SYS_DEF) ||
	   (nowbcol != wantbcol && wantbcol == COLOUR_SYS_DEF)) {
		// If we don't have OP we can't do this - no fall back
		ttputs_noNA(OP);
		nowfcol = nowbcol = COLOUR_SYS_DEF;
	}

	/* SE & UE are the same. gnome terminfo bug */
	if(gnomeSeUeBug && ((nowrev && !wantrev) || (nowul && !wantul))) {
		// Switch them both off, let the code below
		// sort out what should be on.
		ttputs_noNA(SE);
		nowrev = nowul = 0;
	}

	// Reverse video
	// With both reverse & underline there can be an interaction with
	// colour, so if we change either force a reset of the colour.
	// This does nothing if colours have not been changed from default.
	if(nowrev != wantrev) {
		ttputs_noNA(wantrev ? SO : SE);

		nowrev = wantrev;

		nowfcol = nowbcol = COLOUR_UNSET;
	}

	// Underline
	if(nowul != wantul) {
		ttputs_noNA(wantul ? US : UE);

		nowul = wantul;

		nowfcol = nowbcol = COLOUR_UNSET;
	}

	// Set foreground colour
	if(nowfcol != wantfcol && wantfcol != COLOUR_SYS_DEF) {
		ttputs_noNA(ColChange_fg[wantfcol]);
		nowbcol = COLOUR_UNSET;
	}
	nowfcol = wantfcol;

	// Set background colour
	if(nowbcol != wantbcol && wantbcol != COLOUR_SYS_DEF)
		ttputs_noNA(ColChange_bg[wantbcol]);

	nowbcol = wantbcol;
}

/* Change reverse video status.
 * state:	FALSE = normal video, TRUE = reverse video.
 */
void	tcaprev(int state)
{
	wantrev = state;

	if(nowrev != wantrev)
		needNewAttrib = 1;
}

/* Change underline video status.
 * state:	FALSE = normal video, TRUE = underline video.
 *
 * TODO: combin with tcaprev() and use sgr - if available
 */
void	tcapul(int state)
{
	wantul = state;

	if(nowul != wantul)
		needNewAttrib = 1;
}

/* Change screen resolution
 * res, somehow, describes the new resolution.
 * Do nothing.
 */
int	tcapcres(const BYTE* res)
{
	return(TRUE);
}

/* Change palette string */
void	spal(char* dummy)
{
	/*	Does nothing here	*/
}

/* Set foreground colour
 * fc is a number COLOUR_BLACK..COLOUR_WHITE ie 0..7
 * In addition it can be COLOUR_SYS_DEF to indicate the default colour.
 *
 * Only emit something if a colour change is requested.
 * All the work is done in AttribSet().
 *
 * Update wantfcol, compare to nowfcol
 */
void	tcapfcol(int fc)
{
	wantfcol = fc;

	if(wantfcol != nowfcol)
		needNewAttrib = 1;
}

/* Set background colour
 * See tcapfcol() comments.
 *
 * Update wantbcol
 */
void	tcapbcol(int bc)
{
	wantbcol = bc;

	if(wantbcol != nowbcol)
		needNewAttrib = 1;
}

/* Ring the terminal bell.
 * Visible bell is to be preferred, unless the environment variable '$noflash' is true.
 * If the variable '$nobell' is true - don't ring the bell, keep quiet.
 */
void	tcapbeep(void)
{
	if(VB && !noflash)
		tputs(VB, 1, ttputc);
	else
		if( !nobell) {
			if(BL)
				tputs(BL, 1, ttputc);
			else
				ttputc(BEL);
		}

	TTflush();
}

/* Output a string like KS, US, ...
 * Since it is possible that it was not found in termcap: ignore if the passed string is NULL
 */
void	putpad(const char* str)
{
	if(str)
		tputs(str, 1, ttputc);
}

void	putnpad(const char* str, int n)
{
	tputs(str, n, ttputc);
}

#if	FLABEL
/* Label a function key
 *
 * The count is ignored.
 *
 * Command:  label-function-key
 * Bound to: not with termcap
 *
 * Return TRUE
 */
int	fnclabel(int f, Cnt n)
{
	/* on machines with no function keys...don't bother */
	return(TRUE);
}
#endif

/* Look for a token/function-key in FunBuf which contains FunCnt characters.
 * Return values:
 * 0	Not a token or start of one.
 * >0	Found a token which is put into *Token, the return value is the
 *	number of characters 'consumed'.
 * -1	Possible start of a token, ie more wanted.
 *
 * It is assumed that FunCnt is > 0.
 *
 * This used to be able to assume that it was called with FunCnt getting ever larger
 * until there was a match. But we can now also look for sequences - UTF8 ones.
 */
int	TcapFun(const char* FunBuf, int FunCnt, KeyCode* Token)
{
	struct	TkeyType*	tkp;		/* Token struct ptr */
	int			matching = 0;	/* Get a partial match ? */

	// If reading from a pipe or file in a macro started from the command line:
	if(TTopened == TTneverOpened)
		return(0);

	/* It is likely that all of these special keys generate sequences that
	 * start with the same byte, probably escape.
	 * If that is true - test for this and quit early.
	 * The next test down is almost the same thing, but means that things like \r
	 * cause a TermToks scan.
	 */
	if(KeyByte1Same && KeyByte1 != FunBuf[0])
		return(0);

	/* If this is not a control character and all of the 
	 * sequences have a control lead in then return
	 */
	if(ControlLeadIn && ( !isascii((FunBuf[0])) || !iscntrl(FunBuf[0])))
		return(0);

	/* Compare as much as we have to try to get a match.
	 * If we get a partial match - note so that we ask for more.
	 */
	for(tkp = TermToks; tkp->tk_seq != (char**)NULL; tkp++) {
		// Compare the shorter of what we have and how long the escape sequence is:
		size_t	try = (size_t)(FunCnt > tkp->tk_len ? tkp->tk_len : FunCnt);

		if(memcmp(FunBuf, *tkp->tk_seq, try) == 0) {
			// Match not long enough ?
			if(tkp->tk_len > FunCnt) {
				matching = 1;
				continue;
			}

			/* Got a match - set the token & return how many characters
			 * we matched.
			 */
			*Token = tkp->tk_token;
			return(tkp->tk_len);
		}
	}

	/* Match, return back asking for more */
	if(matching)
		return(-1);

	return(0);	// No match
}

/* Return a name for what might be a special key on the keyboard.
 * If not known return NULL
 */
char*	tkeyname(KeyCode c)
{
	struct	TkeyType*	ttp;

	for(ttp = TermToks; ttp->tk_seq; ttp++)
		if(ttp->tk_token == c)
			return(ttp->tk_name);

	return NULL;
}

/* Inspect seq for the name of a termcap character,
 * return the SPEC value for it.
 * Return -1 if it is not known.
 * seq is a NUL terminated string.
 */
KeyCode	tspecial(const BYTE* seq)
{
	struct  TkeyType* ttp;

	// Is it a legacy name like: FNU
	if(seq[0] == 'F' && seq[1] == 'N' && Isletter(seq[2])) {
		for(ttp = TermToks; ttp->tk_legacyName ; ttp++)
			if(ttp->tk_legacyName == seq[2])
				return(ttp->tk_token);
		return(-1);
	}

	// Kludge, won't be found otherwise:
	if( !strcmp("FN0", seq))
		return(UTC_FN_KEY(10));

	// Search the others
	for(ttp = TermToks; ttp->tk_seq ; ttp++)
		if( !strcmp(ttp->tk_name, seq))
			return(ttp->tk_token);

	return(-1);
}

/* Return true if the terminal can do something or
 * does something in some situation.
 * The argument is a TT_DOES_????
 */
int	tDoesWhat(int what)
{
	switch(what) {
	    case TT_DOES_BLANK_UL:
		/* Will the terminal show spaces as underlined.
		 * This seems to be a reasonable indicator - but I suspect
		 * that I am abusing the real intention.
		 */
		return( ! EO);

	    case TT_DOES_SCROLL:
		/* Will the terminal scroll lines by one means or the other ?
		 */
		return(CanIlDl || CanScroll);

	    case TT_DOES_RIGHT_MARGIN:
		/* What happens when a character is output in the right hand column ?
		 * This is the AM boolean
		 * This returns TRUE if it leaves the cursor there - ie on top of the character
		 * just sent, not after it -- on the next line.
		 * I don't know what happens if a double width character is output
		 * it that column.
		 * The point of this was to decide if the last/command line should be
		 * one character shorter - reverse that - always have the last line
		 * shorter - quite what cursor movement happens there is not really defined.
		 * See TTrightMargin()
		 */
		return( ! AM);

	    case TT_DOES_INSERT_CHARS:
		/* Will the terminal do hardware character insert and delete ?
		 * Either by inserting some space at the cursor or entering insert mode.
		 */
		return(CanInsC);

	}

	return(0);	// Keep the compiler happy
}

/* Fix up ttcol (& maybe ttrow) when a character has just been
 * generated (output) at the right hand margin.
 *
 * I did look at AM, if it was set and ttcol is just one beyond the right margin
 * then set it to the right margin. But my understanding of AM was the wrong way
 * round and also even if the cursor does not move to the next line it is not defined
 * what a relative movement does: so the only safe thing to do when the right edge
 * is reached is to assume that the cursor position is undefined and force a
 * complete/absolute cursor movement.
 *
 * If it is more than that (maybe output double width character)
 * then we don't know where the cursor is, so set it invalid/unknown.
 */
void	tcfixMargin(void)
{
	TERMOUTTRACE((TermDebFD, "FixMargin was at row=%d col=%d AllocColumns=%d AM=%d\n", ttrow, ttcol, AllocColumns, AM))

	if(ttcol < AllocColumns)
		return;

// 	// If Automatic Margins - move to start of next line:
// 	if(AM && ttcol == AllocColumns) {
// 		ttcol = 0;
// 		ttrow++;
// 		return;
// 	}

	// Further beyond end of line - safest to assume that we don't know where the cursor is

	// TERMOUTTRACE((TermDebFD, "FixMargin was at row=%d col=%d\n", ttrow, ttcol))

	ttcol = TERM_HUGE;
	ttrow = TERM_HUGE;
}

/* Scroll a some lines up or down, their contents are preserved but are
 * in a different veritical place on the screen.
 * Lines are delted below/above & new lines are created above/below so that
 * lines outside the scroll area are not moved/changed.
 *
 * topLine	the top line that is moved
 *		this line will end up visible in a different place
 * rlines	the number of lines (region) including topLine that are moved
 * moveBy	the number of lines the region is moved down by.
 * 		If negative they are moved up.
 *
 * Thus: 0, 3, 2: deletes 3 & 4; moves: 0->2, 1->3, 2->4, creating 2 blank lines: 0 & 1
 *
 * You should check that the terminal can scroll before calling this.
 *
 * This should be used via TTscroll macro
 *
 *	Terminfo/termcap capabilities for scrolling.
 *
 *	To scroll need one of:
 *	1)	il/il1 & dl/ld1
 *	2)	csr & ind/indn & ri/rin
 *
 *	Prefer (1) to (2)
 *
 *	Terminfo/Termcap names:
 *	il	AL	Insert N lines
 *	il1	al	Insert 1 line
 *	dl	DL	Delete N lines
 *	dl1	dl	Delete 1 line
 *	csr	cs	Sets a scrolling region, top line N1 bottom line N2, leaves cursor position undefined
 *	ind	sf	Scroll up 1 line, cursor must be at bottom of region
 *	indn	SF	Scroll up n lines
 *	ri	sr	Inserts 1 line at 0,0 (top of region) - cursor must be at 0,0
 *	rin	SR	Inserts N lines at 0,0
 *
 *	il & il1 must be at start of line
 *
 *	csr	Defines a scroll region - cursor undefined after csr
 *		Top line X bottom line Y
 *		Then can use rin/ri ind/indn
 *
 *	rin	at 0,0 inserts n lines at the top of the screen
 *		cursor left at 0,0
 *		So: move to bottom line left, delete it/them then rin
 *	ri	As rin but inserts 1 line - use if rin not defined
 *
 *	ind/indn Scroll up when at start of bottom line
 *		Might want to do this when running shell command
 *
 *	ndscr	(documentation sayd ndstr in some places) - Only on Amiga
 *		Scroll does not clear new lines.
 *
 * Remember: top line on screen is line number 0.
 */
void	tcapscroll(int topLine, int rlines, int moveBy)
{
	int	i;

	TERMOUTTRACE((TermDebFD, "tcapscroll(%d, %d, %d) at (%d, %d)\n", topLine, rlines, moveBy, ttrow, ttcol))

	if(CanIlDl) {
		// Can insert/delete line

		TERMOUTTRACE((TermDebFD, "CanIlDl moveBy=%d DL='%s' dl='%s' AL='%s' al='%s'\n", moveBy, (DL?DL:""), (dl?dl:""), (AL?AL:""), (al?al:"")))

		/* Do scrolling by deleting lines in one part of the screen
		 * and then inserting the same number somewhere else.
		 * Delete lines first otherwise we will create too many lines
		 * on the screen and cause it to scroll - or something.
		 *
		 * Scroll up:
		 * * Delete lines at the top
		 * * Insert at the bottom
		 * Scroll down:
		 * * Delete at the bottom
		 * * Insert at the top
		 *
		 * Hopefully can insert/delete multiple lines, if not then do one at a time.
		 */
		if(moveBy < 0) {
			// Up
			moveBy = 0 - moveBy;

			/* Go to the top line that we will delete.
			 * DON'T try to go above line 0 - we could try that if deleting
			 * the top line of the screen.
			 */
			if(topLine - moveBy < 0)
				tcapmove(0, 0);
			else
				tcapmove(topLine - moveBy, 0);

			TERMOUTTRACE((TermDebFD, "Delete %d lines, top=%d\n", moveBy, ttrow))

			if(DL)
				ttputs_noNA(tparm(DL, moveBy));
			else
				for(i = 0; i < moveBy; i++)
					ttputs_noNA(dl);

			// Deleting lines can result in the cursor position being undefined:
			ttrow = TERM_HUGE;

			tcapmove(topLine + rlines - moveBy, 0);
			if(AL)
				ttputs_noNA(tparm(AL, moveBy));
			else
				for(i = 0; i < moveBy; i++)
					ttputs_noNA(al);

			ttrow = TERM_HUGE;	// Force complete CM

			return;
		}

		// Down
		tcapmove(topLine + rlines, 0);

		if(DL)
			ttputs_noNA(tparm(DL, moveBy));
		else
			for(i = 0; i < moveBy; i++)
				ttputs_noNA(dl);

		ttrow = TERM_HUGE;	// Force complete CM

		tcapmove(topLine, 0);

		if(AL)
			ttputs_noNA(tparm(AL, moveBy));
		else
			for(i = 0; i < moveBy; i++)
				ttputs_noNA(al);

		ttrow = TERM_HUGE;	// Force complete CM

		return;
	}

	if(CanScroll) {
		// Can set up scroll region

		TERMOUTTRACE((TermDebFD, "CanScroll moveBy=%d SF='%s' sf='%s' SR='%s' sr='%s'\n", moveBy, (SF?SF:""), (sf?sf:""), (AL?AL:""), (al?al:"")))

		/* Set up a scroll region.
		 * Move to the top/bottom of the region.
		 * Scroll.
		 * Reset scrolling to the whole screen.
		 * Hopefully can scroll multiple lines, if not then do one at a time.
		 */
		if(moveBy < 0) {
			// Up
			moveBy = 0 - moveBy;

			ttputs_noNA(tparm(cs, topLine - moveBy, topLine + rlines - 1));
			ttrow = TERM_HUGE;	// Force complete CM

			tcapmove(topLine + rlines - 1, 0);
			if(SF)
				ttputs_noNA(tparm(SF, moveBy));
			else
				for(i = 0; i < moveBy; i++)
					ttputs_noNA(sf);

		} else {
			// Down
			ttputs_noNA(tparm(cs, topLine, topLine + rlines + moveBy - 1));
			ttrow = TERM_HUGE;	// Force complete CM

			tcapmove(topLine + 0, 0);
			if(SR)
				ttputs_noNA(tparm(SR, moveBy));
			else
				for(i = 0; i < moveBy; i++)
					ttputs_noNA(sr);
		}

		// Reset scroll region to entire terminal:
		ttputs_noNA(tparm(cs, 0, AllocLines));
		ttrow = TERM_HUGE;	// Force complete CM

		return;
	}

	// Should not get here
}

/* Delete places at the current position on the current line.
 * places	The number of screen places to delete, ie
 *		text is moved right.
 *
 * This assumes that the terminal has a delete ability, ie either
 * dc or DC is available.
 */
void	tcapDelete(int places)
{
	TERMOUTTRACE((TermDebFD, "tcapDelete(%d) at %d, %d\n", places, ttrow, ttcol))

	if(places == 1 || !DC)
		while(--places >= 0)
			ttputs_noNA(dc);
	else
		ttputs_noNA(tparm(DC, places));
}

/* Insert places at the current position on the current line or enter insert mode.
 * places	The number of screen places to insert, this means
 *		creating white space at the current position, ie
 *		text is moved left.
 *
 * Prefer to use ic or IC - ie open up the space that we then overwrite
 * with whatever.
 *
 * This assumes that the terminal has a insert ability, ie at least one of
 * ic or IC or im/ei is available.
 *
 * Set ttInInsert TRUE if insert mode was entered, in which case the spaces
 * will not have been put there.
 */
void	tcapInsert(int places)
{
	TERMOUTTRACE((TermDebFD, "tcapInsert(%d) at %d, %d\n", places, ttrow, ttcol))

	if(ic || IC) {
		if((places == 1 || !IC) && ic)
			while(--places >= 0)
				ttputs_noNA(ic);
		else
			ttputs_noNA(tparm(IC, places));
		return;
	}

	tcapEnterInsert();
}

/* Enter Insert mode
 *
 * Set ttInInsert TRUE if we do this.
 * Do not send the sequence if already in insert mode.
 *
 * Beware: Insert mode might get unset by cursor movement.
 */
void	tcapEnterInsert(void)
{
	TERMOUTTRACE((TermDebFD, "tcapEnterInsert (state=%d)\n", ttInInsert))

	if( !ttInInsert && im) {
		ttputs_noNA(im);
		ttInInsert = TRUE;
	}
}

/* Leave Insert mode
 *
 * Unset ttInInsert.
 * Do not send the sequence if not in insert mode.
 *
 * Assumes ei im are set. Unless im we won't be in insert mode.
 */
void	tcapExitInsert(void)
{
	TERMOUTTRACE((TermDebFD, "tcapExitInsert (state=%d)\n", ttInInsert))

	if(ttInInsert)
		ttputs_noNA(ei);

	ttInInsert = FALSE;
}

/* List the keys available to stdout.
 */
void	ListKeys(void)
{
	struct	TkeyType*	tkp;

	termcapRead(0);

	for(tkp = TermToks; tkp->tk_seq != (char**)NULL; tkp++) {
		BYTE	buf[50];
		cmdstr(tkp->tk_token, buf);
		printf("%s\n", buf);
	}
}

#else

void	hello(void)
{
}

#endif	/* TERMCAP */

/* end */
