/* Multibyte characters - definitions and HOWTO
 *
 * ADDW SCCS: @(#)utf8.h	1.31 09/22/20 17:43:21
 *
 * UTF-8 should be written in caps and with a '-' -- according to IANA.
 * Try to stick with this in messages, etc, seen by the user.
 * Can't be done internally since '-' is a minus sign.
 *
 * Text is stored as 8 bit bytes, awareness of multibyte characters is limited to:
 * + reading from files
 * + display to the screen
 * + movement left/right in a line
 *
 * Variable/constant names that relate to multibyte have 'utf8' in them.
 *
 * If a struct BUFFER is b_mode & MDUTF8 then:
 * + it *may* contain UTF-8 characters
 * + any character with the top bit set is part of a valid UTF8 sequence
 * + any new input is checked for UTF-8 correctness
 *
 * LINE.l_utf8 is set to true if the line may contain a multi byte character.
 * The point is that some operations are more expensive in MB mode, so
 * only do it if there is such a character in the line.
 * It is possible that a deletion could remove the last such character,
 * but deletion does not recheck/reset - due to cost.
 * This flag will be set if a MB character is inserted.
 *
 * In UTF8 mode the byte offset of the current character must be the first of
 * a multi byte character.
 *
 * On startup $LANG is checked, if the encoding is something
 * like: en_GB.utf8 then set true the global flag: try_utf8, which is
 * copied to gmode - (MDUTF8).
 * This sets MDUTF8 & checks the file on input. If an invalid MB character
 * is found:
 * + MDUTF8 is cleared
 * + a message warning of the first bad character is emitted, saying where
 *
 * UTF8 was originally defined up to a maximum of 6 bytes.
 * Bits	First		Last		# Bytes	Byte 1		Byte 2		Byte 3		Byte 4		Byte 5		Byte 6
 * 	Code point	Code point
 * 7	U+0000		U+007F		1	0xxxxxxx
 * 11	U+0080		U+07FF		2	110xxxxx	10xxxxxx
 * 16	U+0800		U+FFFF		3	1110xxxx	10xxxxxx	10xxxxxx
 * 21	U+10000		U+1FFFFF	4	11110xxx	10xxxxxx	10xxxxxx	10xxxxxx
 * 26	U+200000	U+3FFFFFF	5	111110xx	10xxxxxx	10xxxxxx	10xxxxxx	10xxxxxx
 * 31	U+4000000	U+7FFFFFFF	6	1111110x	10xxxxxx	10xxxxxx	10xxxxxx	10xxxxxx	10xxxxxx
 *
 * However: in November 2003, UTF-8 was restricted by RFC 3629 to end at U+10FFFF,
 * this was so that Unicode characters could also be represented by one or two UTF-16 code points.
 * This means a max of 4 bytes in a UTF-8 sequence.
 * 4 bytes could encode 0x0 - 0x1FFFFF, a 21 bit code set of 2,097,152 code points (0 - 2,097,151).
 * But only 0x10FFFF are used allowing 1,114,112 code points, remove 2048 surrogate pairs
 * that are needed by UTF-16 (U+D800 .. U+DFFF), this means that there are 1,112,064
 * code points that can be used.
 * These routines respect this restriction, although there are a few
 * constants defined for 5 & 6 byte sequences.
 *
 * As a matter of pure whimsy Emacs can be compiled to handle code points greater than 0x10FFFF.
 * It is unlikly that the Unicode Consortium will raise the restriction - which was put in
 * due to problems with UTF-16.
 *
 * This can be set with CONFIG_MAX_UNICODE in estruct.h; set a value, eg: 0x3FFFFFF.
 * A value of 0 specifies the standard 0x10FFFF.
 * Note that this might result in a change in the definition of UniChar
 * and the bits in K_CODE_MASK.
 *
 * BEWARE: this has only been lighly tested, it seems to work but bugs could be present.
 * Display of such characters will probably not work as expected - the (virtual)
 * terminal will likely not recognise the sequences.
 *
 * Bits (in the table above) is the number of bits in the code point.
 *
 * Note how:
 * + the Byte 1 (start byte) encodes the number of bytes and that subsequent bytes all start 0b10
 * + valid characters start 0b0 or 0b110
 *
 * If an encoding error is found, the following shall happen:
 * + reading from a file: remove the MDUTF8 flag, ie it becomes a normal binary file
 *   No attempt is made to normalise or correct the sequence - this could damage a binary file
 * + reading from the keyboard: reject the character
 *
 * References:
 * http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8
 * http://www.cl.cam.ac.uk/~mgk25/ucs/ISO-10646-UTF-8.html
 * http://www.ietf.org/rfc/rfc3629.txt
 *
 * Note that most of the functions use BYTE rather than char - ensure unsigned.
 */

/* Check if the 1st byte is less than these to get the length.
 * Next is the mask to get data (character) bits from the length byte.
 */
#define	UTF8_B1_LEN1	0x80
#define	UTF8_B1_MASK1	0x7F

#define	UTF8_B1_LEN2	0xE0
#define	UTF8_B1_MASK2	0x1F
#define	UTF8_B1_BITS2	0xC0	// Put top bits

#define	UTF8_B1_LEN3	0xF0
#define	UTF8_B1_MASK3	0x0F
#define	UTF8_B1_BITS3	0xE0	// Put top bits

#define	UTF8_B1_LEN4	0xF8
#define	UTF8_B1_MASK4	0x07
#define	UTF8_B1_BITS4	0xF0	// Put top bits

// The defines for length 5 & 6 are not used in standard config
#define	UTF8_B1_LEN5	0xFC
#define	UTF8_B1_MASK5	0x03
#define	UTF8_B1_BITS5	0xF8	// Put top bits

#define	UTF8_B1_LEN6	0xFE
#define	UTF8_B1_MASK6	0x01
#define	UTF8_B1_BITS6	0xFC	// Put top bits

#define	UTF8_Bn_MASK	0xC0	/* Mask subsequent bytes with this */
#define	UTF8_Bn_TPVALUE	0x80	/* to find this top (bits) value */
#define	UTF8_Bn_VMASK	0x3F	/* Mask to get 6 bits of value */

// The minimum and maximum valid code points for the length of encoding
#define	UTF8_MIN_CP1		0x0000
#define	UTF8_MAX_CP1		0x007F
#define	UTF8_MIN_CP2		0x0080
#define	UTF8_MAX_CP2		0x7FF
#define	UTF8_MIN_CP3		0x800
#define	UTF8_MAX_CP3		0xFFFF
#define	UTF8_MIN_CP4		0x10000
#define	UTF8_MAX_CP4		0x1FFFFF
#define	UTF8_MIN_CP5		0x200000
#define	UTF8_MAX_CP5		0x3FFFFFF
#define	UTF8_MIN_CP6		0x4000000
#define	UTF8_MAX_CP6		0x7FFFFFFF

/* Define the largest code point as per the official standard.
 * Although we might have a different maximum, this number is
 * important as it is the largest that will be understood by, eg,
 * terminal displays.
 */
#define	UTF8_UNICODE_CONSORTIUM_MAX_CODE_POINT	0x10FFFF	// Max code point

// Is the maximum Unicode character value the one specified by the Unicode Consortium ?
#if	CONFIG_MAX_UNICODE == 0
#define	UTF8_MAX_CODE_POINT	UTF8_UNICODE_CONSORTIUM_MAX_CODE_POINT	// Note it is NOT UTF8_MAX_CP4
#else
#define	UTF8_MAX_CODE_POINT	CONFIG_MAX_UNICODE	// Non standard value
#endif

#if	UTF8_MAX_CODE_POINT <= UTF8_MAX_CP4
#define	UTF8_MAX_BYTES		4		// Max bytes in a valid UTF8 character
#elif	UTF8_MAX_CODE_POINT <= UTF8_MAX_CP5
#define	UTF8_MAX_BYTES		5		// Max bytes in a valid UTF8 character
#else	//	UTF8_MAX_CODE_POINT <= UTF8_MAX_CP6
#define	UTF8_MAX_BYTES		6		// Max bytes in a valid UTF8 character
#endif

#define	UTF8_REPLACEMENT	0xFFFD		// Replace bad code points with this - 3 bytes #EF #BF #BD Prints like black diamond
extern	BYTE	utf8_replace[];			// Replacement as characters (bad code point)
extern	BYTE*	utf8_modeReplace;		// What to use in utf8_modecheck(), utf8_replace if not set
#define	UTF8_REPLACEMENT_LENGTH	3		// How long utf8_replace as a string


/* Values in the range of invalid code points: high and low surrogate halves used by UTF-16
 * These are illegal UTF-8 encodings according to RFC 3629 (U+D800 through U+DFFF)
 */
#define	UTF8_B1_LOW_SH	0xD800
#define	UTF8_B1_HIGH_SH	0xDFFF

/* Also illegal */
#define	UTF8_SWAP_BOM	0xFFFE	// Swapped byte order mark
#define	UTF8_ILL_FFFF	0xFFFF	// Just illegal (?)

/* Last 2 code points of each plane is illegal - as non-characters.
 * The next is both a mask and the illegal value
 *
 * This 'illegality' was removed by Corrigendum #9: Clarification About Noncharacters
 * U+FDD0 .. U+FDEF are also non characters, but they are not disallowed by this code.
 * See: docs/UTF-8-test.txt
 */
#define	UTF8_LAST2_PLANE_MASK	0xFFFE

/* clang will generate warnings when utf8.c is compiled, -Wmissing-variable-declarations
 * that say 'no previous external declarations for non-static variable'.
 * Ignore them.
 */
#ifndef	UTF8_EXTERN
#define	UTF8_EXTERN	extern
#endif

/* Record info about where an error was found */
typedef	struct	{
	int	u8errno;
	int	errLen;		/* # bytes in error */
	BufLns	Lineno;
	LineLn	Offset;		/* Byte offset */
	LineLn	CharPos;	/* Character position (offset) */
	UniChar	codepoint;	/* Codepoint - if calculated */
} utf8_errinfo;
UTF8_EXTERN	utf8_errinfo NoteErr;		/* Values recorded by utf8_setmsg() */

UTF8_EXTERN	LineLn	utf8_errOffset;		/* Offset from buffer start of a detected error */
UTF8_EXTERN	int	utf8_errno;		/* What the error is */
#define		UTF8ERR_NO_ERROR	0	/* Not an error - set at start of utf8_modecheck() */
#define		UTF8ERR_TOO_SHORT	1	/* End of multibyte string but more bytes needed */
#define		UTF8ERR_OVER_LONG	2	/* Over long encoding - ie more bytes than needed */
#define		UTF8ERR_INV_CODE_POINT	3	/* Invalid code point */
#define		UTF8ERR_TOO_MANY_BYTES	4	/* Sequence > 4 bytes */
#define		UTF8ERR_BAD_CONT_BYTE	5	/* Continuation byte (ie not first) is bad */
#define		UTF8ERR_UNEXP_CONT_BYTE	6	/* Unexpected continuation byte - ie not following a start byte */
#define		UTF8ERR_NOT_HERE	7	/* Cannot start a UTF-8 character here */
#define		UTF8ERR_DIFF_LEN	8	/* Would make different length UTF-8 character */
UTF8_EXTERN	int	utf8_errLen;		/* Bytes in error sequence */

UTF8_EXTERN	LineLn	utf8_chars;		/* Chars in line counted by utf8_chkbuf() */
extern	unsigned char	utf8_charLen[];		/* Index by 1st byte to get length of utf8 character */

// void	utf8_backToStart(void);
int	utf8_bufmodecheck(BUFFER* bp);		// Check & set UTF-8 mode on bp
int	utf8_check(const BYTE* str);		// Check that a NUL terminated string is UTF8 valid
int	utf8_cheq(const BYTE* c1, const BYTE* c2);	// Compare 2 UTF-8 encoded characters
int	utf8_chkbuf(const BYTE* str, size_t len);	// Check bytes to be UTF8 valid
int	utf8_CheckMangle(const BYTE* pos, BYTE newch, int start);	// See if replacement byte will still result in good UTF-8
size_t	utf8_chrnlen(const BYTE* str, size_t n);// Return # chars in buffer length n
char*	utf8_CurrCharInfo(void);		// About char under cursor return string like: U+xxxx #XX #XX #XX
UniChar	utf8_decode(const BYTE* str);		// Decode a UTF8 string to codepoint
BYTE*	utf8_encode(UniChar uc);		// Return UTF8 string for codepoint uc
void	utf8_errclear(void);			// Clear last error
char*	utf8_errmsg(int fixup);			// Return current error message
int	utf8_isvalid(UniChar uc);		// Check that uc is a valid code point
size_t	utf8_maxlen(const BYTE* str, size_t available_bytes, size_t* max);	// Get at most max chars
BYTE*	utf8_memrev(BYTE* str, size_t len);	// Reverse UTF-8: len bytes (not characters) long
int	utf8_modecheck(BUFFER* bp, int set, int global, int f, Cnt n);	// Check that a buffer contains valid UTF8
char*	utf8_perror(void);			// Return error message about saved error code: utf8_errno
char*	utf8_setmsg(BufLns linenumber);		// Record (set) details of an error - to use later
size_t	utf8_strlen(const BYTE* str);		// Number of characters in \0 terminated string
int	utf8_strncpy(BYTE* to, BYTE* from, size_t len);	// Copy at most len characters, NUL terminate
BYTE*	utf8_strrev(BYTE* str);			// Reverse a string
int	utf8_strtrunc(BYTE* str, size_t len);	// Truncate (with '\0') after len characters


/* Return the length of the current sequence - ie the number of utf8 bytes that follow.
 * This must be given the first byte of the character.
 * It assumes that all characters are well formed (this is checked on input)
 * Beware: the argument may be evaluated more than once.
 * If the result is >1 then:
 * * It is not a continuation byte
 * * It is the start of a multi byte sequence
 */
//#define	CHAR_LEN(c)		((BYTE)(c) < UTF8_B1_LEN1 ? 1 : (BYTE)(c) < UTF8_B1_LEN2 ? 2 : (BYTE)(c) < UTF8_B1_LEN3 ? 3 : 4)
#define	CHAR_LEN(c)		(utf8_charLen[(BYTE)(c)])

/* Is a byte a multi byte continuation */
#define	BYTE_IS_MB_CONT(c)	(((c) & UTF8_Bn_MASK) == UTF8_Bn_TPVALUE)
/* Is a byte part of a MB character */
#define	BYTE_IS_MB_PART(c)	((c) >= UTF8_B1_LEN1)

/* Move a multibyte character from src to dest
 * This also works with non MB characters
 *
 * There are problems with the move if src and dest overlap.
 * Thus 2 functions that are safe when moving in different directions:
 * UTF8_MOVE_DOWN(dest, src)	OK if dest is below src in memory
 * UTF8_MOVE_UP(dest, src)	OK if dest is above src in memory
 * UTF8_MOVE(dest, src)		OK direction is not defined - src and dest must NOT overlap
 *
 * These were originally implemented as macros. Inline functions help with:
 * * a semi-colon being OK/needed after it,
 * * not evaluating the arguments more than once.
 */

#define	UTF8_MOVE(dest, src)		UTF8_MOVE_UP((dest), (src))
static	inline void UTF8_MOVE_DOWN(BYTE* dest, const BYTE* src)	{size_t len = (size_t)CHAR_LEN(*(src)), m_i = 0; for(; m_i < len; m_i++) (dest)[m_i] = (src)[m_i];}
static	inline void UTF8_MOVE_UP(BYTE* dest, const BYTE* src)		{size_t len = (size_t)CHAR_LEN(*(src)); for(; len; len--) (dest)[len - 1] = (src)[len - 1];}

/* Find the first byte of a possibly MB character, moving backwards.
 * cp is a char*.
 * If cp is at the start of a character, do nothing.
 * BEWARE: cp may be evaluated more than once and is modified
 */
#define	UTF8_BACK_FIND_START(cp)	while(BYTE_IS_MB_CONT(*(cp))) (cp)--
/* This is the same, but where the index is 'nx' and an expression
 * that yeilds a character is 'chr' - which presumably involves 'nx' somehow.
 */
#define	UTF8_BACK_FIND_START_NX(chr, index)	while(BYTE_IS_MB_CONT(chr)) (index)--
/* With the above 2 - a move back by 1 will already have been done
 * before checking if more need to be reversed over.
 * They are NOT move-back-one-character macros.
 */

/* end */
