/*
 * ADDW SCCS: @(#)values.c	1.16 10/13/20 12:54:59
 */

/* Value manipulation.
 *
 * (c) Alain D D Williams, September 2020
 *
 * Most of these return void ... makes for bad error handling.
 */

#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"

#if	SHOWMODREL
char	valvers[] = "@(#)values.c	1.16 10/13/20 12:54:59";
#endif

/* Return somewhere to keep a value.
 *
 * type	A V_ value, eg V_INT
 *
 * Take the current head of values chain.
 * If there is not something - allocate one.
 */
VAL*	valGet(int type)
{
	VAL*	vp;

	if(valhead != (VAL*)NULL) {
		vp = valhead;
		valhead = valhead->v_v.v_next;
	} else {
		if((vp = calloc(1, sizeof(VAL))) == (VAL*)NULL)
			Die(OutOfMemory);
	}

	valTypeInit(vp, type);
	vRefSet(vp, 1);

	return(vp);
}

/* Free any data part of a VAL so that it can be reused.
 *
 * This does not free the chain hanging off v_next - this is about data in this
 * value.
 *
 * The type is left as V_UNDEF.
 *
 * TODO: Array & hash types.
 */
void	valFree(VAL* vp)
{
	switch(vtype(vp)) {
	  case V_STR:
		free(vtext(vp));
		break;

	  case V_ARRAY:
		valArraySize(vp, -1);
		// Setting the size frees the a_elems vector
		break;

	  case V_HASH:
		valHashSize(vp, -1);
		// Setting the size frees the h_elems vector
		break;
	}

	vp->v_flags = V_UNDEF;
}

/* Recurse through scan looking for references to top
 *
 * top	the VAL at the top
 * scan	an array or hash to scan
 *
 * Only look at array/hash with a reference count of 1, ie those that only
 * exist in this tree. This means that these that exist elsewhere will not be
 * dug into - so mutual references are not found.
 *
 * Return: number of times that top is found.
 */
int	valFindRefs(VAL* top, VAL* scan)
{
	int	ret = 0;
	int	index;

	/* Look through every a/h element of the a/h that we are scanning
	 * Only look through anonymous a/hs - ie those that are not referred
	 * from elsewhere.
	 * If we find top - just count it.
	 */
	if(vtype(scan) == V_ARRAY) {
		for(index = 0; index < scan->v_v.v_array.a_used; index++)
			if(scan->v_v.v_array.a_elems[index] == top)
				ret++;
			else
				if(vRefGet(scan) == 1 &&
				   (vtype(scan->v_v.v_array.a_elems[index]) == V_ARRAY ||
				    vtype(scan->v_v.v_array.a_elems[index]) == V_HASH))
					ret += valFindRefs(top, scan->v_v.v_array.a_elems[index]);
	} else {
		// It must be a hash
		for(index = 0; index < scan->v_v.v_hash.h_used; index++)
			if(scan->v_v.v_hash.h_elems[index * 2 + 1] == top)
				ret++;
			else
				if(vRefGet(scan) == 1 &&
				   (vtype(scan->v_v.v_hash.h_elems[index * 2 + 1]) == V_ARRAY ||
				    vtype(scan->v_v.v_hash.h_elems[index * 2 + 1]) == V_HASH))
					ret += valFindRefs(top, scan->v_v.v_hash.h_elems[index * 2 + 1]);
	}

	return(ret);
}

/* Release a VAL.
 * Decrement the reference count, free up resource when the count drops to 0.
 *
 * Push it at the start of the free VALues chain.
 *
 * The value that it contains is deallocated - if that is possible for the type.
 * This means much alloc()/free() - especially t_text
 * It would be possible to cache these but:
 * * more complexity
 * * it is contained a long string it could mean resource hogging
 *
 * vp can be NULL with no ill effect.
 */
void	valRelease(VAL* vp)
{
	int	refcnt;
	int	internalRefs = 0;

	// Some are not allocated - straight variables/constants
	if(vp == (VAL*)NULL)
		return;

	refcnt = (int)vRefGet(vp);

	/* Already freed ?
	 * More likely an array/hash with an internal self reference.
	 * <= - just in case
	 */
	if(refcnt <= 0)
		return;

	// If have hash/array in hash/array it will have 2+ refs
	// so if we are not careful it will never get freed.
	// This is to try to find that.
	if(vRefGet(vp) > 1 && (vtype(vp) == V_ARRAY || vtype(vp) == V_HASH))
		internalRefs = valFindRefs(vp, vp);

	/* If there are references that are not there because vp
	 * contains a reference to itself - then we must not release it.
	 * Just reduce the reference count by 1.
	 */
	if(refcnt - internalRefs - 1 > 0) {
		// There are still external references
		vRefSet(vp, (refcnt - 1));
		return;
	}

	/* The reference count is the sum of one that we are releasing
	 * and the internal ones.
	 * So there will no external ones left once the release
	 * is done -- so completely release.
	 */
	vRefSet(vp, 0);

	if((vp->v_flags & V_NOFREE) != 0)
		return;

	// Free contents
	valFree(vp);

	// Put onto chain of free VALs
	vp->v_v.v_next = valhead;
	valhead = vp;
}

/* Return a printable (text) representation of vp.
 *
 * The value might be in vp itself, so beware overwriting.
 * Return probably in playValStr - which is highly volatile.
 */
VAL*	valPrint(VAL* vp)
{
	UniChar	uc;
	char	tinybuf[30];

	switch(vtype(vp)) {
	  case V_BOOL:
		tcpys(vTEXT(playValStr), (vint(vp) ? truem : falsem));
		break;

	  case V_INT:
		mlputli(vint(vp), 10, tinybuf);
		tcpys(vTEXT(playValStr), tinybuf);
		break;

	  case V_STR:	// It is already a string, so just return it
		return(vp);

	  case V_CHAR:
		uc = (UniChar)vint(playValStr);
		if(utf8_isvalid(uc))
			tcpyc(vTEXT(playValStr), uc);
		else
			tcpys(vTEXT(playValStr), utf8_modeReplace);
		break;

	  case V_ARRAY:
	  case V_HASH:
	  case V_ERROR:
	  case V_UNDEF:
		tcpys(vTEXT(playValStr), vtypename(vp));
		break;
	}

	return(playValStr);
}

/* Ensure that vp is of type type.
 *
 * This will trash the value contained. This is about making sure
 * that type can be contained.
 *
 * Non type bits in v_flags are not changed.
 */
void	valChangeType(VAL** vp, int type)
{
	if(vtype(*vp) == type)
		return;

	valRelease(*vp);
	*vp = valGet(type);
}

/* Initialise VAL data structures to hold type type.
 */
void	valTypeInit(VAL* vp, int type)
{
	vp->v_flags = 0;
	vSetType(vp, type);

	switch(type) {
	  case V_STR:
		if((vtext(vp) = calloc(1, NBLOCK)) == (BYTE*)NULL)
			Die(OutOfMemory);

		tsetsize(vTEXT(vp), NBLOCK);
		vTEXT(vp)->t_flags = L_NUL;
		// The calloc() sets the invisible byte to '\0'
		break;

	  case V_ARRAY:
		vp->v_v.v_array.a_elems = (VAL**)NULL;
		vp->v_v.v_array.a_alloc = 0;
		vp->v_v.v_array.a_used = 0;
		valArraySize(vp, 10);
		break;

	  case V_HASH:
		vp->v_v.v_hash.h_elems = (VAL**)NULL;
		vp->v_v.v_hash.h_alloc = 0;
		vp->v_v.v_hash.h_used = 0; 
		break;
	}
}

/* Return a TRUE/FALSE value for vp
 *
 * Return: FALSE (ie 0)
 *	   Any other value is to be taken as TRUE
 */
int	valToBool(VAL* vp)
{
	switch(vtype(vp)) {
	  case V_UNDEF:
		return(FALSE);
	  case V_BOOL:
	  case V_INT:
	  case V_CHAR:
		return(vint(vp) != 0);
	  case V_STR:
		return(stol(vtext(vp)));
	  case V_ARRAY:
	  case V_HASH:
		return(FALSE);
	}

	__NOT_REACHED
}

/* Create an INT VAL* containing int l
 *
 * The ref count is set to zero which means that the value can be passed to
 * valCopy() and what this is assigned to holds the only reference.
 *
 * Return the VAL*
 */
VAL*	valIntToVAL(long l)
{
	VAL*	vp = valGet(V_INT);

	vint(vp) = l;
	vRefSet(vp, 0);

	return(vp);
}

/* Create a STR VAL* containing str
 *
 * The ref count is set to zero which means that the value can be passed to
 * valCopy() and what this is assigned to holds the only reference.
 *
 * Return the VAL*
 */
VAL*	valStrToVAL(const BYTE* str)
{
	VAL*	vp = valGet(V_STR);

	tcpys(vTEXT(vp), str);
	vRefSet(vp, 0);

	return(vp);
}

/* Return a VAL* that contains an INT in vp.
 *
 * Return:
 *	a VAL - that must be released after use
 *	NULL - vp cannot be converted
 */
VAL*	valToInt(VAL* vp)
{
	VAL*	nv;
	long	l;
	char*	tsp;

	if(vtype(vp) == V_INT || vtype(vp) == V_BOOL || vtype(vp) == V_CHAR) {
		vRefSet(vp, vRefGet(vp) + 1);
		return(vp);
	}

	switch(vtype(vp)) {
	  case V_STR:
		l = strtol(vtext(vp), &tsp, 0);
		if(vtext(vp)[0] == '\0' || *tsp != 0) {
			mlwrite("%EBad number convertion '%v'", vp);
			return((VAL*)NULL);
		}
		nv = valGet(V_INT);
		vSetInt(nv, l);
		return(nv);
	}

	return((VAL*)NULL);
}

/* Copy src to dest - change the type of dest to that of src if needs be.
 *
 * This just plays with references.
 */
void	valCopy(VAL** dest, VAL* src)
{
	// Increment the ref count first ...
	vRefSet(src, vRefGet(src) + 1);

	/* in case the src is inside *dest.
	 * This could happen with access to a literal array, eg:
	 *	[ a b c ][1]
	 * If increment too late then "b" could have already been released.
	 */
	valRelease(*dest);

	*dest = src;
}

/* Create a clone of STR src, ie a separate copy.
 *
 * This is mainly for use when src is playValStr - which can be quickly reused.
 *
 * Return the clone.
 */
VAL*	valStrClone(VAL* src)
{
	VAL*	vp = valGet(V_STR);

	tcpyt(vTEXT(vp), vTEXT(src));

	return(vp);
}

/* Return a new reference to src
 *
 * This is like valCopy() but is cheaper as the dest does not need to be
 * released - the dest often having just been created so that it could be
 * copied to.
 */
VAL*	valNewRef(VAL* src)
{
	vRefSet(src, vRefGet(src) + 1);
	return(src);
}

/* Another copy, but the type of dest will try be preserved,
 * the value in src will be converted to a representation
 * that is meaningful in dest.
 *
 * dest will be set to ERROR if numbers cannot be converted.
 *
 * If dest is V_ANY it is converted to that in src and given the
 * src value.
 *
 * Assumption is made that any STR value in src is NUL terminated.
 *
 * If dest is V_UNDEF or V_ERROR no copying/conversion happens.
 *
 * Return:
 *	TRUE  successful copy
 *	FALSE copy failed - could not convert
 */
int	valCopyValue(VAL** dest, VAL* src)
{
	int	ret = TRUE;
	long	l;
	char*	tsp;
	int	destType = vtype(*dest);

	// Change what we want to end up with ?
	if(destType == V_ANY)
		destType = vtype(src);

	// Throw away & get another VAL if in use more than once,
	// simplest to do that as well if need to change type
	if((vRefGet(*dest) > 1 || vtype(*dest) == V_ANY) && destType != V_ARRAY && destType != V_HASH) {
		valRelease(*dest);
		*dest = valGet(destType);
	}

	switch(destType) {
	  case V_BOOL:
		if(vtype(src) == V_BOOL || vtype(src) == V_INT || vtype(src) == V_CHAR)
			vSetInt(*dest, vint(src) != 0);

		if(vtype(src) == V_ERROR || vtype(src) == V_UNDEF || vtype(src) == V_ANY)
			vSetInt(*dest, 0);

		if(vtype(src) == V_STR) {
			// Could be TRUE/FALSE/... or a number
			if(cstrcmp(vtext(src), StrOn) == 0 || cstrcmp(vtext(src), truem) == 0)
				vSetInt(*dest, 1);
			else if(*vtext(src) == '\0' ||
			     cstrcmp(vtext(src), StrOff) == 0 || cstrcmp(vtext(src), falsem) == 0)
				vSetInt(*dest, 0);
			else {
				l = strtol(vtext(src), &tsp, 0);
				if(vtext(src)[0] == '\0' || *tsp != 0) {
					mlwrite("%EBad boolean convertion '%v'", src);
					valChangeType(dest, V_ERROR);
					ret = FALSE;
				} else
					vSetInt(*dest, l != 0);
			}
		}
		break;

	  case V_INT: case V_CHAR:
		if(vtype(src) == V_BOOL || vtype(src) == V_INT || vtype(src) == V_CHAR)
			vSetInt(*dest, vint(src));

		if(vtype(src) == V_ERROR || vtype(src) == V_UNDEF || vtype(src) == V_ANY)
			vSetInt(*dest, 0);

		if(vtype(src) == V_STR) {
			l = strtol(vtext(src), &tsp, 0);
			if(vtext(src)[0] == '\0' || *tsp != 0) {
				mlwrite("%EBad number convertion '%v'", src);
				valChangeType(dest, V_ERROR);
				ret = FALSE;
			} else
				vSetInt(*dest, l);
		}
		break;

	  case V_STR:
		tcpyt(vTEXT(*dest), vTEXT(valPrint(src)));
		break;

	  case V_ARRAY: case V_HASH:
		/* Arrays & hashes copy by reference
		 * Increment first - see valCopy()
		 */
		vRefSet(src, vRefGet(src) + 1);
		valRelease(*dest);
		*dest = src;
		break;
	}

	return(ret);
}


// **** Array functions ****

/* Change the allocated size of an array to elems elements.
 * The used size will not be larger than this.
 *
 * If elems is negative then array is freed.
 *
 * When reallocating: add a few more - stop too many realloc()s
 *
 * If the array vector has not been created - do so
 */
void	valArraySize(VAL* vp, int elems)
{
	if(elems >= 0) {
		// Big enough ?
		if(vp->v_v.v_array.a_alloc >= elems)
			return;

		// A bit of margin:
		if(vp->v_v.v_array.a_alloc + 10 > elems)
			elems = vp->v_v.v_array.a_alloc + 10;
	} else
		elems = 0;

	// Release elements that are no longer wanted
	while(vp->v_v.v_array.a_alloc > elems)
		valRelease(vp->v_v.v_array.a_elems[--vp->v_v.v_array.a_alloc]);

	if(vp->v_v.v_array.a_elems == (VAL**)NULL) {
		// The vector has not been created, do so:
		if(elems > 0 &&
		   (vp->v_v.v_array.a_elems = (VAL**)malloc((size_t)elems * sizeof(VAL))) == (VAL**)NULL)
			Die(OutOfMemory);
		vp->v_v.v_array.a_alloc = 0;	// Set to 0 so that new elements created below
		vp->v_v.v_array.a_used = 0;
	} else {
		// Adjust size of existing:
		if(elems <= 0) {
			free(vp->v_v.v_array.a_elems);
			vp->v_v.v_array.a_elems = (VAL**)NULL;
		} else
			if((vp->v_v.v_array.a_elems = (VAL**)realloc(vp->v_v.v_array.a_elems, (size_t)elems * sizeof(VAL))) == (VAL**)NULL)
				Die(OutOfMemory);
	}

	// Create new elements
	while(vp->v_v.v_array.a_alloc < elems)
		vp->v_v.v_array.a_elems[vp->v_v.v_array.a_alloc++] = valGetUndef();
}

/* Push a new value onto the end of an array
 *
 * The unused elements at the end of array exist, copy newItem over them.
 *
 * This does not check that vp is an array.
 */
void	valArrayPush(VAL* vp, VAL* newItem)
{
	// If full: make bigger
	if(vp->v_v.v_array.a_alloc <= vp->v_v.v_array.a_used)
		valArraySize(vp, vp->v_v.v_array.a_alloc + 1);

	valCopy(&vp->v_v.v_array.a_elems[vp->v_v.v_array.a_used++], newItem);
}

// **** Hash functions ****
/* Make hash vp able to hold elems hash elements.
 *
 * elems  How many key, value pairs are needed
 *        If this is negative the vector will be freed
 *
 * NB: an element is a pair of VAL*s: key, value
 */
void	valHashSize(VAL* vp, int elems)
{
	if(elems >= 0) {
		// Big enough ?
		if(vp->v_v.v_hash.h_alloc > elems)
			return;

		// A bit of margin:
		if(vp->v_v.v_hash.h_alloc + 10 > elems)
			elems = vp->v_v.v_hash.h_alloc + 10;
	} else
		elems = 0;

	// Release elements that are no longer wanted
	while(vp->v_v.v_hash.h_alloc > elems) {
		int	rel = --vp->v_v.v_hash.h_alloc * 2;
		valRelease(vp->v_v.v_hash.h_elems[rel    ]);
		valRelease(vp->v_v.v_hash.h_elems[rel + 1]);
	}

	if(vp->v_v.v_hash.h_elems == (VAL**)NULL) {
		// The vector has not been created, do so:
		if(elems > 0 &&
		   (vp->v_v.v_hash.h_elems = (VAL**)malloc((size_t)elems * sizeof(VAL*) * 2)) == (VAL**)NULL)
			Die(OutOfMemory);
	} else {
		if(elems <= 0) {
			free(vp->v_v.v_hash.h_elems);
			vp->v_v.v_hash.h_elems = (VAL**)NULL;
		} else {
			if((vp->v_v.v_hash.h_elems = (VAL**)realloc(vp->v_v.v_hash.h_elems, (size_t)elems * sizeof(VAL*) * 2)) == (VAL**)NULL)
				Die(OutOfMemory);
		}
	}

	// Create new elements
	while(vp->v_v.v_hash.h_alloc < elems) {
		int	get = vp->v_v.v_hash.h_alloc++ * 2;
		vp->v_v.v_hash.h_elems[get    ] = valGet(V_STR);
		vp->v_v.v_hash.h_elems[get + 1] = valGetUndef();
	}
}

/* Lookup key in hash vp, return a pointer to the VALue.
 *
 * The type of key must be STR
 *
 * create	If true create the key if not found
 *
 * Return:
 *	the value
 *	NULL if
 *	   the key is not found and create is FALSE
 *	   the key is not a STR
 */
VAL**	valHashFind(VAL* vp, VAL* key, int create)
{
	int	elem;

	if(vtype(key) != V_STR)
		return((VAL**)NULL);

	for(elem = 0; elem < vp->v_v.v_hash.h_used; elem++) {
		if(textcmp(vTEXT(vp->v_v.v_hash.h_elems[elem * 2]), vTEXT(key)) == 0)
			return(&vp->v_v.v_hash.h_elems[elem * 2 + 1]);
	}

	if(create == FALSE)
		return((VAL**)NULL);

	return(valHashAdd(vp, key, valGetUndef()));
}

/* Add {key, value} to hash vp.
 *
 * If the key already exists in the hash - the existing value is replaced.
 *
 * Return:
 *	VAL** reference to the VAL* found (& maybe created)
 *	NULL  Error - key is not a: STR, INT, CHAR, BOOL
 */
VAL**	valHashAdd(VAL* vp, VAL* key, VAL* value)
{
	VAL**	f;
	Hash*	h;
	int	KeyNeedsCopy = FALSE;

	if(vtype(key) != V_STR) {
		if(vtype(key) != V_INT || vtype(key) != V_CHAR || vtype(key) != V_BOOL)
			return((VAL**)NULL);

		// Get a string representation of integer types - in playValStr
		key = valPrint(key);
		KeyNeedsCopy = TRUE;
	}

	if((f = valHashFind(vp, key, FALSE)) != (VAL**)NULL) {
		valCopy(f, value);
		return(f);
	}

	h = &vp->v_v.v_hash;	// Easier typing

	// Need to make it bigger ?
	if(h->h_alloc <= h->h_used)
		valHashSize(vp, h->h_used + 1);

	// Insert key, value
	// f is reference to the key
	f = &h->h_elems[h->h_used * 2];

	// If the key is in playValStr make a copy as playValStr contents are volatile
	if(KeyNeedsCopy) {
		key = valStrClone(key);
		valCopy(f, key);
		valRelease(key);
	} else
		valCopy(f, key);

	valCopy(f + 1, value);

	// Note that the hash is one bigger
	h->h_used++;

	return(f + 1);
}

/* Add to hash vp a value with key
 *
 * Return as valHashAdd()
 */
VAL**	valHashAddKs(VAL* vp, char* key, VAL* value)
{
	VAL**	f;
	VAL*	k;

	k = valGet(V_STR);
	tcpys(vTEXT(k), key);
	f = valHashAdd(vp, k, value);
	valRelease(k);

	return(f);
}
