/*
 * ADDW SCCS: @(#)display.c	1.335 10/14/20 20:58:34
 */

/*
 * The functions in this file handle redisplay. There are two halves, the
 * ones that update the virtual display screen, and the ones that make the
 * physical display screen the same as the virtual display screen. These
 * functions use hints that are left in the windows by the commands.
 *
 * The source of the text displayed comes directly from the Windows & thus
 * the BUFFERs. Most of the editor is disconnected from the display.
 *
 * Externally called functions:
 * RePaint()	To be called when everything needs redrawing.
 *		First determines screen size.
 * update(force) Update the screen to a state which is prob similar
 * 		to what it is now.
 *		This will do nothing if there is pending input. Force
 *		ignores any such input.
 * upscreen()	Just calls update().
 *
 * Not all characters have the same width (see estruct.h comments near struct Window).
 * mk_wcwidth() returns the display width needed as:
 *  1 - good old plain US-ASCII characters & many Unicode points, U+A0 (no break space), ...
 *  2 - double width character. Eg: kMandarin yuè U+9605
 *  0 - zero width character. Eg: word joiner U+2060, NUL
 * -1 - control chars (U+01..U+1F), DEL, Unicode control (U+80..U+9F)
 *
 * Chars of width 0 & -1 will be displayed as a single width - but in magenta (as does GNU emacs).
 * When outputting characters to the screen, we cannot trust the screen to correctly
 * interpret something as double width, some will, some won't.
 * Thus a movecursor() will be done whenever a character with width not 1 is output.
 */

#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"
#include	"evers.h"

#if	SHOWMODREL
char	displvers[] = "@(#)display.c	1.335 10/14/20 20:58:34";
#endif

/* Get rid of a function, this is nearly always better except where you have
 * a 16 bit machine.
 */
#define	mlputi(num, base, buf)	mlputli((long)(num), (base), (buf))

/* This works by having a pair of buffers (screens):
 * * vscreen - what the screen contains as the program runs
 * * pscreen - reflection of what is on the physical screen
 *
 * vtputc() and vtpututf8() are called to update vscreen. No real output happens.
 * Updates may be random (different parts of the screen). The same
 * parts maybe more than once or not at all.
 *
 * At some point update() is called, this works out what needs to be done to
 * make pscreen the same as vscreen - as it does so it writes to the terminal
 * cursor movement commands to update the screen.
 * The real work is done in updupd() which calls updateline() for each displayed
 * line.
 *
 * The question is how to do that update quickly when there are multiple bytes.
 * The max # UTF8 bytes needed for a character is 4 - this will fit in a long,
 * so make the buffers use a long for each character position - long integer
 * compares are quick.
 *
 * How to store 1-4 bytes in a long ?
 * Get it's address as a BYTE* and store as a[0] .. a[3]
 * a[0] will store the 1st byte of a UTF8 sequence - vtpututf8()
 * a[0] will store the character as put there by vtputc()
 *
 * We don't care if the machine is big or little endian (or even middle or mixed).
 *
 *
 * vtputc() removes the top bit of non utf8 single character bytes that it stores - thus
 * it can never be confsed with a multi-byte UTF8 string.
 *
 * Type VideoC is either a byte or a long, depends on DISPWIDE.
 * (It might also be long long if we are using a strange CONFIG_MAX_UNICODE).
 */


/* On display we replace some characters with others. Collect them here.
 * ZeroWidthReplacement			A zero or fractional width character
 * FirstHalfDoubleWidthReplacement	When there is only room for a single width character
 *					at the end of line - and we have a double width character.
 * SecondHalfDoubleWidthReplacement	When there is only room for a single width character
 *					at the start of line - and we have a double width character.
 */
static	BYTE	ZeroWidthReplacement[]			= {0xe2, 0x96, 0xa1 };	// □ U+25A1 WHITE SQUARE
static	BYTE	FirstHalfDoubleWidthReplacement[]	= {0xe2, 0x96, 0xb6 };	// ▶ U+25B6 BLACK RIGHT-POINTING TRIANGLE
static	BYTE	SecondHalfDoubleWidthReplacement[]	= {0xe2, 0x97, 0x80 };	// ◀ U+25C0 LEFT POINTING TRIANGLE
#if	UTF8_MAX_CODE_POINT != UTF8_UNICODE_CONSORTIUM_MAX_CODE_POINT
static	BYTE	BeyondAstralReplacement[]		= {0xe2, 0x97, 0xbc };	// ◼ U+25FC BLACK MEDIUM SQUARE
#endif

// updateEdges() assumes that WindowSeparator is a single byte
static	UniChar	EndOfLineMarker				= '$';			// Show end of line
static	UniChar	ShiftedTextMarker			= '$';			// Show that text is not seen due to left/right shift
static	UniChar	WindowSeparator				= '|';			// Between windows

/* The point of these is to give names to special characters
 * rather than just have as literals.
 */
#define	CharCtrlUnderscore	ChrToCtrl('_')	/* C-_ */

/* The position (row and column) of the cursor on the screen.
 * This lets update() communicate with updateline() which might be
 * able to optimise getting there.
 */
static	int	wantRow;
static	int	wantCol;

/* What has been updated ?
 * The point is that if it is just window edges then refresh lines top to bottom,
 * otherwise start at current line and move up/down.
 *
 * This helps update() talk to updupd()
 */
static	int	updateWhat;
#define	UW_TEXT	0x01		// Window contents
#define	UW_MODE	0x02		// Mode line
#define	UW_EDGE	0x04		// Window edge
#define	UW_NUMB	0x08		// Window line numbers

// Not fully used - yet
static	Lflags	NoFlag;		// Just so that CurLineFlagp can always have something to point to
static	Lflags*	CurLineFlagp;	// Flag cur current line - in w_lineFlags

/* Each cell can have some attributes.
 * This is an array the same shape as the VideoC (v_text)
 * This combines desired paintbrush (colour, US, SO) with the character's width
 *
 * Attributes are stored in type VidAttr.
 */

#define	VATT_UL		0x80	// Underline/underscore
#define	VATT_STAND	0x40	// Standout

#define	VATT_CMASK	0x3c	// Mask for colours - 4 bits
#define	VATT_COLR_DEF	0x20	// System default colour - COLOUR_SYS_DEF
#define	VATT_CSHIFT	2	// Shift right for colours - move to bottom 3 bits: 0..7
// If VATT_COLR_DEF - ignore the other bits

#define	VATT_WIDTH	0x03	// Mask - Character width in display: 1 or 2
#define	VATT_BLANK	0x21	// Set empty values to this: plain display, width 1, system default colour

/* Macros to set Video Attributes. Something to be assignedto a VidAttr.
 * In each case the argument is an existing attribute or, if none, va_default()
 */
#define	va_default()		(VATT_BLANK)

#define	va_invalid		0xff					// When we don't know what it is

#define	va_set_ul(c)		(VidAttr)((c) |  VATT_UL)
#define	va_unset_ul(c)		(VidAttr)((c) & ~VATT_UL)
#define	va_get_ul(c)		(VidAttr)((c) &  VATT_UL)		// Result is boolean

#define	va_set_stand(c)		(VidAttr)((c) |  VATT_STAND)
#define	va_unset_stand(c)	(VidAttr)((c) & ~VATT_STAND)
#define	va_get_stand(c)		(VidAttr)((c) &  VATT_STAND)		// Result is boolean

#define	va_get_stand_ul(c)	(VidAttr)((c) & (VATT_STAND | VATT_UL))	// Test if standout and/or underline is set

#define	va_set_colour(c, col)	(VidAttr)(((c) & ~VATT_CMASK) | ((col) << VATT_CSHIFT))
#define	va_get_colour(c)	(VidAttr)(((c) &  VATT_CMASK) >> VATT_CSHIFT)	// Result is 0..7
#define	va_set_colour_sysdef(c)	(VidAttr)(((c) & ~VATT_CMASK) | VATT_COLR_DEF)
#define	va_is_colour_sysdef(c)	(VidAttr)((c) & VATT_COLR_DEF)

#define	va_set_width(c, wid)	(VidAttr)(((c) & ~VATT_WIDTH) | (wid))
#define	va_get_width(c)		(VidAttr)( (c) &  VATT_WIDTH)
#define	va_no_width(c)		(VidAttr)( (c) & ~VATT_WIDTH)		// Width removed - to allow testing of other attributes

/* a & b are attributes for spaces. Are they visibly different ?
 * They are if they differ in standout or underline/underscore
 */
#define	va_diff_space(a, b)	(((a) & (VATT_UL | VATT_STAND)) ^ ((b) & (VATT_UL | VATT_STAND)))

// These 2 still used when displaying files with characters that are control or have meta/top bit set
// They should really be given somewhere to live in penDefaults[], but do the transformation
// as & when needed - hopefully/probably not often.
// va_set_ctrl() reverses the standout attribute. This is because it might be used in a
// mode line which is standout - so make a control character look different by not being standout.
#define	va_set_meta(c)		va_set_ul(c)
#define	va_set_ctrl(c)		(VidAttr)((c) ^ VATT_STAND)

// All the values in here can be computed compile time.
// Although it does not look like it, this is a small BYTE array:
static	VidAttr	penDefaults[] = {
	// A normal window editing area
	[PAL_WINDOW * INKS + INK_PLAIN]=		va_default(),
	[PAL_WINDOW * INKS + INK_WRAP]=			va_set_colour(va_default(), COLOUR_YELLOW),
	[PAL_WINDOW * INKS + INK_WIDTH_NARROW]=		va_set_colour(va_default(), COLOUR_MAGENTA),
	[PAL_WINDOW * INKS + INK_WIDTH_WIDE]=		va_set_colour(va_set_width(va_default(), 2), COLOUR_MAGENTA),
	[PAL_WINDOW * INKS + INK_WIDE_EDGE]=		va_set_colour(va_default(), COLOUR_MAGENTA),
	[PAL_WINDOW * INKS + INK_EOL_MARKER]=		va_set_colour(va_default(), COLOUR_MAGENTA),

	// Mode line under a window
	[PAL_MODE * INKS + INK_PLAIN]=			va_set_stand(va_default()),
	[PAL_MODE * INKS + INK_WRAP]=			va_set_stand(va_set_colour(va_default(), COLOUR_YELLOW)),
	[PAL_MODE * INKS + INK_WIDTH_NARROW]=		va_set_stand(va_set_colour(va_default(), COLOUR_MAGENTA)),
	[PAL_MODE * INKS + INK_WIDTH_WIDE]=		va_set_stand(va_set_colour(va_set_width(va_default(), 2), COLOUR_MAGENTA)),
	[PAL_MODE * INKS + INK_WIDE_EDGE]=		va_set_stand(va_set_colour(va_default(), COLOUR_MAGENTA)),
	[PAL_MODE * INKS + INK_EOL_MARKER]=		va_set_stand(va_set_colour(va_default(), COLOUR_MAGENTA)),

	// Message line: information message
// Making this green meant lots of updates on this line. Fix so that blanks not made green
//	[PAL_MSG_INFO * INKS + INK_PLAIN]=		va_set_colour(va_default(), COLOUR_GREEN),
	[PAL_MSG_INFO * INKS + INK_PLAIN]=		va_default(),
	[PAL_MSG_INFO * INKS + INK_WRAP]=		va_set_colour(va_default(), COLOUR_YELLOW),
	[PAL_MSG_INFO * INKS + INK_WIDTH_NARROW]=	va_set_colour(va_default(), COLOUR_MAGENTA),
	[PAL_MSG_INFO * INKS + INK_WIDTH_WIDE]=		va_set_colour(va_set_width(va_default(), 2), COLOUR_MAGENTA),
	[PAL_MSG_INFO * INKS + INK_WIDE_EDGE]=		va_set_colour(va_default(), COLOUR_MAGENTA),
	[PAL_MSG_INFO * INKS + INK_EOL_MARKER]=		va_set_colour(va_default(), COLOUR_MAGENTA),

	// Message line: warning message
	[PAL_MSG_WARN * INKS + INK_PLAIN]=		va_set_colour(va_default(), COLOUR_CYAN),
	[PAL_MSG_WARN * INKS + INK_WRAP]=		va_set_colour(va_default(), COLOUR_YELLOW),
	[PAL_MSG_WARN * INKS + INK_WIDTH_NARROW]=	va_set_colour(va_default(), COLOUR_MAGENTA),
	[PAL_MSG_WARN * INKS + INK_WIDTH_WIDE]=		va_set_colour(va_set_width(va_default(), 2), COLOUR_MAGENTA),
	[PAL_MSG_WARN * INKS + INK_WIDE_EDGE]=		va_set_colour(va_default(), COLOUR_MAGENTA),
	[PAL_MSG_WARN * INKS + INK_EOL_MARKER]=		va_set_colour(va_default(), COLOUR_MAGENTA),

	// Message line: error message
	[PAL_MSG_ERROR * INKS + INK_PLAIN]=		va_set_colour(va_default(), COLOUR_RED),
	[PAL_MSG_ERROR * INKS + INK_WRAP]=		va_set_colour(va_default(), COLOUR_YELLOW),
	[PAL_MSG_ERROR * INKS + INK_WIDTH_NARROW]=	va_set_colour(va_default(), COLOUR_MAGENTA),
	[PAL_MSG_ERROR * INKS + INK_WIDTH_WIDE]=	va_set_colour(va_set_width(va_default(), 2), COLOUR_MAGENTA),
	[PAL_MSG_ERROR * INKS + INK_WIDE_EDGE]=		va_set_colour(va_default(), COLOUR_MAGENTA),
	[PAL_MSG_ERROR * INKS + INK_EOL_MARKER]=	va_set_colour(va_default(), COLOUR_MAGENTA),

		// Mark
	[PEN_MARK]=					va_set_stand(va_set_colour(va_default(), COLOUR_RED)),
		// Cursor in active/non-active windows
	[PEN_CURSOR]=					va_default(),
	[PEN_NAW_CURSOR]=				va_set_stand(va_set_colour(va_default(), COLOUR_GREEN)),
	[PEN_CURSOR_ALT]=				va_set_colour(va_default(), COLOUR_YELLOW),
		// Line numbers
	[PEN_LINE_NUMBERS]=				va_set_colour(va_default(), COLOUR_YELLOW),
		// Visible tabs
	[PEN_TAB]=					va_set_ul(va_set_colour(va_default(), COLOUR_CYAN)),
		// Current window
	[PEN_CURR_BORDER]=				va_set_stand(va_default()),
		// Non current window
	[PEN_BORDER]=					va_set_stand(va_default()),
		// End of line position - not a wrapping issue
	[PEN_EOL_POSITION]=				va_set_colour(va_default(), COLOUR_CYAN),
		// Notional blank screen; window independent
	[PEN_BLANK_SCREEN]=				va_default(),
		// Where part of the message line is coloured
	[PEN_MSGL_COLOUR]=				va_set_stand(va_set_colour(va_default(), COLOUR_RED)),
};

// These contains the pens in use, filled in by sanitiseColours()
static	VidAttr	pens[sizeof(penDefaults) / sizeof(penDefaults[0])];
static	int	MsgWindowPallet = PAL_WINDOW;


/* This points to somewhere within pens[], where the pallet for window/mode-line/... is found
 * It is easy to confuse indexing into pens[] and inkWell[] which is why they are distinctly
 * named and the indicies into inkWell[] are all INK_xxx
 */
static	VidAttr*inkWell;		// Point to pallet in pens[]

static	VidAttr	vBrush = VATT_BLANK;	// Current virtual attributes
static	VidAttr	pBrush = VATT_BLANK;	// Current physical attributes - will never contain a width attribute



/* One of these for every line of the display.
 * v_attr & v_text are arrays of the screen width.
 */
typedef struct  VIDEO {
	int	v_flag;			/* Flags */
	int	v_scrolled;		/* How many lines the screen has to be scrolled. + is down */
#if	LEGACY_DISPLAY
	int	v_fcolor;		/* current forground color */
	int	v_bcolor;		/* current background color */
	int	v_rfcolor;		/* requested forground color */
	int	v_rbcolor;		/* requested background color */
#endif
	VidAttr*v_attr;			/* Cell attributes */
	VideoC*	v_text;			/* Screen data */
} VIDEO;

// v_flag can contain:
#define VFCHG	0x0001			/* Changed flag			*/
#define	VFEXT	0x0002			/* extended (beyond column 80) - unused	*/
#if	LEGACY_DISPLAY
#define	VFREV	0x0004			/* reverse video status		*/
#define	VFREQ	0x0008			/* reverse video request	*/
#define	VFCOL	0x0010			/* color change requested	*/
#endif


static	VIDEO**	vscreen;		/* Virtual screen. */
#if	MEMMAP == 0
static	VIDEO**	pscreen;		/* Physical screen. */
#endif
static	int	WindowScrolled;		/* A window has scrolled - refresh hint */

/* edgeAt is an array of hints of where a change to a window
 * edge separator (think WindowSeparator) has been changed.
 *
 * This is an array of arrays, dimensions: # screen lines X VID_MAXSEPS,
 * is VID_MAXSEPS per line.
 * 
 * Value is screen offset, these are sorted, 0 means early end of array.
 *
 * In the unlikely event that there are more than 8 windows across
 * a screen: some will not be noticed: unlikely, but only consequence
 * is slighly slower screen updates - they will be updated by updateline().
 * Note that these are short but things like w_leftcol are int,
 * I do not expect screeens wider than 32k.
 */
#define	VID_MAXSEPS	8		// Max edge separators on one line
static	short*	edgeAt;			// Position of a window edge separator - 0 == none
static	short	leftMostEdge;		// Column of the left most edge
static	short	rightMostEdge;		// Column of the right most edge
static	int	edgeSet;		// True if something interesting in edgeAt.



/* When insert/delete lines via scrolling the line contents are OK,
 * however the line numbers will need attention. This is what these are about.
 * Yes: if more than 1 window scrolls at the same time then subsequent ones could result
 * in these global variables being reset. This is not fatal since this is about
 * an optimisation - so what if it get broken on rare occasions ?
 */
static	int	PaintScrollLines;	/* ScrollLines below are valid, set in CheckForScroll() */
static	int	TopScrollLines;		/* Top line of line numbers to redraw */
static	int	BottomScrollLines;	/* Bottom line of line numbers to redraw */


/* Buffer stuff used in mlwrite() for the MessageLine.
 * This was all fixed sized arrays ... yuck, especially when the screen size is dynamic.
 * These arrays were often internal to functions, I have pulled them out & made them
 * global, yuck, but hopefully not so flakey as it used to be.
 * To put things in these buffers, use functions that know how big they are:
 * mbStrcpy(), mlbStrcat()
 * Is is OK to strcpy() between the buffers since they are all the same size.
 *
 * These are also used for file expansion buffers.
 *
 * These are guaranteed to be at least as wide as the screen and at least NFILEN big.
 */


#if	TERMDEBUG
/* Debug function.
 * Return a *static* string that is the first 100-ish characters in the screen buffer.
 * It might stop part way through a multi byte character,
 * it only generates the first byte of multi-byte characters.
 *
 * Private (to this module).
 */
char*	show100(VideoC* v)
{
	static	char	ret[350 + UTF8_MAX_BYTES];
	int		j;
	int		max = 350;
	if(max > AllocColumns)
		max = AllocColumns;

	for(j = 0; j < max; j++)
		ret[j] = (char)v[j];

	ret[j] = '\0';

	return(ret);
}

/* The same as show100() but for attribues.
 * Put them as bytes - preserves column alignment.
 *
 * NUL termination still works as the value should never be 0.
 *
 * How some attributes show:
 * !	21	VATT_BLANK - default - including where the cursor is
 * I	49	VATT_STAND green - non current window cursor
 * a	61	VATT_STAND COLOUR_SYS_DEF (system def colour) - window edge
 * E	45	VATT_STAND COLOUR_RED - mark current window
 * 
	0d	va_default() (VATT_BLANK) COLOUR_YELLOW - line number
 */
char*	show100a(VidAttr* v)
{
	static	char	ret[350 + UTF8_MAX_BYTES];
	int		j;
	int		max = 350;
	if(max > AllocColumns)
		max = AllocColumns;

	for(j = 0; j < max; j++)
		ret[j] = (char)v[j];

	ret[j] = '\0';

	return(ret);
}


/* Debug.
 * Show various things - whatever I want at the time.
 */
void	showit(int ln)
{
	int	i;
	Window*	wp;

	if(!pscreen)
		return;

	fprintf(stderr, "showit at %ld nrow=%d called from %d\n", time(NULL), AllocLines, ln);

	for(wp = wheadp, i = 0; wp != NULL; wp = wp->w_wndp, i++) {
		fprintf(stderr, "window %d wp %p w_flag %#x\n", i, (void*)wp, wp->w_flag);
	}

	for(i = 0; i < AllocLines; ++i) {
		fprintf(stderr, "showit row=%d\n", i);
		fprintf(stderr, "showit row2=%d pscreen=%p\n", i, (void*)pscreen);
		fprintf(stderr, "showit row2a pscreen[%d]=%p\n", i, (void*)pscreen[i]);
	}
}

void pscreen_check(void) {
	int	line;
	VideoC volatile z;
	for(line = 0; line < AllocLines; line++)
		z = *(pscreen[line]->v_attr);
}

/* physical screen line 2 was getting changed - show when it was
 *
 * cmt	comment - about where this was called from
 * row	row being worked on, set -1 if not known
 * n	line number this called from.
 */
void	ShowLine2(const char* comment, int row, int n)
{
	static	int	hash = 0;
	int	value = 0;
	char*	str = show100(pscreen[2]->v_text);

	while(*str)
		value = *str++ ^ value * 33;

	value &= 0xffff;

	if(hash != value)
		TERMOUTTRACE((TermDebFD, "%s working row %d line 2 __LINE__=%d %s\n\"%s\"\n", comment, row, n, (hash != value ? "CHANGED" : ""), show100(pscreen[2]->v_text)))

	hash = value;
}
#else
#define	ShowLine2(comment, row, n)	// nothing
#endif


/* Initialize the data structures used by the display code. The edge vectors
 * used to access the screens are set up. The operating system's terminal I/O
 * channel is set up. All the other things get initialized at compile time.
 * The original window has "WFCHG" set, so that it will get completely
 * redrawn on the first call to "update".
 * VidAttr is cleared (calloc()) to plain & boring.
 *
 * Cooked is true if the tty is already cooked - will happen with WINCH.
 *
 * columns is allocated one more than we need ... in case we put a double width
 * character there.
 *
 * The structures are allocated all at once, rather than (as before) once per screen
 * line. This makes the code shorter, uses less malloc space and should improve memory
 * locality - good for CPU cache.
 * If we have a pscreen: this is put after the vscreen - and everything is allocated
 * twice the size so that there is room.
 */
void	vtinit(int Cooked)
{
	int		i;
	VIDEO*		vp;
	VideoC*		vtp;
	VidAttr*	vap;

	if(Cooked == 0) {
		TTopen();		/* open the screen */
		TTkopen(FALSE);		/* open the keyboard */
		TTrev(FALSE);		/* set reverse video status to normal */
	}

	// In case they were already allocated (benign if not):
	vtfree();

	/* Allocated lines & columns - was used in vtfree(), now by code that
	 * needs work with all allocated lines: race condition might be
	 * different from term.t_mrow.
	 */
	AllocLines   = term.t_nrow;
	AllocColumns = term.t_ncol;

	TERMOUTTRACE((TermDebFD, "vtinit(%d) AllocLines=%d AllocColumns=%d\n", Cooked, AllocLines, AllocColumns))

	size_t	lines = (size_t)AllocLines;	// Lines allocated

#if	MEMMAP == 0
	lines *= 2;			// More needed for pscreen
#endif

	if((vscreen = (VIDEO **) calloc(lines, sizeof(VIDEO *))) == (VIDEO**) NULL)
		Die(MsgOutOfMemoryAlloc, "vscreen");

#if	MEMMAP == 0
	pscreen = vscreen + AllocLines;
#endif

	if((edgeAt = (short*)calloc((size_t)(lines * VID_MAXSEPS), sizeof(short))) == (short*)NULL)
			Die(MsgOutOfMemoryAlloc, "edgeAt");

	if((vp = (VIDEO *) calloc((size_t)lines, sizeof(VIDEO))) == (VIDEO*) NULL)
			Die(MsgOutOfMemoryAlloc, "vscreen row");

	if((vtp = (VideoC*) calloc((size_t)(AllocColumns + 1) * lines, sizeof(VideoC) )) == (VideoC*)NULL)
		Die(MsgOutOfMemoryAlloc, "VideoC");

	if((vap = (VidAttr*)calloc((size_t)(AllocColumns + 1) * lines, sizeof(VidAttr))) == (VidAttr*)NULL)
		Die(MsgOutOfMemoryAlloc, "VidAttr");

	// Assign to screen rows; pscreen after vscreen
	for(i = 0; i < (int)lines; i++) {
		vscreen[i] = vp;

		vp->v_text = vtp;
		vp->v_attr = vap;

		// vp->v_flag = 0;	// Not needed: calloc()

#if	LEGACY_DISPLAY
		// Requested colour
		vp->v_rfcolor = gfcolor;
		vp->v_rbcolor = gbcolor;
#endif

		vp++;
		vtp += AllocColumns + 1;
		vap += AllocColumns + 1;
	}

	/* In case of WINCH or whatever.
	 * In some ways putting it here is not quite right, but it
	 * does mean that sizes are aligned with the virt/phys screens.
	 */
	CheckSizes();

	CheckAutoWidth();	// Calculate wordwrap position
}

#if	USG			/* System V */
#define	WANT_VTFREE
#endif
#if V7 || BSD_UNIX
#define	WANT_VTFREE
#endif
#if	CLEAN
#define	WANT_VTFREE
#endif

#ifdef	WANT_VTFREE
/* Free up all the dynamically allocated video structures.
 * These are now allocated as large, contiguous arrays - so just
 * free the first one.
 * pscreen abuts vscreen.
 */
void	vtfree(void)
{
	if( !vscreen)	// Not allocated (yet) ?
		return;

	free(vscreen[0]->v_attr);
	free(vscreen[0]->v_text);
	free(vscreen[0]);
	free(vscreen);

	free(edgeAt);

	vscreen = (VIDEO**)NULL;
}
#endif

/* Clean up the virtual terminal system, in anticipation for a return to the
 * operating system. Move down to the last line and clear it out (the next
 * system prompt will be written in the line). Shut down the channel to the
 * terminal.
 * Also called when suspending (^z).
 */
void	vttidy(void)
{
	// This will be the case if a startup/command-line macro goes exit-emacs or similar
	if(TTopened == TTneverOpened)
		return;

#if	MSDOS
	if(wbuffs)
		;
	else
		mlerase();
#endif

	msglClear();	/* Remove any status line message */
	TTflush();
	TTkclose();
	TTclose();
}

/* The core of writing to the screen is very simple:
 * + move to row/column: vtmove(row, column)
 * + write a character - which increments the column
 *   2 functions:
 *   + vtputc(ch), ch is U+00 -> U+FF. Tabs expanded; ctrl & meta made visible
 *   + vtpututf8(u8str), u8str is address of 1st byte of UTF8 character
 * Add concept of character attribs that set the current pen used by vtputc() & vtpututf8(),
 * attribs are: Underscore, Standout, colour. Pen keeps attribs until changed.
 * + vtset_ul(on_off)
 * + vtset_stand(on_off)
 * + vtset_colour(0-7)
 * + combinations: set something and print a character/string
 * + vtset_atts(us, stand, colour)
 *
 * Q:
 * + where do max row/col come from ?
 * + vtcol can be -ve & exceed width - look into
 * + same for vtrow ?
 * + how lookup/set char width and where is it done, here or elsewhere
 */


/* Set the virtual cursor to the specified row and column on the virtual
 * screen. There is no checking for nonsense values; this might be a good
 * idea during the early stages.
 */
void	vtmove(int row, int col)
{
	vtrow = row;
	vtcol = col;
}

/* Set PHYSICAL pBrush colour & other attributes to want.
 *
 * Leave it up to the TT routines to work out if anything needs to be done
 * to set the brush correctly.
 *
 * This will be called a lot - try to avoid function calls
 */
void	SetBrush(VidAttr want)
{
	VidAttr	want_no_width = va_no_width(want);
	int	colour;

	// Compare without width. The width is used by display to note how many character
	// positions on the screen that it occupies.
	if(pBrush == want_no_width)
		return;

	colour = va_get_colour(want);

	if(va_is_colour_sysdef(want))
		colour = COLOUR_SYS_DEF;

	TTforg(colour);

	TTul(va_get_ul(want) != 0);

	TTrev(va_get_stand(want) != 0);

	pBrush = want;
}

/* Put a UniChar string at the current position, check that it won't overflow the
 * window RH edge - if it would overflow paint the last character using INK_WRAP.
 * str is a string terminated by UNISTR_END not NUL - we must be able to output U+0.
 */
void	vtputUniStr(Window* wp, const UniChar* str)
{
	VidAttr	cb = vBrush;

	for( ;*str != UNISTR_END; str++) {
		LineLn	margin = wp->w_curcol - wp->w_leftoffset - wp->w_ncolumns;
		if(margin > -1)
			break;
		if(margin > -2)
			vBrush = inkWell[INK_WRAP];
		if(*str <= UTF8_MAX_CP1) {
			// Common case
			vtputc(wp, *str);
		} else {
			/* Code point >= U+80 - multiple bytes
			 * Convert to UTF8 string and output.
			 */
			vtpututf8(wp, utf8_encode(*str));
		}
	}

	vBrush = cb;
}

/* Put a standard sized character string - as vtputUniStr()
 * str will be interpretted as a UTF8 string.
 * The string is NUL terminated, ie cannot output U+0.
 */
void	vtputStr(Window* wp, const unsigned char* str)
{
	UniChar	ucb[2];

	ucb[1] = UNISTR_END;

	while(*str) {
		ucb[0] = utf8_decode(str);
		str += CHAR_LEN(*str);
		vtputUniStr(wp, ucb);
	}
}

/* Write a character to the virtual screen. The virtual column is updated.
 * It is assumed that the character will fit in the window; if the character
 * expands to several on display (eg M-X) the extra characters are checked
 * that they will fit.
 *
 * Although c is UniChar - it should be in the range U+00 - U+FF - ie fit in a BYTE.
 *
 * wp is needed since we need to know the left/right edges of a window (which might
 * not occupy all the screen width) and by how much text in the window is indented.
 *
 * Beware: this is mixing units: screen columns and window columns.
 */
void	vtputc(Window* wp, UniChar c)
{
	register VIDEO* vp;		/* ptr to line being updated */
	UniChar		buf[5];		// Used for alternate character representation, eg: M-x

	vp = vscreen[vtrow];

	if(c >= 0x20 && c < 0x7F) {	// ' ' .. '~'
		/* Normal characters.
		 * Do clear out the rest of the VideoC otherwise comparison to space
		 * fails in updateline().
		 *
		 * The contortion ensures that it gets put into the right byte
		 * regardless of machine byte order.
		 *
		 * See vtpututf8()
		 *
		 * Not sure if this is really needed.
		 */
		vp->v_text[vtcol] = 0;
		*(BYTE*)(&(vp->v_text[vtcol])) = (BYTE)c;
		vp->v_attr[vtcol] = vBrush;
		wp->w_curcol++;
		vtcol++;

	} else if(c < 0x20 || c == CharDelete) {
		// Control character or DEL.

		// Tab width of 1 and it is displayed like any other control character
		if(c == '\t' && wp->w_bufp->b_tabsize > 1) {
			/* Expand the tab - special control character.
			 * Tabs are effectlively variable width.
			 * Stop is going beyond the window edge.
			 */
			LineLn	tocol;
			VidAttr	cb = vBrush;

			// Display tabs by highlighting them ?
			// Need to set standout mode as coloured blanks are, well, blank
			if(wp->w_opts & WO_COLOUR_TABS)
				vBrush = pens[PEN_TAB];

			tocol = ((wp->w_curcol - wp->w_numwidth) / wp->w_bufp->b_tabsize + 1) * wp->w_bufp->b_tabsize + wp->w_numwidth;

			do
				vtputc(wp, ' ');
			while(wp->w_curcol < tocol && wp->w_curcol < wp->w_leftoffset + wp->w_ncolumns);

			// Put as long not got to the tab stop and there is still some screen
			vBrush = cb;	// Undo any tab colourisation

		} else {
			/* DEL is shown as control '?'.
			 * If don't do this we end up in infinite recursion.
			 * US (U+1F) is shown as control '~', strictly wrong but an '_'
			 * that is displayed in reverse or underline is hard to see.
			 */
			c = (c == CharDelete) ? '?' : (c == 0x1F) ? '~' : ChrFromCtrl(c);

			// Alternative way of displaying		
			if(wp->w_opts & WO_META_ATTRIB) {
				VidAttr	cb = vBrush;
				vBrush = va_set_ctrl(vBrush);
				vtputc(wp, c);
				vBrush = cb;
				*CurLineFlagp |= LF_StandOut;	// Note standout of this line
			} else {
				// This could overflow - use function that checks:
				buf[0] = '^';
				buf[1] = c;
				buf[2] = UNISTR_END;
				vtputUniStr(wp, buf);
				// Bug: if it does overflow - all of it ought to be colourised
			}
		}

	} else {
		/* Its a meta character
		 * Display: meta and maybe ctrl as well.
		 */
		int	ctrl;

		c &= 0x7F;	// clobber top bit

		// Also display DEL as control '?' & US as control '~'
		ctrl = c < 0x20 || c == CharDelete;
		if(ctrl)
			c = (c == CharDelete) ? '?' : (c == 0x1F) ? '~' : ChrFromCtrl(c);

		if(wp->w_opts & WO_META_ATTRIB) {
			VidAttr	cb = vBrush;
			vBrush = va_set_meta(vBrush);
			if(ctrl)
				vBrush = va_set_ctrl(vBrush);
			vtputc(wp, c);
			vBrush = cb;
			*CurLineFlagp |= LF_StandOut;	// Note standout of this line
		} else {
			/* Display old fashioned way, eg: M-a or M-^a
			 */
			buf[0] = 'M';
			buf[1] = '-';
			buf[3] = UNISTR_END;
			buf[4] = UNISTR_END;
			if(ctrl) {
				buf[2] = '^';
				buf[3] = c;
			} else {
				buf[2] = c;
			}
			vtputUniStr(wp, buf);
			// Bug: if it does overflow - all of it ought to be colourised
		}
	}
}

/* UTF8 equvalent of vtputc()
 * up is the start of a UTF8 string in a buffer.
 */
void	vtpututf8(Window* wp, const BYTE* up)
{
	VideoC		ch;
	register VIDEO*	vp;			/* ptr to line being updated */
	int		width;			/* Character width */
	int		colour = INK_PLAIN;	/* Colour to display in - index into inkWell[] */

	vp = vscreen[vtrow];

	// Do special processing of tabs, etc
	if(*up <= 0x7F) {
		vtputc(wp, *up);

		return;
	}

	/* Some can have strange widths.
	 * Pass in the info so that the display can handle
	 * If the width is worked out to -1 - set to zero and treat the same.
	 * If the width is not 1 - display in a different colour
	 * Replace (on the screen) a char of width 0 or -1 with a square
	 * so that the user can see something there. Replacing with space is useless
	 * as a coloured space ... does not show coloured at all unless the cursor
	 * rests on it.
	 */
	if((width = mk_wcwidth(utf8_decode(up))) <= 0) {
		width = 0;
		up = ZeroWidthReplacement;
	}

	/* Although this will be called when there is at least 1 position available in
	 * the window it is possible that the 2nd 1/2 of a double width char does overflow it.
	 * Check that this does not happen.
	 * Check not done earlier as mk_wcwidth(utf8_decode()) is needed.
	 * This means that a window of width 1 will never display double width characters - tough!
	 */
	if(width > 1 && wp->w_curcol + width > wp->w_leftoffset + wp->w_ncolumns) {
		up = FirstHalfDoubleWidthReplacement;
		width = 1;
		colour = INK_WIDE_EDGE;
	}

	// Clear out bytes that might not be written to - else ch could be different
	// in spite of containing the same character - ie same when some bytes ignored
	ch = 0;

#if	UTF8_MAX_CODE_POINT != UTF8_UNICODE_CONSORTIUM_MAX_CODE_POINT
	/* Check if the character might not be understood at all by the terminal (emulation).
	 * If it is replace with something that will get shown.
	 */
	if(utf8_decode(up) > UTF8_UNICODE_CONSORTIUM_MAX_CODE_POINT) {
		up = BeyondAstralReplacement;
		width = 0;	// Put in a different colour
	}
#endif

	switch(CHAR_LEN(*up)) {
#if	UTF8_MAX_CODE_POINT > UTF8_MAX_CP4
	// This will not be executed as the character will have been replaced
	case 6:	((BYTE*)&ch)[3] = up[5];
		/* FALLTHROUGH */
	case 5:	((BYTE*)&ch)[3] = up[4];
		/* FALLTHROUGH */
#endif
	case 4:	((BYTE*)&ch)[3] = up[3];
		/* FALLTHROUGH */
	case 3:	((BYTE*)&ch)[2] = up[2];
		/* FALLTHROUGH */
	case 2:	((BYTE*)&ch)[1] = up[1];
		/* FALLTHROUGH */
	case 1:	((BYTE*)&ch)[0] = up[0];
	}

	vp->v_text[vtcol] = ch;

	if(width != 1) {
		/* Set funny width colour for this character, do not change Brush
		 * Because we have substituted zero width chars just above,
		 * record their width as 1.
		 * The only widths that we can display are 1 & 2.
		 */
		vp->v_attr[vtcol] = inkWell[width == 0 ? INK_WIDTH_NARROW : INK_WIDTH_WIDE];
		if(width == 2) {
			// Check overflow - seems OK
			/* If it has width 2 - occupy the next position and put something a character that
			 * cannot exist there.
			 * What we write should not be displayed, but we can test for it in case later
			 * on we end up on it.
			 * The text in the 2nd position will never be output. Put UNISTR_END there as
			 * a marker - if updateline() sees it (which it should not), it moves back
			 * one character position.
			 *
			 * The attribute in char 1 MUST be double width else display fails by only
			 * incrementing one place.
			 */
			wp->w_curcol++;
			vtcol++;
			vp->v_attr[vtcol] = inkWell[INK_WIDTH_WIDE];
			vp->v_text[vtcol] = UNISTR_END;
		}
	} else {
		// Only set the colour if we decided to above.
		// This means that we leave whatever was set by out calling function
		vp->v_attr[vtcol] = inkWell[colour];
	}

	wp->w_curcol++;
	vtcol++;
}

/* Erase from the end of the software cursor to the end of the line on which
 * the software cursor is located.
 * This means setting to spaces with blank attributes.
 * However: for the mode line we allow the character ch to be specified.
 *
 * Limited to the current window.
 */
void	vteeol(Window* wp, VideoC ch, VidAttr at)
{
	VidAttr	OldBrush = vBrush;

	vBrush = at;

	/* We might be on a line of length less than the offset, ie nothing displayed
	 * at all. Set to start erasing at the first screen position.
	 * vtcol (used in vtputc()) will have already been set to 0 in anticipation.
	 */
	if(wp->w_curcol < wp->w_leftoffset)
		wp->w_curcol = wp->w_leftoffset;

	while(wp->w_curcol < wp->w_leftoffset + wp->w_ncolumns)
		vtputc(wp, ch);

	vBrush = OldBrush;
}

/* User routine to force a screen update always finishes complete update.
 *
 * The count is ignored. A default count is not set.
 *
 * Command:  update-screen
 * Bound to: unbound
 *
 * Return TRUE - always
 */
int	upscreen(int f, Cnt n)
{
	update(TRUE);
	return(TRUE);
}

/* Cause the coloured mark to disappear - if it has moved.
 */
void	unmarkmark(Window* wp)
{
	// Unchanged from last time - do nothing
	if(wp->w_markp == wp->w_lastdismarkp && wp->w_marko == wp->w_lastdismarko &&
	     wp->w_markpartoff == wp->w_lastdismarkpo)
		return;

	// Scan/refresh all characters
	wp->w_flag |= WFHARD;
}

/* Show the mark by setting colour attributes on the character position.
 * The mark could just have been set or maybe the
 */
void	markmark(Window* wp)
{
	LINE*	lp;
	LineLn	col;	// Unit is L since cursor might not be displayable (too far left/right)
	int	row;
	VIDEO*	vline;
	VidAttr*mark_attr;

	// No mark to show ?
	if(wp->w_markp == NULL)
		return;

	/* If dot is on top of the mark - do not highlight - "." is more important than the mark,
	 * we must display them both the same way.
	 * With some colour/highlighting combinations it can cause the cursor to disappear.
	 */
	if(wp->w_markp == wp->w_dotp && wp->w_marko == wp->w_doto &&
	    wp->w_markpartoff == wp->w_partoff) {
		// Set the attribute to normal pen, preserve width:
		VidAttr*va = &vscreen[wp->w_disrow]->v_attr[wp->w_discol];
		*va = va_set_width(pens[PEN_CURSOR], va_get_width(*va));

		return;
	}

	// Find the mark. It is possible that it is not displayed.
	// Do not wrap round in the buffer and find a mark near the start of the buffer.
	for(lp = wp->w_topLinep, row = 0; row < wp->w_ntrows; row++, lp = lforw(lp)) {
		// Mark on this line ?
		if(lp != wp->w_markp) {
			/* Test for last line on buffer done here - otherwise
			 * if done in the for() above we will not show the mark
			 * if it is on the end-buffer line.
			 */
			if(lp == wp->w_bufp->b_abotline)
				break;

			continue;
		}

		// Chars as: Hex, M-^x or attributes underline/standout
		col = getdiscol(wp, lp, wp->w_marko,
			(wp->w_opts & WO_HEX ? DisCol_HEX : (wp->w_opts & WO_META_ATTRIB ? DisCol_MetaChar : DisCol_Char))
			) - wp->w_leftoffset;

		if(col < 0 || col >= wp->w_ncolumns - wp->w_numwidth)
			break;

		vline = vscreen[wp->w_toprow + row];
		// Left hand edge of window on screen + position in displayed text + width of numbers + part offset
		// Preserve the width:
		mark_attr = &vline->v_attr[wp->w_leftcol + col + wp->w_numwidth + wp->w_markpartoff];
		*mark_attr = va_set_width(pens[PEN_MARK], va_get_width(*mark_attr));

		// So that the line is looked at and the character redrawn:
		vline->v_flag |= VFCHG;

		break;
	}

	// So that we can detect a change when erasing it
	wp->w_lastdismarkp = wp->w_markp;
	wp->w_lastdismarko = wp->w_marko;
	wp->w_lastdismarkpo = wp->w_markpartoff;
}

/* Colour the cursor in window wp using pen.
 *
 * This is used both for setting the cursor colouring (ie show where it is) and
 * for unsetting a previous colouring before an update is done.
 *
 * When 
 *
 * wp	The window
 *	wp->w_lastcurcol is the cursor position
 * pen	How to colour - non current windows different from the current one
 * find	Find things out about the current cursor and store in wp.
 *	If FALSE use the previously stored values from wp.
 *
 * The complication is hex option where the cursor can occupy up to 11 positions
 * (think: 4 byte UniCode character = 8 hex digits + 3 separating spaces).
 *
 * This is worth doing for showing 2 tracked mode windows onto the same
 * buffer - one with the hex option.
 *
 * Extra complication: not all of the 11 bytes might be displayed in the window.
 *
 * Double width UniCode characters are not a problem - just need to colour
 * the first byte.
 */
void	setCursorAttr(Window* wp, VidAttr pen, int find)
{
	int	chrLen;		// Number of screen positons
	VidAttr* va;		// Attribute to set
	int	col = wp->w_lastcurcol;

	TERMOUTTRACE((TermDebFD, "setCursorAttr('%s', pen=%#x, %d) disrow=%d\n", wp->w_bufp->b_bname, pen, find, wp->w_disrow))

	TERMOUTTRACE((TermDebFD, "col=lastcurcol=%d leftcol=%d lastcurlen=%d discol=%d part=%d CP=%p\n", wp->w_lastcurcol, wp->w_leftcol, wp->w_lastcurlen, wp->w_discol, wp->w_partoff, lgetcaddr(wp->w_dotp, wp->w_doto)))

	if(find) {
		chrLen = 1;	// Default 1 char position

		if(wp != curwp && (wp->w_opts & WO_HEX) /* && (wp->w_bufp->b_flag & BSYNC_CURSOR) */ ) {
			// Here cos wp is not the current one AND it is displayed in hex AND
			// there is another window displaying the buffer.
			// Display the nybles for the whole character in PEN_NAW_CURSOR.

			/* Need to determine the positions of the byte under the cursor.
			 * Since this is a hex window the cursor will be on a nybble,
			 * look in the virtual screen to see where it is: check if the
			 * previous character is a space.
			 * A complication is that the previous character might be to the
			 * left of the (scrolled maybe) window, in which case look right.
			 *
			 * Possible very narrow windows make this harder.
			 *
			 * Do not use w_doto as it is a hard calculation.
			 */
			VideoC*	cp = vscreen[wp->w_disrow]->v_text;

			TERMOUTTRACE((TermDebFD, "look@ '%c'%#x'%c%c'%#x.%#x col=%d leftcol=%d\n", cp[col - 1], cp[col - 1], cp[col], cp[col + 1], cp[col], cp[col + 1], col, wp->w_leftcol))

			/* This was a pain to get right.
			 * The problem is that the left most column could:
			 * * the first nybble of a the byte of start of character (possibly multi byte)
			 * * the first bybble of a continuation byte
			 * * the second nybble, followed space then by a continuation byte or a start byte
			 * * be a space between bytes, followed by a continuation byte or a start byte;
			 *   this can only happen by moving the window left/right - fix this elsewhere by
			 *   making the first column of a hex window start on a first nybble.
			 *
			 * Need to & 0xFF as vtputc() does not clear out top bits.
			 */
			int	dont = 0;	// Don't search - too narrow

			// Find first nybble position on the screen
			if(col - 1 >= wp->w_leftcol) {
				if((cp[col - 1] & 0xFF) != ' ')
					col--;

				chrLen = 2;
			} else {	// At start of visible line
				if(col + 2 < wp->w_leftcol + wp->w_ncolumns) {
					if((cp[col + 2] & 0xFF) == ' ')
						chrLen = 2;	// At start of byte
					else
						if((cp[col + 1] & 0xFF) == ' ')
							col += 2;
				} else
					dont = 1;	// Too short
			}

			TERMOUTTRACE((TermDebFD, "START '%c%c' @ col=%d dont=%d\n", cp[col], cp[col + 1], col, dont))

			if( !dont && (wp->w_bufp->b_mode & MDUTF8)) {

				/* cp[col] is on the first nybble of a byte.
				 * We need to do something with multi byte characters, so
				 * first decide if this is a bulti byte character:
				 */
				if( ! (cp[col] >= '0' && cp[col] <= '7')) {

					BYTE*	chr = lgetcaddr(wp->w_dotp, wp->w_doto);

					// Find start of multi-byte character
					while(cp[col] == '8' || cp[col] == '9' || cp[col] == 'A' || cp[col] == 'B') {
						if(col - 3 >= wp->w_leftcol) {
							col -= 3;	// Look back to the byte before
						} else {
							// Incomplete bytes displayed at starts
							dont = 1;
							TERMOUTTRACE((TermDebFD, "INCOMPLETE\n"))
							break;
						}
					}

					TERMOUTTRACE((TermDebFD, "CHARstart '%c%c' @ col=%d\n", cp[col], cp[col + 1], col))

					// Find how many screen positions:
					UTF8_BACK_FIND_START(chr);
					chrLen = CHAR_LEN(*chr) * 3 - 1;
				}
			}

			/* Limit the computed length so that it does not overflow the window.
			 * This will incorrectly highlight a $ that is on a space in
			 * the middle of several hex bytes - but good enough.s
			 */
			if(col + chrLen >= wp->w_leftcol + wp->w_ncolumns)
				chrLen = wp->w_leftcol + wp->w_ncolumns - col;

			/* Could not find enough to know what to highlight,
			 * just highlight one character position (nybble).
			 * This will affect where the start of the nybbles is off
			 * the left side of the window.
			 * I could prob do a better job and highlight the nybbles
			 * that are visible, but this is a corner case of something
			 * infrequently used - not worth the effort.
			 */
			if(dont) {
				chrLen = 1;
				col = wp->w_lastcurcol;
			}
		}

		wp->w_lastcurlen = chrLen;
		wp->w_lastcurcol = col;
	}

	va = &vscreen[wp->w_disrow]->v_attr[col];

	chrLen = wp->w_lastcurlen;

	// Colour those bytes:
	for( ; --chrLen >= 0; va++)
		*va = va_set_width(pen, va_get_width(*va));

	vscreen[wp->w_disrow]->v_flag |= VFCHG;
}



/* Make sure that the display is right. This is a three part process. First,
 * scan through all of the windows looking for dirty ones. Check the framing,
 * and refresh the screen. Second, make sure that "currow" and "curcol" are
 * correct for the current window. Third, make the virtual and physical
 * screens the same.
 *
 * force: force update past type ahead?
 */
void	update(int force)
{
	register Window* wp;
	register int	check = 1;
	static	Window*	lastWind;	// Was was curwp when last here
	int		i;

	// This can happen if running a macro started from the command line with @
	if(TTopened == TTneverOpened)
		return;

	if(WasSuspd) {
		/* Window size could be changed or become otherwise unknown.
		 * This might have happened when we were NOT waiting for
		 * input from the terminal.
		 */
		sgarbf = TRUE;

		// Now is a safe time to reallocate screen buffers
		WasSuspd = FALSE;
		vtinit(1);
	} else {


#if	TYPEAH
		if(force == FALSE && typahead(FALSE))
			return;
#endif
#if	VISMAC == 0
		if(force == FALSE && kbdmode == PLAY)
			return;
#endif
	}

	/* If screen is garbage, redraw it completely.
	 * Force HARD & other update flags to do this.
	 * Not WFFORCE - try to preserve framing.
	 */
	if(sgarbf != FALSE) {
		updgar();
		for(wp = wmessp; wp != NULL; wp = wp->w_wndp)
			wp->w_flag = WFHARD | WFMODE | WFNWID;
	} 

	/* Look through all the windows restore the cursor position attributes.
	 * On hard refresh - set to default.
	 * w_lastcurcol is needed otherwise it is hard to correctly get the 2nd nybble (HEX option)
	 */
	for(wp = wmessp; wp != NULL; wp = wp->w_wndp) {
		// If repainting - do not reset; in extreme cases (window/screen size change)
		// we could write somewhere that we don't want to.
		if(wp->w_flag & (WFHARD | WFFORCE))
			continue;

		setCursorAttr(wp, va_no_width(wp->w_cursattr), 0);

		// Leave all but the message line alone ?
		// This works nicely since it is the first one looked at.
		if(DisplayMessageLineOnly)
			break;
	}

	/* If the window has changed, then the edges will need updating */
	updateWhat = 0;
	if(lastWind != curwp) {
		updateWhat = UW_EDGE;
		lastWind = curwp;
	}

	// Clear notion of needing scrolling
	for(i = 0; i < AllocLines; i++)
		vscreen[i]->v_scrolled = 0;
	WindowScrolled = 0;

	// Clear note of where any edges may have been changed
	if(edgeSet) {
		memset(edgeAt, 0, (size_t)(AllocLines * VID_MAXSEPS) * sizeof(short));
		edgeSet = 0;
	}

	// Set to something that is way out enough to be updated:
	leftMostEdge = (short)(AllocColumns + 1);
	rightMostEdge = 0;

	// Show if "." is not within displayed width of the current window
	TERMOUTTRACE((TermDebFD, "update() . in '%s' is (%d, %d) offset=%ld %s\n", curwp->w_bufp->b_bname, curwp->w_disrow, curwp->w_discol + curwp->w_partoff, curwp->w_leftoffset,
		((curwp->w_discol + curwp->w_partoff > AllocColumns) ? "**** . is off RH side ****" : "")))

again:
	updatedotpos(curwp);

	/* update any windows that need refreshing */
	for(wp = wmessp; wp != NULL; wp = wp->w_wndp) {
		/* Temp store for lastflag
		 * The point is that this contains CFCPCN that prevents the changing
		 * of w_leftoffset - must allow this with message window.
		 */
		int	savelastflag = lastflag;
		if(wp == wmessp)
			lastflag = 0;

		// Get the pallet that that we paint this window with:
		inkWell = pens + ((wp == wmessp ? MsgWindowPallet : PAL_WINDOW) * INKS);

		unmarkmark(wp);

		// If this window is viewing the same buffer as the current window
		// and we sync cursor position, ensure that the cursor will be positioned in the window.
		// Also check if doing a full display.
		// Also check if the window is being tracked, a vertical movement in
		// the other window must be reflected here.
		if((wp->w_bufp == curwp->w_bufp && wp->w_bufp->b_flag & BSYNC_CURSOR) || (wp->w_flag & WFHARD) || (wp->w_opts & WO_TRACK))
			updatedotpos(wp);

		/* Need to update mode line if it is showing the position
		 * and it might have changed.
		 */
		if((wp->w_opts & WO_POSITION_SHOW) &&
		   ((wp->w_flag & (WFFORCE | WFMOVE | WFEDIT | WFHARD)) |
		    (wp->w_lastcurcol != wp->w_discol + wp->w_partoff))
		  )
			wp->w_flag |= WFMODE;

		if(wp->w_flag) {

			// Number of top line changed, or shown status changed ?
			/* if the window has changed, service it */
			if(reframe(wp))	{		/* check the framing */
				wp->w_flag |= WFNWID;
				updateWhat |= UW_TEXT;
			}

			// Need to check how many characters to display the line # ?
			// Numbers changed UW_NUMB is done by checkNumberWidth()
			if(wp->w_flag & WFNWID)
				checkNumberWidth(wp);

			// If we can get away with only looking at one line then do so:
			if((wp->w_flag & ~(WFMODE)) == WFEDIT)
				updone(wp);	/* update EDITed line */
			else if(wp->w_flag & ~(WFMOVE))
				updall(wp);	/* update all lines */
			CheckForScroll(wp);

			/* Mode line need updating ?
			 * That is, if it has one
			 */
			if((wp->w_flag & WFMODE) && (wp->w_attrib & WATT_MODELINE)) {
				updateWhat |= UW_MODE;
				modeline(wp);	/* update modeline */
				updedge(wp);	/* Any edge */
			}

			wp->w_flag = 0;
			wp->w_force = 0;
		}

		// Update the window's notion of how many lines in the buffer (scroll detection):
		wp->w_lastnumLines = wp->w_bufp->b_numLines;

		markmark(wp);

		if(wp == wmessp) {
			lastflag = savelastflag;

			// Colour text in the message line ?
			if(msgColourFrom >= 0) {
				int	cl = msgColourLen;
				int	cf = msgColourFrom - (int)wmessp->w_leftoffset;
				while(cf < AllocColumns && --cl >= 0)
					vscreen[AllocLines - 1]->v_attr[cf++] = penDefaults[PEN_MSGL_COLOUR];
			}
		}

		// Leave all but the message line alone ?
		if(DisplayMessageLineOnly)
			break;
	}

	/* Previous edge updates are needed to set the edge character to space,
	 * but as edges are shared: that may also set the left edge of the current
	 * window to space.
	 * Rerun on the current window.
	 * Don't if entering a prompt; kind of is DisplayMessageLineOnly.
	 */
	if(curwp != wmessp)
		updedge(curwp);


	/* Look through all the windows highlight the cursor position in all but
	 * the current window. Save the old attributes in w_cursattr.
	 *
	 * Don't set standout mode on the current window, that is where the cursor is and
	 * if that is in reverse the effect is reverse of reverse - which reduces visibility.
	 *
	 * The cursor has a special colour in all except the current window.
	 * Problem with current window cursor: it is displayed in the system default colour,
	 * this could be inverse (standout).
	 * If it is over a character in standout (control character [CC]) it then becomes invisible,
	 * moving it over another CC and the screen looks exactly the same.
	 * To avoid this: check the current character and if it is standout choose a different pen.
	 * This is not perfect: with many CCs it may become one of many,
	 * but this way its position becomes clear when it is moved.
	 *
	 * Just to make it more complicated: we must preserve the width attribute otherwise
	 * the display of double width characters under the cursor is screwed.
	 *
	 * If a non-current window has the hex option, colour all the nybbles that make
	 * up the current character.
	 *
	 * Start with wheadp, not wmessp - don't want a highlight old cursor on the message line.
	 */
	for(wp = wmessp; wp != NULL; wp = wp->w_wndp) {

		// Actually don't do any of this if we are to just update the message line:
		if(DisplayMessageLineOnly)
			break;

		// Save the cursor position that we are about to write, so that
		// we can reset it later on.
		wp->w_lastcurcol = wp->w_discol + wp->w_partoff;

		VidAttr* va = &vscreen[wp->w_disrow]->v_attr[wp->w_lastcurcol];
		wp->w_cursattr = *va;

		// No further colouring for the message line:
		if(wp == wmessp)
			continue;

		VidAttr	pen = PEN_NAW_CURSOR;

		/* What is being avoided here is the cursor being on, or next to, a character
		 * in standout mode. This is often indistinguishable from standard cursor
		 * 'colouring' - thus it is hard to know where the cursor is.
		 * If this is detected use an alternate set of pens for the cursor.
		 *
		 * How far away counts as near ... scan the entire window. This is not
		 * hard, look at the window line flags.
		 */
		if(wp == curwp) {
			int	wline;
			int	wstandout = 0;

			// Look at standout flags on each line
			for(wline = 0; wline < wp->w_ntrows; wline++)
				wstandout |= wp->w_lineFlags[wline];

			// Set alternate pen if a line has standout
			if(wstandout & LF_StandOut)
				pen = PEN_CURSOR_ALT;
			else
				pen = PEN_CURSOR;
		}

		TERMOUTTRACE((TermDebFD, "SET-CURS '%s' %#x\n", wp->w_bufp->b_bname, pens[pen]))
		setCursorAttr(wp, pens[pen], 1);

		wp->w_lasttop = wp->w_toplinenum;
	}

#if	MEMMAP
	/* update the cursor and flush the buffers */
	movecursor(curwp->w_disrow, curwp->w_discol - lbound);
#endif

	/* Note where we will want to place the cursor.
	 * updupd() [ via updateline() ] might be able to optimise getting the
	 * cursor to the right position.
	 */
	wantRow = curwp->w_disrow;
	wantCol = curwp->w_discol + curwp->w_partoff;

	/* update the virtual screen to the physical screen */
	updupd(force);

	/* If its time to age the buffers then age the buffers. the order
	 * is necessary because the old buffers message is more important
	 * than the previous contents of the message line.
	 *
	 * The idea of this "goto again" filth is that we want the "X" to
	 * be displayed on the mode line if buffers are out of date, we
	 * really need to call this routine again. To cut down on the amount
	 * of work, we can jump backwards up the routine a little way.
	 *
	 * If we don't do this then, eg after return from suspend, the user will
	 * be shown a blank screen with a prompt saying that a buffer is
	 * out of date - not nice.
	 */
	if(check && !DisplayMessageLineOnly && agetime == TRUE && checkage(FALSE)) {
		check = 0;	/* Ugh !  What filth !  */
		goto again;
	}

	/* Check to see if checkage put out a message. If it did then the
	 * contents of mlthere will have changed.
	 */

	// mlwrite("w_lefttextcol %d w_leftoffset %ld doto=%ld part=%d margin %d scrsize %d AllocColumns %d w_ncolumns %d",
	//	curwp->w_lefttextcol, (long)curwp->w_leftoffset, (long)curwp->w_doto, curwp->w_partoff,
	//	term.t_margin, term.t_scrsiz, AllocColumns, curwp->w_ncolumns);

	/* Update the cursor and flush the buffers */
	TERMOUTTRACE((TermDebFD, "Restore . to (%d, %d)\n", curwp->w_disrow, curwp->w_discol + curwp->w_partoff))
	movecursor(curwp->w_disrow, curwp->w_discol + curwp->w_partoff);

	TTflush();
}

/* Check if the window is suitable for doing a refresh with physical scrolling
 * of the terminal.
 *
 * There are 4 different ways in which the screen can scroll. Subgroup in 2, within
 * each the scroll is up or down:
 * top	A scroll at the top of the window. This will be as a result of the line
 *	number at window top having changed; prob move up/down in the buffer but
 *	could also be a result of lines above being inserted/removed.
 * mid	A scroll starting in the middle of the window. This as a result of lines
 *	being inserted/deleted in the middle of the window.
 *	The most frequent cause will be newline (\r) & line delete; next most frequent
 *	yank and kill-region.
 *
 * You can get a change that affects both (window top line number & buffer line count),
 * eg: yank/kill-region where the top line that was yanked *is* above the window or the
 * top line killed *was* above the window.
 *
 * Both forms of scroll affect lines from their start point to the bottom line
 * in the window.
 *
 * This updates vscreen[XX]->v_scrolled with the amount that the line has to
 * move up or down to match what there is on the screen (pscreen). The line(s)
 * where the text are moved from will be filled with spaces if they are not replaced
 * by lines scrolled in from elsewhere.
 *
 * Eg: a 4 line window (as shown in the 1st column below) has a newline inserted
 * after 'Bil'. Ben & Weed scroll down; Shed scrolls off the end of the window; thus
 * it ends up as shown in the 2nd column.
 * The v_scolled value to do this is in the 3rd column.
 *
 * 	Bill	Bil	 0
 *	Ben	l	 1
 *	Weed	Ben	 1
 *	Shed	Weed	 0
 *
 * Line after Bill removed:
 * 	Bill	BillBen	 0		This line redrawn after
 *	Ben	Weed	 0
 *	Weed	Shed	-1
 *	Shed	Pot	-1		Pot would have to be drawn after the scroll
 *
 * Scroll up (eg move-window-down)
 * 	Bill	Ben	 0
 *	Ben	Weed	-1
 *	Weed	Shed	-1
 *	Shed	Pot	-1		Pot would have to be drawn after the scroll
 *
 * Scroll up (eg move-window-up)
 *	Bill	Garden	 1		Garden would have to be drawn after the scroll
 *	Ben	Bill	 1
 *	Weed	Ben	 1
 *	Shed	Weed	 0
 *
 * In all the examples, if the movement is more than the height of the window
 * then no scrolling is done.
 *
 * Note that if a line has a movement of 0, it could mean that it is completely
 * untouched by all of this; the line could, however, be overwritten by another
 * line.
 *
 * The use of ldif below assumes that it is a signed type.
 *
 * The philosophy is to try to deduce what has changed in the current window.
 * DoScrolled() actually makes the physical change on the screen (ie for all
 * windows). It doesn't really matter if CheckForScroll() & DoScrolled() get it
 * right as long as DoScrolled() ensures that pscreen is an accurate representation
 * of what is on the screen - as then updupd() & updateline() will correct any mistakes.
 * In practice it gets it right most of the time
 * Regular fixups needed are line numbers (due to top change), updupd() knows about them
 * after scrolling and tries to do them efficiently.
 */
void	CheckForScroll(Window* wp)
{
	BufLns	topldif;	// Difference in top line #
	BufLns	ldif;		// Difference in line #. -ve means scrolled up
	BufLns	ldifa;		// Absolute value of that.
	int	i;
	int	EndLine;

	/* Unless ldif and ldifa are small numbers (screen size) we will abandon.
	 * However they need to be able to hold the max # lines a buffer can hold.
	 * Thus lots of casts to int.
	 */


	LineLn	LnoAdj = wp->w_numwidth - wp->w_leftoffset;	// Adjust due to line # width
	int	iLnoAdj;	// Int version
	if(LnoAdj < 0 || LnoAdj >= AllocColumns)
		LnoAdj = 0;

	iLnoAdj = (int)LnoAdj;	// Use this - int is better than long/whatever

	// Only if it takes up the width of the whole screen
	if(wp->w_ncolumns != AllocColumns)
		return;

	// Terminal can't scroll or we have been asked not to, don't bother:
	if( !TTdoes(TT_DOES_SCROLL) || !ttDoScroll)
		return;

	topldif = wp->w_lasttop - wp->w_toplinenum;	// -ve means lines deleted above

	// Work out if bottom scroll, compensate for line change above
	ldif = wp->w_bufp->b_numLines - wp->w_lastnumLines - topldif;

	// Scroll at the bottom:
	if(ldif != -topldif) {
		/* Looks like scroll at the bottom of the window after line insert/delete
		 * or text yanked/killed.
		 * This sort of change will most likely have happened due to an edit
		 * in the part of the buffer that the window displays. But don't count
		 * on it, eg if a buffer is shown in more than one window.
		 *
		 * If this has happened then the buffer line count will have changed.
		 * This check will fail on a complicated update, eg lines inserted in
		 * one place and removed elsewhere
		 *
		 * When doing compares - need to avoid line number; also line shorter
		 * Number could be scrolled off
		 *
		 * Don't bother comparing v_attr[], these are prob right but if not
		 * will get fixed up later - maybe it will cost a lot; but won't
		 * (I think) happen too often.
		 *
		 * First: how many bytes to compare:
		 */
		size_t	vidBytes = (size_t)(AllocColumns - iLnoAdj) * sizeof(VideoC);

		if((ldifa = ldif) < 0)
			ldifa = 0 - ldif;
		if(ldif == 0 || ldifa > wp->w_ntrows - 1)
			return;

		// Depending on scroll up/down it is vscreen or pscreen that is the top
		// comparison. Set pointers and collapse the 2 loops to 1.

		if(ldif > 0) {
			// More lines in the buffer
			for(i = wp->w_toprow + wp->w_ntrows - 1; i >= wp->w_toprow + (int)ldifa; i--) {

				// Can we get the current line by scrolling one down ?
				if(memcmp(vscreen[i]->v_text + iLnoAdj, pscreen[i - (int)ldifa]->v_text + iLnoAdj, vidBytes)) {
					i = i - (int)ldifa + 1;
					break;
				}
			}

			TopScrollLines = i;	// Top of redraw line # range
			EndLine = wp->w_ntrows + wp->w_toprow - (int)ldifa;
		} else {
			// Fewer lines in the buffer
			for(i = wp->w_ntrows + wp->w_toprow - 1; i >= wp->w_toprow + (int)ldifa; i--) {

				// Can we get the current line by scrolling one up ?
				if(memcmp(vscreen[i - (int)ldifa]->v_text + iLnoAdj, pscreen[i]->v_text + iLnoAdj, vidBytes))
					break;
			}

			i += 2;
			TopScrollLines = i - (int)ldifa;
			EndLine = wp->w_ntrows + wp->w_toprow;
		}

		if(DontReallyWantToScroll(i, EndLine, (int)ldif, iLnoAdj))
			return;

		// Mark all lines from i-1 to the bottom to be scrolled down
		BottomScrollLines = EndLine;
		for(; i < EndLine; i++)
			vscreen[i]->v_scrolled = (int)ldif;

		// Display width of line # is a good flag to use - do we need to scan to update line numbers ?
		PaintScrollLines = (int)LnoAdj;

		WindowScrolled = 1;

		return;
	}

	// If the top line # is the same then it has not scrolled
	// No change, no scroll of the top:
	if(topldif == 0)
		return;

	if((ldifa = topldif) < 0)
		ldifa = 0 - topldif;

	/* Don't do unless preserve at least 1 row
	 * This also catches a move to somewhere very different, in which case
	 * we don't want to set WindowScrolled.
	 */
	if(ldifa > wp->w_ntrows - 1)
		return;

	/* Set the v_scrolled to be the line # difference (ldif) between now
	 * and last time.
	 *
	 * -ve means that the text has scrolled up - the old top line(s) off the top
	 * and something new come in the bottom.
	 *
	 * By coincidence this is what tcapscroll() wants later.
	 *
	 * This same area is the one where line numbers need to be checked:
	 */
	if(topldif < 0) {
		TopScrollLines = -(int)topldif;
		BottomScrollLines = wp->w_ntrows;
	} else {
		TopScrollLines = 0;
		BottomScrollLines = wp->w_ntrows - (int)topldif;
	}

	if(DontReallyWantToScroll(TopScrollLines, BottomScrollLines, (int)topldif, (int)LnoAdj))
		return;

	for(i = TopScrollLines ; i < BottomScrollLines; i++)
		vscreen[i + wp->w_toprow]->v_scrolled = (int)topldif;

	// Display width of line # is a good flag to use - do we need to scan to update line numbers ?
	PaintScrollLines = (int)LnoAdj;

	WindowScrolled = 1;
}

/* If the scroll is for most of the window we might be better off not doing it.
 *
 * The scenario is where the lines that we will need to redraw contain much the
 * same text as they already contain. This can happen where there is a lot of
 * repetition in lines, eg log files.
 *
 * Work out the cost in repainting in the scoll & non scroll cases.
 * This ignores: single/multi byte chars; differences in escape sequences
 * in the 2 cases.
 *
 * Ignore attributes for this purpose.
 *
 * top		First line # that moves intact - lines inserted/deleted above it
 * bottom	Last  line # that moves intact - lines deleted/inserted below it
 * ScrollLines	# Lines scroll up/down. +1 means scroll down, ie line 1 is now line 2
 * LnoWidth	Width of line # at start of line
 *
 * Not entirely sure of this: burning CPU to save some terminal output.
 *
 * Return true if to not bother with a scroll.
 */
int	DontReallyWantToScroll(int top, int bottom, int ScrollLines, int LnoWidth)
{
	int	scrollCost = 0;
	int	noScrollCost = 0;

	// Where do we check:
	int	topAff;		// Top line affected
	int	botAff;
	int	topOpen;	// Top line opened
	int	botOpen;

	int	cols = AllocColumns - 1;

	int	dont;		// Eventual decision

	// How effective is this check ?
	ST_scrollCandidates++;

	if(ScrollLines > 0) {
		topAff = top;
		botAff = bottom - ScrollLines + 1;
		topOpen = bottom;
		botOpen = bottom - ScrollLines + 1;
	} else {
		topAff = top + ScrollLines;
		botAff = bottom - 1;
		topOpen = bottom + ScrollLines;
		botOpen = bottom - 1;
	}


	// fprintf(stderr, "DontReallyWantToScroll(%d, %d, %d, %d) ta=%d ba=%d to=%d bo=%d\n",
	//	top, bottom, ScrollLines, LnoWidth,
	//	topAff,	botAff,	topOpen, botOpen);

	// What if we do not scroll ?
	for( ; topAff <= botAff; topAff++) {
		VideoC*	pc  = pscreen[topAff]->v_text + LnoWidth;	// Start comparing after the line #
		VideoC*	vc  = vscreen[topAff]->v_text + LnoWidth;
		VideoC*	pce = pscreen[topAff]->v_text + cols;		// Right most character to compare
		VideoC*	vce = vscreen[topAff]->v_text + cols;

		// Scan left to right
		for( ; pc < pce ; pc++, vc++) {
			// Different ? Stop scanning
			if(*pc != *vc) {
				pc--;	// Point to last same character
				break;
			}
		}

		// Scan right to left
		// Find where the difference ends
		// Allow trailing spaces since they are cheaply cleared
		for( ; pc < pce ; pce--, vce--)
			if(*pce != *vce && *vce != ' ') {
				pce++;
				break;
			}

		// Note the length of the difference
		noScrollCost += (int)(pce - pc);
	}

	// Calculate cost of scrolling.
	// This is the cost of painting a line that will be blank, so skip over
	// leading/trailing spaces:
	for( ; topOpen <= botOpen; topOpen++) {
		VideoC*	vc  = vscreen[topOpen]->v_text + LnoWidth;
		VideoC*	vce = vscreen[topOpen]->v_text + cols;

		// Find first non space:
		for( ; vc < vce; vc++)
			if(*vc != ' ') {
				vc--;
				break;
			}

		// Find last non space:
		for( ; vc < vce; vce--)
			if(*vce != ' ') {
				vce++;
				break;
			}

		// Note the length of the difference
		scrollCost += (int)(vce - vc);
	}

	// Relative costs is what makes the decision ?
	dont = scrollCost > noScrollCost;
	ST_scrollDont += dont;

	// fprintf(stderr, "DontReallyWantToScroll -> %d (cost: scroll %d > ns %d)\n", dont, scrollCost, noScrollCost);fflush(stderr);

	return(dont);
}

/* Work out how wide the numbers (WO_NUMBER) are.
 *
 * This varies depending on how many digits, will be 0 if not line number option.
 * Set w_lefttextcol - ie where the first text character is put.
 * Also set w_numwidth - a convenience variable.
 * Set WFHARD if the width changes, remove WFNWID.
 */
void	checkNumberWidth(Window* wp)
{
	// Show number as 'nn ' - ie 1 space following, but nn is variable width
	// Find line # of bottom line of screen:
	BufLns	b = wp->w_toplinenum + wp->w_ntrows - 1;

	// With hex option should be offset from start of file TODO

	int	digits = 0;
	if(wp->w_opts & WO_NUMBER) {
		// Quick: how many digits, speed bias to smaller numbers
		if(wp->w_opts & WO_HEX)
			digits = b >= 0x10000 ? (b >= 0x1000000 ? (b >= 0x10000000 ? 8 : 7) : (b >= 0x100000 ? 6 : 5) ) :
				(b >= 0x100 ? (b >= 0x1000 ? 4 : 3) : (b >= 0x10 ? 2 : 1));
		else
			digits = b >= 10000 ? (b >= 1000000 ? (b >= 10000000 ? (b >= 100000000 ? (b >= 1000000000 ? 10 : 9) : 8) : 7) : (b >= 100000 ? 6 : 5) ) :
				(b >= 100 ? (b >= 1000 ? 4 : 3) : (b >= 10 ? 2 : 1));

		digits++;	// Following space
	}

	// Is the requirement what the display is set as ? If not redo margin:
	if(wp->w_numwidth != digits) {
		wp->w_lefttextcol = wp->w_leftcol + digits;
		wp->w_numwidth = digits;
		wp->w_flag |= WFHARD;
		updatedotpos(wp);	// Get the cursor to the right position
		updateWhat |= UW_NUMB;
	}

	wp->w_flag &= ~WFNWID;
}

/* The purpose of this is to fix times when a window does not display any lines of text.
 * This can happen when there is text, but the cursor is displaying the last line
 * in the buffer as the top line of the window.
 *
 * Eg: yank some lines into an empty buffer. Without this the yanked lines are all invisible
 * in lines above the window. This is benign but, to the user, slightly puzzling as to why
 * the window is empty.
 */
void	displaySomething(void)
{
	if(curwp->w_dotp == curbp->b_abotline && curwp->w_topLinep == curbp->b_abotline &&
	    lback(curwp->w_dotp) != curbp->b_btopline) {
		curwp->w_dotp = lback(curwp->w_dotp);
		reframe(curwp);
		curwp->w_dotp = lforw(curwp->w_dotp);
	}
}

/* reframe:	check to see if the cursor (dot) is on in the window
 *		and re-frame it if needed or wanted.
 *
 * Reframing means resetting the top line (w_topLinep) so that "." is visible
 * in the window.
 *
 * Most of Emacs is blisfully ignorant of framing, but a few commands do it:
 * previous-page next-page move-window-up move-window-down split-window-vertically
 * ldelnewline() lnewline()
 *
 * Return TRUE if we had to re-frame.
 */
int	reframe(Window* wp)
{
	register LINE*	lp;
	register int	i;
	int		TopLineNumCalc = 1;	// Need to calculate/scan to determine top line #

	/* Force reframe if "." at bottom but on the top window line.
	 * This can happen if insert file into empty buffer (& other cases).
	 */
	if(wp->w_dotp == wp->w_bufp->b_abotline && wp->w_dotp == wp->w_topLinep)
		wp->w_flag |= WFFORCE;

	/* If not a requested reframe, check for a needed one.
	 * Start looking at the top line in the window,
	 * look down for as many lines as there are in the window.
	 * If we see the cursor: great, got it.
	 */
	if((wp->w_flag & WFFORCE) == 0) {
		lp = wp->w_topLinep;
		for(i = 0; i < wp->w_ntrows; i++) {
			/* Should not come across the first line in the narrow
			 * unless we have just narrowed. If we have restart the scan:
			 */
			if(lp == wp->w_bufp->b_btopline && i != 0 &&
			   wp->w_bufp->b_btopline != wp->w_bufp->b_abotline) {
				wp->w_toplinenum += 1;
				wp->w_topLinep = lp;
				i = -1;
			}

			/* if the line is in the window, no reframe */
			if(lp == wp->w_dotp)
				return(FALSE);

			/* if we are at the end of the (narrowed) buffer, reframe */
			if(lp == wp->w_bufp->b_abotline)
				break;

			/* on to the next line */
			lp = lforw(lp);
		}
	}

	/* Reached here, we need a window refresh,
	 * either because one requested or "." is no longer in the window.
	 * Reposition top & "."
	 */

	// Does the cursor need to be on a particular line number from the top of the window ?
	i = wp->w_force;
	/* how far back to reframe? */
	if(i > 0) {
		/* Lines from window top.
		 * Only one screen worth of lines max
		 */
		if(--i >= wp->w_ntrows)
			i = wp->w_ntrows - 1;
	} else if(i < 0) {
		/* Negative window line # - count from bottom of the window
		 * Don't care if goes -ve; just does nothing below
		 */
		i += wp->w_ntrows;
	} else {
		/* Position not specified, maybe forced reframe.
		 *
		 * There are 2 special cases, when "." is on the line above
		 * the current top of window or when "." is on the line below
		 * the last one shown in the window.
		 * If a special case then ScrollJumpBy ($jump or $jump-no-scroll) come into play.
		 * Values:
		 * 0	scroll to 1/2 the window height
		 * n	scroll by n lines
		 *	Thus if n==1 and "." was on the top line and moved up one
		 *	line - a scroll by 1 leaves "." on the new top line
		 * -n	scroll by window height - n
		 *	Thus if n==-1 and "." was on the top line and moved up one
		 *	line - the whole screen scrolls with the new "." on the
		 *	bottom line of the screen
		 * A value of n (or -n) greater than the window height is set to
		 * the window height.
		 *
		 * Also attempt to ensure as much of the buffer as possible is shown.
		 * This means avoiding too many non-lines taking up window space (ie
		 * lines below the end of buffer). This is not easy.
		 *
		 * $jump or $jump-no-scroll ?
		 * $jump only if the window can scroll: hardware does it, $scroll is TRUE,
		 * window is the only one to take the screen width.
		 */
		int	lns;
		int	scroll = (wp->w_ncolumns == AllocColumns && TTdoes(TT_DOES_SCROLL) && ttDoScroll) ?
				ScrollJumpBy : NoScrollJumpBy;

		// Work out how many lines to move
		if(scroll == 0) {
			lns = wp->w_ntrows / 2;
		} else {
			lns = scroll;
			if(lns < 0)
				lns = wp->w_ntrows + lns + 1;
			if(lns > wp->w_ntrows || lns < 1)
				lns = wp->w_ntrows;
		}

		// mlwrite("%#");
		// mlwrite("%#top=%D scroll=%d height=%d lns=%d .='%l'", wp->w_toplinenum, scroll, wp->w_ntrows, lns, wp->w_dotp);

		/* "." is top line in the buffer ?
		 * "." could be at b_linep if buffer empty. Beware "." at buffer bottom
		 */
		if((wp->w_dotp == wp->w_bufp->b_linep && lforw(wp->w_dotp) == wp->w_bufp->b_linep) ||
		   lback(wp->w_dotp) == wp->w_bufp->b_linep) {
			i = 0;
			TopLineNumCalc = 0;
			wp->w_toplinenum = 1;
			// mlwrite("%#. = line 1");
			
		} else
		/* Is "." just above the top line of window ?
		 * But NOT if the top line is the first in the buffer, else when
		 * we move to the bottom line we think we are at the top.
		 * Beware: not in special case of empty buffer.
		 *
		 * We can deduce the line number, so set it.
		 */

		if(wp->w_dotp == lback(wp->w_topLinep) &&
		   wp->w_topLinep != lforw(wp->w_bufp->b_btopline) &&
		   wp->w_topLinep != wp->w_bufp->b_btopline) {
			/* "." is the line above top
			 * No backup should needed below, let the top line become "."
			 * BUT: where lns ($jump/$jump-no-scroll) is NOT 1 we do need to backup.
			 */
// !!!
			TopLineNumCalc = 0;
			wp->w_toplinenum -= lns;
			i = lns - 1;
			// mlwrite("%#. 1 over top. i=%d wp->w_toplinenum=%D wp->w_disrow=%d", i, wp->w_toplinenum, wp->w_disrow);
		} else {
			/* "." is not visible in the window, or forced reframe
			 *
			 * Find the bottom line currently displayed
			 */
			for(lp = wp->w_topLinep, i = wp->w_ntrows; i > 1; i--, lp = lforw(lp)) {
				if(lp == wp->w_bufp->b_abotline)
					break;
			}
			// mlwrite("%#find bot i=%d hit-bottom=%d l='%l'", i, (int)(lp == wp->w_bufp->b_abotline), lp);

			// "." is the line just under bottom line ?
			if(wp->w_dotp == lforw(lp)) {

				wp->w_toplinenum += lns;

				i = wp->w_ntrows - lns;	// Show $scroll lines at bottom

				TopLineNumCalc = 0;
				// mlwrite("%#. under bottom topn=%d i=%d", wp->w_toplinenum, i);
			}

			/* Have another go at computing i - the # lines between "." and topLine ?
			 */
			if(TopLineNumCalc) {
				/* Is "." is the last line in the buffer ensure that
				 * it is near the bottom of the window.
				 * NB: either on virtual last line or true last line
				 */
				if(wp->w_dotp == wp->w_bufp->b_abotline ||
				   lforw(wp->w_dotp) == wp->w_bufp->b_abotline) {
					i = wp->w_ntrows - lns - LINES_EMPTY_BUFFER_END;
					// mlwrite("%#reset way i=%d",i);
				} else {
					/* If end of buffer is within a page of "." put the
					 * end of buffer at the bottom of the window.
					 */
					for(lp = wp->w_dotp, i = wp->w_ntrows; i > LINES_EMPTY_BUFFER_END && lp != wp->w_bufp->b_abotline; i--, lp = lforw(lp))
						;

					// int iwas = i;
					if(i > LINES_EMPTY_BUFFER_END)
						i = i - LINES_EMPTY_BUFFER_END;
					else
						i = wp->w_ntrows / 2;
					// mlwrite("%#scan to bot iwas=%d i=%d", iwas, i);
				}
			}
		}
	}

	/* Backup to new line at top of window.
	 * Recalculate the screen row that the cursor sits on.
	 */
	lp = wp->w_dotp;
	wp->w_disrow = wp->w_toprow;
	// mlwrite("%#Backup to top. i=%d w_ntrows=%d TopLineNumCalc=%d", i, wp->w_ntrows, TopLineNumCalc);
	while(i > 0 && lback(lp) != wp->w_bufp->b_btopline) {
		++wp->w_disrow;
		--i;
		lp = lback(lp);
	}

	/* and reset the current line at top of window */
	wp->w_topLinep = lp;
	wp->w_flag |= WFHARD | WFNWID | WFMODE;
	wp->w_flag &= ~WFFORCE;

	/* Find the line # of the first line in the window.
	 * If we know it - then avoid this potentially expensive calculation.
	 * For this we use the buffer, not the narrow.
	 */
	if(TopLineNumCalc)
		for(lp = lforw(wp->w_bufp->b_linep), wp->w_toplinenum = 1; lp != wp->w_topLinep; lp = lforw(lp))
			wp->w_toplinenum++;

	// This may be sufficient
	wp->w_lastdismarkp = NULL;

	return(TRUE);
}

/* updone:	update the current line	to the virtual screen
 *
 * wp: window to update current line in
 */
void	updone(Window* wp)
{
	register LINE* lp;	/* line to update */
	register int sline;	/* physical screen line to update */

	/* search down the line we want */
	lp = wp->w_topLinep;
	sline = wp->w_toprow;
	while(lp != wp->w_dotp) {
		++sline;
		lp = lforw(lp);
	}

	updline(wp, lp, sline);
}

/* updall:	update all the lines in a window on the virtual screen
 *
 * wp: window to update lines in
 */
void	updall(Window* wp)
{
	register LINE*	lp;	/* line to update */
	register int	sline;	/* physical screen line to update */

	/* search down the lines, updating them */
	lp = wp->w_topLinep;
	for(sline = wp->w_toprow; sline < wp->w_toprow + wp->w_ntrows; ++sline) {
		// Reached end of buffer/narrow ?
		if(EonLineB(wp->w_bufp, lp))
			lp = (LINE*)NULL;

		updline(wp, lp, sline);

		if(lp)
			lp = lforw(lp);
	}
}

/* A window edge marker has changed.
 * Note that this has happened so that the screen can be updated optimally.
 *
 * Row	Screen row of change
 * Col	Screen column of change
 */
void	NoteEdgeChange(int Row, int Col)
{
	short*	edges;			// Ptr to current edge slot
	int	nedge = VID_MAXSEPS;	// Number of edge slots left

	TERMOUTTRACE((TermDebFD, "NoteEdgeChange: Row=%d Col=%d\n", Row, Col))
	// Find where to mark that an edge exists
	for(edges = edgeAt + (Row * VID_MAXSEPS) ; *edges != 0; edges++) {
		if(*edges == Col)
			return;		// Already marked

		if(*edges > Col) {
			// Make room for the new entry
			if(nedge > 0)
				memmove((char*)&edges[1], (char*)edges, (size_t)(nedge * (int)sizeof(short)));
			break;
		}

		// Run out of room to store an edge ?
		if(--nedge == 0)
			return;
	}

	// Record where if there is room to store the change
	if(nedge > 0) {
		*edges = (short)Col;
		edgeSet = 1;

		// Keep track of left & right most edges:
		if(Col < leftMostEdge)
			leftMostEdge = (short)Col;
		if(Col > rightMostEdge)
			rightMostEdge = (short)Col;
	}
}

/* Update the edges of the window. The current window must look different from others.
 *
 * Set the attributes of the left/right edges of the current window to standout/underline,
 * this is to try to make it stand out - now that the cursor & mark remain showing in non
 * current windows it can be hard to quickly spot the current window.
 */
void	updedge(Window* wp)
{
	/* If the window does not go to the end of the screen, put a RH border
	 * and LH border to make the edge visible.
	 */

	char	EdgeChar = (char)(wp == curwp ? WindowSeparator : ' ');
	VidAttr	EdgeAttr = pens[wp == curwp ? PEN_CURR_BORDER : PEN_BORDER];
	int	RHcol = wp->w_leftcol + wp->w_ncolumns;
	int	LHcol = wp->w_leftcol - 1;
	int	Row;

	for(Row = wp->w_toprow; Row <= wp->w_toprow + wp->w_ntrows ; Row++) {
		VIDEO*	vp = vscreen[Row];
		VIDEO*	pp = pscreen[Row];

		if(RHcol < AllocColumns) {
			vp->v_text[RHcol] = EdgeChar;
			vp->v_attr[RHcol] = EdgeAttr;

			// If we just changed it: note it
			if(vp->v_text[RHcol] != pp->v_text[RHcol])
				NoteEdgeChange(Row, RHcol);
		}

		// Put EdgeChar in LH border of current window - if there is one
		if(LHcol > 0) {
			vp->v_text[LHcol] = EdgeChar;
			vp->v_attr[LHcol] = EdgeAttr;

			// If we just changed it: note it
			if(vp->v_text[LHcol] != pp->v_text[LHcol])
				NoteEdgeChange(Row, LHcol);
		}
	}
}


/* Write a Unicode string to the screen.
 * However: check that the character can be displayed, ie not off the left
 * hand side of the window.
 */
void	vtputUniStrLeftEdge(Window* wp, const UniChar* cp)
{
	int	setWrapColour = wp->w_curcol < wp->w_leftoffset;

	// Can we display any of that ?
	for( ; *cp != UNISTR_END; cp++, wp->w_curcol++) {
		if(wp->w_curcol >= wp->w_leftoffset) {
			// Display in strange colour
			VidAttr OldBrush = vBrush;
			if(setWrapColour)
				vBrush = inkWell[INK_WRAP];

			vtputUniStr(wp, cp);

			vBrush = OldBrush;	// Restore colour
			return;
		}
	}
}

/* Update the line lp in window wp.
 * sline is the screen line.
 *
 * If lp is NULL don't copy/update text - clear the line
 *
 * This takes care to not call vtputc()/vtpututf8() at a character position
 * outside the displayed width of the window. Ie before w_leftoffset columns
 * have been reached and after the RH end of the window.
 */
void	updline(Window* wp, LINE* lp, int sline)
{
	/* and update the virtual line */
	vscreen[sline]->v_flag |= VFCHG;
#if	LEGACY_DISPLAY
	vscreen[sline]->v_flag &= ~VFREQ;
#endif

	/* Flag pointer to the current line.
	 * Remove the standout flag for the line.
	 */
	CurLineFlagp = wp->w_lineFlags + sline - wp->w_toprow;
	*CurLineFlagp &= (Lflags)~(LF_StandOut);

	vtmove(sline, wp->w_leftcol);
	wp->w_curcol = 0;

	if(lp) {
		/* The code below was from updatedotpos() but copied here to avoid too much
		 * function call over head.
		 */

		BYTE*	up = lgetcaddr(lp, 0);
		BYTE*	l_end = lgetcaddr(lp, lused(lp));	// Byte address just beyond Line End 
		VIDEO*	vp = vscreen[sline];

		if(wp->w_opts & WO_NUMBER) {
			// Show line numbers
			char	lnBuf[15];
			int	wid = wp->w_numwidth - 1;	// Width of digits & trailing space
			vBrush = pens[PEN_LINE_NUMBERS];

			// Show offset from start of file, as a hex number
			// TODO

			// Normal hex/decimal line numbers
			BufLns	ln = wp->w_toplinenum + sline - wp->w_toprow;
			sprintf(lnBuf, (wp->w_opts & WO_HEX ? "%*lX " : "%*ld "), wid, (long)ln);

			// Output the number, testing carefully that it is within the left/right margins
			int	i;
			for(i = 0; lnBuf[i]; i++) {
				if(wp->w_curcol >= wp->w_leftoffset && wp->w_curcol < wp->w_leftoffset + wp->w_ncolumns)
					vtputc(wp, lnBuf[i]);
				else	// Don't output, but increment where we would have output to
					wp->w_curcol++;
			}

			/* Reset the pen.
			 * Note that this leaves the space after the number in the line number colour.
			 * OK: it is a space so no one will notice but when line numbers are being added
			 * and just the line numbers inserted - this will result in fewer pen changes.
			 */
			vBrush = inkWell[INK_PLAIN];
		}

		while(up < l_end) {
			int	c;	// Character being looked at

			if(wp->w_curcol >= wp->w_leftoffset + wp->w_ncolumns) {
				/* Don't try to display stuff off the RH side of the window.
				 * Highlight the last character on the line to show that something follows.
				 * This also happens on the mode line, but this is benign.
				 */
				vp->v_attr[wp->w_leftcol + wp->w_ncolumns - 1] = inkWell[INK_WRAP];

				if(vp->v_text[wp->w_leftcol + wp->w_ncolumns - 1] == ' ')
					vp->v_text[wp->w_leftcol + wp->w_ncolumns - 1] = ShiftedTextMarker;

				break;
			}

			c = *up;

			/* Hex display of a character - that could be single or multibyte
			 */
			if(wp->w_opts & WO_HEX) {
				UniChar	hex[UTF8_MAX_BYTES * 3 + 1];
				UniChar*hp = hex;
				int	bytes = lisUtf8(lp) ? CHAR_LEN(c) : 1;

				while(--bytes >= 0) {
					*hp++ = hexdigits[*up >>  4];
					*hp++ = hexdigits[*up & 0xF];
					*hp++ = ' ';
					up++;
				}
				*hp = UNISTR_END;

				/* This checks for left/right visibility */
				vBrush = inkWell[INK_PLAIN];
				vtputUniStrLeftEdge(wp, hex);

				continue;
			}

			if(lisUtf8(lp) && CHAR_LEN(c) > 1) {
				/* Multi byte character: several bytes in buffer,
				 * variable number of places on the display screen - 1 or 2
				 */
				if(wp->w_curcol >= wp->w_leftoffset) {
					vBrush = inkWell[INK_PLAIN];
					vtpututf8(wp, up);
				} else {
					/* We need to work out the width that the character takes.
					 * It is to the left of the window edge and not displayed,
					 * eventually these will add up to bring us over the LH edge.
					 */
					int	width;
					if((width = mk_wcwidth(utf8_decode(up))) <= 0)
						width = 1;
					wp->w_curcol += width;

					/* Might need to fill in for the 2nd 1/2 of a double width
					 * character that did not get displayed.
					 * Display in appropriate colour
					 */
					if(width > 1 && wp->w_curcol - 1 >= wp->w_leftoffset) {
						VidAttr OldBrush = vBrush;
						vBrush = inkWell[INK_WIDE_EDGE];

						wp->w_curcol--;		// Back off to display the triangle

						vtpututf8(wp, SecondHalfDoubleWidthReplacement);
						vBrush = OldBrush;
					}
				}

				up += CHAR_LEN(c);
				continue;
			}

			/* Non UTF8 mode or c < U+80
			 * Output the character if we have reached the LH edge
			 * then work out the space taken by the character - regardless
			 * of it being output.
			 */
			vBrush = inkWell[INK_PLAIN];
			if(wp->w_curcol >= wp->w_leftoffset) {
				vtputc(wp, c);
			} else {
				/* As above: not displayed but must add up the width.
				 * Beware: a tab might start off the LH edge but expanding it
				 * might bring it on, ditto things like M-X & ^C.
				 * Display as much as we can, which might be part of the sequence.
				 * Take care to not write beyond the RH edge.
				 */
				if(c == '\t' && wp->w_bufp->b_tabsize > 1) {
					int	cols = wp->w_bufp->b_tabsize - (int)((wp->w_curcol - wp->w_numwidth) % wp->w_bufp->b_tabsize) - 1;
					// Display tabs by highlighting them ?
					// Need to set standout mode as coloured blanks are, well, blank
					VidAttr	cb = vBrush;
					if(wp->w_opts & WO_COLOUR_TABS)
						vBrush = pens[PEN_TAB];

					while(cols-- >= 0)
						if(wp->w_curcol >= wp->w_leftoffset && wp->w_curcol < wp->w_leftoffset + wp->w_ncolumns)
							vtputc(wp, ' ');
						else
							wp->w_curcol++;

					vBrush = cb;	// Undo any tab colourisation

				} else {
					/* Build up a string of what would have been displayed */
					UniChar		partial[8];	// Contains at least: M-^x\0
					UniChar*	cp = partial;

					if(c > 0x7f && !(wp->w_opts & WO_META_ATTRIB)) {
						/* Knock the top bit off ....
						 * then look to see if a control character (below)
						 */
						c &= 0x7F;
						*cp++ = 'M';	/* M- shows top bit set */
						*cp++ = '-';
					}

					if((c < 0x20 || c == CharDelete) && !(wp->w_opts & WO_META_ATTRIB)) {
						c ^= 0x40;	/* Get to what we display */

						*cp++ = '^';	/* The ^ which shows control */
					}
					*cp++ = c;
					*cp   = UNISTR_END;

					vtputUniStrLeftEdge(wp, partial);

				}
			}
			up++;
		}

		/* With hex option add the newline as a visible character (hex digits).
		 * This makes it obvious that they are there and can be tampered with.
		 *
		 * Complication: if NONL mode suppress for the last line of the buffer.
		 */
		if((wp->w_opts & WO_HEX) &&
		   (((wp->w_bufp->b_mode & MDNONL) == 0) || lforw(lp) != wp->w_bufp->b_abotline)) {
			UniChar	hex[3];
			BYTE	c = '\n';

			hex[0] = hexdigits[c >>  4];
			hex[1] = hexdigits[c & 0xF];
			hex[2] = UNISTR_END;

			/* This checks for left/right visibility */
			vtputUniStrLeftEdge(wp, hex);
		}

		int	did = 0;

		/* If we did not show anything on the line - as there was not enough text to
		 * reach the LH edge - show a $
		 * Do not do this for empty lines or with HEX option.
		 */
		if(lused(lp) && !(wp->w_opts & WO_HEX) &&
		   wp->w_curcol < wp->w_leftoffset && wp->w_ncolumns > MIN_WINDOW_SHOW_LINE_END) {

			VidAttr	cb = vBrush;
			vBrush = inkWell[INK_WRAP];

			/* The notional column in the buffer is still be to the left of the window.
			 * Ensure that we write to a column in the window - for the benefit of vteeol()
			 */
			wp->w_curcol = wp->w_leftoffset;

			vtputc(wp, ShiftedTextMarker);

			did = 1;

			vBrush = cb;	// Undo any tab colourisation

		}

		/* If the last character was written at RH window edge highlight it,
		 * it is a space, change it to a $
		 * The point is that if we don't then the end of line is not visible.
		 * There is code above to do this to the last character, but only if there is text
		 * that follows - this catches when the last character on the line is the last
		 * in the window.
		 * This also works when the last character is a tab.
		 */
		if(up == l_end && lused(lp) && !(wp->w_opts & WO_HEX) &&
			wp->w_curcol == wp->w_leftoffset + wp->w_ncolumns) {
				vp->v_attr[wp->w_leftcol + wp->w_ncolumns - 1] = inkWell[INK_WRAP];

				if(vp->v_text[wp->w_leftcol + wp->w_ncolumns - 1] == ' ')
					vp->v_text[wp->w_leftcol + wp->w_ncolumns - 1] = ShiftedTextMarker;
		}

		/* Show line end option - but not hex ?
		 * Show a $ where the end of line is.
		 * Only if the window is of a minimum width.
		 */
		if((wp->w_opts & (WO_HEX | WO_SHOW_LINE_END)) == WO_SHOW_LINE_END && wp->w_ncolumns > MIN_WINDOW_SHOW_LINE_END &&
		   lused(lp) &&
		   up >= l_end && wp->w_curcol > wp->w_leftoffset && wp->w_curcol < wp->w_leftoffset + wp->w_ncolumns &&
		   !did) {

			// Need to put a $ as coloured blanks are, well, blank
			VidAttr	cb = vBrush;

			/* If we reached the end of the line in the buffer but there is still width
			 * on the screen, show the end of line marker.
			 * Only do this if we wrote something to the line - in case everything is off the LH edge.
			 */
			vBrush = pens[PEN_EOL_POSITION];

			vtputc(wp, EndOfLineMarker);

			vBrush = cb;	// Undo any tab colourisation

		}
	}

	vteeol(wp, ' ', pens[PEN_BLANK_SCREEN]);
//	vteeol(wp, ' ', inkWell[INK_PLAIN]);	// This fights with setting to pens[PEN_BLANK_SCREEN] at the bottom of updateline()

	CurLineFlagp = &NoFlag;	// Delink - in case of other vtputc()
}

/* Ephemerally place character ch in the virtual screen.
 * It is at position lno & loff in window wp.
 *
 * lno & loff must be within bounds of the window - this does no checking.
 *
 * loff is pure window, it must take account of width of line numbers,
 * if any.
 *
 * ch will be displayed as UniCode. Doing a non width 1 character is not wise.
 *
 * pen is a colour, something like PEN_MARK.
 */
void	placeCharPos(Window* wp, int lno, int loff, UniChar ch, int pen)
{
	int	sline = lno + wp->w_toprow;	// Screen line

	vscreen[sline]->v_flag |= VFCHG;	// Ensure that it is updated on screen
	vtmove(sline, wp->w_leftcol + loff);

	vBrush = inkWell[pen];

	if(CHAR_LEN(ch) > 1) {
		// Multi byte character: several bytes in buffer
		vtpututf8(wp, utf8_encode(ch));
	} else {
		vtputc(wp, ch);
	}
}

/* Ephemerally place character ch in window wp.
 * Put it over position loff in line lp.
 * Do nothing if the position is not currently visible in wp.
 *
 * It is ephemeral: it will disappear when the line is next updated
 * in the window - eg at a refresh.
 *
 * Return:
 *	TRUE - placed a character
 *	FALSE - did not place a character - not at a place visible in the window
 */
int	placeChar(Window* wp, LINE* lp, LineLn loff, UniChar ch, int pen)
{
	LINE*	probe = wp->w_topLinep;
	int	lno = 0;
	LineLn	col;

	while(1) {
		// Is the line visible ?
		if(probe == lp)
			break;

		// Run out of lines in the window ?
		if(++lno >= wp->w_ntrows)
			return(FALSE);

		// End of buffer ?
		if(probe == wp->w_bufp->b_abotline)
			return(FALSE);

		probe = lforw(probe);
	}

	col = getdiscol(wp, lp, loff,
		(wp->w_opts & WO_HEX ? DisCol_HEX : (wp->w_opts & WO_META_ATTRIB ? DisCol_MetaChar : DisCol_Char))
			) - wp->w_leftoffset;

	if(col < 0 || col >= wp->w_ncolumns - wp->w_numwidth)
		return(FALSE);

	/* If we are here then the line is visible in the window.
	 * That does not mean that the cursor position is visible as well.
	 */
	placeCharPos(wp, lno, (int)col + wp->w_numwidth, ch, pen);

	return(TRUE);
}

/* Return the display column of byteoffset in line lp.
 *
 * This is relative to the start of line, not to the window LH edge (ie units: L).
 * Often byteoffset will be w_doto.
 *
 * CountMode:
 * DisCol_Char - UTF8 chars have variable width, zero width counts as 1.
 * DisCol_MetaChar - as DisCol_Char but meta/ctrl as attributes underline/standout
 * DisCol_WP - count like a word processor:
 * 	Zero width characters count as 0.
 *	Tabs count as 1 .. b_tabsize positions.
 *	All others count as 1.
 * DisCol_HEX - window has hex option -- the returned value is of the first nybble.
 */
LineLn	getdiscol(const Window* wp, const LINE* lp, LineLn byteoffset, int CountMode)
{
	LineLn	bof;		// Byte offset
	LineLn	curcol;

	/* find the current column
	 * The code below is in updline() - update there as well if
	 * what is below is change.
	 */
	curcol = 0;
	bof = 0;
	while(bof < byteoffset) {
		BYTE	ch = lgetc(lp, bof);

		// Hex: bytes diplayed as: hex hex space
		if(CountMode == DisCol_HEX) {
			curcol += 3;
			bof++;
			continue;
		}

		// Utf8
		if(lisUtf8(lp) && CHAR_LEN(ch) > 1) {
			int	width;

			/* Multi byte character: several bytes in buffer,
			 * variable number of places on the display screen - 1 or 2
			 */
			width = mk_wcwidth(utf8_decode(lgetcaddr(lp, bof)));

			if(CountMode == DisCol_WP) {
				if(width > 0)
					width = 1;
			} else {
				if(width <= 0)
					width = 1;
			}

			// If we are asked about a byteoffset that is in the middle of a
			// multibyte UTF8 char - don't count the width until we have reached
			// the end of it.
			// This might happen with hex option
			if((bof += CHAR_LEN(ch)) <= byteoffset)
				curcol += width;

			continue;
		}

		bof++;

		// Tabs need a special calculation:
		if(ch == '\t' && wp->w_bufp->b_tabsize > 1) {
			curcol = (curcol / wp->w_bufp->b_tabsize + 1) * wp->w_bufp->b_tabsize;
			continue;
		}

		// Meta character with 'M-' prefix ?
		if(ch > 0x7f && CountMode == DisCol_Char) {
			curcol += 2;
			/* knock the top bit off ....
			 * then look to see if a control character (below)
			 */
			ch &= 0x7F;
		}

		// Control character and we display like ^c ?
		if((ch < 0x20 || ch == CharDelete) && CountMode == DisCol_Char) {
			++curcol;	/* For the ^ which shows control */
		} 

		++curcol;
	}

	return(curcol);
}

/* Return a byte offset (w_doto) that corresponds to the display column.
 * If the line is not that long, return the max byte offset.
 *
 * Take into account that not all characters have the same
 * display width.
 *
 * CountMode:
 * DisCol_Char - UTF8 chars have variable width, zero width counts as 1. Meta/ctrl as M- & ^.
 * DisCol_MetaChar - as DisCol_Char but meta/ctrl as attributes underline/standout
 * DisCol_WP - count like a word processor:
 * 	Zero width characters count as 0.
 *	Tabs count as 1-7 positions.
 *	All others count as 1.
 * DisCol_HEX - window with hex option
 */
LineLn	getdoto(const Window* wp, const LINE* lp, LineLn column, int CountMode)
{
	LineLn	bof = 0;	// Byte offset
	LineLn	curcol = 0;

	while(curcol < column) {
		int	ch;

		if(bof >= lused(lp))
			return(bof);

		ch = lgetc(lp, bof);

		// Hex: bytes diplayed as: hex hex space
		if(CountMode == DisCol_HEX) {
			curcol += 3;
			bof++;
			continue;
		}

		if(lisUtf8(lp) && CHAR_LEN(ch) > 1) {
			int	width;

			/* Multi byte character: several bytes in buffer,
			 * variable number of places on the display screen - 1 or 2
			 */
			width = mk_wcwidth(utf8_decode(lgetcaddr(lp, bof)));

			if(CountMode == DisCol_WP) {
				if(width > 0)
					width = 1;
			} else {
				if(width <= 0)
					width = 1;
			}

			// Increment our position by the width of the character
			curcol += width;

			// If the character has width 2 then we might have gone past
			// where we wanted to get to. The point is that we don't move on
			// if the column is in the 'second 1/2' of the character, keep the
			// cursor on the character and not let it display on the next character.
			if(curcol <= column)
				bof += CHAR_LEN(ch);

			continue;
		}

		bof++;

		// Not UTF-8 mode or ch <= UTF8_MAX_CP1 (0x7F)

		if(ch == '\t' && wp->w_bufp->b_tabsize > 1) {
			curcol = (curcol / wp->w_bufp->b_tabsize + 1) * wp->w_bufp->b_tabsize - 1;
			ch = 'a';	// Do not deem control (below)
		} else if(ch > 0x7f && CountMode == DisCol_Char) {
			curcol += 2;
			/* knock the top bit off ....
			 * then look to see if a control character (below)
			 */
			ch &= 0x7F;
		}
		if((ch < 0x20 || ch == CharDelete) && CountMode == DisCol_Char) {
			++curcol;	/* For the ^ which shows control */
		} 

		++curcol;
	}

	return(bof);
}

/* updpos() rewritten
 * This finds the cursor in what is usually the current window and ensures that the window
 * can display it.
 * Also used with cursor synching (HEX option).
 *
 * w_disrow & w_discol are updated.
 *
 * This might cause the window contents to scroll left/right, if so WFHARD
 * will be set.
 */
void	updatedotpos(Window* wp)
{
	register LINE*	lp;
	LineLn		oldOffset;
	LineLn		discol;		// Display column - relative to start of the line

	/* Calculate margin at edge to keep the cursor out of (if we can)
	 * and by how much we scroll - when we do.
	 * Note that the sizes are fractions of the *window* width, not the terminal width.
	 * If the window gets very narrow take care that it still works.
	 */
	int		margin;
	int		scrsiz;

	// Note there is a var (user switch) to set it only be 1
	margin = WindowMargin ? (wp->w_ncolumns > WINDOW_SCROLL_RATIO ? wp->w_ncolumns / WINDOW_SCROLL_RATIO : 1) : 1;
	scrsiz = margin > 1 ? margin * 2 : 1;

	/* Scroll by 1 in the message window - so can see as much as possible.
	 * This window needs to scroll, otherwise there will be buffer overruns on long lines.
	 */
	if(wp == wmessp)
		scrsiz = 1;

	/* Interaction with moveLeftWind()/moveRightWind()
	 * These scroll left/right and can ignore the normal margin restrictions,
	 * defer to them, do not readjust.
	 * CFMARGIN0 when "." must be visible but can touch an edge (search can do this
	 * to avoid a scroll).
	 */
	if(lastflag & (CFCPCN | CFMARGIN0))
		margin = 0;

	/* find the current row */
	lp = wp->w_topLinep;
	wp->w_disrow = wp->w_toprow;
	while(lp != wp->w_dotp) {
		++wp->w_disrow;
		lp = lforw(lp);
	}

	/* Moving a long way can take this completely off the screen.
	 * Force a reframe.
	 * Set w_disrow to something sensible before it causes problems; it is used as
	 * an index into vscreen[].
	 */
	if(wp->w_disrow >= wp->w_toprow + wp->w_ntrows) {
		wp->w_flag |= WFFORCE;
		wp->w_disrow = wp->w_toprow;
	}

	// Chars as: Hex, M-^x or attributes underline/standout
	discol = getdiscol(wp, lp, wp->w_doto,
		(wp->w_opts & WO_HEX ? DisCol_HEX : (wp->w_opts & WO_META_ATTRIB ? DisCol_MetaChar : DisCol_Char)));

	/* If the last command was page/line move up/down do not reset w_leftoffset
	 * This stops the window jerking left/right when moving up/down.
	 * Still need to calculate w_discol, so the cursor will still move to the left on
	 * a short(er) line - otherwise the cursor hovers in positions where there is not
	 * a character - which is kind of confusing to the user.
	 */

	if((lastflag & CFCPCN) == 0) {

		/* Compute how much to set the window's w_leftoffset.
		 * This is the number of columns 'off the left hand side' of the window.
		 * If we are close to the left/right edge of the window - scroll
		 * to keep it away from the edge.
		 * Provide some hysteresis - don't adjust unless we must.
		 */
		oldOffset = wp->w_leftoffset;

		// Fix to if & calc
		// Ensure that this all still fits in the window
		// Check LH side:
		while(discol < wp->w_leftoffset + margin - 1) {
			wp->w_leftoffset -= scrsiz;
		}
		if(wp->w_leftoffset < 0)
			wp->w_leftoffset = 0;

		// Hex displays 'xx ' so minimum scroll size is 3:
		if((wp->w_opts & WO_HEX) && scrsiz < 3)
			scrsiz = 3;

		// Check RH side:
		while(discol + wp->w_partoff - wp->w_leftoffset + wp->w_numwidth >= wp->w_ncolumns - margin)
			wp->w_leftoffset += scrsiz;

		// Constrain a hex window to have the first nybble at window left
		if(wp->w_opts & WO_HEX)
			wp->w_leftoffset -= wp->w_leftoffset % 3;

		// If we changed the offset - refesh the window:
		if(wp->w_leftoffset != oldOffset)
			wp->w_flag |= WFHARD | WFMODE;
	}

	// Convert from window (lineLn) to screen units.
	// w_leftoffset is now right to ensure that w_discol will be within the screen width.
	// Ie this lets us convert from L to S units.
	wp->w_discol = (int)(discol + wp->w_lefttextcol - wp->w_leftoffset);

	// If the line is shorter than the LH offset w_discol could end up -ve -- when bad things happen:
	if(wp->w_discol < wp->w_lefttextcol)
		wp->w_discol = wp->w_lefttextcol;

//TODO -- we ought to store the offset of dot/curcol somewhere so that
// the cursor can be put there. Need to have a new field in Window
// Note that this might result in an offset that is on the 2nd position of a double width character

}

/* updgar:	if the screen is garbage, clear the physical screen and
 *		the virtual screen and force a full update
 */
void	updgar(void)
{
	register VideoC*txt;
	register int	i, j;
	VidAttr*	vap;

	for(i = 0; i < AllocLines; ++i) {
		vscreen[i]->v_flag |= VFCHG;
#if	REVSTA && LEGACY_DISPLAY
		vscreen[i]->v_flag &= ~VFREV;
#endif

#if	LEGACY_DISPLAY
		vscreen[i]->v_fcolor = gfcolor;
		vscreen[i]->v_bcolor = gbcolor;
#endif
#if	MEMMAP == 0
		txt = pscreen[i]->v_text;
		vap = pscreen[i]->v_attr;
		for(j = 0; j < AllocColumns; ++j) {
			*txt++ = ' ';
			*vap++ = pens[PEN_BLANK_SCREEN];
		}
#endif
	}

	TTbacg(gbcolor);		/* Set background colour */

	/* Leave everything alone - just repaint the message line ?
	 * Then don't clear the screen.
	 * However it is important that the above is done - for the
	 * benefit of the message line.
	 * There will be times when the only thing that we control is
	 * the message line, think: shell-command when prompting for
	 * another command and the (unknown to Emacs) output of the
	 * first is still shown.
	 */
	if(DisplayMessageLineOnly) {
		movecursor(AllocLines - 1, 0);
		TTeeol();
		sgarbf = FALSE;		/* Erase-page clears */
		return;
	}

	movecursor(0, 0);		// no more
	TTeeop();			/* Erase the screen. */

	sgarbf = FALSE;			/* Erase-page clears */
	mpresf = FALSE;			/* the message area. */

//	mlerase();			/* needs to be cleared if colored */
}

/* Mark everything out of date
 * Set flags that update() notices rather than calling
 * updall() and modeline() directly.
 * Fill the virtual screen with spaces
 */
void	ReDraw(void)
{
	sgarbf = TRUE;

	/* Request update everything & modeline update() */
	upmode(NULL, WFHARD | WFMODE);

	update(1);
}

/* To be called if we have been suspended - UNIX's ^Z.
 * Also called after a window size change, this is the more
 * difficult of the 2 cases - reallocate buffers in case the window
 * has changed size.
 */
void	RePaint(void)
{
	/* Buffer reallocation (vtinit()) was done from here.
	 * The test for WasSusp is now done in update()
	 * which is called from ReDraw().
	 */

	updgar();
	ReDraw();
}

#if 0
/* When a window is about to be destroyed (ie no longer displayed)
 * we must clear the attributes
 */
void	WinClearCursor(Window* wp)
{
	vscreen[wp->w_disrow]->v_attr[wp->w_discol] = wp->w_cursattr;
	vscreen[wp->w_disrow]->v_flag |= VFCHG;

	wp->w_cursattr = va_default();
}
#endif

/* Cycle (increment) the global foregound colour ($fg-colour) and refresh the screen.
 *
 * The reason for this is that now colours are being set it is possible that the
 * editor starts and the user see the same foreground and background colours.
 *
 * If the colour value is out of the range of valid colours: set to the
 * minimum - which is the system default.
 *
 * If there is an argument, set the global colour to the count.
 * A default count is not set.
 *
 * Command:  cycle-foreground-colour
 * Bound to: FN13
 *
 * Return TRUE - always
 */
int	cyclegfcolor(int f, Cnt n)
{
	if(f)
		gfcolor = (int)n;
	else
		++gfcolor;

	if(gfcolor > COLOUR_MAX || gfcolor < COLOUR_MIN)
		gfcolor = COLOUR_MIN;

	sanitiseColours();

	updgar();
	ReDraw();

	return(TRUE);
}

/* Cycle (increment) the global foregound ($bg-colour)colour and refresh the screen.
 * 
 * See comments in cyclegfcolor().
 *
 * A default count is not set.
 *
 * Command:  cycle-background-colour
 * Bound to: FN14
 *
 * Return TRUE - always
 */
int	cyclegbcolor(int f, Cnt n)
{
	if(f)
		gbcolor = (int)n;
	else
		++gbcolor;

	if(gbcolor > COLOUR_MAX || gbcolor < COLOUR_MIN)
		gbcolor = COLOUR_MIN;

	sanitiseColours();

	updgar();
	ReDraw();

	return(TRUE);
}

/* Sanitise the colour choices.
 * This means fiddling with the foreground colours so that they still work (ie are visible)
 * against the current background colour.
 *
 * Various things have colours anyway (eg INK_WRAP is COLOUR_YELLOW), do not start by upsetting
 * that; if we find something with FG colour SYS_DEF - initialise that to the current FG colour,
 * this might be SYS_DEF or might be another colour.
 *
 * Also: for PEN_TAB (used for visible tabs) check that underlined spaces work.
 *
 * The colours are put into pens[] - which is what is used.
 */
void	sanitiseColours(void)
{
	int	i;

	for(i = 0; i < (int)(sizeof(penDefaults) / sizeof(penDefaults[0])); i++) {
		VidAttr	defAttr = penDefaults[i];
		VidAttr	attr = va_set_width(0, va_get_width(defAttr));
		int	colour;
		int	stand = va_get_stand(defAttr);

		// Extract the colour
		colour = va_get_colour(defAttr);
		if(va_is_colour_sysdef(defAttr))
			colour = gfcolor;

		/* The definition of bad is a FG colour that is the same as the BG colour - so invisible.
		 * Sysdef FG against sysdef BG is OK.
		 * Standout (so) is often/usually implemented as reverse video, ie: wh -> bk, bk -> wh
		 * So: so wh == bk; so bk == wh
		 * But so sysdef FG against sysdef BG is OK - little sense for term manufacturer to make invis.
		 * Sysdef colour is unknown but prob black or white:
		 *   so fg/bg that is black or white - against that is bad
		 *
		 * Some of the colour schemes that this produces might be garish or hard to read,
		 * but the user should be able to see everything.
		 */
		if((colour == gbcolor && colour != COLOUR_SYS_DEF) ||
		   (colour  == COLOUR_SYS_DEF && (gbcolor == COLOUR_BLACK || gbcolor == COLOUR_WHITE)) ||
		   (gbcolor == COLOUR_SYS_DEF && (colour  == COLOUR_BLACK || colour  == COLOUR_WHITE)) ||
		   (stand && gfcolor == COLOUR_BLACK && gbcolor == COLOUR_WHITE) ||
		   (stand && gfcolor == COLOUR_WHITE && gbcolor == COLOUR_BLACK)) {
			colour++;

			// Don't rest on black (straight after SYS_DEF)
			if(colour == COLOUR_BLACK)
				colour++;

			if(colour > COLOUR_MAX)
				colour = COLOUR_RED;
		}

		// Put the clash-avoiding colour into attr:
		if(colour == COLOUR_SYS_DEF)
			attr = va_set_colour_sysdef(attr);
		else
			attr = va_set_colour(attr, colour);

		if(va_get_stand(defAttr))
			attr = va_set_stand(attr);

		// Special test for tabs ... we want to underline spaces,
		// buy not everything does that, if so make standout (== reverse)
		if(i == PEN_TAB && !TTdoes(TT_DOES_BLANK_UL))
			attr = va_set_stand(attr);
		else
			if(va_get_ul(defAttr))
				attr = va_set_ul(attr);

		pens[i] = attr;
	}
}

/* Updates the edges between windows.
 * These are vertical lines all in the pens PEN_CURR_BORDER
 * or PEN_BORDER -- which are, currently, the same thing.
 * Saving much pen changing is the big reason for doing this all at once.
 *
 * The interest in this function is identifying the vertical lines.
 * There may be several entries in edgeAt, but not necessarily aligned.
 *
 * This assumes that WindowSeparator is a single byte.
 *
 * It is assumed that updateEdges() is called after the current row has been
 * updated - so don't do it again.
 */
void	updateEdges(void)
{
	int	Col = leftMostEdge;
	int	nextCol;

	TERMOUTTRACE((TermDebFD, "updateEdges() leftMostEdge=%d rightMostEdge=%d\n", leftMostEdge, rightMostEdge))

	/* Scan columns left to right.
	 * We know the first column with an edge, as we scan determine
	 * the next edge column.
	 */
	for( ; Col <= rightMostEdge; Col = nextCol) {
		int	Row;
		nextCol = AllocColumns + 1;

		TERMOUTTRACE((TermDebFD, "updateEdges() Col=%d nextCol=%d\n", Col, nextCol))
		// Scan all rows
		for(Row = 0; Row < AllocLines - 1; Row++) {
			short*	edges;
			int	nedge;

			// Scan the edge indices in the row
			for(edges = edgeAt + (Row * VID_MAXSEPS), nedge = VID_MAXSEPS ; nedge > 0 && *edges != 0; edges++, nedge--) {
				TERMOUTTRACE((TermDebFD, "updateEdges() Row=%d *edges=%d Col=%d\n", Row, *edges, Col))
				// An edge right of what we are looking for ?
				if(*edges > Col) {
					// It might be the next one to look for
					if(nextCol > *edges)
						nextCol = *edges;

					break;
				}

				/* Output the edge character that we have found.
				 * Set the brush - this should only do something for the first one,
				 * then the edge character, the set the physical screen.
				 *
				 * Not current row since it has already been done.
				 */
				if(*edges == Col && Row != curwp->w_disrow) {
					TERMOUTTRACE((TermDebFD, "updateEdges() Row=%d Col=%d '%c'\n", Row, Col, (char)vscreen[Row]->v_text[Col]))
					movecursor(Row, Col);
					SetBrush(vscreen[Row]->v_attr[Col]);
					TTputc(vscreen[Row]->v_text[Col]);
					++ttcol;

					pscreen[Row]->v_attr[Col] = vscreen[Row]->v_attr[Col];
					pscreen[Row]->v_text[Col] = vscreen[Row]->v_text[Col];
				}
			}
		}
	}
}

/* updupd:	update the physical screen from the virtual screen
 *
 * force: forced update flag
 */
void	updupd(int force)
{
	register int i, j;

	// Leave all but the message line alone ?
	if(DisplayMessageLineOnly) {
		updateline(AllocLines - 1, force);
		return;
	}

	/* If a window has scrolled, do that first and sort out
	 * whatever other differences there are later.
	 */
	if(WindowScrolled)
		DoScrolled();

	// Great for debugging scrolling: separate physical screen update
	// from rest of repainting (done after the next keystroke).
	//	if(WindowScrolled)return;

	/* Update the current used line before all others, likely the one
	 * of most interest, especially on a slow (tty) line.
	 */
	if( !updateline(curwp->w_disrow, force))
		return;

	// The message line
	if( !updateline(AllocLines - 1, force))
		return;

	/* If a bunch of lines have been scrolled and they have line numbers
	 * update the numbers. This should not take long and doing them top to
	 * bottom is an efficient way.
	 */
	if(PaintScrollLines)
		for( ; TopScrollLines <= BottomScrollLines; )
			if( !updateline(TopScrollLines++, force))
				return;

	/* Update any edges all at once.
	 * The gain for doing this is that we change the pen colour once,
	 * moving down edges is cheap.
	 * Should really check for pending input before this, don't bother.
	 */
	if(edgeSet)
		updateEdges();

	/* After the current line, move away from it: up and down simultaneously.
	 * This results in the area of interest being updated first.
	 *
	 * However: if the edges or line numbers have been updated
	 * then update the screen top -> bottom. This allows for cursor movement
	 * optimisation - probably.
	 * The above used to have a no-text-changed condition, but often the message
	 * line gets changed - eg add/remove line numbers.
	 *
	 * Ignore window mode lines, do them as we come across them.
	 *
	 * This will stop if there is any pending input.
	 * We will always update a line completely before checking for pending input.
	 *
	 * Do not update the message line, already done.
	 */

	if(updateWhat & (UW_NUMB | UW_EDGE)) {
		// Edge only update. Top to bottom:
		for(i = 0; i < AllocLines - 1; i++) {

			if(!updateline(i, force))
				break;
		}
	} else {
		/* Text in a window updated, there could be a lot of it.
		 * Start with row with cursor, move away from it.
		 * Do this in pairs of lines, up/down, reduces movement a bit
		 */
		for(i = curwp->w_disrow, j = curwp->w_disrow + 1 ; i >= 0 || j < AllocLines - 1; i -= 2, j += 2) {

			if(i - 1 >= 0 && !updateline(i - 1, force))
				break;
			if(i     >= 0 && !updateline(i,     force))
				break;

			if(j     < AllocLines - 1 && !updateline(j,     force))
				break;
			if(j + 1 < AllocLines - 1 && !updateline(j + 1, force))
				break;
		}
	}
}

/* Find the lines that have scrolled and physically scroll them.
 * Probably only one region that has scrolled, but possible that
 * others have scrolled, maybe by different amounts.
 */
void	DoScrolled(void)
{
	int		row;
	int		topScrollRow;
	int		nscrolled = 0;

	TERMTRACE((TermDebFD, "DoScrolled(), TermDebug=%d\n", TermDebug))

	topScrollRow = 0;	// Unecessary code to stop the compiler from wingeing

	// Look for contiguous regions where the scroll (v_scrolled) is the same.
	// Search top to bottom.
	// Beware: fiddle with row inside the loop:
	for(row = 0; ; row++) {
		int	doScroll = 0;	// Scroll now
		int	topSpace;	// Top line # of space opened up

		if(row >= AllocLines) {
			// Need for scrolling not seen, break loop
			if(nscrolled == 0)
				break;

			doScroll = 1;
		} else {
			// Check if scrolling by same amount
			// which could be nothing

			if(vscreen[row]->v_scrolled != nscrolled) {
				if(nscrolled == 0) {
					topScrollRow = row;
					nscrolled = vscreen[row]->v_scrolled;
				} else
					doScroll = 1;
			}
		}

		// Check if need to look further
		if(!doScroll)
			continue;

		TERMOUTTRACE((TermDebFD, "row=%d topScrollRow=%d nscrolled=%d\n", row, topScrollRow, nscrolled))

		// Perform a physical scroll in whichever direction, then
		// update what we know the physical screen looks like, ie
		// scroll our internal representation.
		// Brute force is to copy the buffers, smart is to switch pointers;
		// the structure is almost all pointers - so just swap them.
		// The easiest way of doing this is to move up/down swapping.
		// Mark lines as changed so that a check/redraw is done.
		// scrolled will be non zero.
		if(nscrolled < 0) {
			// Moving rows up
			int	i;

			TTscroll(topScrollRow - 1, row - topScrollRow + 1, nscrolled);

			if((i = topScrollRow + nscrolled - 1) < 0)
				i = 0;	// Will happen if scroll off the top of screen

			for( ; i < row + nscrolled; i++) {
				VIDEO*	pt = pscreen[i];
				VIDEO*	pb = pscreen[i - nscrolled];
				VIDEO	ptmp = *pt;
				*pt = *pb;
				*pb = ptmp;
				vscreen[i]->v_flag |= VFCHG;
				vscreen[i - nscrolled]->v_flag |= VFCHG;
			}

			topSpace = i;
			nscrolled = 0 - nscrolled;

		} else {
			// Moving rows down
			int	i;

			TTscroll(topScrollRow, row - topScrollRow, nscrolled);

			for(i = row + nscrolled - 1; i >= topScrollRow + nscrolled; i--) {
				VIDEO*	pt = pscreen[i - nscrolled];
				VIDEO*	pb = pscreen[i];
				VIDEO	ptmp = *pt;
				*pt = *pb;
				*pb = ptmp;
				pb->v_scrolled = pt->v_scrolled = 0;	// So don't see it again
				vscreen[i]->v_flag |= VFCHG;
			}

			// Top of rows to fill with spaces
			topSpace = topScrollRow;
		}

		// Clear to spaces the new rows opened up
		for( ; --nscrolled >= 0; topSpace++) {
			int	i;
			VidAttr*pa = pscreen[topSpace]->v_attr;
			VideoC*	pv = pscreen[topSpace]->v_text;
			vscreen[topSpace]->v_flag |= VFCHG;

			for(i = 0; i <= AllocColumns; i++, pa++, pv++) {
				*pa = pens[PEN_BLANK_SCREEN];
				*pv = ' ';
			}
		}

		// Note that the vscreen has not been touched, thus if we guessed
		// completely wrong in the scrolling all will be put right later on
		// at cost of more being sent to the terminal than if we had got
		// it right.

		// Set to scan for another region
		nscrolled = 0;

		// Look at the current row again:
		row--;
	}
}


#if	MEMMAP == 0
/* Variables shared by updateline() & updateCP()
 *
 * The point of UpdateRow is to be able to delay cursor movement until a
 * character (or clear line, ...) is output. Before this the cursor could be
 * moved & then promptly moved back.
 * This means that ttrow & ttcol cannot always be used as an indication of
 * which part of the screen is being looked at.
 * Doing movecursor() to where it already is will not output anything.
 *
 * Beware: LineWidth is not the same as AllocColumns
 * It is:
 * * 0 based (ie subtract 1)
 * * the number that we want to draw on the line,
 *   this is one less on the message/command line as we want to prevent an unwanted
 *   screen scroll by writing to the last position on the screen.
 */
static	VideoC*		virtcp;		/* Pointer to what we want the screen to be - characters */
static	VideoC*		physcp;		/* What it currently is */
static	VidAttr*	virtap;		/* Pointer to attributes that we want */
static	VidAttr*	physap;		/* What attributes currently are */
static	int		LineWidth;	/* Width of line being drawn*/
static	int		UpdateRow;	/* Screen row being updated by updateline() */

/* Update a character point.
 * The cursor is at the right place.
 *
 * Use: virtcp, physcp virtap, physap above as the place to update.
 * Yes: these should be passed as arguments, but keep the code simpler
 * by using global variables - that are maintained in updateline().
 */
void	updateCP(void)
{
	int	width = va_get_width(*virtap);

	// Do (delayed) cursor movement
	TERMOUTTRACE((TermDebFD, "updateCP %d, %d => %d, %d width=%d phys:'%c' %#2.2x -> virt:'%c' %#2.2x\n", ttrow, ttcol, UpdateRow, (int)(physcp - pscreen[UpdateRow]->v_text), width, *physcp, *physap, *virtcp, *virtap))
	TTmove(UpdateRow, (int)(physcp - pscreen[UpdateRow]->v_text));

	/* Ensure that the physical brush correct for the character that we are about
	 * to output.
	 */
	SetBrush(*virtap);

	switch(CHAR_LEN(*virtcp)) {
#if	UTF8_MAX_CODE_POINT > UTF8_MAX_CP4
	case 6:	TTputc(((BYTE*)virtcp)[0]);
		TTputc(((BYTE*)virtcp)[1]);
		TTputc(((BYTE*)virtcp)[2]);
		TTputc(((BYTE*)virtcp)[3]);
		TTputc(((BYTE*)virtcp)[4]);
		TTputc(((BYTE*)virtcp)[5]);
		break;
	case 5:	TTputc(((BYTE*)virtcp)[0]);
		TTputc(((BYTE*)virtcp)[1]);
		TTputc(((BYTE*)virtcp)[2]);
		TTputc(((BYTE*)virtcp)[3]);
		TTputc(((BYTE*)virtcp)[4]);
		break;
#endif
	case 4:	TTputc(((BYTE*)virtcp)[0]);
		TTputc(((BYTE*)virtcp)[1]);
		TTputc(((BYTE*)virtcp)[2]);
		TTputc(((BYTE*)virtcp)[3]);
		break;
	case 3:	TTputc(((BYTE*)virtcp)[0]);
		TTputc(((BYTE*)virtcp)[1]);
		TTputc(((BYTE*)virtcp)[2]);
		break;
	case 2:	TTputc(((BYTE*)virtcp)[0]);
		TTputc(((BYTE*)virtcp)[1]);
		break;
	case 1:	TTputc(((BYTE*)virtcp)[0]);
	}

	/* Characters with funny widths have been looked at in vtpututf8()
	 * The width that we see here is 1 or 2,
	 * 0 width chars having been replaced with a pink square width 1
	 *
	 * There are a few characters that we think is width 1 but *some*
	 * terminals display as 2, eg U+10FFFF
	 * Can't do anything about it since we can't eyeball what the
	 * terminal (emulator) is doing.
	 *
	 * If we reach the right hand edge, TTrightMargin() will fix up
	 * the cursor position.
	 */
	if((ttcol += width) >= LineWidth)
		TTrightMargin();

	*physcp++ = *virtcp++;
	*physap++ = *virtap++;
// bad write line above

	// Wide - just step over
	if(width == 2) {
		*physcp++ = *virtcp++;
		*physap++ = *virtap++;
	}
}
#endif	// MEMMAP

/* Update a single line. This does not know how to use insert or delete
 * character sequences; we are using VT52 functionality. Update the physical
 * row and column variables. It does try an exploit erase to end of line. The
 * RAINBOW version of this routine uses fast video.
 *
 * Returns TRUE.
 */
#if	MEMMAP
/*	UPDATELINE specific code for the IBM-PC and other compatables */

/*
 * row: row of screen to update
 * vp1: virtual screen image
 */
int	updateline(int row, int force)
{
	struct VIDEO* vp1 = vscreen[row];

#if	LEGACY_DISPLAY
	scwrite(row, vp1->v_text, vp1->v_rfcolor, vp1->v_rbcolor);
	vp1->v_fcolor = vp1->v_rfcolor;
	vp1->v_bcolor = vp1->v_rbcolor;
#else
	if(vp1->v_flag & VFREQ)
		scwrite(row, vp1->v_text, 0, 7);
	else
		scwrite(row, vp1->v_text, 7, 0);
#endif
	vp1->v_flag &= ~(VFCHG | VFCOL);	/* flag this line as changed */

	return(TRUE);
}

#else	/* MEMMAP */


/* Update the physical screen from the virtual one
 * row: row of screen to update
 * force: set to TRUE to ignore typeahead (pending input)
 *
 * Starts with a test to check if the line needs updating, then if there is pending input.
 *
 * Returns TRUE when updated
 *	   FALSE if there is input in typahead.
 */
int	updateline(int row, int force)
{
    struct VIDEO* virtscreen = vscreen[row];
    struct VIDEO* physscreen = pscreen[row];
    UpdateRow = row;

    // Can the terminal scroll left/right and does config desire it ($scroll-lr)
    // This might get set false below
    int		  tInsertChars = TTdoes(TT_DOES_INSERT_CHARS) && ttDoScrollLr;

    TERMOUTTRACE2((TermDebFD, "updateline(%d, %d) . = (%d, %d)\n", row, force, ttrow, ttcol))

#if	RAINBOW
/*	UPDATELINE specific code for the DEC rainbow 100 micro	*/

    register VideoC* virtcp;
    register VideoC* physcp;
    register int nch;

#if	LEGACY_DISPLAY
    /* since we don't know how to make the rainbow do this, turn it off */
    flags &= (~VFREV & ~VFREQ);
#endif

    virtcp = &virtscreen->v_text[0];                    /* Use fast video. */
    physcp = &physscreen->v_text[0];
    putline(row+1, 1, virtcp);
    nch = AllocColumns;

    do
        {
        *physcp = *virtcp;
        ++physcp;
        ++virtcp;
        }
    while(--nch);
    *flags &= ~VFCHG;
#else	/* RAINBOW */

/*	UPDATELINE code for all other versions		*/

// Some of these are now global, shared with updateCP()
// 	VideoC*		virtcp;		/* Pointer to what we want the screen to be - characters */
// 	VideoC*		physcp;		/* What it currently is */
	VideoC*		EndVirtTxt;	/* Virtscreen text ends (right hand) just before this */
	VideoC*		EndPhysTxt;	/* Physical screen text ends just before this */
	VideoC*		RightmostChanged;/* Rightmost character that has changed */
	VideoC*		lastNonSpace;	/* Last non space on physical screen */


// 	VidAttr*	virtap;		/* Pointer to attributes that we want */
// 	VidAttr*	physap;		/* What attributes currently are */
	VidAttr*	endVattr;	/* Probe into virtap */
	VidAttr*	endPattr;	/* Probe into physap */

	int		nbflag = FALSE;	/* Non-blanks to the right flag */
#if	LEGACY_DISPLAY
	int		rev;		/* reverse video flag */
	int 		req;		/* reverse video request flag */
#if	TERMCAP && FIXSG && REVSTA
	int		sook = 0;
#endif
#endif

	/* Actually one less than the line/screen width.
	 * Zero based index of the last position.
	 */
//	int	LineWidth = AllocColumns - 1;
	LineWidth = AllocColumns - 1;
	if(row == AllocLines - 1)	// Bottom (command/message) line is shorter
		LineWidth--;

	// Check if something needs to be done:
	if((virtscreen->v_flag & VFCHG) == 0)
		return(TRUE);

#if	TYPEAH
	// Is there pending input ?
	if(force == FALSE && typahead(FALSE))
		return(FALSE);
#endif

	/* Set up pointers to virtual and physical lines */
	virtcp = virtscreen->v_text;
	physcp = physscreen->v_text;
	virtap = virtscreen->v_attr;
	physap = physscreen->v_attr;

	/* Debug: show lines at start of update
	 * Convention is that:
	 *   text is surrounded by " " except just before return when it is surrounded by < >
	 *   attributes are surrounded by ' '
	 */
	TERMOUTTRACE4((TermDebFD, "SUL row=%d phys/virt\n", row))
	TERMOUTTRACE4((TermDebFD, "\"%s\"\n", show100(physscreen->v_text)))
	TERMOUTTRACE4((TermDebFD, "\"%s\"\n", show100(virtscreen->v_text)))
	TERMOUTTRACE4((TermDebFD, "'%s'\n",  show100a(physscreen->v_attr)))
	TERMOUTTRACE4((TermDebFD, "'%s'\n",  show100a(virtscreen->v_attr)))

#if	LEGACY_DISPLAY
	TTforg(gfcolor);
	TTbacg(gbcolor);
	TTforg(virtscreen->v_rfcolor);
	TTbacg(virtscreen->v_rbcolor);
#endif

	// **** The code immediately below is not used with modern displays ****

#if	LEGACY_DISPLAY
#if	REVSTA || COLOR
	/* if we need to change the reverse video status of the
	 * current line, we need to re-write the entire line
	 * ......NOW ......
	 */
	rev = (virtscreen->v_flag & VFREV) == VFREV;
	req = (virtscreen->v_flag & VFREQ) == VFREQ;
	if(rev && SG == 0)
		rev = 0;
	if((rev != req)
	    || (virtscreen->v_fcolor != virtscreen->v_rfcolor) || (virtscreen->v_bcolor != virtscreen->v_rbcolor)
#if	HP150
	/* the HP150 has some reverse video problems */
	    || req || rev
#endif
			) {
		movecursor(row, 0);	/* Go to start of line. */
		/* set rev video if needed */
		if(rev != req)
			TTrev(req);		

		/* scan through the line and dump it to the screen and
		   the virtual screen array				*/

		/*
		 * the following fix came from george@rebel.UUCP 
		 * (George M. Sipe)pp
		 *
		 * Subject: MicroEMACS 3.8l standout bugs (fix)
		 * Summary: patch to fix standout on embedded attribute
		 *          terminals
		 */
#if	TERMCAP && FIXSG && REVSTA
		if(req && SG > 0) {		/* Skip over 'spaces' - hopefully a rare need these days */
			ttcol += SG;
			virtcp += SG;
			physcp += SG;
			virtap += SG;
			physap += SG;
			EndVirtTxt = &virtscreen->v_text[LineWidth - SG];
		} else
#endif
			EndVirtTxt = &virtscreen->v_text[LineWidth];

		/* Output a character: which could be UTF8 up to 4 bytes
		 */
		while(virtcp < EndVirtTxt) {
			VidAttr	va;
			int	width = va_get_width(*virtap);

			/* Suppress a wide character that would take us over the edge
			 * Put a space - in case something else there
			 */
			if(virtcp + width > EndVirtTxt) {
				movecursor(row, (int)(physcp - physscreen->v_text));
				TTputc(' ');

				// Do this copy else we will be forever refreshing it
				++ttcol;
				*physcp++ = *virtcp++;
				*physap++ = *virtap++;

				break;
			}

			/* Ensure that the physical brush correct for the character that we are about
			 * to output.
			 */
			if(pBrush != va_no_width(*virtap))
				SetBrush(*virtap);

			switch(CHAR_LEN(*virtcp)) {
#if	UTF8_MAX_CODE_POINT > UTF8_MAX_CP4
			case 6:	TTputc(((BYTE*)virtcp)[0]);
				TTputc(((BYTE*)virtcp)[1]);
				TTputc(((BYTE*)virtcp)[2]);
				TTputc(((BYTE*)virtcp)[3]);
				TTputc(((BYTE*)virtcp)[4]);
				TTputc(((BYTE*)virtcp)[5]);
				break;
			case 5:	TTputc(((BYTE*)virtcp)[0]);
				TTputc(((BYTE*)virtcp)[1]);
				TTputc(((BYTE*)virtcp)[2]);
				TTputc(((BYTE*)virtcp)[3]);
				TTputc(((BYTE*)virtcp)[4]);
				break;
#endif
			case 4:	TTputc(((BYTE*)virtcp)[0]);
				TTputc(((BYTE*)virtcp)[1]);
				TTputc(((BYTE*)virtcp)[2]);
				TTputc(((BYTE*)virtcp)[3]);
				break;
			case 3:	TTputc(((BYTE*)virtcp)[0]);
				TTputc(((BYTE*)virtcp)[1]);
				TTputc(((BYTE*)virtcp)[2]);
				break;
			case 2:	TTputc(((BYTE*)virtcp)[0]);
				TTputc(((BYTE*)virtcp)[1]);
				break;
			case 1:	TTputc(((BYTE*)virtcp)[0]);
			}

			/* Characters with funny widths have been looked at in vtpututf8()
			 * The width that we see here is 1 or 2,
			 * 0 width chars having been replaced with a pink square width 1
			 *
			 * There are a few characters that we think is width 1 but *some*
			 * terminals display as 2, eg U+10FFFF
			 * Can't do anything about it since we can't eyeball what the
			 * terminal (emulator) is doing.
			 */
			ttcol += width;
			*physcp++ = *virtcp++;
			*physap++ = *virtap++;

			// Wide - just step over
			if(width == 2) {
				*physcp++ = *virtcp++;
				*physap++ = *virtap++;
			}
		}
		/* turn rev video off */
		if(rev != req)
			TTrev(FALSE);

		/* update the needed flags */
		virtscreen->v_flag &= ~VFCHG;
		if(req)
			virtscreen->v_flag |= VFREV;
		else
			virtscreen->v_flag &= ~VFREV;

		virtscreen->v_fcolor = virtscreen->v_rfcolor;
		virtscreen->v_bcolor = virtscreen->v_rbcolor;

		// SetBrush(va_default());	// Reset, else can get probs
		return(TRUE);
	}
#endif	// REVSTA || COLOR
#endif	// LEGACY_DISPLAY

	// **** Modern display below ****

	/* Find the position of the last non white space.
	 * Note: *white* space, so look at attributes.
	 * OK to let this go to the left of line start, only used for comparison
	 */
	lastNonSpace = physscreen->v_text + LineWidth;
	endPattr     = physscreen->v_attr + LineWidth;
	while(lastNonSpace >= physscreen->v_text && *lastNonSpace == ' ' && !va_diff_space(*endPattr, pens[PEN_BLANK_SCREEN])) {
		lastNonSpace--;
		endPattr--;
	}
	TERMOUTTRACE2((TermDebFD, "lNS=%d\n", (int)(lastNonSpace - physscreen->v_text)))

	/* Advance past any common chars at the left
	 * Ie find the first character that differs.
	 */
	while(virtcp < virtscreen->v_text + LineWidth && *virtcp == *physcp && *virtap == *physap) {
		++virtcp;
		++physcp;
		++virtap;
		++physap;
	}
	TERMOUTTRACE3((TermDebFD, "1st difference @ %ld (%#2.2x/%#2.2x) LineWidth=%d  phys/virt\n", (long)(virtcp - virtscreen->v_text), *physcp, *virtcp, LineWidth))
	TERMOUTTRACE4((TermDebFD, "\"%s\"\n", show100(physscreen->v_text)))
	TERMOUTTRACE4((TermDebFD, "\"%s\"\n", show100(virtscreen->v_text)))
	TERMOUTTRACE4((TermDebFD, "'%s'\n",  show100a(physscreen->v_attr)))
	TERMOUTTRACE4((TermDebFD, "'%s'\n",  show100a(virtscreen->v_attr)))

	/* If both lines are the same, no update needs to be done.
	 * This can still happen, even though we only call this routine on changed
	 * lines. A hard update is always done when a line splits, a massive
	 * change is done, or a buffer is displayed twice. This optimizes out most
	 * of the excess updating. A lot of computes are used, but these tend to
	 * be hard operations that do a lot of update, so I don't really care.
	 */
	if(virtcp == &virtscreen->v_text[LineWidth]) {

		/* Try to optimise (== minimise # characters output) the eventual
		 * placing of the cursor.
		 * The scenario is the user typing and hitting space at the end of
		 * line - this is where we are: the line looks the same but the cursor
		 * is just to the right of where it was. Emacs will generate a cursor move
		 * to get it to the right place. We can also get there by outputting
		 * the current character -- which also moved the cursor as side effect.
		 * Usually this will be a space (or tab), but could be something else
		 * in overwrite mode.
		 *
		 * # 3 below ? 1 is usually sufficient; allow 3 for typeahead reasons.
		 */
		if(row == ttrow && wantCol > ttcol && wantCol - 3 <= ttcol) {
			/* The use of ttrow & ttcol is justified here as the point is to
			 * optimise if we are in about the right place.
			 */
			virtcp = virtscreen->v_text + ttcol;
			physcp = physscreen->v_text + ttcol;
			virtap = virtscreen->v_attr + ttcol;
			physap = physscreen->v_attr + ttcol;
			EndVirtTxt = RightmostChanged = virtscreen->v_text + wantCol - 1;
			tInsertChars = 0;

			goto StartPaint;	// Too hard to do otherwise!

		}

 		virtscreen->v_flag &= ~VFCHG;		/* flag this line as updated */
		TERMOUTTRACE((TermDebFD, "<%s>\n",  show100(physscreen->v_text)))
		TERMOUTTRACE((TermDebFD, "'%s'\n", show100a(physscreen->v_attr)))

		return(TRUE);
	}

	/* Character insert/delete can screw if there are double width characters.
	 * Scan to see if there are any after the changed point, if so unset
	 * the ability to do it.
	 */
	for( ; virtap < virtscreen->v_attr + LineWidth; virtap++, physap++)
		if(va_get_width(*physap) == 2 || va_get_width(*virtap) == 2)
			tInsertChars = 0;
	virtap = virtscreen->v_attr + (int)(virtcp - virtscreen->v_text);
	physap = physscreen->v_attr + (int)(virtcp - virtscreen->v_text);

	/* If the difference is close to the start of the line set the pointers
	 * back to the start.
	 * The point is that the cursor movement to get there will often be a small distance and
	 * it will be cheaper to output the first few chars again for the sake
	 * of simpler cursor movement.
	 * Don't if a multibyte character - only test a couple of them.
	 *
	 * Don't if the cursor is a line or two away and a nearby column anyway (as in redrawing line numbers).
	 * Again use or ttrow & ttcol is justified as we want to optimise screen movement.
	 */
	if(virtcp <= virtscreen->v_text + 3 && CHAR_LEN(*virtscreen->v_text) == 1 && CHAR_LEN(*virtcp) == 1 &&
	   !(abs(row - ttrow) <= 2 && ttcol <= 6)) {
		virtcp = virtscreen->v_text;
		physcp = physscreen->v_text;
		virtap = virtscreen->v_attr;
		physap = physscreen->v_attr;
	}

	/* Should not happen, but if we end up on the 2nd character position of a double width
	 * character, back up to the starting position of that character.
	 */
	if(*virtcp == UNISTR_END) {
		--virtcp;
		--physcp;
		--virtap;
		--physap;
	}

	/* So we now know that something has changed and we have skipped over
	 * the start of the line where stuff is not changed.
	 */

	/* Find out if there is a match on the right
	 * Start at the right hand end and work back to the first change.
	 * Hopefully just a small part of the line needs updating.
	 * End....Txt are pointers to the last position that we need worry about
	 */
	EndVirtTxt = virtscreen->v_text + LineWidth;
	EndPhysTxt = physscreen->v_text + LineWidth;
	endVattr =   virtscreen->v_attr + LineWidth;
	endPattr =   physscreen->v_attr + LineWidth;

	/*
	 * more code from george@rebel.UUCP (George M. Sipe)
	 */
#if	TERMCAP && FIXSG && REVSTA && LEGACY_DISPLAY
	if(req && SG > 0)			/* Adjust for 'spaces' */
		sook = (virtcp - &virtscreen->v_text[0]) > 0;
	else
		sook = FALSE;
#endif

	/* Now start at the right hand side of the screen
	 * Work left moving over characters and attributes that
	 * have not changed.
	 * Stop where we find a difference.
	 *
	 * We don't need to test about going back to the start of the buffer
	 * since we know that we will meet a change - we found that earlier.
	 *
	 * nbflag will be TRUE if any of the unchanged rightmost chars is non blank,
	 * but spaces do not always look the same.
	 * If this is TRUE then can't use hardware clear-to-line-end to
	 * 'fix' any of the rightmost changed characters
	 */
	while(*EndVirtTxt == *EndPhysTxt && *endVattr == *endPattr) {
		/* Note if any nonblank or strange attribute */
		if(*EndVirtTxt != ' '
			// Found space: do they look the same ?
			|| va_diff_space(*endVattr, pens[PEN_BLANK_SCREEN])
		 )
			nbflag = TRUE;					/* in right match. */

		endVattr--;
		endPattr--;
		EndVirtTxt--;
		EndPhysTxt--;
	}

	/* Note the right-most character that changed, needs redrawing.
	 */
	RightmostChanged = EndVirtTxt;
	TERMOUTTRACE3((TermDebFD, "RightmostChanged=%ld (EndVirtTxt) LineWidth=%d nbflag=%d v'%c'(%#lx)%2.2x p'%c'(%#lx)%2.2x\n", (RightmostChanged - virtscreen->v_text), LineWidth, nbflag, (char)*EndVirtTxt, (long)*EndVirtTxt, *endVattr, (char)*EndPhysTxt, (long)*EndPhysTxt, *endPattr)) 

	/* Erase to EOL ?
	 * We have found the rightmost changed character, if !nbflag then
	 * everything to the right is blanks, so see if chars to the left of
	 * this have been made blanks. This would have happened if the line had
	 * ''been shortened'' - blanks replaced what was at the end.
	 */
	if(nbflag == FALSE && eolexist == TRUE
#if	LEGACY_DISPLAY
			&& (req != TRUE)
#endif
		 ) {

		TERMOUTTRACE2((TermDebFD, "RIGHT-DIFF @ %d: %#lx ~ %#lx & %2.2x ~ %2.2x\n", (int)(RightmostChanged - virtscreen->v_text), (long)*EndVirtTxt, (long)*EndPhysTxt, *endVattr, *endPattr))

		while(RightmostChanged > virtcp && *RightmostChanged == ' '
			// Found space: do they look the same ?
			&& !va_diff_space(*endVattr, pens[PEN_BLANK_SCREEN])
		      ) {
			--RightmostChanged;
			--endVattr;
		}

		TERMOUTTRACE2((TermDebFD, "EOL-2-BLANKS blanks=%d starting at %d\n", (int)(EndVirtTxt - RightmostChanged), (int)(RightmostChanged - virtscreen->v_text)))

		if((int)(EndVirtTxt - RightmostChanged) <= 3)	/* Use only if erase is */
			RightmostChanged = EndVirtTxt;		/* fewer characters. */
	}

	/* Go to start changed text on the line */
	movecursor(row, (int)(virtcp - &virtscreen->v_text[0]));

#if	LEGACY_DISPLAY
#if	REVSTA
#if	TERMCAP && FIXSG
	if(!sook)
		TTrev(rev);
#else
	TTrev(rev);
#endif
#endif
#endif	// LEGACY_DISPLAY

StartPaint:
#if	0	// Rough code below - does not work (yet)

	/* Start painting.
	 * Before we do check for the commonish situation of characters having been inserted
	 * at the start of line or in the middle and the whole lot to the end is just
	 * shifted to the right. Eg1:
	 * phys: CDEF
	 * virt: ABCDEF		-- AB insert at start
	 * Eg2:
	 * phys: ABCDEF
	 * virt: ABXYCDEF	-- XY inserted in the middle
	 *
	 * In both cases physcp will point to C (in phys)
	 * and RightmostChanged to F (in virt).
	 * virtcp to A (Eg1) and X (Eg2)
	 *
	 * What we try to do is find F (in phys) then progressively look left
	 * matching E D C.
	 * If length(XY) > TTCostIns2Move
	 * then	move to after F; delete len(XY) 
	 *	move to C; insert XY
	 *	set RightmostChanged to B
	 * (may need to remove on Right end)
	 *
	 * Maybe ought to also deal with shifts to the left, eg:
	 * phys: ABCDEF
	 * virt: CDEF
	 *
	 * The reason for all of this is that the code below
	 * phys: CDABCDEF
	 * virt: ABCDABCDEF
	 * the code below may break this up into several insert/dels when
	 * all that is needed is insert AB.
	 */
	if(tInsertChars) {
		VideoC*	phcEnd = (RightmostChanged - virtcp) + physcp;
		VidAttr*phaEnd = (RightmostChanged - virtcp) + physap;
		VidAttr*viaEnd = (RightmostChanged - virtcp) + virtap;
		while(phcEnd > physcp) {
			// Found 'F' ?
			if(*RightmostChanged != *phcEnd || *viaEnd != *phaEnd) {
				phcEnd--;
				phaEnd--;
				continue;
			}

			int	endCol = (int)(phcEnd - physscreen->v_text) + 1;

			// Matched the last character, what is length of match ?
			VideoC*	vicStart = RightmostChanged;
			VidAttr*viaStart = (RightmostChanged - virtcp) + virtap;
			while(*vicStart == *phcEnd && *viaStart == *phaEnd) {
				vicStart--;
				viaStart--;
				phcEnd--;
				phaEnd--;
			}

			// Enough to be worth while ?
			if(RightmostChanged - vicStart < TTCostIns2Move)
				break;	// No, not worth it

			TTmove(row, endCol);					// after F
			TTdelchars((int)(RightmostChanged - vicStart));
			TTmove(row, (int)(physcp - physscreen->v_text));	// to C

			while(ttcol < endCol)
				updateCP();

			TTexitins();

			// No more updates needed:
			RightmostChanged = virtcp - 1;
			break;
		}
	}
#endif	// 0

	/* We now have a narrow region in the middle of the line of characters that
	 * have changed. Some, between the LH & RH markers might not have changed,
	 * if a few are unchanged just rewrite them, if enough skip over them.
	 * Move left to right rewriting them.
	 */

Paint:	// Restart the Paint loop
	while(virtcp <= RightmostChanged) {		/* Ordinary. */
		int	width = va_get_width(*virtap);

		/* Suppress a wide character that would take us over the edge
		 * Put a space - in case something else there.
		 * This should not happen if vtpututf8() has done its job correctly.
		 */
		if(virtcp + width > RightmostChanged + 2) {
			// SetBrush(va_default());
			TTmove(row, (int)(physcp - physscreen->v_text));
			TTputc(' ');

			// Do this copy else we will be forever refreshing it
			++ttcol;
			*physcp++ = *virtcp++;
			*physap++ = *virtap++;

			break;
		}

		/* Look if there is a region of sameness in the middle the changed region.
		 * If there are enough to make movement worth while - do the jump.
		 *
		 * What this is about is when we have written at least one character at virtcp,
		 * ie it has moved on from the span of chars at the start of line that
		 * are the same.
		 */
		int	same;		// # char positions the same
		int	sameCost;	// Cost of these chars
		int	sameTo;		// Screen column of where to jump to after same sequence

		for(same = sameCost = 0; same <= RightmostChanged - virtcp; ) {
			TERMOUTTRACE((TermDebFD, "SAME COST same=%d EVT=%ld RMC=%ld same+ph=%d\n", same, (EndVirtTxt - virtscreen->v_text), (RightmostChanged - virtscreen->v_text), same+(int)(physcp - physscreen->v_text)))
			if(virtcp[same] != physcp[same] ||
			   virtap[same] != physap[same])
				break;

			sameCost += CHAR_LEN(virtcp[same]);	// Some cost more than others. Correctly deal multi byte
			if(va_get_width(virtap[same]) == 2)	// Some characters have width 2
				same++;				// (length 0 displayed as width 1)
			same++;
		}

		sameTo = (int)(virtcp - virtscreen->v_text) + same;

		/* A crude assessment of whether a move/jump-right would be more efficient
		 * than outputting the sequence of same characters.
		 * Jumping to a tab stop is a short sequence so recognise that.
		 */
		if(sameCost > 5 || (sameCost > 1 && (sameTo % 8 == 0))) {
			TERMOUTTRACE((TermDebFD, "SAME row=%d same=%d sameCost=%d col=%d->%d '%c%c%c' ->  '%c%c%c'\n", row, same, sameCost, (int)(virtcp - virtscreen->v_text), sameTo,
				physcp[0], physcp[1], physcp[2], virtcp[0], virtcp[1], virtcp[2]))

			// Skip over the same characters
			virtcp += same;
			virtap += same;
			physcp += same;
			physap += same;

			continue;
		}

		/* Look to see if we can find the current content left/right shifted a bit.
		 * Ie that some text as been deleted/inserted or a window moved left/right.
		 */
		int	shift;

		/* Look see if characters have been deleted, eg 'ab' have been spanned or
		 * updated and have a difference at [2] (where *physcp == 'c' *virtcp == 'g'), 4 chars deleted
		 * abcdefghiX |		physcp
		 * abghiY     |		virtcp
		 *           ^EndVirtTxt
		 * Look in phys to find 'ghi'
		 * so: delete 4 chars at [2]
		 *     insert 4 chars at [6] (EndVirtTxt - 4)
		 * See: Design/CharInsDel
		 *
		 * Simultaneously:
		 * Look see if characters have been inserted, eg *physcp == 'g', 4 chars inserted:
		 * abghi		physcp
		 * abcdefghi		virtcp
		 *
		 * Look at both and see which gives the longest match.
		 *
		 * Beware: use EndVirtTxt not RightmostChanged as RightmostChanged might
		 * be truncated if all/part of it will be replaced by trailing spaces.
		 *
		 * This works if there is one such change per line - which will be the
		 * common case, hopefully.
		 *
		 * In the code below NUL bytes & BELs are used to help make it
		 * easier to follow the generated terminal output. Debug only.
		 *
		 * Double width characters. Don't do them.
		 * Inserting/deleting a double width char the hardware might
		 * do two places - or not. I did try doing a cursor movement after each
		 * one (and also before) but sometimes the line just scrolled left/right
		 * more than I expected. Probably possible to fix but at a cost that
		 * doesn't make trying worth while. This was especially bad with windows
		 * left/right - characters could not fall off the right hand edge.
		 * So if we see one: give up, unfortunately.
		 * Some of the code acts as if they were still handled, but this doesn't
		 * add to the cost.
		 */
		if(tInsertChars && (*virtcp != *physcp || *virtap != *physap)) {
			TERMOUTTRACE3((TermDebFD, "EVT=%ld RightmostChanged=%ld virtcp=%ld EVT='%c%c%c'\n", (EndVirtTxt - virtscreen->v_text), (RightmostChanged - virtscreen->v_text), (virtcp - virtscreen->v_text), EndVirtTxt[-2], EndVirtTxt[-1], EndVirtTxt[0]))
			// Offset of where difference found
			int	diff = (int)(virtcp - virtscreen->v_text);

			// Scan looking for initial matching character
			for(shift = 1; shift < (int)(EndVirtTxt - virtcp); shift++) {
				int	maxMatch = (int)(EndVirtTxt - virtcp) - shift + 1;	// Max poss match
				int	delMatch = 0;	// Match length after shift deleted characters
				int	insMatch = 0;	// Match length after shift inserted characters
				int	delGain = 0;	// Gain by deleting
				int	insGain = 0;	// Gain by inserting
				int	delLend = 0;	// Flag: need to fiddle at line end on delete. Init for compiler
				int	insLend = 0;	// Flag: need to fiddle at line end on insert. Init for compiler

				TERMOUTTRACE4((TermDebFD, "SHIFT row=%d shift=%d diff=%d MM=%d poffset=%ld voffset=%ld p='%c' v='%c' p[%d]='%c'\n", row, shift, diff, maxMatch, physcp - physscreen->v_text, virtcp - virtscreen->v_text + shift, (char)*physcp, (char)*virtcp, shift, physcp[shift]))
				// Possible deleted sequence ?
				if(*virtcp == physcp[shift]) {
					// Characters moving left
					// 1st char match, but only interesting if it is start of a long match
					VideoC*	pp = &physcp[shift];
					VideoC*	vp = virtcp;
					VidAttr*pa = &physap[shift];
					VidAttr*va = virtap;

					TERMOUTTRACE4((TermDebFD, "del test init gain P:'%c'%#2.2x != V:'%c'%#2.2x  pa[-1]=%#2.2x pa[0]=%#2.2x\n", *pp, *pa, *vp, *va, pa[-1], pa[0]))
					if(pa[-1] != pa[0])
						delGain += TTCostAttChange;

					// How much matches ?
					for( ; delMatch < maxMatch; delMatch++, pp++, vp++, pa++, va++) {
						if(*pp != *vp || *pa != *va)
							break;
					}

					// Compute:
					// do we need to insert text on the right of what we shifted ?
					delLend = physscreen->v_text + diff + delMatch + shift <= lastNonSpace;
					// Gain of deleting as opposed to redrawing
					delGain += (shift + delMatch) - (TTCostDel + (delLend ? shift + TTCostIns2Move : 0));

					//delMatch=0;
				}

				// Possible inserted sequence ?
				if(virtcp[shift] == *physcp) {
					// Characters moving right
					// 1st char match, but only interesting if it is start of a long match
					VideoC*	pp = physcp;
					VideoC*	vp = &virtcp[shift];
					VidAttr*pa = physap;
					VidAttr*va = &virtap[shift];

					TERMOUTTRACE4((TermDebFD, "ins test init gain P:'%c'%#2.2x != V:'%c'%#2.2x  va[-1]=%#2.2x va[0]=%#2.2x\n", *pp, *pa, *vp, *va, va[-1], va[0]))
					// Attribute changes are quite expensive:
					if(va[-1] != va[0])
						insGain += TTCostAttChange;

					// How much matches ?
					for( ; insMatch < maxMatch; insMatch++, pp++, vp++, pa++, va++) {
						if(*pp != *vp || *pa != *va)
							break;
					}

					/* On a long line where there is a character in the last screen
					 * column: linsMatch will not include this last char as it will be
					 * highlit to show that it is the last or that something follows.
					 * Because it is highlit this character cannot be moved, it must be
					 * redrawn as its display attributes need to be changed.
					 */

					/* Maybe that was completely unnecessary
					 * What can happen is that we have just decided to shift a
					 * sequence that is just replacing what is there.
					 * Eg: physical: blank line
					 * virtual: a few chars at the start, rest blank.
					 * Don't want to insert a few spaces to save updating
					 * the load of blanks.
					 * This happens more often than you might think.
					 *
					 * This can also be triggered by the non-current window cursor
					 * being there, drat.
					 */

					int	i = insMatch;
					pp = &physcp[shift];
					vp = &virtcp[shift];
					pa = &physap[shift];
					va = &virtap[shift];
					for( ; i > 0; i--, pp++, vp++, pa++, va++)
						if(*pp != *vp || *pa != *va)
							break;

					//insMatch=0;
					if(i == 0) {
						TERMOUTTRACE4((TermDebFD, "Suppress unnecessary deleted sequence. insMatch=%d i=%d\n", insMatch, i))
						insMatch = 0;
					}

					// Need to fiddle at line end & what do we gain ?
					insLend = physscreen->v_text + diff + insMatch <= lastNonSpace;
					TERMOUTTRACE4((TermDebFD, "insGain=%d (%d) - (%d)\n", insGain, (shift + insMatch), (TTCostIns + shift + (insLend ? TTCostDel2Move : 0))))
					insGain += (shift + insMatch) - (TTCostIns + shift + (insLend ? TTCostDel2Move : 0));

				}

				TERMOUTTRACE((TermDebFD, "SHIFT RES row=%d RightmostChanged=%ld delMatch=%d delGain=%d delLend=%d insMatch=%d insGain=%d insLend=%d diff=%d shift=%d maxMatch=%d\n", row, (RightmostChanged - virtscreen->v_text), delMatch, delGain, delLend, insMatch, insGain, insLend, diff, shift, maxMatch))

				/* Did that result in gain ?
				 * Also - diff insert/delete match anything ?
				 */
				if((delGain <= 0 || delMatch == 0) && (insGain <= 0 || insMatch == 0))
					continue;

				/* It might be better to see if there is a longer match ...
				 * but, gut feeling, that would be rare.
				 */
				// Do which gained the most:
				if(delGain >= insGain) {
					// Delete characters - ie text moved left
					// TTputc('\0');TTputc('\0');	// Debug identify delete

					TERMOUTTRACE((TermDebFD, "\nSHIFT DO DEL row=%d diff=%d shift=%d delMatch=%d copy-to=%d delcol=%d lNS=%d\n", row, diff, shift, delMatch, (int)(physcp - physscreen->v_text), (int)(EndVirtTxt - virtscreen->v_text) - delMatch, (int)(lastNonSpace - physscreen->v_text)))
					/* Delete shift chars from the physical screen & its internal shadow
					 * This will leave/move-left m characters
					 */
					TERMOUTTRACE((TermDebFD, "phys/virt (cw offset=%d):\n", (int)curwp->w_leftoffset))
					TERMOUTTRACE((TermDebFD, "\"%s\"\n", show100(physscreen->v_text)))
					TERMOUTTRACE((TermDebFD, "\"%s\"\n", show100(virtscreen->v_text)))
					TERMOUTTRACE((TermDebFD, "'%s'\n",  show100a(physscreen->v_attr)))
					TERMOUTTRACE((TermDebFD, "'%s'\n",  show100a(virtscreen->v_attr)))

					TTmove(row, (int)(physcp - physscreen->v_text));
					TTdelchars(shift);
					memmove((char*)physcp, (char*)(physcp + shift), (size_t)(AllocColumns - shift - diff) * sizeof(VideoC));
					memmove((char*)physap, (char*)(physap + shift), (size_t)(AllocColumns - shift - diff) * sizeof(VidAttr));
					TERMOUTTRACE((TermDebFD, "\"%s\"\n",  show100(physscreen->v_text)))
					TERMOUTTRACE((TermDebFD, "'%s'\n",   show100a(physscreen->v_attr)))

					/* Replace the width removed above otherwise what follows is
					 * also left shifted.
					 * Don't bother if what follows is blank, it is probable that
					 * the last bit of line will be cleared below.
					 */
					TERMOUTTRACE((TermDebFD, "TestINS %d <= %d\n", diff + delMatch + shift, (int)(lastNonSpace - physscreen->v_text)))
					if(delLend) {
						int	clearEm = shift;

						physcp = physscreen->v_text + diff + delMatch;
						physap = physscreen->v_attr + diff + delMatch;

						TTmove(row, diff + delMatch);
						TTinschars(clearEm);

						// Reflect the insert in our shadow of the physical screen.
						// Screen columns not LineWidth - we are dealing with the physical display
						int	len = AllocColumns - (diff + delMatch + shift);

						TERMOUTTRACE((TermDebFD, "Copy ON INS (%d, %d, %d) ttInInsert=%d\n", (int)(physcp - physscreen->v_text) + shift, (int)(physcp - physscreen->v_text), len, ttInInsert))
						memmove((char*)(physcp + shift), (char*)(physcp), (size_t)len * sizeof(VideoC));
						memmove((char*)(physap + shift), (char*)(physap), (size_t)len * sizeof(VidAttr));

						if(ttInInsert) {
							/* Insert mode - need to write what is to fill the space,
							 * ie the screen space only gets opened up when we write to it.
							 */
							int	endCol = diff + delMatch + clearEm;
							virtcp = virtscreen->v_text + diff + delMatch;
							virtap = virtscreen->v_attr + diff + delMatch;

							while(ttcol < endCol)
								updateCP();

							TTexitins();
						} else {
							/* Space opened up.
							 * Set opened space in physical shadow to spaces and let
							 * normal update fill it in, this will let it do optimisations.
							 */
							while(--clearEm >= 0) {
								*physcp++ = ' ';
								*physap++ = pens[PEN_BLANK_SCREEN];
							}
						}
					} else {
						// Spaces inserted
						// Use LineWidth to avoid writing to the last position on the screen

						int	len = shift;
						physcp = physscreen->v_text + LineWidth - shift;
						physap = physscreen->v_attr + LineWidth - shift;
						TERMOUTTRACE((TermDebFD, "Clear %d SP at %d\n", len, (int)(physcp - physscreen->v_text)))

						while(--len >= 0) {
							// Update shadow
							*physcp++ = ' ';
							*physap++ = pens[PEN_BLANK_SCREEN];
						}

						// Adjust where the last space on the line is:
 						TERMOUTTRACE((TermDebFD, "lNS %d -> %d ('%c%c%c')\n", (int)(lastNonSpace - physscreen->v_text), (int)(lastNonSpace - physscreen->v_text) - shift, lastNonSpace[-shift-2], lastNonSpace[-shift-1], lastNonSpace[-shift-0]))
						lastNonSpace -= shift;

 					}

				} else {
					// Do insert - ie text moved right
					// TTputc('\0');	// Debug identify insert
					TERMOUTTRACE((TermDebFD, "\nSHIFT DO INS row=%d diff=%d insMatch=%d shift=%d phys[%d]='%c'(%#lx) phys[%d]='%c'(%#lx)\n", row, diff, insMatch, shift, diff, (char)physcp[0], (long)physcp[0], (diff+insMatch), (char)physcp[insMatch], (long)physcp[insMatch]))
					TERMOUTTRACE((TermDebFD, "phys/virt (cw offset=%d):\n", (int)curwp->w_leftoffset))
					TERMOUTTRACE((TermDebFD, "\"%s\"\n",  show100(physscreen->v_text)))
					TERMOUTTRACE((TermDebFD, "\"%s\"\n",  show100(virtscreen->v_text)))
					TERMOUTTRACE((TermDebFD, "'%s'\n",   show100a(physscreen->v_attr)))
					TERMOUTTRACE((TermDebFD, "'%s'\n",   show100a(virtscreen->v_attr)))
//continue;
					/* Delete at end of line so that there is room for the inserted text
					 * Don't do if at end of line, ie diff+insMatch > blanks.
					 *
					 * Also don't (need to) do if inserting below will just push
					 * them off the edge of the screen - where they disappear into oblivion.
					 * Don't, for now, test for this.
					 *
					 * Delete shift chars, then insert shift chars elsewhere - thus line width preserved.
					 */
					TERMOUTTRACE((TermDebFD, "TestDEL %d <= %d\n", diff + insMatch, (int)(lastNonSpace - physscreen->v_text)))
					if(insLend) {
						TTmove(row, diff + insMatch);
						TTdelchars(shift);	// move needs to be for rest of line; leaves shift bytes @ line end - bad
						// Make room for insMatch bytes in the physical shadow
						physcp = physscreen->v_text + diff + insMatch;
						physap = physscreen->v_attr + diff + insMatch;

						// Screen columns not LineWidth - we are dealing with the physical display
						int	len = AllocColumns - (shift + diff + insMatch);
						
						TERMOUTTRACE((TermDebFD, "memmove(%d, %d, %d)\n", (int)(physcp - physscreen->v_text), (int)(physcp - physscreen->v_text) + shift, len))
						if(len > 0) {
							memmove((char*)(physcp + 0), (char*)(physcp + shift), (size_t)len * sizeof(VideoC));
							memmove((char*)(physap + 0), (char*)(physap + shift), (size_t)len * sizeof(VidAttr));
							TERMOUTTRACE((TermDebFD, "\"%s\"\n", show100(physscreen->v_text)))
							TERMOUTTRACE((TermDebFD, "'%s'\n",  show100a(physscreen->v_attr)))
						}

						TERMOUTTRACE((TermDebFD, "clear from %d, for %d\n", AllocColumns - shift, shift))
						physcp = physscreen->v_text + AllocColumns - shift;
						physap = physscreen->v_attr + AllocColumns - shift;
						VideoC* le = physscreen->v_text + AllocColumns;
						while(physcp <= le) {
							*physcp++ = ' ';
							*physap++ = pens[PEN_BLANK_SCREEN];
						}
						TERMOUTTRACE((TermDebFD, "\"%s\"\n", show100(physscreen->v_text)))
						TERMOUTTRACE((TermDebFD, "'%s'\n",  show100a(physscreen->v_attr)))
					}

					// Open up diff columns for new text
					// Virtual: this means copying up to make the empty space
					TTmove(row, diff);
					physcp = physscreen->v_text + diff;
					physap = physscreen->v_attr + diff;
					virtcp = virtscreen->v_text + diff;
					virtap = virtscreen->v_attr + diff;

					TERMOUTTRACE((TermDebFD, "memmove(%d, %d, %d)\n", (int)(physcp - physscreen->v_text) + shift, (int)(physcp - physscreen->v_text), AllocColumns - (shift + diff)))
					memmove((char*)(physcp + shift), (char*)(physcp), (size_t)(AllocColumns - (shift + diff)) * sizeof(VideoC));
					memmove((char*)(physap + shift), (char*)(physap), (size_t)(AllocColumns - (shift + diff)) * sizeof(VidAttr));
					TERMOUTTRACE((TermDebFD, "\"%s\"\n", show100(physscreen->v_text)))
					TERMOUTTRACE((TermDebFD, "'%s'\n",  show100a(physscreen->v_attr)))

					/* Do that opening up on the screen
					 * Then write the (new) inserted text
					 * Depending on how insert works on the tty we overwrite or insert,
					 * but either way we write them here.
					 *
					 * Use the end column, not a count. updateCP() will
					 * do two places on a double width character.
					 */
					TTinschars(shift);
					if(ttInInsert) {
						int	endCol = ttcol + shift;
						while(ttcol < endCol)
							updateCP();

						TERMOUTTRACE((TermDebFD, "SHIFT last-copy-ttcol=%d '%c' ttInInsert=%d\n", ttcol-1, (char)physcp[-1], ttInInsert))
						TTexitins();	// Might do nothing - depends on terminal
					} else {
						/* Space opened up.
						 * Set opened space in physical shadow to spaces and let
						 * normal update fill it in, this will let it do optimisations.
						 */
						int	i = shift;
						while(--i >= 0) {
							*physcp++ = ' ';
							*physap++ = pens[PEN_BLANK_SCREEN];
						}
					}
				}

				TERMOUTTRACE((TermDebFD, "\"%s\"\n", show100(physscreen->v_text)))
				TERMOUTTRACE((TermDebFD, "'%s'\n",  show100a(physscreen->v_attr)))
				ST_LRscrolls++;

				// TTputc('\a');
				/* Look from where we were and update, ie rescan what we have just
				 * done - the above code is hard to debug/get-right.
				 * It should decide that a lot more of it is the same.
				 */
				physcp = physscreen->v_text + diff;
				physap = physscreen->v_attr + diff;
				virtcp = virtscreen->v_text + diff;
				virtap = virtscreen->v_attr + diff;

				goto Paint;

			}
		}

		// Write the character to the terminal:		
		updateCP();
	}

	/* Hardware erase to End Of Line.
	 * Note that if we do what we think the physical screen contains must also be updated,
	 * so set spaces and default attribs - especially not reverse video.
	 *
	 * Do NOT do this if there were some non ordinary-attribute blank characters at
	 * the end of the line.
	 */
	if(RightmostChanged != EndVirtTxt && !nbflag && physcp <= lastNonSpace) {
		movecursor(UpdateRow, (int)(physcp - pscreen[UpdateRow]->v_text));
		TTeeol();
		TERMOUTTRACE2((TermDebFD, "EEOL updateline() loop col=%d->%d  lNS=%d\n", (int)(virtcp - virtscreen->v_text), (int)(EndVirtTxt - virtscreen->v_text), (int)(lastNonSpace - physscreen->v_text)))

		while(virtcp <= EndVirtTxt) {
			*physcp++ = ' ';
			*physap++ = pens[PEN_BLANK_SCREEN];	// See below on colour choice
			virtap++;	// *virtap++ = pens[PEN_BLANK_SCREEN];	// See below on colour choice
			virtcp++;
		}
	}

	/* Don't leave the line with the brush at anything other than default,
	 * especially standout or underline.
	 * It can cause problems.
	 * Since different windows can have different colours (message line), the
	 * notion of a screen default is not clear.
	 */
	SetBrush(pens[PEN_BLANK_SCREEN]);

	TERMOUTTRACE((TermDebFD, "EUL row=%d phys/virt\n", row))
	TERMOUTTRACE((TermDebFD, "<%s>\n",  show100(physscreen->v_text)))
	TERMOUTTRACE((TermDebFD, "<%s>\n",  show100(virtscreen->v_text)))
	TERMOUTTRACE((TermDebFD, "'%s'\n", show100a(physscreen->v_attr)))
	TERMOUTTRACE((TermDebFD, "'%s'\n", show100a(virtscreen->v_attr)))

#if	LEGACY_DISPLAY
#if	REVSTA
#if	TERMCAP && FIXSG
	if(!sook) TTrev(rev);
#else
	TTrev(FALSE);
#endif
#endif
#endif
	virtscreen->v_flag &= ~VFCHG;		/* flag this line is updated */
	return(TRUE);
#endif	/* RAINBOW */
}
#endif	/* MEMMAP */

/* Special putc() for mode stuff: does a 'is it in the window' check
 */
void	mode_putc(Window* wp, UniChar c)
{
	if(wp->w_curcol < wp->w_ncolumns)
		vtputc(wp, c);
}

/* Redisplay the mode line for the window pointed to by the "wp".
 *
 * This is the only routine that has any idea of how the modeline is formatted.
 * You can change the modeline format by hacking at this routine. Called by
 * "update" any time there is a dirty window.
 */
void	modeline(Window* wp)
{
    register int mode_n;	/* cursor position count */
    register BUFFER *bp;
    register int i;		/* loop index */
    BYTE	lchar;		/* character to draw line in buffer with */
    register int firstm;	/* is this the first mode? */
    UniChar	statCh;		/* Status character - used all over */
    BYTE*	statStr;	// Status string
    LineLn	leftoffset;	/* So that it can be temporarily set to zero */
    char	numbuf[20];	/* Somewhere to put a number string */
    short	BufferTabSize;	/* Save buffer's b_tabsize */

    bp = wp->w_bufp;

    /* Show column# line#/total-line# in the buffer.
     *
     * I did think of showing the byte position in the buffer, but that
     * would have been a lot of computation work for every update.
     * It would have been byte position as the character position is more expensive to calculate.
     *
     * Do this now rather than just before adding to the mode line.
     * This is because b_tabsize is fiddled with a few lines below and I
     * don't like the complication of resetting it and restoring it
     * just for calculating the cursor's column.
     *
     * Columns counted from 0.
     */
    char	posBuf[50];
    if(wp->w_opts & WO_POSITION_SHOW) {
	BufLns	lineNo = wp->w_toplinenum;
	LineLn	colNo;
	LINE*	lp = wp->w_topLinep;

	while(lp != wp->w_dotp) {
		lineNo++;
		lp = lforw(lp);
	}

	// HEX option show byte not character position:
	colNo = ((wp->w_opts & WO_HEX) ? wp->w_doto : getdiscol(wp, wp->w_dotp, wp->w_doto, DisCol_Char));

	// Show as hex in HEX option:
	sprintf(posBuf, (wp->w_opts & WO_HEX ? " %lX %lX/%lX " : " %ld %ld/%ld "), (long)colNo, (long)lineNo, (long)bp->b_numLines);
    }


    /* Set the buffer tab size to 1. That will prevent tabs in file names, etc,
     * from being expanded. Keep the current value so that we can restore.
     */
    BufferTabSize = bp->b_tabsize;
    bp->b_tabsize = 1;

    /* Save the current attribute brush & set standout */
//    VidAttr	oldBrush = vBrush;
    inkWell = pens + PAL_MODE * INKS;
    vBrush = inkWell[INK_PLAIN];

    SetBrush(va_default());

    mode_n = wp->w_toprow+wp->w_ntrows;		/* Location. */

#if	LEGACY_DISPLAY
    vscreen[mode_n]->v_flag |= VFCHG | VFREQ | VFCOL;/* Redraw next time. */
    vscreen[mode_n]->v_rfcolor = gfcolor;	/* Originally black on */
    vscreen[mode_n]->v_rbcolor = gbcolor;	/* white.....*/
#else
    vscreen[mode_n]->v_flag |= VFCHG;		/* Redraw next time. */
#endif

    vtmove(mode_n, wp->w_leftcol);
    wp->w_curcol = 0;

    leftoffset = wp->w_leftoffset;		/* Mode line does not scroll left/right */
    wp->w_leftoffset = 0;

    if(wp == curwp)				/* mark the current buffer */
	lchar = '-';
    else {
#if	REVSTA
#if	TERMCAP && FIXSG
	if(revexist && SG > 0)			/* Initial spaces. */
	    for(i = 0; i < SG; ++i)
		mode_putc(wp, ' ');
#endif
	lchar = revexist ? ' ' : '=';
#else
	lchar = '=';
#endif
    }

#if	UNIX
    /* The idea of this nastiness is that the mode line is changed to give
     * some visual indication when running as root.
     */
    if(suidRoot)
    	statCh = '#';
    else
    	statCh = lchar;
    mode_putc(wp, statCh);
#else
    mode_putc(wp, lchar);
#endif

    // View is separate from modified:
    mode_putc(wp, (bp->b_mode & MDVIEW) ? '%' : lchar);

    // Distinguish between changed as read from .emacssave than it just being changed:
    mode_putc(wp, (bp->b_flag & BFROMSAVE) ? 'S' : ((bp->b_flag & BFCHG) ? '*' : lchar));

    if(bp->b_flag & BRENEWED) {
	/* Auto renewed - show the user that the file is being reread
	 * For a bit of fun cycle: r R ᴙ Я (ᴙ = Cyrillic Capital YA, Я = Capital reversed R)
	 * There is also 'ᴚ ʁ ɹ' if I ever get bored. Missing reverse ɹ - maybe ∟ ⍳ ⎿ 𐑃 𐑊 ι ւ ɩ
	 * Also: ᖆ ᖇ ᖈ ᖉ - all a pun on RENEW & reflections/rotations
	 * Not completely silly: it helps to show that renew has happened again.
	 * Only if the terminal understands UTF-8.
	 */
	statStr = (try_utf8 && (bp->b_flag & BRENEWEDAGAIN1) != 0) ?
		((bp->b_flag & BRENEWEDAGAIN) ? "Я" : "ᴙ") :
		((bp->b_flag & BRENEWEDAGAIN) ? "R" : "r");
    }
    else if((bp->b_flag & BFOLD))	// never true if not running under UNIX
	statStr =  bp->b_lsType == '.' ? "x" : "X";
    else if(bp->b_flag & BFTRUNC)
	statStr = "T";
    else if(bp->b_flag & BFWEIRD)
	statStr = "+";
    else
	statStr = (BYTE[]){ lchar, '\0' };
    vtputStr(wp, statStr);		// A bit long winded, but just the once

    if(bp->b_mode & MDASAVE)
	mode_putc(wp, 's');
    else
        mode_putc(wp, lchar);

    // If narrow show '<>' else '--' or '  '
    if(bp->b_btopline != bp->b_linep || bp->b_abotline != bp->b_linep) {
	vtputStr(wp, "<>");
    } else {
	mode_putc(wp, lchar);
	mode_putc(wp, lchar);
    }

    mode_putc(wp, ' ');

    if(kbdmode == RECORD)
    	mode_putc(wp, '[');

    vtputStr(wp, PROGNAME);
    
    if(kbdmode == RECORD)
    	mode_putc(wp, ']');

    mode_putc(wp, ' ');
    vtputStr(wp, Version);
    mode_putc(wp, ' ');

	/* Display the modes */
	firstm = TRUE;
	for(i = 0; i < ModeBuf.md_mode_cnt; i++) {	/* add in the mode flags */

		if((1 << i) == MDVIEW || (1 << i) == MDASAVE)
			continue;

		if(bp->b_mode & (1 << i)) {
			mode_putc(wp, firstm ? '(' : ' ');

			firstm = FALSE;

			char	codec = modecode[i];
			char*	codes = modeNameExplan[i].mn_name;

			// Renew append mode show differently:
			if((1 << i) == MDRENEW && (bp->b_flag & BRENEWAPPEND) != 0) {
				codec = 'r';
				codes = renewAppend;
			}

			if(beterse)
				mode_putc(wp, codec);
			else
				vtputStr(wp, codes);
		}
	}
	if(firstm != TRUE)
	    vtputStr(wp, ") ");

    // Some window options need showing:
    if(wp->w_opts & WO_TRACK)
        vtputStr(wp, "TRACK ");

    // Show the left scroll:
    sprintf(numbuf, ">>%ld  ", (long)leftoffset);
    vtputStr(wp, numbuf);

    vtputStr(wp, bp->b_bname);

    mode_putc(wp, ' ');
    mode_putc(wp, ' ');

    if(bp->b_fnameClean[0] != '\0') {
	/* File name
	 * If this is the help buffer the filename is a topic, say so.
	 * If there is no associated file name (think command buffer)
	 * then don't output anything at all.
	 */
	char*	what = (char*)NULL;
	if(bp->b_fname[0] != '\0')
		what = " File: ";
	if((bp->b_flag & BHELP) != 0)
		what = " Topic: ";
	if(what != NULL)
		vtputStr(wp, what);

	vtputStr(wp, bp->b_fnameClean);
	mode_putc(wp, ' ');
    }

    if(wp->w_opts & WO_POSITION_SHOW) {
	mode_putc(wp, lchar);
	vtputStr(wp, posBuf);
    }

    vteeol(wp, lchar, vBrush);

    wp->w_leftoffset = leftoffset;
//    vBrush = oldBrush;

    // Restore to what it was:
    bp->b_tabsize = BufferTabSize;

}

/* Update the mode lines.
 * If bp is not NULL update windows that display that buffer,
 * otherwise update all windows.
 */
void	upmode(BUFFER* bp, int flag)
{
	register Window* wp;

	wp = wheadp;
	while(wp != NULL) {
		if( !bp || wp->w_bufp == bp)
			wp->w_flag |= (char)flag;
		wp = wp->w_wndp;
	}
}

/* Send a command to the terminal to move the hardware cursor to row "row"
 * and column "col". The row and column arguments are origin 0. Optimize out
 * random calls.
 *
 * Some optimisations in TTmove() (tcap: tcapmove()) to use \h \n ... instead of
 * full blown cursor movements sometimes fall down on double width characters.
 * This is because of different notions of where the cursor is: 1st or 2nd char
 * position in the double width character. Terminal (emulation) s/ware seems to
 * take it to be the 1st whereas here we take it to be the 2nd. This causes problems
 * with where you want to go to next.
 * The simplest way out is to trash the ttcol value forcing a full cursor movement
 * if the character that the cursor is currently on is a double width character.
 * Ideally this should be invisible to this code (ie deal with in TTmove()) but
 * that would mean letting the lower level look at pscreen - which would cause
 * more complication than I want to think about.
 *
 * Sometimes this is called twice to move to the same place. The second call will
 * be ignored as we know that we are already there. However the above double width
 * fix trashing the known cursor position means that the cursor movement will
 * be output again. Avoid this by remembering how much we had output after the
 * last move; if this has not changed then don't trash the position since it
 * must be right.
 *
 * Losing the position means an uncertain column position, we still know the line.
 * So trashing column is the right thing to do, it means that we can still just
 * emit \r to get to the start of line.
 */
void	movecursor(int row, int col)
{
	static	ST_val	lastMoved;	// Output char count after the last move

	if(ttcol < AllocColumns && ttrow < AllocLines) {
		// Kill optimisation if the cursor is on a double width character
		int	width = va_get_width(pscreen[ttrow]->v_attr[ttcol]);

		if(width == 2 && lastMoved != ST_termOut)
			ttcol = TERM_HUGE;
	}

	// In fact tcap's TTmove will emit nothing if the cursor is in the right place
	if(row != ttrow || col != ttcol) {
		TTmove(row, col);
		lastMoved = ST_termOut;
	}
}

/* Force erase the message line. This is a special routine because the message line
 * is not considered to be part of the virtual screen. It always works
 * immediately; the terminal buffer is flushed via a call to the flusher.
 */
void	mlferase(void)
{
	register int save_discmd;

	save_discmd = discmd;
	discmd = TRUE;
	mlerase();
	discmd = save_discmd;
}

/* Erase the message line.
 * Display the last/empty line in the message window
 *
 * Do nothing if there is a prompt being shown or commands are not shown.
 *
 * Note that the message line is the full width, not left/right chopped as are other windows.
 */
void	mlerase(void)
{
	if(discmd == FALSE)
		return;

	// Set to first (blank) line
	wmessp->w_topLinep = wmessp->w_dotp = bmlPromptHistp->b_dotp = lforw(bmlPromptHistp->b_linep);

	lsetused(wmessp->w_dotp, 0);
	wmessp->w_doto = 0;

	wmessp->w_flag |= WFHARD;

	mpresf = FALSE;
	ccounted = 0;
}

/* Flush the message line out, other lines will not be written.
 * Ignore any pending input.
 *
 * If the terminal has not been initialised write the message
 * line to stdout, do not attempt to use the visual code.
 * The reason for this is to be able to delay vtinit() until after
 * the startup scripts have been run.
 */
void	mlFlush(void)
{
	// Unless it is in raw mode cannot use cursor positioning, etc
	if(TTopened != TTfullyOpened) {
		if(EarlyLog) {
			// Use fwrite() rather than puts() as it will write NUL bytes
			fwrite(ltext(wmessp->w_dotp), 1, (size_t)lused(wmessp->w_dotp), stdout);
			putchar('\n');
		}

		return;
	}

	if(vscreen == (VIDEO**)NULL)		// Screen not initialised
		return;

	int	save = DisplayMessageLineOnly;
	DisplayMessageLineOnly = 1;

	wmessp->w_doto = lused(wmessp->w_dotp);	// Cursor to end of line

	update(1);

	DisplayMessageLineOnly = save;
}

/* Append an empty LINE to the prompt/message window.
 * This is because we want to write something there.
 *
 * We might not want to touch the message line, but write to [Message History].
 * Eg: a message is generated using mlwrite() by checkage() while a prompt is
 * being entered. So the ability to save/restore state is important.
 *
 * Remember that the [Prompt Line] buffer is scratch, it is a mess, just a
 * place to put things!
 *
 * The line will have the NUL flag set and the UTF-8 if curbp is in UTF-8 mode.
 *
 * Mark the message window as needing a complete refresh.
 *
 * rep - mark the entry as replaceable as soon as there is something better
 *
 * This ought to return true/false - it can fail
 */
void	msgPush(int rep)
{
	// Create a new line at the buffer end
	if(lforw(bmlPromptHistp->b_dotp) == bmlPromptHistp->b_linep)
		BufAppendLine(bmlPromptHistp);

	// Throw away old stuff, we need little in the buffer.
	// Remove lines at the top of the buffer
	trimbuffer(bmlPromptHistp, -10);

	wmessp->w_topLinep = wmessp->w_dotp = bmlPromptHistp->b_dotp = bmlPromptHistp->b_topLinep = lforw(bmlPromptHistp->b_dotp);

	// Save state & push
	msgTop->ms_doto       = wmessp->w_doto;
	msgTop->ms_force      = wmessp->w_force;
	msgTop->ms_cursattr   = wmessp->w_cursattr;
	msgTop->ms_leftoffset = wmessp->w_leftoffset;
	msgTop->ms_pallet     = MsgWindowPallet;
	msgTop++;

	// Set nice defaults for the new line:
	wmessp->w_doto = 0;
	wmessp->w_force = 0;
	wmessp->w_cursattr = va_default();
	wmessp->w_leftoffset = 0;
	msgTop->ms_pallet = PAL_WINDOW;

	// Clear it out & set NUL flag:
	lsetused(wmessp->w_dotp, 0);
	lsetNUL(wmessp->w_dotp);
	wmessp->w_flag |= WFHARD;

	if(curbp->b_mode & MDUTF8)
		lsetUtf8(wmessp->w_dotp);
}

/* Return to the previous window and buffer
 *
 * BEWARE: if a window or buffer has been killed this could
 * result in setting invalidated values for curwp & curbp.
 * So: do the pop before the delete/kill.
 *
 * msgPop() must always be called to match msgPush() else we run out
 * of stack. But so doing will result in the message just put there being
 * wiped - so to leave it there set showMsg TRUE.
 *
 * What showMsg is about is that the bottom line has 2 uses: Prompt line &
 * Message line. if showMsg is FALSE the current message is not shown, the
 * previous one is restored.
 */
void	msgPop(int showMsg)
{
	// Pop state
	msgTop--;

	if(showMsg == FALSE) {
		// Pop back previous
		if(lback(bmlPromptHistp->b_dotp) != bmlPromptHistp->b_linep)
			wmessp->w_topLinep = wmessp->w_dotp = bmlPromptHistp->b_dotp = bmlPromptHistp->b_topLinep = lback(bmlPromptHistp->b_dotp);

		// Restore previous values
		wmessp->w_doto       = msgTop->ms_doto;
		wmessp->w_force      = msgTop->ms_force;
		wmessp->w_cursattr   = msgTop->ms_cursattr;
		wmessp->w_leftoffset = msgTop->ms_leftoffset;
		MsgWindowPallet      = msgTop->ms_pallet;
		if(wmessp->w_doto > lused(wmessp->w_dotp))
			wmessp->w_doto = lused(wmessp->w_dotp);
		if(wmessp->w_leftoffset > lused(wmessp->w_dotp))
			wmessp->w_leftoffset = 0;	// Should not happen
	}

	// Else it ends up at what is the offset of the end of "." - but on the wrong line ???
	bmlPromptHistp->b_doto = 0;

	wmessp->w_flag |= WFHARD;
}

/* Write a message into the message line window (wmessp),
 * push a new line in buffer [Message Line].
 *
 * This message is appended to buffer [Msg History], unless %h - see below.
 *
 * A small class of printf like format items is handled.
 *
 * fmt: format string for output
 *
 */
void	mlwrite(const CHAR* fmt, ...)
{
	va_list	ap;
	va_start(ap, fmt);
	msgWrite(fmt, ap);
	va_end(ap);
}



/* This does the actual work of mlwrite()
 *
 * If global CursorInPrompt is TRUE:
 * * don't write to the message line - another message/dialogue could be there
 * * only put the line to history if it is an error - this is because there could
 *   be many trivial messages. Original motivation was to suppress 'Read xx lines'
 *   messages caused by RENEW mode.
 *
 * MsgWindowPallet is changed written if fmt contains %N %I %W %E.
 *
 * If fmt contains %W or %E the global flag ShowingImportantMessage is set.
 * The idea is to stop an initial message overwriting the message.
 *
 * If EarlyLog (also set by %E & %W) flush the message - else the message might
 * not be seen.
 *
 * If fmt contains %A/%+A the last history line is appended to, regardless of %h.
 * %+A does not put the space. Bug. Won't fix.
 */
void	msgWrite(const CHAR* fmt, va_list ap)
{
	Window*	savwp = curwp;
	BUFFER*	savbp = curbp;		/* Save independently as curbp is not always curwp->w_bufp !! */
	int	flags;
	LINE*	msgHere;		/* Message written here */

	// Make message line current
	curwp = wmessp;
	curbp = bmlPromptHistp;

	// Set to default:
	mlNormalPallet();

	/* Make the current window the message window, which is set to display
	 * the buffer [Msg History]. WRONG
	 */
	msgPush(1);
	msgHere = wmessp->w_dotp;

	// So that Unicode characters in the message display properly
	if(gmode & MDUTF8)
		lsetUtf8(wmessp->w_dotp);


	ccounted = ccount;	// Countdown to clear message set to $ccount

	/* This can be used to detect if a message has been put to the mode line.
	 * Why ? To avoid overwriting it.
	 */
	mlmsgcount++;

	tsetempty(&ltxt(msgHere));	// Empty - in case %A given, which is dealed with below

	flags = vprintToText(&ltxt(msgHere), fmt, ap);

	// Don't blat important message
	if(flags & PTL_IMPORTANT)
		ShowingImportantMessage = 1;

	// Setting message line Pallet ?
	if(flags & PTL_MSG_PALLET)
		MsgWindowPallet = flags & PTL_PALLET_MASK;

	if(flags & PTL_BELL)
		TTbeep();	// Ring the bell - checked before the flush below

	/* Find the last history line & append it there.
	 * msgHere is not pointing there so vprintToText() will not
	 * have done the right thing - it can't.
	 */
	if(flags & PTL_APPEND) {
		// This would do the wrong thing if message history were
		// emtpy - which it won't be - %A not in first message
		lcatl(lback(bmlHistp->b_linep), msgHere);

		// Copy back so that the whole (original + appended) get displayed
		lcpyl(msgHere, lback(bmlHistp->b_linep));
	} else
		/* Add to the history buffer.
		 * Not harmful if this fails.
		 *
		 * Nasty interaction: we have diverted curwp to display
		 * a message, temporarily reconnect in case the user is
		 * viewing the history buffer.
		 */
		if((flags & PTL_NO_HISTORY) == 0 || (flags & PTL_ERROR) != 0) {
			Window*	savcurwp = curwp;	// Don't be clever guessing where it is
			curwp = savwp;

			addHistoryText(bmlHistp, &ltxt(msgHere), AHL_NONE);
			curwp = savcurwp;

			// Move "." to the last line unless [Message Buffer] is displayed:
			if(bmlHistp->b_nwnd == 0) {
				bmlHistp->b_dotp = lback(bmlHistp->b_linep);
				bmlHistp->b_doto = lused(bmlHistp->b_dotp);
			}
		}

#if	TERMDEBUG
	if(flags & PTL_DEBUG_FILE) {
		// Seen %! - append the line to the file DEBUG
		if(TermDebFD == (FILE*)NULL) {
			// This is not explicitly closed - let exit() do that
			TermDebFD = fopen("DEBUG", "w+");
			setlinebuf(TermDebFD);
		}

		fprintf(TermDebFD, "msgHere=%s\n", ltext(msgHere));
	}
#endif

	/* Cursor to start of line - so at least the start can be seen in a narrow window.
	 * This is a whatever-you-do-is-wrong: is the start or end of message most interesting.
	 * Recommend that the Message History is looked at if needs be.
	 */
	curwp->w_doto = 0;

	/* Flush the message window if %F but "." is not in the message window OR
	 * the terminal has not (yet) opened &: $early-log or there has been some error.
	 */
	if(((flags & PTL_FLUSH) != 0 && !CursorInPrompt) ||
	   (TTopened == TTneverOpened && EarlyLog))
		mlFlush();

	/* Don't output.
	 * If not to be written to history - empty it.
	 *
	 * Also cope with the nasty situation of a write to the message line
	 * when the prompt line is in use (they are the same line). QV MGC_PROMPTING
	 */
	/* if we are not currently echoing on the command line, abort this */
	if( (flags & PTL_NO_MSGLINE) || discmd == FALSE || CursorInPrompt) {
		msgPop(FALSE);
	} else {
		mpresf = TRUE;	// Say something on the message line
		msgPop(TRUE);
	}

	// Reset window pointer for return
	curwp = savwp;
	curbp = savbp;
}

/* This does the actual work of msgWrite()
 * Do a formatted print to line. Any previous content is cleared/lost unless %A.
 *
 * Format codes:
 * %A	Append to line, ie not clear it down. MUST be first 2 characters in fmt
 * %+A	As %A, but if line is not empty put a space to separate from what is already there
 * %c	char (ASCII - ie fits in a byte)
 * %C	UniChar (Unicode - ie possibly needs encoding as multi byte)
 * %u	UTF-8 char - address to a byte that is a character, this might be multi byte
 * %d	decimal int
 * %D	decimal long
 * %o	octal int
 * %p	hex pointer
 * %x	hex int
 * %X	hex long
 * %f	fixed point int; eg: 1234 prints as 12.34
 * %l	LINE - the text therein
 * %t	TEXT - the text therein
 * %v	VAL  - the text therein or something printable
 * %s	string - NUL terminated
 * %e	integer to be interpretted as an errno value and printed as a string via strerror()
 *
 * %%	%
 *
 * Control codes - not a format codes.
 * The requested action is not done, but a PTL_xxx flag is returned.
 * %h	Do not append the line to buffer [Msg History]
 * %#	Do not write to the message line. Message is put to history - unless %h.
 * %F	Flush the message line at the end of this write.
 *	Only really makes sense if line is in the message window or if running
 *	Emacs under a debugger, eg gdb.
 * %B	Ring the terminal bell (or visual bell or what ever is defined to happen)
 *	Only really makes sense if line is in the message window.
 *	This results in PTL_BELL being returned - this does not ring the bell itself.
 * %!	Append the line to the file DEBUG. Only if TERMDEBUG is #defined;
 *	This is always recognised, not always actioned.
 *
 * Messages can be suppressed (see $early-log), %E or %W will reset this.
 *
 * %N	Normal window colours/default colours	PAL_WINDOW
 * %I	Info	green				PAL_MSG_INFO
 * %W	Warning	cyan				PAL_MSG_WARN
 * %E	Error	red				PAL_MSG_ERROR
 *
 * %c %C %u may result in the value being a NUL byte, which is put into line.
 * This might cause premature termination by some users of line.
 *
 * %?	Where ? is none of the above produces undefined results.
 *
 * Return is more complicated than I would want, it the set of flags:
 * PTL_NO_HISTORY	Seen %h - do not write to history
 * PTL_NO_MSGLINE	Seen %# - do not write to message line
 * PTL_ERROR		Seen %E - the message was an error - write to history
 * PTL_IMPORTANT	Seen %W/%E - important message
 * PTL_FLUSH		Seen %F - Flush the message window
 * PTL_PALLET_MASK	Covers the PAL_ values that are wanted
 * PTL_APPEND		Seen %A/%+A - line was appended to rather than emptied
 * PTL_DEBUG_FILE	Seen %! - append to DEBUG file
 * The reason for the complication is that this function:
 * * can be used as AKA sprintfToLine()
 * * understands %N %I %W %E - that could most easily done by changing globals
 *   MsgWindowPallet & ShowingImportantMessage - but do not we might not want
 *   to update fiddle with the display.
 * * Some of these are irrelevant when called via printToText() - so what ?
 */
int	vprintToText(TEXT* text, const CHAR* fmt, va_list ap)
{
	int	c;			/* current char in format string */
	CHAR	tinybuf[30];		/* Small buffer into which put next %? */
	CHAR*	tinyp;			/* Tiny buffer to append	*/
	BYTE*	play;
	UniChar	uc;

	int	flags = 0;		// Flags returned

	// Clear the text, unless the format starts: %A or %+A
	if(fmt[0] == '%' && (fmt[1] == 'A' || (fmt[1] == '+' && fmt[2] == 'A'))) {
		if(fmt[1] == '+') {
			// Space if there is something in the text
			if(tused(text) != 0)
				tcatc(text, ' ');
			fmt++;
		}
		fmt += 2;	// Skip %A
		flags = PTL_APPEND;
	} else
		tsetused(text, 0);

	/* Get an item at a time and add to output if there is room */

	while((c = *fmt) != '\0') {
		if(c != '%') {
			// Append the character - which might be multi byte
			tcatm(text, fmt, CHAR_LEN(c));
			// In case it is a MB char don't just add 1:
			fmt += CHAR_LEN(c);
		} else {
			fmt++;		// Skip the %
			tinyp = tinybuf;
			tinybuf[0] = tinybuf[1] = '\0';
			c = (int) *fmt++;
			switch(c) {
			  // case 'A': // Append, now't to do
			  //	continue;

			  case 'c':
				tcatc(text, (CHAR)va_arg(ap, int));
				continue;

			  case 'C':
				uc = va_arg(ap, UniChar);
				tinyp = utf8_isvalid(uc) ? utf8_encode(uc) : (CHAR*)"?";
				tcatm(text, tinyp, CHAR_LEN(*tinyp));
				continue;

			  case 'u':
				// Copy to tinybuf & NUL terminate
				play = (BYTE*)va_arg(ap, BYTE*);
				tcatm(text, play, CHAR_LEN(*play));
				continue;

			  case 'd':
				mlputi(va_arg(ap, int), 10, tinybuf);
				break;

			  case 'D':
				mlputli(va_arg(ap, long), 10, tinybuf);
				break;

			  case 'l':
				tcatl(text, va_arg(ap, LINE*));
				continue;

			  case 't':
				tcatt(text, va_arg(ap, TEXT*));
				continue;

			  case 'v':	// A VAL - can be many things
				tcatt(text, vTEXT(valPrint(va_arg(ap, VAL*))));
				continue;

			  case 'o':
				mlputi(va_arg(ap, int), 8, tinybuf);
				break;

			  case 'p':
				mlputp(va_arg(ap, CHAR*), tinybuf);
				break;

			  case 'x':
				mlputi(va_arg(ap, int), 16, tinybuf);
				break;

			  case 'X':
				mlputli(va_arg(ap, long), 16, tinybuf);
				break;

			  case 's':
				tinyp = va_arg(ap, CHAR *);
				break;

			  case 'f':
				mlputf(va_arg(ap, int), tinybuf);
				break;

			  case 'h':
				flags |= PTL_NO_HISTORY;
				continue;

			  case '#':
				flags |= PTL_NO_MSGLINE;
				continue;

			  case '!':
				flags |= PTL_DEBUG_FILE;
				continue;

			  case 'N':
				flags |= PTL_MSG_PALLET | PAL_WINDOW;
				continue;

			  case 'I':
				flags |= PTL_MSG_PALLET | PAL_MSG_INFO;
				continue;

			  case 'W':
				flags |= PTL_MSG_PALLET | PAL_MSG_WARN | PTL_IMPORTANT;
				EarlyLog = TRUE;
				continue;

			  case 'E':
				flags |= PTL_MSG_PALLET | PAL_MSG_ERROR | PTL_IMPORTANT | PTL_ERROR;
				EarlyLog = TRUE;
				continue;

			  case 'F':
				flags |= PTL_FLUSH;	// Flush message window
				continue;

			  case 'e':
				tinyp = strerror(va_arg(ap, int));
				break;

			  case 'B':
				flags |= PTL_BELL;	// Ring the bell
				continue;

			  case '%':
				tcatc(text, '%');
				continue;

			  default:
				/* Take care not to copy a NUL byte or leave us beyond
				 * end of input.
				 * This will screw if the character after % is multi byte
				 */
				tinybuf[0] = (CHAR)c;
				if(c == '\0')
					fmt--;
				break;
			} /* switch */

			tcats(text, tinyp);
		} /* if */
	} /* while */

	return(flags);
}


/* Formatted print to a text.
 * The same format codes as mlwrite().
 *
 * Returns the flags PTL_xx
 */
int	printToText(TEXT* text, const CHAR* fmt, ...)
{
	int	flags;

	va_list	ap;
	va_start(ap, fmt);
	flags = vprintToText(text, fmt, ap);
	va_end(ap);

	return(flags);
}

/* Force a string out to the message line regardless of the
 * current $discmd setting. This is needed when $debug is TRUE
 * and for the write-message and clear-message-line commands
 *
 * s: string to force out. This will not be interpretted for %<code>
 */
void	mlforce(const CHAR* fmt, ...)
{
	register int oldcmd;	/* original command display flag */

	oldcmd = discmd;	/* save the discmd value */
	discmd = TRUE;		/* and turn display on */

	va_list	ap;
	va_start(ap, fmt);
	msgWrite(fmt, ap);
	va_end(ap);

	discmd = oldcmd;	/* and restore the original setting */
}

/* Write out a string. Update the physical cursor position. This assumes that
 * the characters in the string all have width "1"; if this is not the case
 * things will get screwed up a little.
 *
 * The second argument is a pointer to a buffer.
 *
 * Return the number of characters written to the buffer.
 */
#if	0	/* No longer needed */
int	mlputs(CHAR* s, CHAR* buf)
{
	register int	c;
	register int	count = 0;

	while((c = *s++) != '\0') {
		if(buf == (char*) NULL) {
			TTputc(c);
			++ttcol;
		} else
			*buf++ = (char)c;
		count++;
        }
        return(count);
}
#endif

/* Write out a character. This is trivial. Update the physical cursor
 * position by 1.
 *
 * If the second argument is a valid pointer then write to the buffer, not
 * to the screen.
 *
 * Return the number of characters output (=1);
 */
#if	0
int	mlputc(CHAR c, CHAR* buf)
{
	if(buf == (char*) NULL) {
		TTputc(c);
		++ttcol;
	} else
		*buf = c;

	return(1);
}
#endif

/* Write out an integer, in the specified radix (up to 16).
 *
 * Return the number of characters written to buf.
 */
#ifndef	mlputi			/* Not done by mlputli() ? */
int	mlputi(int i, unsigned int r, CHAR* buf)
{
	register int  q;
	register int  count = 0;

	if(i < 0) {
		i = -i;
		// count++;
		*buf++ = '-';
	}

	q = i / r;

	if(q != 0) {
		count += mlputi(q, r, buf);
		buf += count;
	}

	q = hexdigits[i % r];
	*buf++ = q;
	*buf = '\0';
	
	return(++count);
}
#endif	/* mlputi */

int	mlputp(void* l, char* buf)
{
	return(sprintf(buf, "%p", l));
}

/* Do the same except as a long integer.
 * Output l as radix r into buffer buf (which is always not NULL).
 *
 * This was done as a recursive routine ... had a bug where the last digit
 * of -ve numbers was truncated - because count was used to increment buf, but
 * this was 1 more than necessary as count had already been incremented making l
 * positive. This can be seen commented out in the ifdefed out version of mlputi()
 * above.
 *
 * Rewritten to be iterative, not recursive.
 *
 * Negative numbers will only have a '-' at the start if the radix is 10.
 * Others (16 and 8 are used, 2 could be) will come showing the top bit set and
 * might be quite long, 64 digits in the case of base 2.
 */
int	mlputli(long l, unsigned int r, CHAR* buf)
{
	char		tmp[66];	// Long enough for even base 2.
	char*		tp = tmp;
	int		len = 0;
	unsigned long	ul;

	if(l < 0 && r == 10) {
		ul = (unsigned long)(0-l);
		*buf++ = '-';
		len = 1;
	} else
		ul = (unsigned long)l;

	// Build up digits in reverse order in tmp
	do {
		*tp++ = hexdigits[(int)(ul % (unsigned)r)];
		ul = ul / (unsigned)r;
	} while(ul);

	// Reverse copy to output:
	len += (int)(tp - tmp);
	while(--tp >= tmp)
		*buf++ = (CHAR)*tp;

	*buf = '\0';

	return(len);
}

/* Write out a scaled integer with two decimal places into buffer buf.
 * Return the # bytes output.
 */
int	mlputf(int s, CHAR* buf)
{
	register int 	i;		/* integer portion of number	*/
	register int 	f;		/* fractional portion of number */
	register int	count;		/* number of characters output	*/

	/* break it up */
	i = s / 100;
	f = s % 100;

	/* send out the integer portion */
	count = mlputi(i, 10, buf);

	buf += count;

	*buf++ = '.';
	*buf++ = (CHAR)((f / 10) + '0');
	*buf++ = (CHAR)((f % 10) + '0');
	*buf = '\0';

	return(count + 3);
}

#if RAINBOW

void	putline(int row, int col, CHAR buf[])
{
	int	n;

	n = strlen(buf);
	if(col + n - 1 > AllocColumns)
		n = AllocColumns - col + 1;
	Put_Data(row, col, n, buf);
}
#endif

/* **** Transition on message line manipulation
 * Here for a bit, some might even eventually disappear.
 *
 * Look at mapkey() - needs work
 *	   respawn() - \n\r ?? also pipecmd() filterBuffer()
 *	   gtenv() printf \n before exit ... bad
 *	   quickexit() dittob
 *	   Die(), sigeat() quit()
 */

/* Cursor position
 * Interesting dual use:
 * * logical write
 * * physically place the cursor here
 */
void	msglCpos(int col)
{
	ttrow = TERM_HUGE;	// Disable optimisation - ttrow might be wrong

	// Might get called in prompt from command line macro
	if(TTopened == TTneverOpened)
		return;

	TTmove(AllocLines - 1, col);
}

/* Flush before input
 * This must ensure that the cursor is physically where it logically is.
 * This ought to be implicit before a cursor read
 */
void	msglFlush(void)
{
	TTflush();
}

/* Clear the message line.
 * This must only be called when physical write to the screen is being done,
 * eg shutdown.
 */
void	msglClear(void)
{
	TERMOUTTRACE((TermDebFD, "MESSAGE-CLEAR\n"))

	movecursor(AllocLines - 1, 0);
	TTeeol();	/* Remove any status line message */
}

/* Trash the command/message line.
 * This should be called when we have lost control of what this line contains,
 * ie when a sub shell has run and output something about which we know nothing.
 *
 * This does not need to be called if a full screen refresh is being done, ie
 * sgarbf. But we want to prompt for another line without refresh everything.
 */
void	msglGarbage(void)
{
	TERMOUTTRACE((TermDebFD, "MESSAGE line is garbage\n"))

	VideoC* phcp = pscreen[AllocLines - 1]->v_text;
	int	i = AllocColumns - 1;

	while(--i >= 0)
		*phcp++ = UNISTR_END;
}

/* Set the pallet for the message line to normal.
 */
void	mlNormalPallet(void)
{
	MsgWindowPallet = PAL_WINDOW;
}

/* end */
